(***********************************************************************
 * File: Exc.v
 * $Revision: 364 $
 * $Date: 2010-11-04 14:16:55 -0200 (Thu, 04 Nov 2010) $
 * $LastChangedBy: jdcampo $
 * ********************************************************************
 * This module introduces a way of representing a partial funcion in
   Coq.

   The intended partial funcion f : A -> B is represented as a total
   function f' : A -> (exc B E) that can yield either a value of type
   B or a default value --i.e., an exception-- of type E.

 ************************************************************************)


(**************************************************************************)
(*                    Definition of Exception types                       *)
(**************************************************************************)

Set Implicit Arguments.

Section Exc_Type.

(* The type representing the codomain of the partial function *)
Inductive exc (V E : Set) : Set :=
  | Value : V -> exc V E
  | Error : E -> exc V E.

Variable V1 V2 V3 E1 E2 : Set.

Definition is_Value (e : exc V1 E1) : Prop :=
  match e with
  | Value _ => True
  | Error _ => False
  end.

Definition is_Error (e : exc V1 E1) : Prop :=
  match e with
  | Value _ => False
  | Error _ => True
  end.

(* Analysing the result e of a partial function *)
Definition try_with 
  (C : Set) (e : exc V1 E1) 
  (f : V1 -> C) (g : E1 -> C) : C := 
  match e with
  | Value x => f x
  | Error y => g y
  end.

(* Analysing the result e of a partial function, discarding error
   value *)
Definition try_with'
  (C : Set) (e : exc V1 E1) 
  (f : V1 -> C) (g : C) : C := try_with e f (fun _ : E1 => g).

End Exc_Type.