(***********************************************************************
 * File: Execution_isolation.v
 * $Revision: 454 $
 * $Date: 2011-01-21 09:51:03 -0200 (Fri, 21 Jan 2011) $
 * $LastChangedBy: jdcampo $
 ***********************************************************************
 * This is the proof that Full isolation (os_isolation) is safe 
 **********************************************************************)

Require Import Exec_invariant.
Require Import Write_Isolation.
Require Import Isolation.
Require Import Traces.
Require Import Hypervisor_imp_hcalls.
Require Import Actions.
Require Import State_equivalence.

Section Execution_Isolation.

Variable ctxt : context. 


Lemma no_switch_same_aos : forall (s1 s2:state) (a:action),
  (forall o, a <> Switch o) ->
  valid_state ctxt s1 -> valid_exec ctxt s1 a s2 ->
  st_active_os s1 = st_active_os s2.
Proof.
  intros s1 s2 a H_a_not_switch H_s1_valid H_valid_exec.
  induction a; inversion_clear H_valid_exec; unfold post_action in H0.
  
  unfold post_silent, same_state in H0.
  destruct H0 as [new_cache [new_tlb [H_cache'_valid [H_tlb'_valid
    H_assign]]]]; auto.
  unfold content_of_address.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s1 cmp = state_component_interp s2 cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_OneComp H_interp H_aos System_Cache Active_OS.
  rewrite <- H_aos; try (intro; discriminate); auto.


  unfold post_read, post_all_read in H0.
  unfold pre_action, pre_read, pre_all_read in H.
  destruct H as [[ma [H_ma_map H_ptype]] [H_v_accessible H_os_running]].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s1) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s1) ma); try discriminate.
  exists v0; auto.
  destruct H0 with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with Active_OS;
    try (intros [H_abs H_aos1]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_aos1];
    rewrite H_aos1; auto.

  unfold post_read_hyper, post_all_read in H0.
  unfold pre_action, pre_read_hyper, pre_all_read in H.
  destruct H as [[ma [H_ma_map H_ptype]] H_hyper_running].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s1) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s1) ma); try discriminate.
  exists v0; auto.
  destruct H0 with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with Active_OS;
    try (intros [H_abs H_aos1]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_aos1];
    rewrite H_aos1; auto.

  unfold post_write, post_all_write in H0.
  elim H0.
  intros ma H_va_mapped [newmem H_update H_s2].
  unfold update_rw_page in H_update.
  destruct H_update as [pg H_pg_map H_eq_newmem].
  destruct H_s2 as [pg' H_pg'_map_cache
    [new_cache H_new_cache [new_tlb H_new_tlb H_s2]]].
  elim valid_state_included_cache_mem with ctxt s1 v pg'; auto.
  intros ma' H_ma'_map H_pg'_map_mem.
  assert (va_mapped_to_ma s1 (st_active_os s1) v ma) as H_ma_map.
  apply valid_state_included_tlb_mem with ctxt; auto.
  assert (ma = ma').
  elim va_mapped_same_ma with s1 v ma ma'; auto.
  subst ma'; clear H_ma'_map.
  rewrite <- H_pg_map in H_pg'_map_mem.
  injection H_pg'_map_mem; intro; subst pg'; clear H_pg'_map_mem.
  rewrite H_s2; auto.

  unfold post_write_hyper, post_all_write in H0.
  elim H0.
  intros ma H_va_mapped [newmem H_update H_s2].
  unfold update_rw_page in H_update.
  destruct H_update as [pg H_pg_map H_eq_newmem].
  destruct H_s2 as [pg' H_pg'_map_cache
    [new_cache H_new_cache [new_tlb H_new_tlb H_s2]]].
  elim valid_state_included_cache_mem with ctxt s1 v pg'; auto.
  intros ma' H_ma'_map H_pg'_map_mem.
  assert (va_mapped_to_ma s1 (st_active_os s1) v ma) as H_ma_map.
  apply valid_state_included_tlb_mem with ctxt; auto.
  assert (ma = ma').
  elim va_mapped_same_ma with s1 v ma ma'; auto.
  subst ma'; clear H_ma'_map.
  rewrite <- H_pg_map in H_pg'_map_mem.
  injection H_pg'_map_mem; intro; subst pg'; clear H_pg'_map_mem.
  rewrite H_s2; auto.

  destruct H0 as [ma1 H_ma1_hyp_map [pt_addr1
    H_pt_addr1_curr_map [newmem1 H_update1 [system1 H_system1_map H_s2]]]].
  rewrite H_s2; auto.

  destruct H0 as [system H_system_map [pt_addr1 H_pt_addr1_map
    [newmem1 H_update1 [os H_os_map H_s2]]]].
  rewrite H_s2; auto.

  unfold post_new_hyper in H0.
  destruct H0 as [system1 H_system1_map [pt_addr H_os_curr_pg [newmem1 H_drop1 H_s2]]].
  rewrite H_s2; auto.

  destruct H0 as [pt_addr H_curr_map [newmem H_drop [pt_addr' H_pt_addr'_map [newmem' H_drop' H_s2]]]].
  simpl in H_s2.
  rewrite H_s2; auto.

  destruct H0 as [pt_addr H_curr_map [newmem H_drop H_s2]].
  rewrite H_s2; auto.
 
  destruct H0 as [pt_addr H_curr_map [newmem H_drop H_s2]].
  rewrite H_s2; auto.
  
  elim H_a_not_switch with o; reflexivity.

  destruct H0 as [os1 H_os1_map H_s2].
  rewrite H_s2; auto.

  destruct H0 as [os1 H_os1_map H_s2].
  rewrite H_s2; auto.
  
  destruct H0 as [aos1 H_aos1_map H_s1'].
  rewrite H_s1'; simpl.
  reflexivity.

  elim H0 with Active_OS.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  assumption.
  intros [H_abs1 H_abs2].
  contradiction H_abs1; split; intro; discriminate.

  destruct H0 as [[H_trusted_aos1 H_state1_assign] | [H_untrusted_aos1 H_state1_assign]].
  elim H_state1_assign with Active_OS.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  assumption.
  intros [H_abs1 H_abs2].
  contradiction H_abs1; split; intro; discriminate.
  elim H_state1_assign with Active_OS.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  assumption.
  intros [H_abs1 H_abs2].
  contradiction H_abs1; split; intro; discriminate.
  
  destruct H0 as [ma1 [pg1 H_pg1_map [H_pg1_owner
    [H_pg1_content [p2m1 H_post_hyp1 [os1 H_os1_map H_s1']]]]]].
  rewrite H_s1'; simpl; reflexivity.
  destruct H0 as [p2m1 H_p2m1_map [ma1 H_ma1_hyp_map [system1 H_system1_map H_s1']]].
  rewrite H_s1'; simpl; reflexivity.
  destruct H0 as [ma1 [pg1 H_pg1_map [H_pg1_owner
    [H_pg1_content [p2m1 H_post_hyp1 [os1 H_os1_map H_s1']]]]]].
  rewrite H_s1'; simpl; reflexivity.
  destruct H0 as [p2m1 H_p2m1_map [ma1 H_ma1_hyp_map [system1 H_system1_map H_s1']]].
  rewrite H_s1'; simpl; reflexivity.
Qed.
  
(* Execution of the AOS does not modify data of other OS. *) 
Lemma os_data_isolation : forall (s s':state) (a:action) (o:os_ident),
  valid_state ctxt s -> valid_exec ctxt s a s' ->
  ~(hyper_running s) -> o <> (st_active_os s) ->
  map_apply os_ident_eq (st_oss s) o =
  map_apply os_ident_eq (st_oss s') o.
Proof.
  intros s s' a o H_s_valid H_valid_exec H_not_hyper_running H_o_not_active.

  induction a; inversion_clear H_valid_exec; unfold post_action in H0.

  destruct H0 as [new_cache [new_tlb [H_cache'_valid [H_tlb'_valid
    H_assign]]]]; auto.
  unfold content_of_address.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s cmp = state_component_interp s' cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_OneComp H_interp H_oss System_Cache Guest_OSs.
  rewrite <- H_oss; try (intro; discriminate); auto.

  unfold post_silent, same_state in H0; try (subst s'; reflexivity).
    unfold post_read, post_all_read in H0.
  unfold pre_action, pre_read, pre_all_read in H.
  destruct H as [[ma [H_ma_map H_ptype]] [H_v_accessible H_os_running]].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v0; auto.
  destruct H0 with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with Guest_OSs;
    try (intros [H_abs H_aos1]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_aos1];
    rewrite H_aos1; auto.

  unfold post_read_hyper, post_all_read in H0.
  unfold pre_action, pre_read_hyper, pre_all_read in H.
  destruct H as [[ma [H_ma_map H_ptype]] H_hyper_running].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v0; auto.
  destruct H0 with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with Guest_OSs;
    try (intros [H_abs H_aos1]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_aos1];
    rewrite H_aos1; auto.

  unfold post_write, post_all_write in H0.
  elim H0.
  intros ma H_va_mapped [newmem H_update H_s'].
  unfold update_rw_page in H_update.
  destruct H_update as [pg H_pg_map H_eq_newmem].
  destruct H_s' as [pg' H_pg'_map_cache [new_cache H_new_cache
    [new_tlb H_new_tlb H_s']]].
  elim valid_state_included_cache_mem with ctxt s v pg'; auto.
  intros ma' H_ma'_map H_pg'_map_mem.
  assert (va_mapped_to_ma s (st_active_os s) v ma) as H_ma_map.
  apply valid_state_included_tlb_mem with ctxt; auto.
  assert (ma = ma').
  elim va_mapped_same_ma with s v ma ma'; auto.
  subst ma'; clear H_ma'_map.
  rewrite <- H_pg_map in H_pg'_map_mem.
  injection H_pg'_map_mem; intro; subst pg'; clear H_pg'_map_mem.
  rewrite H_s'; auto.

  unfold post_write, post_all_write in H0.
  elim H0.
  intros ma H_va_mapped [newmem H_update H_s'].
  unfold update_rw_page in H_update.
  destruct H_update as [pg H_pg_map H_eq_newmem].
  destruct H_s' as [pg' H_pg'_map_cache [new_cache H_new_cache
    [new_tlb H_new_tlb H_s']]].
  elim valid_state_included_cache_mem with ctxt s v pg'; auto.
  intros ma' H_ma'_map H_pg'_map_mem.
  assert (va_mapped_to_ma s (st_active_os s) v ma) as H_ma_map.
  apply valid_state_included_tlb_mem with ctxt; auto.
  assert (ma = ma').
  elim va_mapped_same_ma with s v ma ma'; auto.
  subst ma'; clear H_ma'_map.
  rewrite <- H_pg_map in H_pg'_map_mem.
  injection H_pg'_map_mem; intro; subst pg'; clear H_pg'_map_mem.
  rewrite H_s'; auto.

  unfold pre_action, pre_new_untrusted, hyper_running in H.
  destruct H as (H_os_untrusted, (ma', H_hyp_map, (H_ptype, (H_ma'_owner, (H_os_acc, (H_hcall, H_activity)))))).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  destruct H0 as [ma1 H_ma1_hyp_map [pt_addr1
    H_pt_addr1_curr_map [newmem1 H_update1 [system1 H_system1_map H_s']]]].
  rewrite H_s'; simpl.
  rewrite lemma_map_add_apply_distinct; auto.

  unfold pre_action, pre_new_hyper, hyper_running in H.
  destruct H as (H_waiting, H').
  unfold hyper_running in H_not_hyper_running.
  rewrite H_waiting in H_not_hyper_running; elim H_not_hyper_running; auto.

  unfold pre_action, pre_del_untrusted, hyper_running in H.
  destruct H as (H_os_untrtusted, (H_os_hcall, (system, H_system_map, (p2m, H_hyp_mapping,
    (pt_addr, H_pt_addr_map, (H_pre_all_del, (H_activity, H_accessible))))))).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  destruct H0 as [pt_addr H_curr_map [newmem H_drop H_s']].
  rewrite H_s'; auto.

  unfold pre_action, pre_del_hyper, pre_all_del, hyper_running in H.
  destruct H as (pt_addr, H_curr_map, (H_pre', (H_activity, H_not_accessible))). 
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  unfold pre_action, pre_switch, hyper_running in H.
  destruct H as (H_activity, H_pre').
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.
 
  destruct H0 as [os1 H_os1_map H_s'].
  rewrite H_s'; simpl.
  rewrite lemma_map_add_apply_distinct; auto.

  unfold pre_action, pre_lswitch_untrusted, hyper_running in H.
  destruct H as [H_o0_untrusted [aos H_aos_map [H_hcall [H_act H_all_lswitch]]]].
  unfold hyper_running in H_not_hyper_running.
  rewrite H_act in H_not_hyper_running; elim H_not_hyper_running; auto.

  destruct H0 as [aos1 H_aos1_map H_s1'].
  rewrite H_s1'; simpl.
  rewrite lemma_map_add_apply_distinct; auto.

  destruct H0 with Guest_OSs as [[H_trusted_aos1 H_state1_assign] | [H_untrusted_aos1 H_state1_assign]].
  unfold state_component_interp in H_state1_assign.
  rewrite H_state1_assign; reflexivity.
  contradiction H_untrusted_aos1.
  split; intro; discriminate.

  unfold pre_action, pre_chmod, hyper_running in H.
  destruct H as (H_activity, H_hcall).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  destruct H0 as [ma1 [pg1 H_pg1_map [H_pg1_owner
    [H_pg1_content [p2m1 H_post_hyp1 [os1 H_os1_map H_s1']]]]]].
  rewrite H_s1'; simpl; rewrite lemma_map_add_apply_distinct; auto.
  destruct H0 as [p2m1 H_p2m1_map [ma1 H_ma1_hyp_map [system1 H_system1_map H_s1']]].
  rewrite H_s1'; simpl; rewrite lemma_map_add_apply_distinct; auto.

  unfold pre_action, pre_page_pin_untrusted, pre_all_pin, hyper_running in H.
  destruct H as (H_ctxt, (H_hcall, (H_activity, H_pre'))).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  unfold pre_action, pre_page_unpin_untrusted, pre_all_unpin, hyper_running in H.
  destruct H as (H_ctxt, (H_hcall, (H_activity, H_pre'))).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.
Qed.

Lemma read_hyper_osi_equiv : forall (s s':state) (osi:os_ident) (va:vadd),
  valid_state ctxt s -> valid_exec ctxt s (Read_hyper va) s' ->
  osi_equivalent ctxt s s' osi.
Proof.
  intros s s' osi va H_s_valid H_valid_exec.
  unfold osi_equivalent.
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  
  unfold post_read_hyper, post_all_read in H_post.
  unfold pre_action, pre_read_hyper, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_ma_map H_ptype]] H_hyper_running].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v; auto.
  unfold post_action, post_read_hyper, post_all_read in H_post.
  destruct H_post with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with Active_OS;
    try (intros [H_abs H_cmp]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_cmp];
    rewrite <- H_cmp; auto; clear H_cmp H_triv.
  elim H_assign with OS_Activity;
    try (intros [H_abs H_cmp]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_cmp];
    rewrite <- H_cmp; auto; clear H_cmp H_triv.
  elim H_assign with Guest_OSs;
    try (intros [H_abs H_cmp]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_cmp];
    rewrite <- H_cmp; auto; clear H_cmp H_triv.

  unfold osi_equivalent_hyper_mapping, osi_hyper_mapping_eq,
    osi_equivalent_pt_pages, osi_pt_pages_eq, padd_page.
  elim H_assign with Hypervisor_SC;
    try (intros [H_abs H_cmp]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_cmp];
    rewrite <- H_cmp; auto; clear H_cmp H_triv.
  elim H_assign with System_Memory;
    try (intros [H_abs H_cmp]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_cmp];
    rewrite <- H_cmp; auto; clear H_cmp H_triv.

  repeat split; auto.
  
  case (os_ident_eq (st_active_os s) osi).
  left; repeat split; auto.
  right; repeat split; auto.

  (* s eq_pt s' *)
  intros pt_padd pt_page mp [pt_addr H_pt_addr_map H_pt_page_map] H_pt_page_content.
  exists pt_page.
  exists pt_addr; auto.
  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; try split; auto.

  (* s' eq_pt s *)
  intros pt_padd pt_page mp [pt_addr H_pt_addr_map H_pt_page_map] H_pt_page_content.
  exists pt_page.
  exists pt_addr; auto.
  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; try split; auto.
Qed.

Lemma write_hyper_osi_equiv : forall (s s':state) (osi:os_ident) (va:vadd) (val:value),
  valid_state ctxt s -> valid_exec ctxt s (Write_hyper va val) s' ->
  osi_equivalent ctxt s s' osi.
Proof.
  intros s s' osi va val H_s_valid H_valid_exec.

  unfold osi_equivalent.

  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold post_write, post_all_write in H_post.
  elim H_post.
  intros ma H_va_mapped [newmem H_update H_s'].
  unfold update_rw_page in H_update.
  destruct H_update as [pg H_pg_map H_eq_newmem].
  destruct H_s' as [pg' H_pg'_map_cache [new_cache H_new_cache [new_tlb H_new_tlb H_s']]].
  elim valid_state_included_cache_mem with ctxt s va pg'; auto.
  intros ma' H_ma'_map H_pg'_map_mem.
  assert (va_mapped_to_ma s (st_active_os s) va ma) as H_ma_map.
  apply valid_state_included_tlb_mem with ctxt; auto.
  assert (ma = ma').
  elim va_mapped_same_ma with s va ma ma'; auto.
  subst ma'; clear H_ma'_map.
  rewrite <- H_pg_map in H_pg'_map_mem.
  injection H_pg'_map_mem; intro; subst pg'; clear H_pg'_map_mem.
  unfold osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page.

  unfold pre_action, pre_write_hyper in H_pre.
  destruct H_pre as [H_hyper_running [ma' H_ma'_map [H_pg_type H_pg_owner]]].
  assert (ma'=ma).
  apply va_mapped_same_ma with s va; auto.
  subst ma'; clear H_ma'_map.
  unfold maddress_owner in H_pg_owner.
  rewrite <- H_pg_map in H_pg_owner.

  rewrite H_s', H_eq_newmem; simpl.
  repeat split.
  
  case (os_ident_eq (st_active_os s) osi).
  left; repeat split; auto.
  right; repeat split; auto.
  
  (* s eq_hyp s' *)
  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  exists ma'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_ma'.
  rewrite <- H_ma_eq_ma' in H_pg'_map.
  rewrite <- H_pg_map in H_pg'_map.
  injection H_pg'_map; intro H_pg'_eq_pg.
  subst pg'; clear H_pg'_map.
  subst ma'.

  unfold hypervisor_mapping_madd in H_ma'_map.
  destruct H_ma'_map as [p2m H_p2m_map H_ma_hyp_map]. 

  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper.
  unfold valid_hypervisor in H_valid_hyper.
  destruct H_valid_hyper with osi pa' p2m as [[ma' H_ma'_map H_ma_owner] | H_error]; auto.
  rewrite <- H_ma_hyp_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_pg_map in H_ma_owner.
  rewrite H_pg_owner in H_ma_owner.
  discriminate H_ma_owner.
  rewrite <- H_ma_hyp_map in H_error.
  discriminate H_error.

  (* s' eq_hyp s *)
  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  exists ma'; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  intro H_ma_eq_ma'.
  subst ma'.

  unfold hypervisor_mapping_madd in H_ma'_map.
  destruct H_ma'_map as [p2m H_p2m_map H_ma_hyp_map]. 

  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper.
  unfold valid_hypervisor in H_valid_hyper.
  destruct H_valid_hyper with osi pa' p2m as [[ma' H_ma'_map H_ma_owner] | H_error]; auto.
  rewrite <- H_ma_hyp_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_pg_map in H_ma_owner.
  rewrite H_pg_owner in H_ma_owner.
  discriminate H_ma_owner.
  rewrite <- H_ma_hyp_map in H_error.
  discriminate H_error.

  (* s eq_pt s' *)
  intros pt_padd pt_page mp [pt_addr H_pt_addr_map H_pt_page_map] H_pt_page_content.
  exists pt_page.
  exists pt_addr; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  
  rewrite <- H_ma_eq_pt_addr in H_pt_page_map.
  rewrite <- H_pg_map in H_pt_page_map.
  injection H_pt_page_map; intros H_pt_page_eq_pg.
  rewrite H_pt_page_eq_pg in H_pt_page_content.
  unfold page_type in H_pg_type.
  rewrite <- H_pg_map in H_pg_type.
  rewrite <- H_pt_page_content in H_pg_type.
  discriminate H_pg_type.

  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; try split; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_ma'.
  rewrite <- H_ma_eq_ma' in H_ma'_hyp_map.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m H_p2m_map H_ma'_hyp_map].
  destruct H_valid_hyper with osi pa' p2m as [[ma'' H_ma''_map H_ma_owner] | H_error]; auto.
  rewrite <- H_ma'_hyp_map in H_ma''_map.
  injection H_ma''_map; intro; subst ma''; clear H_ma''_map.
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_pg_map in H_ma_owner.
  rewrite H_pg_owner in H_ma_owner.
  discriminate H_ma_owner.
  rewrite <- H_ma'_hyp_map in H_error.
  discriminate H_error.

  (* s' eq_pt s *)
  intros pt_padd pt_page mp [pt_addr H_pt_addr_map H_pt_page_map] H_pt_page_content.
  exists pt_page.
  exists pt_addr; auto.
  rewrite lemma_map_add_apply_distinct in H_pt_page_map; auto.
  intro H_ma_eq_pt_addr.
  
  rewrite <- H_ma_eq_pt_addr in H_pt_page_map.
  rewrite lemma_map_add_apply in H_pt_page_map.
  injection H_pt_page_map; intro H_pt_page_value.
  rewrite H_pt_page_value in H_pt_page_content.
  simpl in H_pt_page_content.
  discriminate H_pt_page_content.

  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; try split; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  intro H_ma_eq_ma'.
  rewrite <- H_ma_eq_ma' in H_ma'_hyp_map.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m H_p2m_map H_ma'_hyp_map].
  destruct H_valid_hyper with osi pa' p2m as [[ma'' H_ma''_map H_ma_owner] | H_error]; auto.
  rewrite <- H_ma'_hyp_map in H_ma''_map.
  injection H_ma''_map; intro; subst ma''; clear H_ma''_map.
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_pg_map in H_ma_owner.
  rewrite H_pg_owner in H_ma_owner.
  discriminate H_ma_owner.
  rewrite <- H_ma'_hyp_map in H_error.
  discriminate H_error.
Qed.

Lemma new_hyper_osi_equiv : forall (s s':state) (osi:os_ident) (va:vadd) (ma:madd),
  valid_state ctxt s -> valid_exec ctxt s (New_hyper va ma) s' ->
  osi_equivalent ctxt s s' osi.
Proof.
  intros s s' osi va ma H_s_valid H_valid_exec.

  unfold osi_equivalent.

  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).

  unfold osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page.
  unfold post_action, post_new_hyper in H_post.
  destruct H_post as [aos H_aos_map [pt_addr H_curr_pt_page [newmem H_update H_s']]].
  rewrite H_s'; simpl.
  split.
  case (os_ident_eq (st_active_os s) osi); [left | right]; split; auto.
  
  unfold update_pt_page in H_update.
  destruct H_update as [pt_page H_pt_page_map [oldmap H_pt_page_content H_newmem]].
  rewrite H_newmem.

  unfold pre_action, pre_new_hyper in H_pre.
  destruct H_pre as [H_hyper_running [H_ma_type [H_ma_owner H_va_not_accessible]]].

  repeat split.
  (* s eq_hyp s' *)
  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  exists ma'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma'.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_map.
  rewrite <- H_pt_page_map in H_pg'_map.
  injection H_pg'_map; intro H_pg'_eq_pt_page.
  rewrite <- H_pg'_eq_pt_page in H_pt_page_content.
  rewrite <- H_pt_page_content in H_pg'_content.
  discriminate H_pg'_content.

  (* s' eq_hyp s *)
  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  exists ma'; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  intro H_pt_addr_eq_ma'.
  subst ma'.
  rewrite lemma_map_add_apply in H_pg'_map.
  injection H_pg'_map; intro H_pg'_value.
  rewrite H_pg'_value in H_pg'_content.
  discriminate H_pg'_content.

  (* s eq_pt s' *)
  intros pt_padd pt_page' mp [pt_addr' H_pt_addr'_map H_pt_page'_map] H_pt_page'_content.
  case (os_ident_eq (st_active_os s) osi).
  (* osi is AOS *)
  intro; subst osi.
  unfold os_current_vadd_mapping_madd in H_curr_pt_page.
  destruct H_curr_pt_page as [system H_system_map [p2m H_p2m_map H_pt_addr_map]].
  case (padd_eq pt_padd (os_current_page system)).
  intro H_pt_padd_current.
  rewrite H_pt_padd_current in H_pt_addr'_map.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m' H_p2m'_map H_pt_addr'_map].
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_page'_map.
  injection H_pt_page'_map; intro; subst pt_page'; clear H_pt_page'_map.
  rewrite <- H_pt_page_content in H_pt_page'_content.
  injection H_pt_page'_content; intro; subst mp; clear H_pt_page'_content.

  exists (Page (PT_c (map_add vadd_eq oldmap va ma)) (page_owned_by pt_page)).
  exists pt_addr.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  rewrite H_pt_padd_current.
  assumption.
  rewrite lemma_map_add_apply.
  reflexivity.
  exists (map_add vadd_eq oldmap va ma); auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  assert (va<>va') as H_va_ne_va'.
  intro H_va_eq_va'.
  rewrite <- H_va_eq_va' in H_va'_accessible.
  apply H_va_not_accessible; auto.
  exists ma'.
  rewrite lemma_map_add_apply_distinct; auto.
  
  split; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma'.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite <- H_pt_page_map in H_pg'_type.
  rewrite <- H_pt_page_content in H_pg'_type.
  discriminate H_pg'_type.
  (* pt_padd is not current page, so it does not change *)
  intro H_pt_padd_not_current.
  exists pt_page'.
  exists pt_addr'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_pt_addr'.
  apply H_pt_padd_not_current.
  apply hyp_mapped_same_paddr with ctxt s (st_active_os s) pt_addr; auto.
  rewrite H_pt_addr_eq_pt_addr'; assumption.
  unfold hypervisor_mapping_madd; exists p2m; auto.
  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  assert (va<>va') as H_va_ne_va'.
  intro H_va_eq_va'.
  rewrite <- H_va_eq_va' in H_va'_accessible.
  apply H_va_not_accessible; auto.
  exists ma'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma'.
  unfold page_type in H_pg'_type.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_type.
  rewrite <- H_pt_page_map in H_pg'_type.
  rewrite <- H_pt_page_content in H_pg'_type.
  discriminate H_pg'_type.

  (* osi not active *)
  intro H_osi_not_active.
  exists pt_page'.
  exists pt_addr'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_pt_addr'.

  rewrite <- H_pt_addr_eq_pt_addr' in H_pt_page'_map.
  rewrite <- H_pt_page_map in H_pt_page'_map.
  injection H_pt_page'_map; intro H_pt_page'_eq_pt_page; clear H_pt_page'_map.

  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m' H_p2m'_map H_pt_addr'_map].
  destruct H_valid_hyper with osi pt_padd p2m' as [[pt_addr'' H_pt_addr''_map H_pt_addr'_owner] | H_error]; auto.
  rewrite <- H_pt_addr'_map in H_pt_addr''_map.
  injection H_pt_addr''_map; intro; subst pt_addr''; clear H_pt_addr''_map.
  unfold maddress_owner in H_pt_addr'_owner.
  rewrite <- H_pt_addr_eq_pt_addr' in H_pt_addr'_owner.
  rewrite <- H_pt_page_map in H_pt_addr'_owner.
  clear H_valid_hyper.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold os_current_vadd_mapping_madd in H_curr_pt_page.
  destruct H_curr_pt_page as [aos' H_aos'_map [p2m H_p2m_map H_pt_addr_map]].
  rewrite <- H_aos_map in H_aos'_map; injection H_aos'_map; intro; subst aos'; clear H_aos'_map.
  unfold hypervisor_mapping in H_p2m_map.
  destruct H_valid_hyper with (st_active_os s) (os_current_page aos) p2m as [[pt_addr'' H_pt_addr''_map H_pt_addr_owner] | H_error]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr''_map.
  injection H_pt_addr''_map; intro; subst pt_addr''; clear H_pt_addr''_map.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner.
  rewrite H_pt_addr'_owner in H_pt_addr_owner.
  injection H_pt_addr_owner; intro H_osi_active.
  apply H_osi_not_active; auto.
  rewrite <- H_pt_addr_map in H_error.
  discriminate H_error.
  rewrite <- H_pt_addr'_map in H_error.
  discriminate H_error.

  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; auto.
  split; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma'.
  unfold page_type in H_pg'_type.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_type.
  rewrite <- H_pt_page_map in H_pg'_type.
  rewrite <- H_pt_page_content in H_pg'_type.
  discriminate H_pg'_type.

  (* s' eq_pt s *)
  intros pt_padd pt_page' mp [pt_addr' H_pt_addr'_map H_pt_page'_map] H_pt_page'_content.
  case (os_ident_eq (st_active_os s) osi).
  (* osi is AOS *)
  intro; subst osi.
  unfold os_current_vadd_mapping_madd in H_curr_pt_page.
  destruct H_curr_pt_page as [system H_system_map [p2m H_p2m_map H_pt_addr_map]].
  case (padd_eq pt_padd (os_current_page system)).
  intro H_pt_padd_current.
  rewrite H_pt_padd_current in H_pt_addr'_map.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m' H_p2m'_map H_pt_addr'_map].
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite lemma_map_add_apply in H_pt_page'_map.
  exists pt_page.
  exists pt_addr; auto.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  rewrite H_pt_padd_current; auto.
  exists oldmap; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  assert (va<>va') as H_va_ne_va'.
  intro H_va_eq_va'.
  rewrite <- H_va_eq_va' in H_va'_accessible.
  apply H_va_not_accessible; auto.
  exists ma'.
  injection H_pt_page'_map; intro H_pt_page'_value.
  rewrite H_pt_page'_value in H_pt_page'_content.
  simpl in H_pt_page'_content.
  injection H_pt_page'_content; intro H_mp_value.
  rewrite H_mp_value in H_ma'_map.
  rewrite lemma_map_add_apply_distinct in H_ma'_map; auto.
  split; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  intro H_pt_addr_eq_ma'.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
  (* pt_padd is not current page, so it does not change *)
  intro H_pt_padd_not_current.
  exists pt_page'.
  exists pt_addr'; auto.
  rewrite lemma_map_add_apply_distinct in H_pt_page'_map; auto.
  intro H_pt_addr_eq_pt_addr'.
  apply H_pt_padd_not_current.
  apply hyp_mapped_same_paddr with ctxt s (st_active_os s) pt_addr; auto.
  rewrite H_pt_addr_eq_pt_addr'; assumption.
  unfold hypervisor_mapping_madd; exists p2m; auto.
  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  assert (va<>va') as H_va_ne_va'.
  intro H_va_eq_va'.
  rewrite <- H_va_eq_va' in H_va'_accessible.
  apply H_va_not_accessible; auto.
  exists ma'; auto.
  split; auto.
  unfold page_type in H_pg'_type.
  assert (pt_addr<>ma') as H_pt_addr_ne_ma'.
  intro H_pt_addr_eq_ma'.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_type; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.

  (* osi not active *)
  intro H_osi_not_active.
  exists pt_page'.
  exists pt_addr'; auto.
  rewrite lemma_map_add_apply_distinct in H_pt_page'_map; auto.
  intro H_pt_addr_eq_pt_addr'.

  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m' H_p2m'_map H_pt_addr'_map].
  destruct H_valid_hyper with osi pt_padd p2m' as [[pt_addr'' H_pt_addr''_map H_pt_addr'_owner] | H_error]; auto.
  rewrite <- H_pt_addr'_map in H_pt_addr''_map.
  injection H_pt_addr''_map; intro; subst pt_addr''; clear H_pt_addr''_map.
  unfold maddress_owner in H_pt_addr'_owner.
  rewrite <- H_pt_addr_eq_pt_addr' in H_pt_addr'_owner.
  rewrite <- H_pt_page_map in H_pt_addr'_owner.
  clear H_valid_hyper.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold os_current_vadd_mapping_madd in H_curr_pt_page.
  destruct H_curr_pt_page as [aos' H_aos'_map [p2m H_p2m_map H_pt_addr_map]].
  rewrite <- H_aos_map in H_aos'_map; injection H_aos'_map; intro; subst aos'; clear H_aos'_map.
  unfold hypervisor_mapping in H_p2m_map.
  destruct H_valid_hyper with (st_active_os s) (os_current_page aos) p2m as [[pt_addr'' H_pt_addr''_map H_pt_addr_owner] | H_error]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr''_map.
  injection H_pt_addr''_map; intro; subst pt_addr''; clear H_pt_addr''_map.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner.
  rewrite H_pt_addr'_owner in H_pt_addr_owner.
  injection H_pt_addr_owner; intro H_osi_active.
  apply H_osi_not_active; auto.
  rewrite <- H_pt_addr_map in H_error.
  discriminate H_error.
  rewrite <- H_pt_addr'_map in H_error.
  discriminate H_error.

  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; auto.
  split; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  intro H_pt_addr_eq_ma'.
  rewrite H_pt_addr_eq_ma' in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
Qed.

Lemma del_hyper_osi_equiv : forall (s s':state) (osi:os_ident) (va:vadd),
  valid_state ctxt s -> valid_exec ctxt s (Del_hyper va) s' ->
  osi_equivalent ctxt s s' osi.
Proof.
  intros s s' osi va H_s_valid H_valid_exec.
  unfold osi_equivalent.
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).

  unfold osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page.
  unfold post_action, post_del_hyper, post_all_del in H_post.
  destruct H_post as [pt_addr H_curr_pt_page [newmem H_newmem H_s']].
  rewrite H_s'; simpl.

  split.
  case (os_ident_eq (st_active_os s) osi); [left|right]; split; auto.
  split; auto.
  
  unfold drop_va_pt_page in H_newmem.
  destruct H_newmem as [pt_page H_pt_page_map [oldmap H_pt_page_content H_newmem]].
  rewrite H_newmem.

  unfold pre_action, pre_del_hyper, pre_all_del in H_pre.
  destruct H_pre as [pt_addr' H_curr_pt_page' [[mp H_mp_map [ma H_ma_map]] [H_hyper_running H_va_not_accessible]]].

  unfold current_vadd_mapping_madd, os_current_vadd_mapping_madd in *.
  destruct H_curr_pt_page as [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]].
  destruct H_curr_pt_page' as [aos' H_aos'_map [p2m' H_p2m'_map H_pt_addr'_map]].
  rewrite <- H_aos_map in H_aos'_map; injection H_aos'_map; intro; subst aos'; clear H_aos'_map.
  unfold hypervisor_mapping in H_p2m_map, H_p2m'_map.
  rewrite <- H_p2m_map in H_p2m'_map; injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_pt_addr_map in H_pt_addr'_map; injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold vadd_mapping in H_mp_map.
  destruct H_mp_map as [pt_page' H_pt_page'_map H_pt_page_content'].
  rewrite <- H_pt_page_map in H_pt_page'_map; injection H_pt_page'_map; intro; subst pt_page'; clear H_pt_page'_map.
  rewrite <- H_pt_page_content in H_pt_page_content'; injection H_pt_page_content'; intro; subst mp; clear H_pt_page_content'.
  
  repeat split.
  
  (* s eq_hyp s' *)
  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  exists ma'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma'.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_map.
  rewrite <- H_pt_page_map in H_pg'_map.
  injection H_pg'_map; intro H_pg'_eq_pt_page.
  rewrite <- H_pg'_eq_pt_page in H_pt_page_content.
  rewrite <- H_pt_page_content in H_pg'_content.
  discriminate H_pg'_content.

  (* s' eq_hyp s *)
  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  exists ma'; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  intro H_pt_addr_eq_ma'.
  subst ma'.
  rewrite lemma_map_add_apply in H_pg'_map.
  injection H_pg'_map; intro H_pg'_value.
  rewrite H_pg'_value in H_pg'_content.
  discriminate H_pg'_content.

  (* s eq_pt s' *)
  intros pt_padd pt_page' mp [pt_addr' H_pt_addr'_map H_pt_page'_map] H_pt_page'_content.
  case (os_ident_eq (st_active_os s) osi).
  (* osi is active *)
  intro; subst osi.
  case (padd_eq pt_padd (os_current_page aos)).
  intro H_pt_padd_current.
  rewrite H_pt_padd_current in H_pt_addr'_map.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m' H_p2m'_map H_pt_addr'_map].
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_page'_map.
  injection H_pt_page'_map; intro; subst pt_page'; clear H_pt_page'_map.
  rewrite <- H_pt_page_content in H_pt_page'_content.
  injection H_pt_page'_content; intro; subst mp; clear H_pt_page'_content.

  exists (Page (PT_c (map_drop vadd_eq oldmap va)) (page_owned_by pt_page)).
  exists pt_addr.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  rewrite H_pt_padd_current; auto.
  rewrite lemma_map_add_apply; auto.
  exists (map_drop vadd_eq oldmap va); auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  assert (va<>va') as H_va_ne_va'.
  intro H_va_eq_va'.
  rewrite <- H_va_eq_va' in H_va'_accessible.
  apply H_va_not_accessible; auto.
  exists ma'.
  rewrite map_drop_apply_distinct; auto.
  
  split; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma'.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite <- H_pt_page_map in H_pg'_type.
  rewrite <- H_pt_page_content in H_pg'_type.
  discriminate H_pg'_type.
  (* pt_padd is not current page, so it does not change *)
  intro H_pt_padd_not_current.
  exists pt_page'.
  exists pt_addr'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_pt_addr'.
  apply H_pt_padd_not_current.
  apply hyp_mapped_same_paddr with ctxt s (st_active_os s) pt_addr; auto.
  rewrite H_pt_addr_eq_pt_addr'; assumption.
  unfold hypervisor_mapping_madd; exists p2m; auto.
  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  assert (va<>va') as H_va_ne_va'.
  intro H_va_eq_va'.
  rewrite <- H_va_eq_va' in H_va'_accessible.
  apply H_va_not_accessible; auto.
  exists ma'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma'.
  unfold page_type in H_pg'_type.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_type.
  rewrite <- H_pt_page_map in H_pg'_type.
  rewrite <- H_pt_page_content in H_pg'_type.
  discriminate H_pg'_type.

  (* osi is not active *)
  intro H_osi_not_active.
  exists pt_page'.
  exists pt_addr'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_pt_addr'.

  rewrite <- H_pt_addr_eq_pt_addr' in H_pt_page'_map.
  rewrite <- H_pt_page_map in H_pt_page'_map.
  injection H_pt_page'_map; intro H_pt_page'_eq_pt_page; clear H_pt_page'_map.

  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap H_valid_currpg
      H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m' H_p2m'_map H_pt_addr'_map].
  destruct H_valid_hyper with osi pt_padd p2m' as [[pt_addr'' H_pt_addr''_map H_pt_addr'_owner] | H_error]; auto.
  rewrite <- H_pt_addr'_map in H_pt_addr''_map.
  injection H_pt_addr''_map; intro; subst pt_addr''; clear H_pt_addr''_map.
  unfold maddress_owner in H_pt_addr'_owner.
  rewrite <- H_pt_addr_eq_pt_addr' in H_pt_addr'_owner.
  rewrite <- H_pt_page_map in H_pt_addr'_owner.
  clear H_valid_hyper.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap H_valid_currpg
      H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  destruct H_valid_hyper with (st_active_os s) (os_current_page aos) p2m as [[pt_addr'' H_pt_addr''_map H_pt_addr_owner] | H_error]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr''_map.
  injection H_pt_addr''_map; intro; subst pt_addr''; clear H_pt_addr''_map.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner.
  rewrite H_pt_addr'_owner in H_pt_addr_owner.
  injection H_pt_addr_owner; intro H_osi_active.
  apply H_osi_not_active; auto.
  rewrite <- H_pt_addr_map in H_error.
  discriminate H_error.
  rewrite <- H_pt_addr'_map in H_error.
  discriminate H_error.

  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; auto.
  split; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma'.
  unfold page_type in H_pg'_type.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_type.
  rewrite <- H_pt_page_map in H_pg'_type.
  rewrite <- H_pt_page_content in H_pg'_type.
  discriminate H_pg'_type.

  (* s' eq_pt s *)
  intros pt_padd pt_page' mp [pt_addr' H_pt_addr'_map H_pt_page'_map] H_pt_page'_content.
  case (os_ident_eq (st_active_os s) osi).
  (* osi is AOS *)
  intro; subst osi.
  case (padd_eq pt_padd (os_current_page aos)).
  intro H_pt_padd_current.
  rewrite H_pt_padd_current in H_pt_addr'_map.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m' H_p2m'_map H_pt_addr'_map].
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite lemma_map_add_apply in H_pt_page'_map.
  exists pt_page.
  exists pt_addr; auto.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  rewrite H_pt_padd_current; auto.
  exists oldmap; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  assert (va<>va') as H_va_ne_va'.
  intro H_va_eq_va'.
  rewrite <- H_va_eq_va' in H_va'_accessible.
  apply H_va_not_accessible; auto.
  exists ma'.
  injection H_pt_page'_map; intro H_pt_page'_value.
  rewrite H_pt_page'_value in H_pt_page'_content.
  simpl in H_pt_page'_content.
  injection H_pt_page'_content; intro H_mp_value.
  rewrite H_mp_value in H_ma'_map.
  rewrite map_drop_apply_distinct in H_ma'_map; auto.
  split; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  intro H_pt_addr_eq_ma'.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
  (* pt_padd is not current page, so it does not change *)
  intro H_pt_padd_not_current.
  exists pt_page'.
  exists pt_addr'; auto.
  rewrite lemma_map_add_apply_distinct in H_pt_page'_map; auto.
  intro H_pt_addr_eq_pt_addr'.
  apply H_pt_padd_not_current.
  apply hyp_mapped_same_paddr with ctxt s (st_active_os s) pt_addr; auto.
  rewrite H_pt_addr_eq_pt_addr'; assumption.
  unfold hypervisor_mapping_madd; exists p2m; auto.
  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  assert (va<>va') as H_va_ne_va'.
  intro H_va_eq_va'.
  rewrite <- H_va_eq_va' in H_va'_accessible.
  apply H_va_not_accessible; auto.
  exists ma'; auto.
  split; auto.
  unfold page_type in H_pg'_type.
  assert (pt_addr<>ma') as H_pt_addr_ne_ma'.
  intro H_pt_addr_eq_ma'.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_type; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.

  (* osi not active *)
  intro H_osi_not_active.
  exists pt_page'.
  exists pt_addr'; auto.
  rewrite lemma_map_add_apply_distinct in H_pt_page'_map; auto.
  intro H_pt_addr_eq_pt_addr'.

  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap H_valid_currpg
      H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m' H_p2m'_map H_pt_addr'_map].
  destruct H_valid_hyper with osi pt_padd p2m' as [[pt_addr'' H_pt_addr''_map H_pt_addr'_owner] | H_error]; auto.
  rewrite <- H_pt_addr'_map in H_pt_addr''_map.
  injection H_pt_addr''_map; intro; subst pt_addr''; clear H_pt_addr''_map.
  unfold maddress_owner in H_pt_addr'_owner.
  rewrite <- H_pt_addr_eq_pt_addr' in H_pt_addr'_owner.
  rewrite <- H_pt_page_map in H_pt_addr'_owner.
  clear H_valid_hyper.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  destruct H_valid_hyper with (st_active_os s) (os_current_page aos) p2m as
    [[pt_addr'' H_pt_addr''_map H_pt_addr_owner] | H_error]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr''_map.
  injection H_pt_addr''_map; intro; subst pt_addr''; clear H_pt_addr''_map.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner.
  rewrite H_pt_addr'_owner in H_pt_addr_owner.
  injection H_pt_addr_owner; intro H_osi_active.
  apply H_osi_not_active; auto.
  rewrite <- H_pt_addr_map in H_error.
  discriminate H_error.
  rewrite <- H_pt_addr'_map in H_error.
  discriminate H_error.

  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; auto.
  split; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  intro H_pt_addr_eq_ma'.
  rewrite H_pt_addr_eq_ma' in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
Qed.

Lemma switch_osi_equiv : forall (s1 s1' s2 s2':state) (osi o:os_ident),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  valid_exec ctxt s1 (Switch o) s1' -> valid_exec ctxt s2 (Switch o) s2' ->
  osi_equivalent ctxt s1 s2 osi -> osi <> (st_active_os s1) ->
  osi_equivalent ctxt s1' s2' osi.
Proof.
  intros s1 s1' s2 s2' osi o H_s1_valid H_s2_valid H_valid_exec1 H_valid_exec2 H_s1_eq_s2 H_osi_not_active.
  unfold osi_equivalent in *.
  inversion_clear H_valid_exec1 as (s_vexec1, s'_vexec1, act_vexec1, H_pre1, H_post1).
  inversion_clear H_valid_exec2 as (s_vexec2, s'_vexec2, act_vexec2, H_pre2, H_post2).

  unfold osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page in *.
  unfold post_action, post_switch,
    state_component_assign, state_component_update in H_post1, H_post2.

  rewrite H_post1, H_post2; simpl.
  destruct H_s1_eq_s2 as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].

  repeat split; auto.

  case (os_ident_eq o osi).
  intro H_o_eq_osi.
  subst o.

  destruct H_s1_active_s2 as [[H_s1_active [H_s2_active H_s1_activity_s2]] | [H_s1_not_active H_s2_not_active]].
  cut False; try contradiction.
  apply H_osi_not_active; auto.
  left; repeat split; auto.
  unfold pre_action, pre_switch in H_pre1, H_pre2.
  destruct H_pre1 as [H_hyper_running1 H_pre1].
  destruct H_pre2 as [H_hyper_running2 H_pre2].
  rewrite H_hyper_running1; auto.
  
  intro H_o_ne_osi.
  right; split; auto.
Qed.

Lemma lswitch_untrusted_osi_equiv : forall (s1 s1' s2 s2':state) (pa:padd) (osi o:os_ident),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  valid_exec ctxt s1 (Lswitch_untrusted o pa) s1' ->
  valid_exec ctxt s2 (Lswitch_untrusted o pa) s2' ->
  osi_equivalent ctxt s1 s2 osi ->
  osi_equivalent ctxt s1' s2' osi.
Proof.
  intros s1 s1' s2 s2' pa osi o H_s1_valid H_s2_valid H_valid_exec1 H_valid_exec2 H_s1_eq_s2.
  unfold osi_equivalent in *.
  inversion_clear H_valid_exec1 as (s_vexec1, s'_vexec1, act_vexec1, H_pre1, H_post1).
  inversion_clear H_valid_exec2 as (s_vexec2, s'_vexec2, act_vexec2, H_pre2, H_post2).

  unfold osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page in *.
  unfold post_action, post_lswitch_untrusted, post_all_lswitch in H_post1, H_post2.
  destruct H_post1 as [aos1 H_aos1_map H_s1'].
  destruct H_post2 as [aos2 H_aos2_map H_s2'].

  rewrite H_s1', H_s2'; simpl.
  destruct H_s1_eq_s2 as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].

  repeat split; auto.

  case (os_ident_eq o osi).
  intros H_o_eq_osi.
  subst o.
  repeat rewrite lemma_map_add_apply.
  reflexivity.
  intros H_o_ne_osi.
  repeat rewrite lemma_map_add_apply_distinct; auto.
Qed.  

Lemma chmod_osi_equiv : forall (s s':state) (osi:os_ident),
  valid_state ctxt s -> valid_exec ctxt s Chmod s' ->
  (osi <> st_active_os s) -> osi_equivalent ctxt s s' osi.
Proof.
  intros s s' osi H_s_valid H_valid_exec H_osi_not_active.
  unfold osi_equivalent, osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page.
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold post_action, post_chmod in H_post.
  
  assert (forall cmp' : state_component,
              cmp' <> Exec_Mode /\ cmp' <> OS_Activity ->
              state_component_interp s cmp' = state_component_interp s' cmp') as
    H_interp.

  destruct H_post as [[H_trusted H_assign] | [H_not_trusted H_assign]].

  intros cmp H_cmp_not_activity.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with cmp.
  intros [H_cmp H_interp]; auto.
  intros [H_abs H_interp]; contradiction H_abs.

  intros cmp H_cmp_not_activity.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with cmp.
  intros [H_cmp H_interp]; auto.
  intros [H_abs H_interp]; contradiction H_abs.
  unfold state_component_interp in H_interp.

  replace (st_active_os s') with (st_active_os s).
  replace (st_oss s') with (st_oss s).
  replace (st_hypervisor s') with (st_hypervisor s).
  replace (st_memory s') with (st_memory s).

  repeat split; auto.

  intros pt_padd pt_page mp [pt_addr H_pt_addr_map H_pt_page_map] H_pt_page_content.
  exists pt_page.
  exists pt_addr; auto.
  exists mp; auto.
  intros va pa ma1 pg H_ma1_map H_pg_map H_va_accessible H_ma1_type H_ma1_hyp_map.
  exists ma1; auto.
  intros pt_padd pt_page mp [pt_addr H_pt_addr_map H_pt_page_map] H_pt_page_content.
  exists pt_page.
  exists pt_addr; auto.
  exists mp; auto.
  intros va pa ma1 pg H_ma1_map H_pg_map H_va_accessible H_ma1_type H_ma1_hyp_map.
  exists ma1; auto.

  (* Same Memory *)
  elim H_interp with System_Memory; auto.
  split; intro; discriminate.

  (* Same Hypervisor *)
  elim H_interp with Hypervisor_SC; auto.
  split; intro; discriminate.

  (* Same Guest OSs *)
  elim H_interp with Guest_OSs; auto.
  split; intro; discriminate.
  
  (* Same Active OS *)
  elim H_interp with Active_OS; auto.
  split; intro; discriminate.
Qed.

Lemma other_os_running_same_hyp_mapping : forall (s s':state) (a:action) (osi:os_ident),
  valid_state ctxt s -> valid_exec ctxt s a s' ->
   ~(hyper_running s) -> osi <> (st_active_os s) ->
   map_apply os_ident_eq (st_hypervisor s) osi =
   map_apply os_ident_eq (st_hypervisor s') osi.
Proof.
  intros s s' a osi H_s_valid H_valid_exec H_not_hyper_running H_osi_not_active.

  induction a; inversion_clear H_valid_exec; unfold post_action in H0.

  destruct H0 as [new_cache [new_tlb [H_cache'_valid [H_tlb'_valid
    H_assign]]]]; auto.
  unfold content_of_address.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s cmp = state_component_interp s' cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_OneComp H_interp H_hyp System_Cache Hypervisor_SC.
  rewrite <- H_hyp; try (intro; discriminate); auto.

  unfold post_read, post_all_read in H0.
  unfold pre_action, pre_read, pre_all_read in H.
  destruct H as [[ma [H_ma_map H_ptype]] [H_v_accessible H_os_running]].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v0; auto.

  destruct H0 with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with Hypervisor_SC;
    try (intros [H_abs H_aos1]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_aos1];
    rewrite H_aos1; auto.

  unfold post_read_hyper, post_all_read in H0.
  unfold pre_action, pre_read_hyper, pre_all_read in H.
  destruct H as [[ma [H_ma_map H_ptype]] H_hyper_running].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v0; auto.

  destruct H0 with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with Hypervisor_SC;
    try (intros [H_abs H_aos]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_aos];
    rewrite H_aos; auto.

  unfold post_write, post_all_write in H0.
  elim H0.
  intros ma H_va_mapped [newmem H_update H_s'].
  unfold update_rw_page in H_update.
  destruct H_update as [pg H_pg_map H_eq_newmem].
  destruct H_s' as [pg' H_pg'_map_cache
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]].
  elim valid_state_included_cache_mem with ctxt s v pg'; auto.
  intros ma' H_ma'_map H_pg'_map_mem.
  assert (va_mapped_to_ma s (st_active_os s) v ma) as H_ma_map.
  apply valid_state_included_tlb_mem with ctxt; auto.
  assert (ma = ma').
  elim va_mapped_same_ma with s v ma ma'; auto.
  subst ma'; clear H_ma'_map.
  rewrite <- H_pg_map in H_pg'_map_mem.
  injection H_pg'_map_mem; intro; subst pg'; clear H_pg'_map_mem.
  rewrite H_s'; auto.

  unfold post_write_hyper, post_all_write in H0.
  elim H0.
  intros ma H_va_mapped [newmem H_update H_s'].
  unfold update_rw_page in H_update.
  destruct H_update as [pg H_pg_map H_eq_newmem].
  destruct H_s' as [pg' H_pg'_map_cache
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]].
  elim valid_state_included_cache_mem with ctxt s v pg'; auto.
  intros ma' H_ma'_map H_pg'_map_mem.
  assert (va_mapped_to_ma s (st_active_os s) v ma) as H_ma_map.
  apply valid_state_included_tlb_mem with ctxt; auto.
  assert (ma = ma').
  elim va_mapped_same_ma with s v ma ma'; auto.
  subst ma'; clear H_ma'_map.
  rewrite <- H_pg_map in H_pg'_map_mem.
  injection H_pg'_map_mem; intro; subst pg'; clear H_pg'_map_mem.
  rewrite H_s'; auto.

  unfold pre_action, pre_new_untrusted, hyper_running in H.
  destruct H as (H_os_untrusted, (ma', H_hyp_map, (H_ptype, (H_ma'_owner, (H_os_acc, (H_hcall, H_activity)))))).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  destruct H0 as [ma1 H_ma1_hyp_map [pt_addr1
    H_pt_addr1_curr_map [newmem1 H_update1 [system1 H_system1_map H_s']]]].
  rewrite H_s'; auto.

  unfold pre_action, pre_new_hyper, hyper_running in H.
  destruct H as (H_waiting, H').
  unfold hyper_running in H_not_hyper_running.
  rewrite H_waiting in H_not_hyper_running; elim H_not_hyper_running; auto.

  unfold pre_action, pre_del_untrusted, hyper_running in H.
  destruct H as (H_os_untrtusted, (H_os_hcall, (system, H_system_map, (p2m, H_hyp_mapping,
    (pt_addr, H_pt_addr_map, (H_pre_all_del, (H_activity, H_accessible))))))).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  destruct H0 as [pt_addr H_curr_map [newmem H_drop H_s']].
  rewrite H_s'; auto.

  unfold pre_action, pre_del_hyper, pre_all_del, hyper_running in H.
  destruct H as (pt_addr, H_curr_map, (H_pre', (H_activity, H_not_accessible))). 
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  unfold pre_action, pre_switch, hyper_running in H.
  destruct H as (H_activity, H_pre').
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.
 
  destruct H0 as [os1 H_os1_map H_s'].
  rewrite H_s'; auto.

  unfold pre_action, pre_lswitch_untrusted, hyper_running in H.
  destruct H as [H_o0_untrusted [aos H_aos_map [H_hcall [H_act H_all_lswitch]]]].
  unfold hyper_running in H_not_hyper_running.
  rewrite H_act in H_not_hyper_running; elim H_not_hyper_running; auto.

  destruct H0 as [aos1 H_aos1_map H_s1'].
  rewrite H_s1'; reflexivity.

  destruct H0 with Hypervisor_SC as [[H_trusted_aos1 H_state1_assign] | [H_untrusted_aos1 H_state1_assign]].
  unfold state_component_interp in H_state1_assign.
  rewrite H_state1_assign; reflexivity.
  contradiction H_untrusted_aos1.
  split; intro; discriminate.

  unfold pre_action, pre_chmod, hyper_running in H.
  destruct H as (H_activity, H_hcall).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  destruct H0 as [ma1 [pg1 H_pg1_map [H_pg1_owner
    [H_pg1_content [p2m1 H_post_hyp1 [os1 H_os1_map H_s1']]]]]].
  rewrite H_s1'; simpl; rewrite lemma_map_add_apply_distinct; auto.
  destruct H0 as [p2m1 H_p2m1_map [ma1 H_ma1_hyp_map [system1 H_system1_map H_s1']]].
  rewrite H_s1'; simpl; rewrite lemma_map_add_apply_distinct; auto.

  unfold pre_action, pre_page_pin_untrusted, pre_all_pin, hyper_running in H.
  destruct H as (H_ctxt, (H_hcall, (H_activity, H_pre'))).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  unfold pre_action, pre_page_unpin_untrusted, pre_all_unpin, hyper_running in H.
  destruct H as (H_ctxt, (H_hcall, (H_activity, H_pre'))).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.
Qed.

Lemma other_os_running_same_mem : forall (s s':state) (a:action) (osi:os_ident),
  valid_state ctxt s -> valid_exec ctxt s a s' ->
   ~(hyper_running s) -> osi <> (st_active_os s) ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
      owner_of_page pg osi ->
      Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof.
  intros s s' a osi H_s_valid H_valid_exec H_not_hyper_running H_osi_not_active
    ma pg H_pg_map H_pg_owner.

  generalize (write_isolation_proof ctxt s s').
  intros H_write_isolation.
  unfold write_isolation, content_of_address in H_write_isolation.
  apply H_write_isolation with a osi; auto.
Qed.

Lemma other_os_running_hyper_equiv : forall (s s':state) (a:action) (osi:os_ident),
  valid_state ctxt s -> valid_exec ctxt s a s' ->
   ~(hyper_running s) -> osi <> (st_active_os s) ->
   osi_hyper_mapping_eq s s' osi /\ osi_hyper_mapping_eq s' s osi.
Proof.
  intros s s' a osi H_s_valid H_valid_exec H_not_hyper_running H_osi_not_active.

  generalize H_valid_exec; intro H_valid_exec'.
  inversion_clear H_valid_exec' as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold osi_equivalent_hyper_mapping, osi_hyper_mapping_eq, padd_page.
  split.
  
  intros pa pg [ma H_ma_map H_pg_map] [val H_pg_content].
  unfold hypervisor_mapping_madd.
  destruct H_ma_map as [p2m H_p2m_map H_ma_map].

  exists ma.
  exists p2m; auto.
  rewrite H_p2m_map.
  apply other_os_running_same_hyp_mapping with a; auto.
  
  apply other_os_running_same_mem with s a osi; auto.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  destruct H_valid_hyper with osi pa p2m as [[ma' H_ma'_map H_ma_owner] | H_error]; auto.
  unfold owner_of_page.
  assert (ma'=ma).
  apply hyp_mapped_same_maddr with s osi pa; auto.
  unfold hypervisor_mapping_madd; exists p2m; auto.
  unfold hypervisor_mapping_madd; exists p2m; auto.

  subst ma'; clear H_ma'_map.
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_pg_map in H_ma_owner; auto.

  rewrite <- H_ma_map in H_error; discriminate H_error.

  (* s' eq s *)
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  intros pa pg [ma H_ma_map H_pg_map] [val H_pg_content].
  destruct H_ma_map as [p2m H_p2m_map H_ma_map].
  rewrite <- other_os_running_same_hyp_mapping with s s' a osi in H_p2m_map; auto.

  destruct H_valid_hyper with osi pa p2m as [[ma' H_ma'_map H_ma_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  assert (ma'=ma).
  apply hyp_mapped_same_maddr with s osi pa; auto.
  unfold hypervisor_mapping_madd; exists p2m; auto.
  unfold hypervisor_mapping_madd; exists p2m; auto.
  subst ma'; clear H_ma'_map.

  exists ma; auto.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.

  unfold maddress_owner in H_ma_owner.
  assert (exists pg1, Value madd pg1 = map_apply madd_eq (st_memory s) ma).
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v; auto.
  destruct H as [pg1 H_pg1_map].
  assert (Value madd pg1 = map_apply madd_eq (st_memory s') ma).
  apply other_os_running_same_mem with s a osi; auto.
  unfold owner_of_page.
  rewrite <- H_pg1_map in H_ma_owner; auto.
  rewrite <- H_pg_map in H.
  rewrite <- H_pg1_map.
  auto.
Qed.

Lemma other_os_running_pt_equiv : forall (s s':state) (a:action) (osi:os_ident),
  valid_state ctxt s -> valid_exec ctxt s a s' ->
   ~(hyper_running s) -> osi <> (st_active_os s) ->
   osi_pt_pages_eq ctxt s s' osi /\ osi_pt_pages_eq ctxt s' s osi.
Proof.
  intros s s' a osi H_s_valid H_valid_exec H_not_hyper_running H_osi_not_active.

  generalize H_valid_exec; intro H_valid_exec'.
  inversion_clear H_valid_exec' as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold osi_pt_pages_eq, padd_page.
  split.
  
  intros pt_padd pt_page mp1 [pt_addr H_pt_addr_map H_pt_page_map] H_pt_page_content.
  destruct H_pt_addr_map as [p2m H_p2m_map H_pt_addr_map].
  exists pt_page.
  exists pt_addr.
  unfold hypervisor_mapping_madd in H_pt_addr_map.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  rewrite <- other_os_running_same_hyp_mapping with s s' a osi; auto.
  
  apply other_os_running_same_mem with s a osi; auto.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  destruct H_valid_hyper with osi pt_padd p2m as [[ma' H_ma'_map H_ma_owner] | H_error]; auto.
  unfold owner_of_page.
  assert (ma'=pt_addr).
  apply hyp_mapped_same_maddr with s osi pt_padd; auto.
  unfold hypervisor_mapping_madd; exists p2m; auto.
  unfold hypervisor_mapping_madd; exists p2m; auto.

  subst ma'; clear H_ma'_map.
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_pt_page_map in H_ma_owner; auto.

  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
  exists mp1; auto.
  intros va pa ma1 pg H_ma1_map H_pg_map H_va_accessible H_pt_type H_ma1_hyp_map.
  exists ma1; auto.
  split.
  unfold hypervisor_mapping_madd in *.
  destruct H_ma1_hyp_map as [p2m' H_p2m'_map H_ma1_hyp_map].
  rewrite <- H_p2m_map in H_p2m'_map; injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  exists p2m; auto.
  rewrite <- other_os_running_same_hyp_mapping with s s' a osi; auto.
  rewrite <- other_os_running_same_mem with s s' a osi ma1 pg; auto.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  destruct H_valid_hyper with osi pa p2m as [[ma' H_ma'_map H_ma_owner] | H_error];
    try (rewrite <- H_ma1_map in H_error; discriminate H_error); auto.
  assert (ma'=ma1).
  apply hyp_mapped_same_maddr with s osi pa; auto.
  unfold hypervisor_mapping_madd; exists p2m; auto.
  subst ma'; clear H_ma'_map.
  unfold maddress_owner in H_ma_owner.
  unfold owner_of_page.
  rewrite <- H_pg_map in H_ma_owner; auto.
  unfold hypervisor_mapping_madd in H_ma1_hyp_map.
  destruct H_ma1_hyp_map as [p2m' H_p2m'_map H_ma1_hyp_map].
  rewrite <- H_p2m_map in H_p2m'_map; injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_ma1_hyp_map in H_error; discriminate H_error.

  (* s' eq s *)
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  intros pt_padd pt_page mp1 [pt_addr H_pt_addr_map H_pt_page_map] H_pt_page_content.
  destruct H_pt_addr_map as [p2m H_p2m_map H_pt_addr_map].
  exists pt_page.
  exists pt_addr.
  unfold hypervisor_mapping_madd in *.
  exists p2m; auto.
  rewrite other_os_running_same_hyp_mapping with s s' a osi; auto.
  destruct H_valid_hyper with osi pt_padd p2m as [[ma' H_ma'_map H_ma_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  rewrite other_os_running_same_hyp_mapping with s s' a osi; auto.
  assert (ma'=pt_addr).
  apply hyp_mapped_same_maddr with s osi pt_padd; auto.
  unfold hypervisor_mapping_madd; exists p2m; auto.
  rewrite other_os_running_same_hyp_mapping with s s' a osi; auto.
  unfold hypervisor_mapping_madd; exists p2m; auto.
  rewrite other_os_running_same_hyp_mapping with s s' a osi; auto.
  subst ma'; clear H_ma'_map.

  unfold maddress_owner in H_ma_owner.
  assert (exists pg1, Value madd pg1 = map_apply madd_eq (st_memory s) pt_addr).
  induction (map_apply madd_eq (st_memory s) pt_addr); try discriminate.
  exists v; auto.
  destruct H as [pg1 H_pg1_map].
  assert (Value madd pg1 = map_apply madd_eq (st_memory s') pt_addr).
  apply other_os_running_same_mem with s a osi; auto.
  unfold owner_of_page.
  rewrite <- H_pg1_map in H_ma_owner; auto.
  rewrite <- H_pt_page_map in H.
  rewrite <- H_pg1_map.
  auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error; auto.
  
  exists mp1; auto.
  intros va pa ma1 pg H_ma1_map H_pg_map H_va_accessible H_pt_type H_ma1_hyp_map.
  destruct H_ma1_hyp_map as [p2m' H_p2m'_map H_ma1_hyp_map].
  rewrite <- H_p2m_map in H_p2m'_map; injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  exists ma1; auto.
  split.
  unfold hypervisor_mapping_madd in *.
  exists p2m; auto.
  rewrite other_os_running_same_hyp_mapping with s s' a osi; auto.
  
  destruct H_valid_hyper with osi pa p2m as [[ma' H_ma'_map H_ma_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  rewrite other_os_running_same_hyp_mapping with s s' a osi; auto.
  assert (ma'=ma1).
  apply hyp_mapped_same_maddr with s osi pa; auto.
  unfold hypervisor_mapping_madd; exists p2m; auto.
  rewrite other_os_running_same_hyp_mapping with s s' a osi; auto.
  unfold hypervisor_mapping_madd; exists p2m; auto.
  rewrite other_os_running_same_hyp_mapping with s s' a osi; auto.
  unfold hypervisor_mapping_madd in H_ma1_hyp_map.
  subst ma'; clear H_ma'_map.

  unfold maddress_owner in H_ma_owner.
  assert (exists pg1, Value madd pg1 = map_apply madd_eq (st_memory s) ma1).
  induction (map_apply madd_eq (st_memory s) ma1); try discriminate.
  exists v; auto.
  destruct H as [pg1 H_pg1_map].
  assert (Value madd pg1 = map_apply madd_eq (st_memory s') ma1).
  apply other_os_running_same_mem with s a osi; auto.
  unfold owner_of_page.
  rewrite <- H_pg1_map in H_ma_owner; auto.
  rewrite <- H_pg_map in H.
  rewrite <- H_pg1_map.
  auto.
  rewrite <- H_ma1_hyp_map in H_error; discriminate H_error; auto.
Qed.

Lemma other_os_running_equiv : forall (s s':state) (a:action) (osi:os_ident),
  valid_state ctxt s -> valid_exec ctxt s a s' ->
   ~(hyper_running s) -> osi <> (st_active_os s) ->
   osi_equivalent ctxt s s' osi.
Proof.
  intros s s' a osi H_s_valid H_valid_exec H_not_hyper_running H_osi_not_active.
  unfold osi_equivalent.

  elim other_os_running_hyper_equiv with s s' a osi; auto.
  intros H_eq_hyp1 H_eq_hyp2.
  elim other_os_running_pt_equiv with s s' a osi; auto.
  intros H_eq_pt1 H_eq_pt2.
  repeat split; auto.
  right; split; auto.

  assert ((forall o, a <> Switch o) \/ (exists o, a = Switch o)).
  induction a; try (left; intros o0 H_abs; discriminate).
  right.
  exists o; auto.
  destruct H as [H_switch | H_no_switch].

  rewrite <- no_switch_same_aos with s s' a; auto.
  destruct H_no_switch as [o H_switch]; subst a.
  inversion_clear H_valid_exec.
  unfold pre_action, pre_switch in H.
  destruct H as [H_abs H'].
  contradiction H_abs.

  apply os_data_isolation with a; auto.
Qed.

Inductive osi_action (osi:os_ident) (s:state) : action -> Prop  :=
| osi_active : forall a:action, (st_active_os s = osi) -> st_aos_activity s = OSA_running ->
    osi_action osi s a
| new_untrusted_osi  : forall va pa, osi_action osi s (New_untrusted osi va pa)
| del_untrusted_osi  : forall va, osi_action osi s (Del_untrusted osi va)
| lswitch_untrusted_osi  : forall pa, osi_action osi s (Lswitch_untrusted osi pa)
| pin_untrusted_osi  : forall pa ptype, osi_action osi s (Page_pin_untrusted osi pa ptype)
| unpin_untrusted_osi  : forall pa, osi_action osi s (Page_unpin_untrusted osi pa)
| chmod_osi : (st_active_os s = osi) -> osi_action osi s Chmod
| switch_from_osi : forall o, (st_active_os s = osi) -> osi_action osi s (Switch o)
| switch_to_osi : osi_action osi s (Switch osi).
  

Ltac not_osi_action H_not_osi_action :=
  apply H_not_osi_action; intro H_abs;
    destruct H_abs as [[H_osi_active' H_abs] | [H_triv H_abs]];
      [try discriminate H_abs | destruct H_abs as [H_new | [H_del | [H_lswitch | [H_pin | H_unpin]]]];
        [destruct H_new as [va [pa H_abs]]; discriminate H_abs |
          destruct H_del as [va H_abs]; discriminate H_abs |
            destruct H_lswitch as [pa H_abs]; discriminate H_abs |
              destruct H_pin as [pa [ptype H_abs]]; discriminate H_abs |
                destruct H_unpin as [pa H_abs]; discriminate H_abs]].

Lemma new_untrusted_not_osi_eq : forall (s s':state) (osi o:os_ident) (va:vadd) (pa:padd),
  valid_state ctxt s -> valid_exec ctxt s (New_untrusted o va pa) s' ->
  o <> osi -> osi_equivalent ctxt s s' osi.
Proof.
  intros s s' osi o va pa H_s_valid H_valid_exec H_o_ne_osi.
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold post_action, post_new_untrusted, post_all_new in H_post.
  unfold osi_equivalent.
  destruct H_post as [ma H_ma_map [pt_addr H_pt_addr_map
    [newmem H_update [system H_system_map H_s']]]].

  unfold osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page.
  rewrite H_s'; simpl.

  unfold hypervisor_mapping_madd in H_ma_map.
  destruct H_ma_map as [p2m H_p2m_map H_ma_map].
  unfold os_current_vadd_mapping_madd in H_pt_addr_map.
  destruct H_pt_addr_map as [system' H_system'_map [p2m' H_p2m'_map H_pt_addr_map]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  unfold hypervisor_mapping in H_p2m'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  unfold update_pt_page in H_update.
  destruct H_update as [pt_page H_pt_page_map [oldmap H_pt_page_content H_update]].

  repeat split.
  case (os_ident_eq (st_active_os s) osi); [left|right]; split; auto.

  rewrite lemma_map_add_apply_distinct; auto.

  (* s eq_hyp s' *)
  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  exists ma'; auto.
  rewrite H_update.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma'.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_map.
  rewrite <- H_pt_page_map in H_pg'_map.
  injection H_pg'_map; intro H_pg'_eq_pt_page.
  rewrite H_pg'_eq_pt_page in H_pg'_content.
  rewrite <- H_pt_page_content in H_pg'_content.
  discriminate H_pg'_content.
  
  (* s' eq_hyp s *)
  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  exists ma'; auto.
  rewrite H_update in H_pg'_map.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  intro H_pt_addr_eq_ma'.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_map.
  rewrite lemma_map_add_apply in H_pg'_map.
  injection H_pg'_map; intro H_pg'_value.
  rewrite H_pg'_value in H_pg'_content.
  unfold page_content in H_pg'_content.
  discriminate H_pg'_content.

  (* s eq_pt s' *)
  intros pt_padd pt_page' mp [pt_addr' H_pt_addr'_map H_pt_page'_map] H_pt_page'_content.
  exists pt_page'.
  exists pt_addr'; auto.
  rewrite H_update.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_pt_addr'.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m_osi H_p2m_osi_map H_pt_addr'_map].
  destruct H_valid_hyper with osi pt_padd p2m_osi as [[ma' H_ma'_map H_pt_addr'_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  rewrite <- H_pt_addr'_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  destruct H_valid_hyper with o (os_current_page system) p2m as [[ma' H_ma'_map H_pt_addr_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  rewrite <- H_pt_addr_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  rewrite H_pt_addr_eq_pt_addr' in H_pt_addr_owner.
  rewrite H_pt_addr_owner in H_pt_addr'_owner.
  injection H_pt_addr'_owner.
  intro H_o_eq_osi; apply H_o_ne_osi; auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
  rewrite <- H_pt_addr'_map in H_error; discriminate H_error.

  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; auto.
  split; auto.
  rewrite H_update.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma'.
  unfold page_type in H_pg'_type.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_type.
  rewrite <- H_pt_page_map in H_pg'_type.
  rewrite <- H_pt_page_content in H_pg'_type.
  discriminate H_pg'_type.
  
  (* s' eq_pt s *)
  intros pt_padd pt_page' mp [pt_addr' H_pt_addr'_map H_pt_page'_map] H_pt_page'_content.
  exists pt_page'.
  exists pt_addr'; auto.
  rewrite H_update in H_pt_page'_map.
  rewrite lemma_map_add_apply_distinct in H_pt_page'_map; auto.
  intro H_pt_addr_eq_pt_addr'.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m_osi H_p2m_osi_map H_pt_addr'_map].
  destruct H_valid_hyper with osi pt_padd p2m_osi as [[ma' H_ma'_map H_pt_addr'_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  rewrite <- H_pt_addr'_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  destruct H_valid_hyper with o (os_current_page system) p2m as [[ma' H_ma'_map H_pt_addr_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  rewrite <- H_pt_addr_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  rewrite H_pt_addr_eq_pt_addr' in H_pt_addr_owner.
  rewrite H_pt_addr_owner in H_pt_addr'_owner.
  injection H_pt_addr'_owner.
  intro H_o_eq_osi; apply H_o_ne_osi; auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
  rewrite <- H_pt_addr'_map in H_error; discriminate H_error.

  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; auto.
  split; auto.
  rewrite H_update in H_pg'_map, H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  intro H_pt_addr_eq_ma'.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_map, H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_map, H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
Qed.

Lemma del_untrusted_not_osi_eq : forall (s s':state) (osi o:os_ident) (va:vadd),
  valid_state ctxt s -> valid_exec ctxt s (Del_untrusted o va) s' ->
  o <> osi -> osi_equivalent ctxt s s' osi.
Proof.
  intros s s' osi o va H_s_valid H_valid_exec H_o_ne_osi.
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold post_action, post_del_untrusted, post_all_del in H_post.
  unfold osi_equivalent.
  destruct H_post as [system H_system_map [p2m H_p2m_map
    [pt_addr H_pt_addr_map [newmem H_drop H_s']]]].
  unfold osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page.
  rewrite H_s'; simpl.

  unfold hypervisor_mapping in H_p2m_map.
  unfold drop_va_pt_page in H_drop.
  destruct H_drop as [pt_page H_pt_page_map [oldmap H_pt_page_content H_newmem]].
  subst newmem.

  repeat split.
  case (os_ident_eq (st_active_os s) osi); [left|right]; split; auto.
  
  rewrite lemma_map_add_apply_distinct; auto.

  (* s eq_hyp s' *)
  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  exists ma'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma'.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_map.
  rewrite <- H_pt_page_map in H_pg'_map.
  injection H_pg'_map; intro H_pg'_eq_pt_page.
  rewrite H_pg'_eq_pt_page in H_pg'_content.
  rewrite <- H_pt_page_content in H_pg'_content.
  discriminate H_pg'_content.
  
  (* s' eq_hyp s *)
  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  exists ma'; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  intro H_pt_addr_eq_ma'.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_map.
  rewrite lemma_map_add_apply in H_pg'_map.
  injection H_pg'_map; intro H_pg'_value.
  rewrite H_pg'_value in H_pg'_content.
  unfold page_content in H_pg'_content.
  discriminate H_pg'_content.

  (* s eq_pt s' *)
  intros pt_padd pt_page' mp [pt_addr' H_pt_addr'_map H_pt_page'_map] H_pt_page'_content.
  exists pt_page'.
  exists pt_addr'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_pt_addr'.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m_osi H_p2m_osi_map H_pt_addr'_map].
  destruct H_valid_hyper with osi pt_padd p2m_osi as [[ma' H_ma'_map H_pt_addr'_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  rewrite <- H_pt_addr'_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  destruct H_valid_hyper with o (os_current_page system) p2m as [[ma' H_ma'_map H_pt_addr_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  rewrite <- H_pt_addr_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  rewrite H_pt_addr_eq_pt_addr' in H_pt_addr_owner.
  rewrite H_pt_addr_owner in H_pt_addr'_owner.
  injection H_pt_addr'_owner.
  intro H_o_eq_osi; apply H_o_ne_osi; auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
  rewrite <- H_pt_addr'_map in H_error; discriminate H_error.

  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; auto.
  split; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma'.
  unfold page_type in H_pg'_type.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_type.
  rewrite <- H_pt_page_map in H_pg'_type.
  rewrite <- H_pt_page_content in H_pg'_type.
  discriminate H_pg'_type.
  
  (* s' eq_pt s *)
  intros pt_padd pt_page' mp [pt_addr' H_pt_addr'_map H_pt_page'_map] H_pt_page'_content.
  exists pt_page'.
  exists pt_addr'; auto.
  rewrite lemma_map_add_apply_distinct in H_pt_page'_map; auto.
  intro H_pt_addr_eq_pt_addr'.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m_osi H_p2m_osi_map H_pt_addr'_map].
  destruct H_valid_hyper with osi pt_padd p2m_osi as [[ma' H_ma'_map H_pt_addr'_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  rewrite <- H_pt_addr'_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  destruct H_valid_hyper with o (os_current_page system) p2m as [[ma' H_ma'_map H_pt_addr_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  rewrite <- H_pt_addr_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  rewrite H_pt_addr_eq_pt_addr' in H_pt_addr_owner.
  rewrite H_pt_addr_owner in H_pt_addr'_owner.
  injection H_pt_addr'_owner.
  intro H_o_eq_osi; apply H_o_ne_osi; auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
  rewrite <- H_pt_addr'_map in H_error; discriminate H_error.

  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; auto.
  split; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  intro H_pt_addr_eq_ma'.
  rewrite <- H_pt_addr_eq_ma' in H_pg'_map, H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_map, H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
Qed.

Lemma lswitch_untrusted_not_osi_eq : forall (s s':state) (osi o:os_ident) (pa:padd),
  valid_state ctxt s -> valid_exec ctxt s (Lswitch_untrusted o pa) s' ->
  o <> osi -> osi_equivalent ctxt s s' osi.
Proof.
  intros s s' osi o pa H_s_valid H_valid_exec H_o_ne_osi.
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold post_action, post_lswitch_untrusted, post_all_lswitch in H_post.
  destruct H_post as [system H_system_map H_s'].

  unfold osi_equivalent.
  unfold osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page.
  rewrite H_s'; simpl.

  repeat split.
  case (os_ident_eq (st_active_os s) osi); [left|right]; split; auto.

  rewrite lemma_map_add_apply_distinct; auto.

  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  exists ma'; auto.

  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  exists ma'; auto.

  intros pt_padd pt_page' mp [pt_addr' H_pt_addr'_map H_pt_page'_map] H_pt_page'_content.
  exists pt_page'.
  exists pt_addr'; auto.
  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; auto.
  
  intros pt_padd pt_page' mp [pt_addr' H_pt_addr'_map H_pt_page'_map] H_pt_page'_content.
  exists pt_page'.
  exists pt_addr'; auto.
  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; auto.
Qed.

Lemma pin_untrusted_not_osi_eq : forall (s s':state) (osi o:os_ident) (pa:padd) (t:ptype),
  valid_state ctxt s -> valid_exec ctxt s (Page_pin_untrusted o pa t) s' ->
  o <> osi -> osi_equivalent ctxt s s' osi.
Proof.
  intros s s' osi o pa t H_s_valid H_valid_exec H_o_ne_osi.
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold post_action, post_page_pin_untrusted, post_all_pin in H_post.
  unfold osi_equivalent.
  destruct H_post as [ma [pg H_pg_map [H_pg_owner [H_pg_content
    [p2m H_p2m_map [system H_system_map H_s']]]]]].

  unfold osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page.
  subst s'; simpl.

  unfold hypervisor_mapping in H_p2m_map.

  repeat split.
  case (os_ident_eq (st_active_os s) osi); [left|right]; split; auto.
  rewrite lemma_map_add_apply_distinct; auto.

  (* s eq_hyp s' *)
  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  exists ma'.
  unfold hypervisor_mapping_madd in *.
  destruct H_ma'_map as [p2m_osi H_p2m_osi H_ma'_map].
  exists p2m_osi; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_ma'.
  rewrite <- H_ma_eq_ma' in H_pg'_map.
  rewrite <- H_pg_map in H_pg'_map.
  injection H_pg'_map; intro H_pg'_eq_pg.
  rewrite H_pg'_eq_pg in H_pg'_content.
  rewrite H_pg_content in H_pg'_content.
  discriminate H_pg'_content.
  
  (* s' eq_hyp s *)
  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].

  assert (ma<>ma') as H_ma_ne_ma'.
  intro H_ma_eq_ma'.
  rewrite <- H_ma_eq_ma' in H_pg'_map.
  rewrite lemma_map_add_apply in H_pg'_map.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_ma'_map.
  destruct H_ma'_map as [p2m_osi H_p2m_osi_map H_ma'_map].
  rewrite lemma_map_add_apply_distinct in H_p2m_osi_map; auto.
  destruct H_valid_hyper with osi pa' p2m_osi as [[ma'' H_ma''_map H_ma'_owner] | H_error];
    try (rewrite <- H_ma'_map in H_error; discriminate H_error); auto.
  rewrite <- H_ma'_map in H_ma''_map.
  injection H_ma''_map; intro; subst ma''; clear H_ma''_map.
  unfold maddress_owner in H_ma'_owner.
  rewrite <- H_ma_eq_ma' in H_ma'_owner.
  rewrite <- H_pg_map in H_ma'_owner.
  rewrite H_pg_owner in H_ma'_owner.
  discriminate H_ma'_owner.

  exists ma'.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  unfold hypervisor_mapping_madd in *.
  destruct H_ma'_map as [p2m_osi H_p2m_osi_map H_ma'_map].
  rewrite lemma_map_add_apply_distinct in H_p2m_osi_map; auto.
  exists p2m_osi; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.

  (* s eq_pt s' *)
  intros pt_padd pt_page' mp [pt_addr' H_pt_addr'_map H_pt_page'_map] H_pt_page'_content.
  assert (ma<>pt_addr') as H_ma_ne_pt_addr'.
  intro H_ma_eq_pt_addr'.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m_osi H_p2m_osi_map H_pt_addr'_map].
  destruct H_valid_hyper with osi pt_padd p2m_osi as [[ma' H_ma'_map H_pt_addr'_owner] | H_error];
    try (rewrite <- H_pt_addr'_map in H_error; discriminate H_error); auto.
  rewrite <- H_pt_addr'_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  unfold maddress_owner in H_pt_addr'_owner.
  rewrite <- H_ma_eq_pt_addr' in H_pt_addr'_owner.
  rewrite <- H_pg_map in H_pt_addr'_owner.
  rewrite H_pg_owner in H_pt_addr'_owner.
  discriminate H_pt_addr'_owner.

  exists pt_page'.
  exists pt_addr'; auto.
  unfold hypervisor_mapping_madd in *.
  destruct H_pt_addr'_map as [p2m_osi H_p2m_osi_map H_pt_addr'_map].
  rewrite lemma_map_add_apply_distinct; auto.
  exists p2m_osi; auto.
  rewrite lemma_map_add_apply_distinct; auto.

  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; auto.
  split; auto.
  unfold hypervisor_mapping_madd in *.
  destruct H_ma'_hyp_map as [p2m_osi H_p2m_osi_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply_distinct; auto.
  exists p2m_osi; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_ma'.
  unfold page_type in H_pg'_type.
  rewrite <- H_ma_eq_ma' in H_pg'_type.
  rewrite <- H_pg_map in H_pg'_type.
  rewrite H_pg_content in H_pg'_type.
  discriminate H_pg'_type.
  
  (* s' eq_pt s *)
  intros pt_padd pt_page' mp [pt_addr' H_pt_addr'_map H_pt_page'_map] H_pt_page'_content.

  assert (ma<>pt_addr') as H_ma_ne_pt_addr'.
  intro H_ma_eq_pt_addr'.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m_osi H_p2m_osi_map H_pt_addr'_map].
  rewrite lemma_map_add_apply_distinct in H_p2m_osi_map; auto.
  destruct H_valid_hyper with osi pt_padd p2m_osi as [[ma' H_ma'_map H_pt_addr'_owner] | H_error];
    try (rewrite <- H_pt_addr'_map in H_error; discriminate H_error); auto.
  clear H_valid_hyper.
  rewrite <- H_pt_addr'_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  rewrite <- H_ma_eq_pt_addr' in H_pt_addr'_owner.
  unfold maddress_owner in H_pt_addr'_owner.
  rewrite <- H_pg_map in H_pt_addr'_owner.
  rewrite H_pg_owner in H_pt_addr'_owner.
  discriminate H_pt_addr'_owner.

  exists pt_page'.
  exists pt_addr'.
  unfold hypervisor_mapping_madd in *.
  destruct H_pt_addr'_map as [p2m_osi H_p2m_osi_map H_pt_addr'_map].
  rewrite lemma_map_add_apply_distinct in H_p2m_osi_map; auto.
  exists p2m_osi; auto.
  rewrite lemma_map_add_apply_distinct in H_pt_page'_map; auto.

  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; auto.
  split.
  unfold hypervisor_mapping_madd in *.
  destruct H_ma'_hyp_map as [p2m_osi H_p2m_osi_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply_distinct in H_p2m_osi_map; auto.
  exists p2m_osi; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  intro H_ma_eq_ma'.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m_osi H_p2m_osi_map H_pt_addr'_map].
  rewrite lemma_map_add_apply_distinct in H_p2m_osi_map; auto.
  destruct H_ma'_hyp_map as [p2m_osi' H_p2m_osi'_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply_distinct in H_p2m_osi'_map; auto.
  rewrite <- H_p2m_osi_map in H_p2m_osi'_map.
  injection H_p2m_osi'_map; intro; subst p2m_osi'; clear H_p2m_osi'_map.
  destruct H_valid_hyper with osi pa' p2m_osi as [[ma'' H_ma''_map H_ma'_owner] | H_error];
    try (rewrite <- H_ma'_hyp_map in H_error; discriminate H_error); auto.
  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  rewrite <- H_ma'_hyp_map in H_ma''_map.
  injection H_ma''_map; intro; subst ma''; clear H_ma''_map.
  unfold maddress_owner in H_ma'_owner.
  rewrite <- H_ma_eq_ma' in H_ma'_owner.
  rewrite <- H_pg_map in H_ma'_owner.
  rewrite H_pg_owner in H_ma'_owner.
  discriminate H_ma'_owner.
Qed.

Lemma unpin_untrusted_not_osi_eq : forall (s s':state) (osi o:os_ident) (pa:padd),
  valid_state ctxt s -> valid_exec ctxt s (Page_unpin_untrusted o pa) s' ->
  o <> osi -> osi_equivalent ctxt s s' osi.
Proof.
  intros s s' osi o pa H_s_valid H_valid_exec H_o_ne_osi.
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold post_action, post_page_unpin_untrusted, post_all_unpin in H_post.
  unfold osi_equivalent.
  destruct H_post as [p2m H_p2m_map [ma H_ma_map [system H_system_map H_s']]].

  unfold osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page.
  subst s'; simpl.

  unfold hypervisor_mapping in H_p2m_map.

  repeat split.
  case (os_ident_eq (st_active_os s) osi); [left|right]; split; auto.
  rewrite lemma_map_add_apply_distinct; auto.

  (* s eq_hyp s' *)
  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  unfold hypervisor_mapping_madd in H_ma'_map.
  destruct H_ma'_map as [p2m_osi H_p2m_osi_map H_ma'_map].
  exists ma'.
  unfold hypervisor_mapping_madd.
  exists p2m_osi; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_ma'.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  destruct H_valid_hyper with osi pa' p2m_osi as [[ma'' H_ma''_map H_ma'_owner] | H_error];
    try (rewrite <- H_ma'_map in H_error; discriminate H_error); auto.
  rewrite <- H_ma'_map in H_ma''_map.
  injection H_ma''_map; intro; subst ma''; clear H_ma''_map.
  unfold maddress_owner in H_ma'_owner.
  rewrite <- H_ma_eq_ma' in H_ma'_owner.
  destruct H_valid_hyper with o pa p2m as [[ma'' H_ma''_map H_ma_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  rewrite <- H_ma_map in H_ma''_map.
  injection H_ma''_map; intro; subst ma''; clear H_ma''_map.
  unfold maddress_owner in H_ma_owner.
  rewrite H_ma_owner in H_ma'_owner.
  injection H_ma'_owner; intro; contradiction H_o_ne_osi.
  
  (* s' eq_hyp s *)
  intros pa' pg' [ma' H_ma'_map H_pg'_map] [val' H_pg'_content].
  assert (ma<>ma') as H_ma_ne_ma'.
  intro H_ma_eq_ma'.
  rewrite <- H_ma_eq_ma' in H_pg'_map.
  rewrite lemma_map_add_apply in H_pg'_map.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_ma'_map.
  destruct H_ma'_map as [p2m_osi H_p2m_osi_map H_ma'_map].
  rewrite lemma_map_add_apply_distinct in H_p2m_osi_map; auto.
  destruct H_valid_hyper with osi pa' p2m_osi as [[ma'' H_ma''_map H_ma'_owner] | H_error];
    try (rewrite <- H_ma'_map in H_error; discriminate H_error); auto.
  rewrite <- H_ma'_map in H_ma''_map.
  injection H_ma''_map; intro; subst ma''; clear H_ma''_map.
  unfold maddress_owner in H_ma'_owner.
  rewrite <- H_ma_eq_ma' in H_ma'_owner.
  destruct H_valid_hyper with o pa p2m as [[ma'' H_ma''_map H_ma_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  rewrite <- H_ma_map in H_ma''_map.
  injection H_ma''_map; intro; subst ma''; clear H_ma''_map.
  unfold maddress_owner in H_ma_owner.
  rewrite H_ma_owner in H_ma'_owner.
  injection H_ma'_owner; intro; contradiction H_o_ne_osi.

  exists ma'.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  unfold hypervisor_mapping_madd in *.
  destruct H_ma'_map as [p2m_osi H_p2m_osi_map H_ma'_map].
  rewrite lemma_map_add_apply_distinct in H_p2m_osi_map; auto.
  exists p2m_osi; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.

  (* s eq_pt s' *)
  intros pt_padd pt_page' mp [pt_addr' H_pt_addr'_map H_pt_page'_map] H_pt_page'_content.
  assert (ma<>pt_addr') as H_ma_ne_pt_addr'.
  intro H_ma_eq_pt_addr'.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m_osi H_p2m_osi_map H_pt_addr'_map].
  destruct H_valid_hyper with osi pt_padd p2m_osi as [[ma' H_ma'_map H_pt_addr'_owner] | H_error];
    try (rewrite <- H_pt_addr'_map in H_error; discriminate H_error); auto.
  rewrite <- H_pt_addr'_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  unfold maddress_owner in H_pt_addr'_owner.
  rewrite <- H_ma_eq_pt_addr' in H_pt_addr'_owner.
  destruct H_valid_hyper with o pa p2m as [[ma'' H_ma''_map H_ma_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  rewrite <- H_ma_map in H_ma''_map.
  injection H_ma''_map; intro; subst ma''; clear H_ma''_map.
  unfold maddress_owner in H_ma_owner.
  rewrite H_ma_owner in H_pt_addr'_owner.
  injection H_pt_addr'_owner; intro; contradiction H_o_ne_osi.
  
  exists pt_page'.
  exists pt_addr'.
  unfold hypervisor_mapping_madd in *.
  destruct H_pt_addr'_map as [p2m_osi H_p2m_osi_map H_pt_addr'_map].
  rewrite lemma_map_add_apply_distinct; auto.
  exists p2m_osi; auto.
  rewrite lemma_map_add_apply_distinct; auto.

  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; auto.
  split; auto.
  unfold hypervisor_mapping_madd in *.
  destruct H_ma'_hyp_map as [p2m_osi H_p2m_osi_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply_distinct; auto.
  exists p2m_osi; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_ma'.
  rewrite <- H_ma_eq_ma' in H_pg'_map.
  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping_madd in H_ma'_map.
  destruct H_ma'_hyp_map as [p2m_osi H_p2m_osi_map H_ma'_hyp_map].
  destruct H_valid_hyper with osi pa' p2m_osi as [[ma'' H_ma''_map H_ma'_owner] | H_error];
    try (rewrite <- H_ma'_hyp_map in H_error; discriminate H_error); auto.
  rewrite <- H_ma'_hyp_map in H_ma''_map.
  injection H_ma''_map; intro; subst ma''; clear H_ma''_map.
  unfold maddress_owner in H_ma'_owner.
  rewrite <- H_ma_eq_ma' in H_ma'_owner.
  destruct H_valid_hyper with o pa p2m as [[ma'' H_ma''_map H_ma_owner] | H_error];
    try (rewrite <- H_ma_map in H_error; discriminate H_error); auto.
  rewrite <- H_ma_map in H_ma''_map.
  injection H_ma''_map; intro; subst ma''; clear H_ma''_map.
  unfold maddress_owner in H_ma_owner.
  rewrite H_ma_owner in H_ma'_owner.
  injection H_ma'_owner; intro; contradiction H_o_ne_osi.
  
  (* s' eq_pt s *)
  intros pt_padd pt_page' mp [pt_addr' H_pt_addr'_map H_pt_page'_map] H_pt_page'_content.

  assert (ma<>pt_addr') as H_ma_ne_pt_addr'.
  intro H_ma_eq_pt_addr'.
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m_osi H_p2m_osi_map H_pt_addr'_map].
  rewrite lemma_map_add_apply_distinct in H_p2m_osi_map; auto.
  rewrite H_ma_eq_pt_addr' in H_pt_page'_map.
  rewrite lemma_map_add_apply in H_pt_page'_map.
  injection H_pt_page'_map; intro H_pt_page'_value.
  rewrite H_pt_page'_value in H_pt_page'_content.
  simpl in H_pt_page'_content.
  discriminate H_pt_page'_content.

  exists pt_page'.
  exists pt_addr'.
  unfold hypervisor_mapping_madd in *.
  destruct H_pt_addr'_map as [p2m_osi H_p2m_osi_map H_pt_addr'_map].
  rewrite lemma_map_add_apply_distinct in H_p2m_osi_map; auto.
  exists p2m_osi; auto.
  rewrite lemma_map_add_apply_distinct in H_pt_page'_map; auto.

  exists mp; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va'_accessible H_pg'_type H_ma'_hyp_map.
  exists ma'; auto.
  split.
  unfold hypervisor_mapping_madd in *.
  destruct H_ma'_hyp_map as [p2m_osi H_p2m_osi_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply_distinct in H_p2m_osi_map; auto.
  exists p2m_osi; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  intro H_ma_eq_ma'.
  unfold page_type in H_pg'_type.
  rewrite H_ma_eq_ma' in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
Qed.

Lemma not_osi_action_state_eq : forall osi s s' a,
  valid_state ctxt s -> valid_exec ctxt s a s' ->
  ~(osi_action osi s a) -> osi_equivalent ctxt s s' osi.
Proof.
  intros osi s s' a H_s_valid H_valid_exec H_not_osi_action.
  generalize H_valid_exec; intro H_valid_exec';
    inversion_clear H_valid_exec' as
      (s_vexec, s'_vexec, act_vexec, H_pre, H_post);
      unfold pre_action in H_pre; unfold post_action in H_post;
        induction a.

  (* Silent action *)
  unfold post_silent, same_state in H_post.
  destruct H_post as [new_cache [new_tlb [H_cache'_valid [H_tlb'_valid
    H_assign]]]]; auto.

  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s cmp = state_component_interp s' cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_ThreeComps H_interp H_aos H_oss H_activity System_Cache
    Active_OS Guest_OSs OS_Activity.
  Simp_Inv_TwoComps H_interp H_hyper H_mem System_Cache
    Hypervisor_SC System_Memory.

  unfold osi_equivalent, aos_equivalent_pt_pages, osi_equivalent_pt_pages, osi_pt_pages_eq,
    aos_equivalent_hyper_mapping, osi_equivalent_hyper_mapping, 
    osi_hyper_mapping_eq, padd_page.
  rewrite <- H_aos; try (intro; discriminate).
  rewrite <- H_oss; try (intro; discriminate). 
  rewrite <- H_activity; try (intro; discriminate).
  rewrite <- H_hyper; try (intro; discriminate).
  rewrite <- H_mem; try (intro; discriminate).
  elim osi_equivalent_refl with ctxt s osi.
  intro H_eq.
  unfold osi_equivalent, aos_equivalent_pt_pages, osi_equivalent_pt_pages, osi_pt_pages_eq,
    aos_equivalent_hyper_mapping, osi_equivalent_hyper_mapping, 
    osi_hyper_mapping_eq, padd_page in H_eq; auto.

  (* Read action *)
  assert (st_active_os s <> osi) as H_osi_not_active.
  intro H_osi_active.
  apply H_not_osi_action.
  constructor; auto.
  unfold pre_read in H_pre.
  destruct H_pre as [H_pre [H_v_accessible H_os_running]]; auto.
  apply other_os_running_equiv with (Read v); auto.
  unfold pre_read in H_pre.
  destruct H_pre as [H_pre [H_accessible H_os_running]].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_os_running in H_abs; discriminate H_abs.
  
  (* Read hyper action *)
  apply read_hyper_osi_equiv with v; auto.

  (* Write action *)
  assert (st_active_os s <> osi) as H_osi_not_active.
  intro H_osi_active.
  apply H_not_osi_action.
  constructor; auto.
  unfold pre_write in H_pre.
  destruct H_pre; auto.

  apply other_os_running_equiv with (Write v v0); auto.
  unfold pre_write in H_pre.
  destruct H_pre as [H_act H_pre].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.

  (* Write hyper *)
  apply write_hyper_osi_equiv with v v0; auto.

  (* New untrusted *)
  assert (o <> osi) as H_o_not_osi.
  intro; subst o.
  apply H_not_osi_action.
  constructor 2.
  apply new_untrusted_not_osi_eq with o v p; auto.

  (* New trusted *)
  unfold pre_new_trusted in H_pre.
  destruct H_pre as [ma H_ma_map [H_ma_ptype [H_ma_owner [H_v_accessible [H_aos_trusted H_os_running]]]]]; auto.
  assert (st_active_os s <> osi) as H_osi_not_active.
  intro H_osi_active.
  apply H_not_osi_action.
  unfold os_running in H_os_running.
  constructor; auto.
  apply other_os_running_equiv with (New_trusted v p); auto.
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_os_running in H_abs; discriminate H_abs.

  (* New hyper *)
  apply new_hyper_osi_equiv with v m; auto.
  
  (* Del untrusted *)
  assert (o <> osi) as H_o_not_osi.
  intro; subst o.
  apply H_not_osi_action.
  constructor 3.
  apply del_untrusted_not_osi_eq with o v; auto.
  
  (* Del trusted *)
  unfold pre_del_trusted, pre_all_del in H_pre.
  destruct H_pre as [pt_addr H_pt_addr_map
    [[mp H_mp_map [ma H_ma_map]] [H_os_running [H_trusted H_v_accessible]]]]; auto.
  assert (st_active_os s <> osi) as H_osi_not_active.
  intro H_osi_active.
  apply H_not_osi_action.
  unfold os_running in H_os_running.
  constructor; auto.
  apply other_os_running_equiv with (Del_trusted v); auto.
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_os_running in H_abs; discriminate H_abs.

  (* Del hyper *)
  apply del_hyper_osi_equiv with v; auto.

  (* Switch *)
  assert (o<>osi) as H_o_ne_osi.
  intro H_o_eq_osi.
  apply H_not_osi_action.
  subst o.
  constructor 9.
  assert (st_active_os s <> osi) as H_osi_not_active.
  intro H_osi_active.
  apply H_not_osi_action.
  constructor 8; auto.
  
  unfold pre_switch in H_pre.
  unfold osi_equivalent, osi_equivalent_hyper_mapping,
    osi_equivalent_pt_pages, osi_hyper_mapping_eq,
    osi_pt_pages_eq, padd_page.

  unfold post_switch in H_post.
  rewrite H_post; simpl.

  repeat split; auto.
  intros pt_padd pt_page mp [pt_addr H_pt_addr_map H_pt_page_map] H_pt_page_content.
  exists pt_page.
  exists pt_addr; auto.
  exists mp; auto.
  intros va pa ma1 pg H_ma1_map H_pg_map H_va_accessible H_pt_type H_ma1_hyp_map.
  exists ma1; auto.
  intros pt_padd pt_page mp [pt_addr H_pt_addr_map H_pt_page_map] H_pt_page_content.
  exists pt_page.
  exists pt_addr; auto.
  exists mp; auto.
  intros va pa ma1 pg H_ma1_map H_pg_map H_va_accessible H_pt_type H_ma1_hyp_map.
  exists ma1; auto.

  (* Lswitch trusted *)
  unfold pre_lswitch_trusted, pre_all_lswitch in H_pre.
  destruct H_pre as [H_os_running [H_trusted [ma H_ma_map [H_ma_ptype H_ma_owner]]]].
  unfold hypervisor_mapping_madd in H_ma_map.
  destruct H_ma_map as [p2m H_p2m_map H_ma_map].
  assert (st_active_os s <> osi) as H_osi_not_active.
  intro H_osi_active.
  apply H_not_osi_action.
  unfold os_running in H_os_running.
  constructor; auto.
  apply other_os_running_equiv with (Lswitch_trusted p); auto.
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_os_running in H_abs; discriminate H_abs.

  (* Lswitch untrusted *)
  apply lswitch_untrusted_not_osi_eq with o p; auto.
  intro; subst o.
  apply H_not_osi_action.
  constructor 4.

  (* Hcall *)
  unfold pre_hcall in H_pre.
  destruct H_pre as [H_os_running H_untrusted]. 
  assert (st_active_os s <> osi) as H_osi_not_active.
  intro H_osi_active.
  apply H_not_osi_action.
  unfold os_running in H_os_running.
  constructor; auto.
  apply other_os_running_equiv with (Hcall h); auto.
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_os_running in H_abs; discriminate H_abs.

  (* Ret_ctrl *)
  unfold pre_ret_ctrl in H_pre.
  assert (st_active_os s <> osi) as H_osi_not_active.
  intro H_osi_active.
  apply H_not_osi_action.
  unfold os_running in H_pre.
  constructor; auto.
  apply other_os_running_equiv with Ret_ctrl; auto.
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_pre in H_abs; discriminate H_abs.

  (* Chmod *)
  unfold pre_chmod in H_pre.
  destruct H_pre as [H_hyper_running H_no_hcall].
  assert (st_active_os s <> osi) as H_osi_not_active.
  intro H_osi_active.
  apply H_not_osi_action.
  constructor 7; auto.
  apply chmod_osi_equiv; auto.

  (* Page pin trusted *)
  unfold pre_page_pin_trusted in H_pre.
  destruct H_pre as [H_os_running [H_trusted H_pre]].
  assert (st_active_os s <> osi) as H_osi_not_active.
  intro H_osi_active.
  apply H_not_osi_action.
  unfold os_running in H_os_running.
  constructor; auto.
  apply other_os_running_equiv with (Page_pin_trusted p p0); auto.
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_os_running in H_abs; discriminate H_abs.
  
  (* Page unpin trusted *)
  unfold pre_page_unpin_trusted in H_pre.
  destruct H_pre as [H_os_running [H_trusted H_pre]].
  assert (st_active_os s <> osi) as H_osi_not_active.
  intro H_osi_active.
  apply H_not_osi_action.
  unfold os_running in H_os_running.
  constructor; auto.
  apply other_os_running_equiv with (Page_unpin_trusted p); auto.
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_os_running in H_abs; discriminate H_abs.

  (* Page pin untrusted *)
  apply pin_untrusted_not_osi_eq with o p p0; auto.
  intro; subst o.
  apply H_not_osi_action.
  constructor 5.

  (* Page unpin untrusted *)
  apply unpin_untrusted_not_osi_eq with o p; auto.
  intro; subst o.
  apply H_not_osi_action.
  constructor 6.
Qed.
  
Definition full_isolation (s1 s1':state) (osi:os_ident) : Prop :=
  forall (s2 s2':state) (a:action),
     valid_state ctxt s1 ->
     valid_state ctxt s2 ->
     osi_equivalent ctxt s1 s2 osi ->
     valid_exec ctxt s1 a s1' ->
     valid_exec ctxt s2 a s2' ->
     osi_equivalent ctxt s1' s2' osi.

Theorem full_isolation_proof :
  forall s1 s1' osi, full_isolation s1 s1' osi.
Proof.
  unfold full_isolation.
  intros s1 s1' osi s2 s2' a H_s1_valid H_s2_valid H_s1_eqosi_s2
    H_valid_exec1 H_valid_exec2.
  case (os_ident_eq (st_active_os s1) osi).

  (* OSI is the Active OS *)
  assert ((forall o, a <> Switch o) \/ (exists o, a = Switch o)).
  induction a; try (left; intros o0 H_abs; discriminate).
  right.
  exists o; auto.
  destruct H as [H_no_switch | H_switch].

  intro H_osi_active.
  apply aos_osi_equivalent; auto.
  rewrite <- H_osi_active.
  apply no_switch_same_aos with a; auto.

  apply os_isolation with s1 s2 a; auto.
  apply osi_aos_equivalent with osi; auto.

  (* Switch action *)
  intro H_osi_active.
  destruct H_switch as [o H_switch].

  inversion_clear H_valid_exec1; inversion_clear H_valid_exec2.
  rewrite H_switch in H0, H2.
  simpl in H0, H2.
  unfold post_switch in H0, H2.

  unfold osi_equivalent, osi_equivalent_hyper_mapping, osi_equivalent_pt_pages,
    osi_hyper_mapping_eq, osi_pt_pages_eq, padd_page.
  
  rewrite H0, H2; simpl.
  destruct H_s1_eqosi_s2 as [H_active [H_osi [[H_hyp1 H_hyp2] [H_pt1 H_pt2]]]].
  repeat split; auto.

  case (os_ident_eq o osi).
  intro; subst o.
  left; repeat split; auto.
  destruct H_active as [[H_osi_active1 [H_osi_active2 H_activity]] |
    [H_osi_no_active1 H_osi_no_active2]]; auto.
  contradiction.

  (* If o <> osi *)
  intro H_o_ne_osi; right; auto.
  
  (* If osi is not the active OS. *)
  intro H_osi_not_active1.
  assert (st_active_os s2 <> osi) as H_osi_not_active2.
  intro H_osi_active2.
  unfold osi_equivalent in H_s1_eqosi_s2.
  destruct H_s1_eqosi_s2 as [H_active [H_osi [[H_hyp1 H_hyp2] [H_pt1 H_pt2]]]].
  destruct H_active as [[H_osi_active1 [H_osi_active' H_activity]] | [H_osi_no_active1 H_osi_no_active2]].
  apply H_osi_not_active1; assumption.
  apply H_osi_no_active2; assumption.

  assert (valid_state ctxt s1') as H_s1'_valid.
  generalize valid_state_invariant.
  intros H_valid_state_invariant.
  elim H_valid_state_invariant with ctxt s1 s1' a; auto.
  unfold valid_state; auto.
  assert (valid_state ctxt s2') as H_s2'_valid.
  generalize valid_state_invariant.
  intros H_valid_state_invariant.
  elim H_valid_state_invariant with ctxt s2 s2' a; auto.
  unfold valid_state; auto.
  
  generalize H_valid_exec1, H_valid_exec2.
  intros H_valid_exec1' H_valid_exec2'.
  induction a; inversion_clear H_valid_exec1' as (s_vexec1, s'_vexec1, act_vexec1, H_pre1, H_post1);
    inversion_clear H_valid_exec2' as (s_vexec2, s'_vexec2, act_vexec2, H_pre2, H_post2);
      unfold post_action, pre_action in *.

  (* Silent *)
  unfold post_silent, same_state in H_post1, H_post2.
  destruct H_post1 as [new_cache1 [new_tlb1 [H_cache1'_valid [H_tlb1'_valid
    H_assign1]]]]; auto.
  destruct H_post2 as [new_cache2 [new_tlb2 [H_cache2'_valid [H_tlb2'_valid
    H_assign2]]]]; auto.

  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s1 cmp = state_component_interp s1' cmp) as H_interp1.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign1 with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_ThreeComps H_interp1 H_aos1 H_oss1 H_activity1 System_Cache
    Active_OS Guest_OSs OS_Activity.
  Simp_Inv_TwoComps H_interp1 H_hyper1 H_mem1 System_Cache
    Hypervisor_SC System_Memory.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s2 cmp = state_component_interp s2' cmp) as H_interp2.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign2 with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_ThreeComps H_interp2 H_aos2 H_oss2 H_activity2 System_Cache
    Active_OS Guest_OSs OS_Activity.
  Simp_Inv_TwoComps H_interp2 H_hyper2 H_mem2 System_Cache
    Hypervisor_SC System_Memory.

  unfold osi_equivalent, aos_equivalent_pt_pages, osi_equivalent_pt_pages, osi_pt_pages_eq,
    aos_equivalent_hyper_mapping, osi_equivalent_hyper_mapping, 
    osi_hyper_mapping_eq, padd_page.
  rewrite <- H_aos1; try (intro; discriminate).
  rewrite <- H_oss1; try (intro; discriminate). 
  rewrite <- H_activity1; try (intro; discriminate).
  rewrite <- H_hyper1; try (intro; discriminate).
  rewrite <- H_mem1; try (intro; discriminate).
  rewrite <- H_aos2; try (intro; discriminate).
  rewrite <- H_oss2; try (intro; discriminate). 
  rewrite <- H_activity2; try (intro; discriminate).
  rewrite <- H_hyper2; try (intro; discriminate).
  rewrite <- H_mem2; try (intro; discriminate).
  
  unfold aos_equivalent_pt_pages, osi_equivalent_pt_pages, osi_pt_pages_eq,
    aos_equivalent_hyper_mapping, osi_equivalent_hyper_mapping, 
    osi_hyper_mapping_eq, padd_page in H_s1_eqosi_s2; auto.
  
  (* Read *)
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  apply other_os_running_equiv with (Read v); auto.
  unfold pre_read in H_pre1.
  destruct H_pre1 as [H_pre1 [H_v_accessible H_act]].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.
  apply osi_equivalent_trans with s2; auto.
  apply other_os_running_equiv with (Read v); auto.
  destruct H_pre2 as [H_pre2 [H_v_accessible H_act]].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.

  (* Read Hyper *)
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  apply read_hyper_osi_equiv with v; auto.
  apply osi_equivalent_trans with s2; auto.
  apply read_hyper_osi_equiv with v; auto.

  (* Write *)
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  apply other_os_running_equiv with (Write v v0); auto.
  unfold pre_write in H_pre1.
  destruct H_pre1 as [H_act H_pre1].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.
  apply osi_equivalent_trans with s2; auto.
  apply other_os_running_equiv with (Write v v0); auto.
  destruct H_pre2 as [H_act H_pre2].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.
  
  (* Write Hyper *)
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  apply write_hyper_osi_equiv with v v0; auto.
  apply osi_equivalent_trans with s2; auto.
  apply write_hyper_osi_equiv with v v0; auto.
  
  (* New untrusted *)
  apply new_os_isolation with s1 s2 o v p; auto.

  (* New trusted *)
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  apply other_os_running_equiv with (New_trusted v p); auto.
  unfold pre_new_trusted in H_pre1.
  destruct H_pre1 as [pt_addr H_current [H_ptype [H_owner [H_accessible [H_trusted H_act]]]]].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.
  apply osi_equivalent_trans with s2; auto.
  apply other_os_running_equiv with (New_trusted v p); auto.
  destruct H_pre2 as [pt_addr H_current [H_ptype [H_owner [H_accessible [H_trusted H_act]]]]].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.

  (* New hyper *)
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  apply new_hyper_osi_equiv with v m; auto.
  apply osi_equivalent_trans with s2; auto.
  apply new_hyper_osi_equiv with v m; auto.

  (* Del untrusted *)
  apply del_untrusted_os_isolation with s1 s2 o v; auto.

  (* Del trusted *)
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  apply other_os_running_equiv with (Del_trusted v); auto.
  unfold pre_del_trusted in H_pre1.
  destruct H_pre1 as [pt_addr H_current [H_pre_all [H_act H_pre1]]].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.
  apply osi_equivalent_trans with s2; auto.
  apply other_os_running_equiv with (Del_trusted v); auto.
  destruct H_pre2 as [pt_addr H_current [H_pre_all [H_act H_pre2]]].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.

  (* Del hyper *)
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  apply del_hyper_osi_equiv with v; auto.
  apply osi_equivalent_trans with s2; auto.
  apply del_hyper_osi_equiv with v; auto.

  (* Switch *)
  apply switch_osi_equiv with s1 s2 o; auto.

  (* Lswitch trusted *)
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  apply other_os_running_equiv with (Lswitch_trusted p); auto.
  unfold pre_lswitch_trusted in H_pre1.
  destruct H_pre1 as [H_act H_pre1].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.
  apply osi_equivalent_trans with s2; auto.
  apply other_os_running_equiv with (Lswitch_trusted p); auto.
  destruct H_pre2 as [H_act H_pre2]. 
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.

  (* Lswitch untrusted *)
  apply lswitch_untrusted_osi_equiv with s1 s2 p o; auto.

  (* Hcall *)
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  apply other_os_running_equiv with (Hcall h); auto.
  unfold pre_hcall in H_pre1.
  destruct H_pre1 as [H_act H_pre1].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.
  apply osi_equivalent_trans with s2; auto.
  apply other_os_running_equiv with (Hcall h); auto.
  destruct H_pre2 as [H_act H_pre2]. 
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.

  (* Ret ctrl *)
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  apply other_os_running_equiv with Ret_ctrl; auto.
  unfold pre_ret_ctrl in H_pre1.
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_pre1 in H_abs; discriminate H_abs.
  apply osi_equivalent_trans with s2; auto.
  apply other_os_running_equiv with Ret_ctrl; auto.
  unfold pre_ret_ctrl in H_pre2.
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_pre2 in H_abs; discriminate H_abs.

  (* Chmod *)
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  apply chmod_osi_equiv; auto.
  apply osi_equivalent_trans with s2; auto.
  apply chmod_osi_equiv; auto.

  (* Pin trusted *)
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  apply other_os_running_equiv with (Page_pin_trusted p p0); auto.
  unfold pre_page_pin_trusted in H_pre1.
  destruct H_pre1 as [H_act H_pre1].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.
  apply osi_equivalent_trans with s2; auto.
  apply other_os_running_equiv with (Page_pin_trusted p p0); auto.
  destruct H_pre2 as [H_act H_pre2].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.

  (* Unpin trusted *)
  apply osi_equivalent_trans with s1.
  apply osi_equivalent_symm.
  apply other_os_running_equiv with (Page_unpin_trusted p); auto.
  unfold pre_page_pin_trusted in H_pre1.
  destruct H_pre1 as [H_act H_pre1].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.
  apply osi_equivalent_trans with s2; auto.
  apply other_os_running_equiv with (Page_unpin_trusted p); auto.
  destruct H_pre2 as [H_act H_pre2].
  unfold os_running, hyper_running in *.
  intro H_abs; rewrite H_act in H_abs; discriminate H_abs.

  (* Pin untrusted *)
  apply pin_untrusted_os_isolation with s1 s2 o p p0; auto.

  (* Unpin untrusted *)
  apply unpin_untrusted_os_isolation with s1 s2 o p; auto.
Qed.

End Execution_Isolation.

Section Safe_full_isolation.

Variable osi : os_ident.
Variable ctxt : context.

CoInductive step_equivalent_traces : state_stream -> state_stream -> Prop :=
  Step_equivalent: forall (str1 str2:state_stream)
    (a:action) (s1 s2:state),
    step_equivalent_traces str1 str2 -> osi_equivalent ctxt s1 s2 osi -> 
    step_equivalent_traces (Cons (Ex_State s1 a) str1) (Cons (Ex_State s2 a) str2).


CoInductive same_actions : state_stream -> state_stream -> Prop :=
  Osi_actions_active: forall (str1 str2:state_stream)
    (a:action) (s1 s2:state),
    same_actions str1 str2 ->
    same_actions (Cons (Ex_State s1 a) str1) (Cons (Ex_State s2 a) str2).

Theorem safe_full : forall (str1 str2:state_stream),
  same_actions str1 str2 -> run ctxt str1 -> run ctxt str2 ->
  osi_equivalent ctxt (ex_state (hd str1)) (ex_state (hd str2)) osi ->
  step_equivalent_traces str1 str2.
Proof.
  cofix str1.
  intros str2 str3 H_same_actions H_run2 H_run3 H_first_osi_eq.
  inversion_clear H_same_actions in H_run2 H_run3 H_first_osi_eq .
  simpl in H_first_osi_eq.
  constructor; auto.

  unfold run, init_state in H_run2, H_run3.
  simpl in H_run2, H_run3.
  destruct H_run2; destruct H_run3. 

  apply str1; auto.
  
  unfold run.
  unfold init_state.
  split.
  inversion_clear H1.
  unfold system_transition in H5.
  simpl in H5.
  apply valid_state_invariant with s1 a; auto.
  inversion H1; auto.

  unfold run.
  unfold init_state.
  split.
  inversion_clear H3.
  unfold system_transition in H5.
  simpl in H5.
  apply valid_state_invariant with s2 a; auto.
  inversion H3; auto.

  generalize full_isolation_proof.
  intros H_full_isolation_proof.
  unfold full_isolation in H_full_isolation_proof.
  apply H_full_isolation_proof with s1 s2 a; auto.
  
  inversion H1.
  unfold system_transition in H7.
  simpl in H7; auto.
  inversion H3.
  unfold system_transition in H7.
  simpl in H7; auto.
Qed.


End Safe_full_isolation.
  
