(***********************************************************************
 * File: OS.v
 * $Revision: 364 $
 * $Date: 2010-11-04 14:16:55 -0200 (Thu, 04 Nov 2010) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * This file specifies what is meant by an OS
 * 
 *************************************************************************)

Require Export Memory.

(** Hypercall *)

(** When an untrusted OS wants to do something that requires privileges,
    it makes an hypercall to ask the Hypervisor to do it. 
    The possible requests are: *)

Inductive Hyperv_call: Set :=
(** New mapping : Add a pair to the memory mapping *)
| Hyperv_call_new : vadd -> padd -> Hyperv_call
(** Delete a pair from the memory mapping *)
| Hyperv_call_del : vadd -> Hyperv_call  
(** Local Switch, i.e inside a guest OS *)
| Hyperv_call_lswitch : padd -> Hyperv_call
(** Remove the label classification of a page. Not implemented by VLX *)
| Hyperv_call_unpin : padd -> Hyperv_call    
(** Classification  of a page (PT/RW...). Not implemented by VLX *)
| Hyperv_call_pin : padd -> ptype -> Hyperv_call.

(** *** OS and Hypervisor *)

Parameter os_ident : Set.
Parameter os_ident_eq : forall oi1 oi2 : os_ident, {oi1 = oi2} + {oi1 <> oi2}.

(** Operating Systems (dynamic information) *)

Record os : Set :=
  OS { (** hypercall to Hypervisor *)	
       os_hcall : option Hyperv_call;  
       (** The current page table of an OS *)
       os_current_page : padd 
      }.

(** There are two execution modes :
- Hypervisor and trusted OSes can execute in supervisor mode (svc) 
- untrusted OSes must run in user mode (usr).  *)

Inductive exec_mode: Set :=
| EXMD_usr : exec_mode
| EXMD_svc : exec_mode.

(** An active OS can be running (executing) or waiting. 
    When it is waiting, it means that the hypervisor is executing. *)

Inductive os_activity: Set := 
| OSA_running : os_activity
| OSA_waiting : os_activity.

