(*****************************************************
 * File: Soundness.v
 * $Revision: 455 $
 * $Date: 2011-01-25 20:29:37 -0200 (Tue, 25 Jan 2011) $
 * $LastChangedBy: jdcampo $
 *****************************************************
 * An OS  execution is not affected by the virtualization platform
 *****************************************************)

Require Import Exec_invariant.
Require Import Traces.
Require Import State_equivalence.
Require Import Erasure.

Section Soundness.

Variable ctxt : context.
Variable osi : os_ident.


(*************************************************)

CoInductive str_erasure : state_stream -> state_stream -> Prop :=
  Erased_trace : forall (s:exec_state)(str estr:state_stream),
                     str_erasure str estr ->
                     osi_exists osi (ex_state s) ->
	                   let es  := state_erase ctxt osi (ex_state s) in
                     let ea  := action_erase osi (ex_state s) (ex_action s) in
                     let ns  := Ex_State es ea in
                     str_erasure (Cons s str) (Cons ns estr).

CoInductive step_equivalent_traces_upto_aos : state_stream -> state_stream -> Prop :=
  Step_equivalent: forall (str1 str2:state_stream)
    (a1 a2:action) (s1 s2:state),
    step_equivalent_traces_upto_aos str1 str2 -> osi_equivalent_upto_aos ctxt osi s1 s2 -> 
    step_equivalent_traces_upto_aos (Cons (Ex_State s1 a1) str1) (Cons (Ex_State s2 a2) str2).


Lemma erasure_run_is_trace : forall (str1 str2:state_stream), 
  str_erasure str1 str2 ->  run ctxt str1 -> trace ctxt str2.
Proof.
  cofix str.
  intros str1 str2 H_erasure [H_s_valid H_trace].
  inversion H_erasure.
  rewrite <- H1 in H_s_valid.
  unfold init_state in H_s_valid; simpl in H_s_valid.
  rewrite <- H1 in H_trace; inversion H_trace.
  constructor; auto.
(* Trace *)
  apply (str str0 estr); auto.
  unfold run; split; auto.
  unfold init_state.
  apply valid_state_invariant with (ex_state s) (ex_action s); auto.
(* Valid step *)
  unfold system_transition in H6.
  unfold es, ns; simpl.
  unfold system_transition; simpl.
  unfold es, ea.
  inversion_clear H in H6.
  simpl in H6; simpl.
  unfold es0.
  apply erasure_step; auto.
Qed.

Lemma erasure_run_is_run : forall (str1 str2:state_stream), 
  str_erasure str1 str2 ->  run ctxt str1 -> run ctxt str2.
Proof.
  intros str1 str2 H_erasure; unfold run.
  inversion H_erasure.
  simpl; unfold init_state.
  intros (H_valid,H_trace).
  inversion H_trace.
  split.
(* valid state *)
  unfold ns; simpl.
  unfold es.
  apply erasure_valid; auto.
(* erased trace *)
  unfold ns, es, ea.
  apply (erasure_run_is_trace (Cons s str)).
  constructor; auto.
  unfold run; split; auto.
Qed.


Theorem Os_Soundness : forall (str1:state_stream),
  run ctxt str1 ->
  forall str2, str_erasure str1 str2  ->
    step_equivalent_traces_upto_aos str1 str2.
Proof.
  cofix str.
  intros str1 H_run str2 H_erasure.
  inversion H_erasure.
  replace s with (Ex_State (ex_state s) (ex_action s));
    try (induction s; simpl; auto; fail).

  rewrite <- H1 in H_run; unfold run in H_run; simpl in H_run.
  destruct H_run as [H_s_valid H_trace].
  inversion_clear H_trace. unfold system_transition in H4; inversion_clear H4.

  constructor. 
  apply str; auto.
  split; unfold init_state; auto.
  apply valid_state_invariant with (ex_state s) (ex_action s); auto.
  constructor; auto.
  unfold es; apply erase_state_equiv; auto.
Qed.
  

End Soundness.
