(***********************************************************************
 * File: Exec_invariant.v
 * $Revision: 482 $
 * $Date: 2011-03-05 09:41:24 -0200 (Sat, 05 Mar 2011) $
 * $Author: jdcampo $
 *******************************************************************
 * In this module the invariance of a secure state wrt the actions is
 * proved.
 ******************************************************************)
Require Import Omega.
Require Import List.
Require Export State_proof.
Require Export State_effects.
Require Export Exec.
Require Export Cache_invariant.

Set Undo 20000.

Section Secure_State.

Variable ctxt : context.

Theorem silent_secure : 
  forall (s s':state),
  valid_state ctxt s ->
  valid_exec ctxt s Silent s'->
  valid_state ctxt s'.
Proof.
  intros s s' H_valids H_valid_exec.
  generalize H_valids; Intros_Valid_Hyp;
    destruct H_va_has_pa as [H_va_has_pa [H_cache_valid H_tlb_valid]].
    
  inversion_clear H_valid_exec.
  unfold post_action, post_silent, same_state in H0.

  destruct H0 as [new_cache [new_tlb [H_cache'_valid [H_tlb'_valid
    H_assign]]]]; auto.
  unfold state_components_assign, state_components_update in H_assign.
  unfold valid_state.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s cmp = state_component_interp s' cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_ThreeComps H_interp H_aos H_oss H_exec_mode System_Cache
    Active_OS Guest_OSs Exec_Mode.
  Simp_Inv_ThreeComps H_interp H_activity H_hyper H_mem System_Cache
    OS_Activity Hypervisor_SC System_Memory.
  rewrite <- H_aos; try (intro; discriminate).
  rewrite <- H_oss; try (intro; discriminate). 
  rewrite <- H_exec_mode; try (intro; discriminate).
  rewrite <- H_activity; try (intro; discriminate).
  rewrite <- H_hyper; try (intro; discriminate).
  rewrite <- H_mem; try (intro; discriminate).
  do 10 (try split); auto.
Qed.
  

Theorem Read_secure : 
  forall (s s':state),
  forall va,
  valid_state ctxt s ->
  valid_exec ctxt s (Read va) s'->
  valid_state ctxt s'.
Proof.
  intros s s' va H_valids H_valid_exec.
  generalize H_valid_exec; intro H_valid_exec'.
  inversion_clear H_valid_exec' as [s_vexec s'_vexec act_vexec H_pre H_post].
  unfold post_action, post_read, post_all_read in H_post.
  unfold pre_action, pre_read, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_ma_map H_ptype]] [H_va_accessible H_os_running]].

  generalize H_ma_map; intro H_ma_map'.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map'.
  destruct H_ma_map' as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map']].

  assert (Some ma = va_mapped_to_ma_cache s va) as H_ma_cache_map.
  unfold va_mapped_to_ma_cache.
  unfold va_mapped_to_ma in H_ma_map.
  assert (exists it, it = map_apply vadd_eq (st_tlb s) va) as [it H_it_map].
  induction (map_apply vadd_eq (st_tlb s) va).
  exists (Value vadd v); auto.
  exists (Error madd e); auto.
  rewrite <- H_it_map.
  induction it.
  assert (va_mapped_to_ma s (st_active_os s) va v).
  apply valid_state_included_tlb with ctxt; auto.
  rewrite va_mapped_same_ma with s va ma v; auto.
  rewrite <- H_aos_map.
  rewrite <- H_p2m_map.
  rewrite <- H_pt_addr_map.
  rewrite <- H_pt_page_map.
  rewrite <- H_mp_map.
  rewrite <- H_ma_map'; auto.

  assert (exists pg , Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v; auto.
  assert (Some pg = va_mapped_to_pg_cache s va) as H_pg_map_cache.
  unfold va_mapped_to_pg_cache.
  assert (exists it, it = map_apply vadd_eq (st_cache s) va) as [it H_it_map].
  induction (map_apply vadd_eq (st_cache s) va).
  exists (Value vadd v); auto.
  exists (Error page e); auto.
  rewrite <- H_it_map.
  induction it.
  elim valid_state_included_cache with ctxt s va v; auto.
  intros ma' H_ma'_map H_pg_map'.
  assert (ma = ma').
  apply va_mapped_same_ma with s va; auto.
  subst ma'; clear H_ma'_map.
  rewrite <- H_pg_map in H_pg_map'; injection H_pg_map'; intro; subst v; auto.
  rewrite <- H_ma_cache_map.
  rewrite <- H_pg_map; auto.

  generalize H_valids; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [H_valid_cache H_valid_tlb]].
  unfold state_components_assign, state_components_update in H_post.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s cmp = state_component_interp s' cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  destruct H_post with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_ThreeComps H_interp H_aos H_oss H_exec_mode System_Cache Active_OS Guest_OSs Exec_Mode.
  Simp_Inv_ThreeComps H_interp H_activity H_hyper H_mem System_Cache OS_Activity Hypervisor_SC System_Memory.
  unfold valid_state.
  rewrite <- H_aos; try (intro; discriminate).
  rewrite <- H_oss; try (intro; discriminate). 
  rewrite <- H_exec_mode; try (intro; discriminate).
  rewrite <- H_activity; try (intro; discriminate).
  rewrite <- H_hyper; try (intro; discriminate).
  rewrite <- H_mem; try (intro; discriminate).
  do 10 (try split); auto.
  apply exec_valid_cache_tlb with ctxt s (Read va); auto.
Qed.


Theorem Read_hyper_secure : 
  forall (s s':state),
  forall va,
  valid_state ctxt s ->
  valid_exec ctxt s (Read_hyper va) s'->
  valid_state ctxt s'.
Proof.
  intros s s' va H_valids H_valid_exec.
  generalize H_valid_exec; intro H_valid_exec'.
  inversion_clear H_valid_exec' as [s_vexec s'_vexec act_vexec H_pre H_post].
  unfold post_action, post_read, post_all_read in H_post.
  unfold pre_action, pre_read_hyper, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_ma_map H_ptype]] H_hyper_running].

  generalize H_ma_map; intro H_ma_map'.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map'.
  destruct H_ma_map' as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map']].

  assert (Some ma = va_mapped_to_ma_cache s va) as H_ma_cache_map.
  unfold va_mapped_to_ma_cache.
  unfold va_mapped_to_ma in H_ma_map.
  assert (exists it, it = map_apply vadd_eq (st_tlb s) va) as [it H_it_map].
  induction (map_apply vadd_eq (st_tlb s) va).
  exists (Value vadd v); auto.
  exists (Error madd e); auto.
  rewrite <- H_it_map.
  induction it.
  assert (va_mapped_to_ma s (st_active_os s) va v).
  apply valid_state_included_tlb with ctxt; auto.
  rewrite va_mapped_same_ma with s va ma v; auto.
  rewrite <- H_aos_map.
  rewrite <- H_p2m_map.
  rewrite <- H_pt_addr_map.
  rewrite <- H_pt_page_map.
  rewrite <- H_mp_map.
  rewrite <- H_ma_map'; auto.

  assert (exists pg , Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v; auto.
  assert (Some pg = va_mapped_to_pg_cache s va) as H_pg_map_cache.
  unfold va_mapped_to_pg_cache.
  assert (exists it, it = map_apply vadd_eq (st_cache s) va) as [it H_it_map].
  induction (map_apply vadd_eq (st_cache s) va).
  exists (Value vadd v); auto.
  exists (Error page e); auto.
  rewrite <- H_it_map.
  induction it.
  elim valid_state_included_cache with ctxt s va v; auto.
  intros ma' H_ma'_map H_pg_map'.
  assert (ma = ma').
  apply va_mapped_same_ma with s va; auto.
  subst ma'; clear H_ma'_map.
  rewrite <- H_pg_map in H_pg_map'; injection H_pg_map'; intro; subst v; auto.
  rewrite <- H_ma_cache_map.
  rewrite <- H_pg_map; auto.

  generalize H_valids; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [H_valid_cache H_valid_tlb]].
  unfold state_components_assign, state_components_update in H_post.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s cmp = state_component_interp s' cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  destruct H_post with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_ThreeComps H_interp H_aos H_oss H_exec_mode System_Cache Active_OS Guest_OSs Exec_Mode.
  Simp_Inv_ThreeComps H_interp H_activity H_hyper H_mem System_Cache OS_Activity Hypervisor_SC System_Memory.
  unfold valid_state.
  rewrite <- H_aos; try (intro; discriminate).
  rewrite <- H_oss; try (intro; discriminate). 
  rewrite <- H_exec_mode; try (intro; discriminate).
  rewrite <- H_activity; try (intro; discriminate).
  rewrite <- H_hyper; try (intro; discriminate).
  rewrite <- H_mem; try (intro; discriminate).
  do 10 (try split); auto.
  apply exec_valid_cache_tlb with ctxt s (Read_hyper va); auto.
Qed.


(*************************
 ** Write is Secure
 ************************)

(* Lemmas *)
Lemma update_rw_same_owner :
  forall (smem smem':system_memory) (ma:madd) (val:value),
    update_rw_page ma val smem smem' ->
    maddress_owner smem ma = maddress_owner smem' ma.
Proof.
  intros smem smem' ma val up_rw.
  elim up_rw.
  intros pg map_smem map_smem'.
  rewrite map_smem'.
  induction smem; simpl.
  simpl in map_smem.
  discriminate.

  destruct (madd_eq (item_index a) ma); simpl.
  unfold maddress_owner.
  simpl.
  rewrite e.
  elim (madd_eq ma ma).
  intro H.
  cut (item_info a = pg).
  intros iteminfo_a.
  rewrite iteminfo_a.
  reflexivity.

  unfold map_apply in map_smem.
  rewrite e in map_smem.
  generalize map_smem.
  elim (madd_eq ma ma).
  intro H''.
  intro pg_eq_iia.
  symmetry.
  injection pg_eq_iia.
  auto.

  contradiction.
  intros; cut False; auto.

  unfold maddress_owner at 2.
  simpl.
  destruct (madd_eq ma (item_index a)).
  symmetry in e; contradiction.

  (* Work with the left side of the eq *)

  unfold maddress_owner.
  rewrite <- map_smem.

  rewrite lemma_map_add_apply.
  auto.
Qed.
  

Lemma update_rw_same_ptype :
  forall (ma:madd) (val:value) (smem smem':system_memory),
    page_type smem ma = Some RW ->
    update_rw_page ma val smem smem' ->
    page_type smem ma = page_type smem' ma.
Proof.
  intros ma val.
  induction smem; intros smem' H_pagetype up_rw; elim up_rw; simpl.
  intros; discriminate.

  intros pg Hmapap Hsmem'; rewrite Hsmem'.
  case (madd_eq (item_index a) ma).
  intro Haeqma; rewrite <- Haeqma.
  generalize H_pagetype.
  unfold page_type.
  rewrite <- Haeqma.
  simpl.
  intro H_pagetype'.
  rewrite H_pagetype'.
    
  case (madd_eq (item_index a) (item_index a)).
  simpl.
  auto.

  intro abs.  elim abs. trivial.

  intro H_anema.
  rewrite H_pagetype.
  unfold page_type.
  simpl.
  case (madd_eq ma (item_index a)).
  intro Hmaeqa; symmetry in Hmaeqa; elim (H_anema Hmaeqa).

  intro H_anema'.
  rewrite lemma_map_add_apply.
  auto.
Qed.  


Lemma update_rw_same_owner_all_mem :
  forall (smem smem':system_memory) (ma:madd) (val:value),
    update_rw_page ma val smem smem' ->
    forall ma', maddress_owner smem ma' = maddress_owner smem' ma'.
Proof.
  intros smem smem' ma val H_up_rw ma'.
  case (madd_eq ma ma').
  intro H_ma_eq_ma'.
  apply update_rw_same_owner with val.
  rewrite <- H_ma_eq_ma'.
  assumption.

  (* Case ma <> ma' *)
  intro H_ma_ne_ma'.
  elim H_up_rw.
  intros pg H_pg_map H_smem'_map.
  rewrite H_smem'_map.
  unfold maddress_owner.
  assert (map_apply madd_eq (map_add madd_eq smem ma (Page (RW_c (Some val)) (page_owned_by pg))) ma' = map_apply madd_eq smem ma').
  rewrite lemma_map_add_apply_distinct; (reflexivity || assumption).

  rewrite H.
  reflexivity.
Qed.

Lemma update_rw_same_ptype_all_mem :
  forall (smem smem':system_memory) (ma:madd) (val:value),
    page_type smem ma = Some RW ->
    update_rw_page ma val smem smem' ->
    forall ma', page_type smem ma' = page_type smem' ma'.
Proof.
  intros smem smem' ma val H_ma_ptype H_up_rw ma'.
  case (madd_eq ma ma').
  intro H_ma_eq_ma'.
  apply update_rw_same_ptype with val.
  rewrite <- H_ma_eq_ma'.
  assumption.
  rewrite <- H_ma_eq_ma'.
  assumption.

  (* Case ma <> ma' *)
  intro H_ma_ne_ma'.
  elim H_up_rw.
  intros pg H_pg_map H_smem'_map.
  rewrite H_smem'_map.
  unfold page_type.
  assert (map_apply madd_eq (map_add madd_eq smem ma (Page (RW_c (Some val)) (page_owned_by pg))) ma' = map_apply madd_eq smem ma').
  rewrite lemma_map_add_apply_distinct; (reflexivity || assumption).

  rewrite H.
  reflexivity.
Qed.
  
Lemma post_all_write_valid_hyper :
  forall (s s':state) (va:vadd) (val:value),
    valid_hypervisor (st_hypervisor s) (st_memory s) ->
    (exists ma, va_mapped_to_ma s (st_active_os s) va ma) ->
    post_all_write va val s s' ->
    valid_hypervisor (st_hypervisor s') (st_memory s'). 
Proof.
  intros s s' va val H_valid_hyper H_ex_va_mapped H_postallwrt.
  destruct H_ex_va_mapped as (ma, H_va_mapped).
  
  unfold post_all_write in H_postallwrt.
  destruct H_postallwrt as [ma' H_ma'_map [new_mem H_update [pg H_pg_map
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]]]].
  rewrite H_s'; simpl.

  unfold valid_hypervisor.
  intros osi pa p2m H_p2m_map.
  elim H_valid_hyper with osi pa p2m.
  intros H_ex_ma''.
  left.
  destruct H_ex_ma'' as (ma'', H_ma''_map, H_ma''_owner).
  exists ma''; try assumption.

  (* Prove that maddress owner is the same forall ma'', using
     update_rw_same_owner_all_mem *)

  rewrite <- H_ma''_owner.
  symmetry.
  apply update_rw_same_owner_all_mem with ma' val; auto.

  intros H_pa_not_mapped; right; assumption.
  assumption.
Qed.

Lemma post_all_write_valid_virtual_mapping :
  forall (s s':state) (va:vadd) (val:value),
    valid_state ctxt s ->
    (exists2 ma, va_mapped_to_ma s (st_active_os s) va ma &
        page_type (st_memory s) ma = Some RW) ->
    post_all_write va val s s' ->
    valid_virtual_mapping (st_memory s') ctxt.
Proof.
  intros s s' va val H_s_valid H_ex_va_mapped H_postallwrt.

  generalize H_s_valid; Intros_Valid_Hyp.

  destruct H_ex_va_mapped as (ma, H_va_mapped, H_ma_ptype).
  unfold post_all_write in H_postallwrt.
  destruct H_postallwrt as [ma2 H_ma2_map [smem' H_up_rw [pg H_pg_map
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]]]].
  rewrite H_s'; simpl.

  assert (ma = ma2).
  apply va_mapped_same_ma with s va; auto.
  apply valid_state_included_tlb_mem with ctxt; auto.
  subst ma2.

  unfold valid_virtual_mapping.
  intros osi pt_addr mp pt H_pt_map H_mp_cont H_pt_owner va' ma'.
  intros H_va'_map.
  elim H_valid_virtmap with osi pt_addr mp pt va' ma';auto.
  intros pg' H_pg'_map H_valid_virtmap'.

  case (madd_eq ma ma').
  (* If the updated page is the image of the current mapping *)
  intros H_ma_eq_ma'.
  elim H_up_rw.
  intros pg_old H_pgold_map H_map_add.

  exists (Page (RW_c (Some val)) (page_owned_by pg_old)).
  
  inversion_clear H_map_add.
  rewrite H_ma_eq_ma'.
  rewrite lemma_map_add_apply; reflexivity.

  (* Prove pg and pg_old are the same (given ma=ma'). *)
  assert (pg' = pg_old).
  rewrite H_ma_eq_ma' in H_pgold_map.
  rewrite <- H_pgold_map in H_pg'_map.
  injection H_pg'_map; auto.
  subst pg'.
  simpl; auto.

  (* If the updated page is not the image of the current mapping *)
  intros H_ma_ne_ma'.
  elim H_up_rw.
  intros pg'' H_pg''_map H_map_add.
  exists pg'; auto.
  rewrite H_map_add.
  rewrite lemma_map_add_apply_distinct; auto.

  elim H_up_rw.
  intros pg' H_pg'_map H_map_add.

  rewrite H_pt_map.
  rewrite H_map_add.

  cut (ma <> pt_addr).
  intro H_ma_ne_ptaddr.
  rewrite lemma_map_add_apply_distinct; auto.

  cut (page_type smem' pt_addr = Some PT).
  intros H_pt_page_type H_ma_eq_pt_addr.
  rewrite <- H_ma_eq_pt_addr in H_pt_page_type.
  rewrite H_map_add in H_pt_page_type.
  generalize H_pt_page_type.
  unfold page_type.
  rewrite lemma_map_add_apply.
  simpl.
  intro; discriminate.

  unfold page_type.
  rewrite <- H_pt_map.
  destruct H_map_add.
  rewrite <- H_mp_cont.
  auto.
Qed.

Lemma post_all_write_valid_current_page :
  forall (s s':state) (va:vadd) (val:value),
    valid_state ctxt s ->
    (exists2 ma, va_mapped_to_ma s (st_active_os s) va ma &
        page_type (st_memory s) ma = Some RW) ->
    post_all_write va val s s' ->
    valid_current_page (st_hypervisor s') (st_memory s') (st_oss s').
Proof.
  intros s s' va val H_s_valid H_ex_va_mapped H_postallwrt.
  generalize H_s_valid; Intros_Valid_Hyp.

  destruct H_ex_va_mapped as (ma, H_va_mapped, H_ma_ptype).
  unfold post_all_write in H_postallwrt.
  destruct H_postallwrt as [ma2 H_ma2_map [smem' H_up_rw [pg H_pg_map
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]]]].
  rewrite H_s'; simpl.

  unfold valid_current_page.
  intros osi system H_system_map p2m H_p2m_map.
  elim H_valid_currpg with osi system p2m.
  intros pt_addr H_pt_map (H_pt_ptype, H_pt_owner).

  exists pt_addr.
  assumption.

  assert (ma = ma2).
  apply va_mapped_same_ma with s va; auto.
  apply valid_state_included_tlb_mem with ctxt; auto.
  subst ma2.

  split.
  
  cut (page_type smem' pt_addr = page_type (st_memory s) pt_addr).
  intros H_ptype'_eq_ptype.
  rewrite H_ptype'_eq_ptype.
  assumption.

  symmetry.
  apply update_rw_same_ptype_all_mem with ma val.
  assumption.
  assumption.

  cut (maddress_owner smem' pt_addr = maddress_owner (st_memory s) pt_addr).
  intros H_owner'_eq_owner.
  rewrite H_owner'_eq_owner.
  assumption.

  symmetry.
  apply update_rw_same_owner_all_mem with ma val; auto.
  assumption.

  assumption.
Qed.



Lemma post_all_write_va_has_valid_pa :
  forall (s s':state) (va:vadd) (val:value),
    valid_state ctxt s ->
    (exists2 ma, va_mapped_to_ma s (st_active_os s) va ma &
        page_type (st_memory s) ma = Some RW) ->
    post_all_write va val s s' ->
    va_has_valid_pa (st_memory s') (st_hypervisor s') ctxt.
Proof.
  intros s s' va val H_s_valid H_ex_va_mapped H_postallwrt.
  destruct H_ex_va_mapped as (ma, H_va_mapped, H_ma_ptype).

  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [H_cache_valid H_tlb_valid]].

  unfold post_all_write in H_postallwrt.
  destruct H_postallwrt as [ma2 H_ma2_map [smem' H_up_rw [pg H_pg_map
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]]]].
  rewrite H_s'; simpl.

  assert (ma = ma2).
  apply va_mapped_same_ma with s va; auto.
  apply valid_state_included_tlb_mem with ctxt; auto.

  subst ma2.

  destruct H_up_rw as (ma_pg,H_mapg,H_newmem_def).
  simpl in H_newmem_def.

  unfold va_has_valid_pa.
  intro osi.

  intros osi_pt_padd osi_pt_addr osi_pt_page
    osi_mp osi_va osi_ma.
  intros H_osi_pt_page_addr H_pt_padd_madd
    H_pt_page_cont H_osi_ma_va H_osi_va.

  unfold va_mapped_to_ma in H_va_mapped.
  destruct H_va_mapped as (aos_cpt_madd,H_cpt_madd,H_va_mapped).
  destruct H_va_mapped as (aos_cur_map,H_curr_map,H_va_ma).
  destruct H_curr_map as (aos_cur_page,H_cur_page,H_curr_map).
  
  (*osi_pt_addr cant be ma, they have different type.*)
  assert (osi_pt_addr<>ma) as H_ptmadd_no_ma.
  cut (page_type (st_memory s') osi_pt_addr
     <>page_type (st_memory s') ma); intro H_diff_types.
  intro H_absurd.
  rewrite H_absurd in H_diff_types.
  elim H_diff_types;auto.
  rewrite H_s', H_newmem_def in H_diff_types; simpl in H_diff_types.
  assert (page_type smem' osi_pt_addr=Some PT) as H_cp_type.
  unfold page_type.
  rewrite <- H_osi_pt_page_addr.
  rewrite <- H_pt_page_cont;auto.

  rewrite H_newmem_def in H_cp_type.
  rewrite H_cp_type in H_diff_types.
  
  unfold page_type in H_diff_types.
  rewrite lemma_map_add_apply in H_diff_types.
  simpl in H_diff_types.
  inversion H_diff_types;auto.

  destruct H_va_has_pa with osi osi_pt_padd osi_pt_addr osi_pt_page
    osi_mp osi_va osi_ma as (pa,H_pa);auto.
  rewrite H_newmem_def in H_osi_pt_page_addr.
  rewrite lemma_map_add_apply_distinct
    in H_osi_pt_page_addr;auto.
  exists pa;auto.
Qed.

Lemma post_write_all_secure:
  forall (s s':state) (a:action),
  forall (va:vadd)(val:value),
  valid_state ctxt s ->
  valid_exec ctxt s a s' ->
  (exists2 ma, va_mapped_to_ma s (st_active_os s) va ma &
     page_type (st_memory s) ma = Some RW) ->

  post_all_write va val s s' ->
  valid_state ctxt s'.
Proof.
  intros s s' a va val H_s_valid H_valid_exec.
  generalize H_s_valid; Intros_Valid_Hyp.
  intros H_va_mapped H_postallwrt.
  generalize H_postallwrt; intro H_postallwrt'.
  destruct H_postallwrt' as [ma2 H_ma2_map [smem' H_up_rw [pg H_pg_map
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]]]].

  do 10 (try split); try (rewrite H_s'; simpl; assumption).

  (* Prove valid_hypervisor *)
  apply post_all_write_valid_hyper with s va val; auto.
  elim H_va_mapped; intros ma H_va_mapped' H_page_type; exists ma; auto.

  (* Prove valid_virtual_mapping *)
  apply post_all_write_valid_virtual_mapping with s va val; auto.

  (* Prove valid_current_page *)
  apply post_all_write_valid_current_page with s va val; auto.
  
  (* Prove va_has_valid_pa *)
  apply post_all_write_va_has_valid_pa with s va val; auto.

  (* Prove cache and tlb validity *)
  apply exec_valid_cache_tlb with ctxt s a; auto.
Qed.

Theorem Write_secure : 
  forall (s s':state),
  forall va val,
  valid_state ctxt s ->
  valid_exec ctxt s (Write va val) s'->
  valid_state ctxt s'.
Proof.
  intros s s' va val H_valids H_valid_exec.
  
  generalize H_valid_exec; intro H_valid_exec';
    inversion_clear H_valid_exec'. 
  unfold post_action in H0.
  unfold post_write in H0.
  unfold pre_action in H.
  unfold pre_write in H.
  destruct H as (H_os_running, (H_os_accessible, H_ex_va_mapped)).
  apply post_write_all_secure with s (Write va val) va val; auto.
Qed.


Theorem Write_Hyper_secure : 
  forall (s s':state),
  forall va val,
  valid_state ctxt s ->
  valid_exec ctxt s (Write_hyper va val) s'->
  valid_state ctxt s'.
Proof.
  intros s s' va val H_valids H_valid_exec.
  generalize H_valid_exec; intro H_valid_exec';
    inversion_clear H_valid_exec'. 
  unfold post_action in H0.
  unfold post_write_hyper in H0.
  unfold pre_action in H.
  unfold pre_write_hyper in H.
  destruct H as (H_hyp_running, H_ex_ma).
  apply post_write_all_secure with s (Write_hyper va val) va val; auto.

  elim H_ex_ma.
  intros ma H_va_mapped H_ma_owner.
  exists ma.
  assumption.

  apply H_ma_owner.
Qed.


(*************************
 ** New is Secure
 ************************)

(*Lemas*)


(*Translate the "component_assign" to equalities, when oss and memory are updated*)

Lemma Unfold_Oss_Mem_Assign :
  forall newmem newoss s s',
  state_components_assign System_Memory Guest_OSs newmem newoss s s' ->
  st_active_os s' = st_active_os s /\
  st_oss s' = newoss /\
  st_exec_mode s'=st_exec_mode s /\
  st_aos_activity s'=st_aos_activity s /\
  st_hypervisor s'=st_hypervisor s /\
  st_memory s'=newmem.
Proof.

  intros newmem newoss s s' H_oss_mem_update.
  repeat split.

  (*st_active_os doesnt change*)
  elim H_oss_mem_update with Active_OS;auto.
  intro H_assign;destruct H_assign as (H_t,H_same_aos);auto.
  intro absurd;destruct absurd as (absurd1,absurd2).
  elim absurd1;split;discriminate.
  

  (*the st_oss change*)
  elim H_oss_mem_update with Guest_OSs.
  intro absurd;destruct absurd as (absurd1,absurd2).
  destruct absurd1 as (H_true,absurd1);elim absurd1;auto.
  intros H_assign.
  destruct H_assign as (H_t, H_assign).
  clear H_t; destruct H_assign as (H_newmem,H_oss_add);auto.

  (*st_exec_mode doesnt change*)  
  elim H_oss_mem_update with Exec_Mode.
  intro H_assign;destruct H_assign as (H_t,H_same_exmd);auto.
  intro absurd;destruct absurd as (absurd1,absurd2).
  elim absurd1;split;discriminate.

  (*st_aos_activity doesnt change*)
  elim H_oss_mem_update with OS_Activity.
  intro H_assign;destruct H_assign as (H_t,H_same_activ);auto.
  intro absurd;destruct absurd as (absurd1,absurd2).
  elim absurd1;split;discriminate.

  (*st_hypervisor doesnt_change*)
  elim H_oss_mem_update with Hypervisor_SC.
  intro H_assign;destruct H_assign as (H_t,H_same_hyper);auto.
  intro absurd;destruct absurd as (absurd1,absurd2).
  elim absurd1;split;discriminate.
  
  (*the st_memory changes*)
  elim H_oss_mem_update with Guest_OSs.
  intro absurd;destruct absurd as (absurd1,absurd2).
  destruct absurd1 as (H_true,absurd1).
  elim absurd1;auto.
  intros H_assign.
  destruct H_assign as (H_t, H_assign).
  clear H_t; destruct H_assign;auto.
Qed.

(*Translate the "component_assign" to equalities, when memory is updated*)

Lemma Unfold_Mem_Assign :
  forall newmem s s',
  state_component_assign System_Memory newmem s s' ->
  
  st_active_os s' = st_active_os s /\
  st_oss s' = st_oss s /\
  st_exec_mode s'=st_exec_mode s /\
  st_aos_activity s'=st_aos_activity s /\
  st_hypervisor s'=st_hypervisor s /\
  st_memory s'=newmem
  .
Proof.
  intros newmem s s' H_mem_update.
  repeat split.

  (*st_active_os doesnt change*)
  elim H_mem_update with Active_OS;intro H_assign.
  destruct H_assign;discriminate.
  destruct H_assign;auto.

  (*st_oss doesnt change*)
  elim H_mem_update with Guest_OSs;intro H_assign.
  destruct H_assign;discriminate.
  destruct H_assign;auto.

  (*st_exec_mode doest change*)  
  elim H_mem_update with Exec_Mode;intro H_assign.
  destruct H_assign;discriminate.
  destruct H_assign;auto.

  (*st_aos_activity doesnt change*)
  elim H_mem_update with OS_Activity;intro H_assign.
  destruct H_assign;discriminate.
  destruct H_assign;auto.

  (*hypervisor doesnt change*)
  elim H_mem_update with Hypervisor_SC;intro H_assign.
  destruct H_assign;discriminate.
  destruct H_assign;auto.
  
  (*the st_memory changes*)
  elim H_mem_update with System_Memory.
  intros H_assign.
  destruct H_assign as (H_t, H_assign).
  clear H_t; destruct H_assign;auto.
  intro absurd;destruct absurd as (absurd1,absurd2).
  elim absurd1;auto.

Qed.

  (* Lemma that proves valid_hypervisor constraint 
  for new actions*)

Lemma valid_hyper_new: forall va ma osi osi_cpt_madd s s',
  valid_state ctxt s -> st_hypervisor s' = st_hypervisor s ->
  os_current_vadd_mapping_madd s osi osi_cpt_madd ->
  update_pt_page osi_cpt_madd va ma (st_memory s) (st_memory s') ->
  valid_hypervisor (st_hypervisor s') (st_memory s').
Proof.

  intros va ma osi osi_cpt_madd s s' H_valid_s 
    H_same_hyper H_osi_cpt_madd H_mem_pt_newmem. 

  rewrite H_same_hyper.
  unfold valid_hypervisor.
  intros osi' any_pa osi'_p2m H_is_osi'_p2m.

  (*s is a valid state, so valid_hypervisor holds*)
  cut (valid_hypervisor (st_hypervisor s) (st_memory s)).
  intro H_valid_hyper_s.
  2:apply H_valid_s.
  unfold valid_hypervisor in H_valid_hyper_s.
  elim H_valid_hyper_s with osi' any_pa osi'_p2m;auto.
  intro H_exist_in_s.
  cut (forall ma, maddress_owner (st_memory s) ma
    = maddress_owner (st_memory s') ma).
  intro H_same_owners.
  destruct H_exist_in_s as (s_ma,H_s_ma_pa,H_s_ma_owner).
  left.
  exists s_ma;auto.
  
  rewrite <-(H_same_owners s_ma);auto.
  intro s_ma.

  (*Proof for both cases: new maddr = curr_pt, and new maddr <> curr_pt*)
  case (madd_eq s_ma osi_cpt_madd).
  intro H_case_osi_cpt.

  (*In this case rewrite with the data updated in the system memory*)
  unfold update_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem as (osi_curr_page, H_osi_curr_page, H_mem_pt_newmem).
  destruct H_mem_pt_newmem as (osi_curr_map, H_osi_curr_map,H_update_osi_cpt).
  rewrite H_update_osi_cpt.
  rewrite H_case_osi_cpt.
  unfold maddress_owner.
  rewrite lemma_map_add_apply.
  rewrite <- H_osi_curr_page;trivial.
  intro H_case_no_pt.

  (*In this case the data didn't change in system memory*)
  unfold update_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem as (osi_curr_page, H_osi_curr_page, H_mem_pt_newmem).
  destruct H_mem_pt_newmem as (osi_curr_map, H_osi_curr_map,H_update_osi_cpt).
  rewrite H_update_osi_cpt.

  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct;auto.

Qed.

  (* Lemma that proves valid_virtual_mapping constraint
  for new trusted and untrusted actions*)

Lemma valid_virtual_mapping_new_trusted_untrusted: 

  forall osi va pa s s',
  valid_state ctxt s ->
  post_all_new osi va pa None s s' ->

   (exists2 ma : madd,
   hypervisor_mapping_madd (st_hypervisor s) osi pa ma &
   page_type (st_memory s) ma <> None /\
   maddress_owner (st_memory s) ma = Os_PO osi /\
   os_accessible ctxt va ) ->
  valid_virtual_mapping (st_memory s') ctxt.
Proof.

  intros osi va pa s s' H_valid_s H_post_new H_pre_new.

  (*Use the post-conditions, so the relation between*)
  (*s and s' component can be established*)


  destruct H_post_new as (ma,H_ma_pa,H_post_new).
  destruct H_post_new as (osi_cpt_madd, H_osi_cpt_madd,H_post_new).
  destruct H_post_new as (newmem, H_mem_pt_newmem, H_post_new).
  destruct H_post_new as (os, H_is_os,H_s').
  simpl in H_s'.
  
  rewrite H_s'; simpl.

  (*More hypothesis, about ma pa and va extracted from preconditions*)
  destruct H_pre_new as (ma',H_ma'_pa,H_pre_new).
  destruct H_pre_new as (H_ma_page_type,H_pre_new).
  destruct H_pre_new as (H_ma_owner,H_acces_va).

  (*unify all about ma and ma' because they are the same*)
  assert (ma'=ma) as H_ma_ma'.

  (*prove that ma and ma' are the same*)
  unfold hypervisor_mapping_madd in H_ma'_pa.
  unfold hypervisor_mapping_madd in H_ma_pa.
  destruct  H_ma'_pa as (p2m', H_p2m_ma',H_ma'_pa).
  destruct  H_ma_pa as (p2m, H_p2m_ma,H_ma_pa).
  cut (p2m=p2m').
  intro H_p2m_p2m'.
  rewrite H_p2m_p2m' in H_ma_pa.
  rewrite <- H_ma_pa in H_ma'_pa.
  inversion H_ma'_pa;trivial.
  rewrite <- H_p2m_ma in H_p2m_ma'.
  inversion H_p2m_ma';trivial.

  rewrite H_ma_ma' in H_ma_page_type.
  rewrite H_ma_ma' in H_ma_owner.
  clear H_ma'_pa H_ma_ma'.

  (*Ended proof ma = ma'*)
  unfold update_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem 
    as (osi_curr_page, H_osi_curr_page, H_mem_pt_newmem).
  destruct H_mem_pt_newmem 
    as (osi_curr_map, H_osi_curr_map,H_update_osi_cpt).


  (*Start main Proof*)

  unfold valid_virtual_mapping.
  intros osi' osi'_ptaddr mp osi'_pt H_osi'_pt_map.

  (*Prove by cases, first on the os considered in *)
  (*valid_virtual_mapping, (osi'). It can be the  *)
  (*one involved in new action, or not (osi) *)
  case (os_ident_eq osi' osi).
  


  (*Case osi'=osi*)
  intro H_osi'_osi.
  intros H_osi'_pt_cont H_osi'_pt_owner mp_va mp_ma H_mp_va_ma.

  (*Current page owner is used several times on the proof*)
  assert (page_owned_by osi_curr_page = Os_PO osi) as H_owner_osi_cpt.

  assert (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)) 
    as H_valid_curr.
  apply H_valid_s.
  
  unfold hypervisor_mapping_madd in H_ma_pa.
  destruct H_ma_pa as (p2m,H_p2m,H_ma_pa).
  
  elim H_valid_curr with osi os p2m;auto.
  intros osi_pt_madd' H_osi_pt' H_osi_pt'_prop.
  
  assert (osi_pt_madd'=osi_cpt_madd) as H_cpt'_cpt.

  unfold os_current_vadd_mapping_madd in H_osi_cpt_madd.
  destruct (H_osi_cpt_madd) as (os'',H_os'',H_os''_prop).

  assert (os''=os) as H_os''_os.
  rewrite <- H_is_os in H_os'';inversion H_os'';auto.
  rewrite H_os''_os in H_os''_prop;clear os'' H_os'' H_os''_os.
  destruct H_os''_prop as (p2m'',H_p2m'',H_osi_cpt).

  assert (p2m''=p2m) as H_p2m''_p2m.
  unfold hypervisor_mapping in H_p2m''.
  rewrite <- H_p2m'' in H_p2m;inversion H_p2m;auto.

  rewrite H_p2m''_p2m in H_osi_cpt.
  rewrite <- H_osi_cpt in H_osi_pt';inversion H_osi_pt';auto.

  rewrite H_cpt'_cpt in H_osi_pt'_prop.
  unfold maddress_owner in H_osi_pt'_prop.
  rewrite <- H_osi_curr_page in H_osi_pt'_prop.
  apply H_osi_pt'_prop.
  (*End of curr_page owner demostration*)
  
  (*Cases of mp_ma are considered, where mp_ma is the image 
  on mp, the mapping considered in valid_virtual_mapping.
  It can be the same of that being updated, current page 
  madd, or not*)
  case (madd_eq osi_cpt_madd mp_ma).
 
  (*In case it is*)
  intro H_mpma_cptmadd.

  (*The corresponding page is the current page updated*) 
  exists (Page (PT_c (map_add vadd_eq osi_curr_map va ma))
                     (page_owned_by osi_curr_page)).

  (*Prove that is the corresponding in newmem*)
  rewrite H_update_osi_cpt.
  rewrite H_mpma_cptmadd.
  rewrite lemma_map_add_apply;auto.
  
  (*Now the accessibility constraint needs to be proved*)
  simpl.
  
  (*To prove accessibility, cases of mp_va are considered*)
  
  (*mp_va (pre-image in mp) can be the same that va or not*)
  case (vadd_eq va mp_va).

  (*In case mp_va=va*)
  intro H_mpva_va.
  (*Pre-conditions are used, that ensures accessibility from aos*)
  rewrite H_mpva_va in H_acces_va.
  split.
  intro.
  rewrite H_osi'_osi.
    
  auto.

  unfold not.
  intro H_absurd.
  elim H_absurd;auto.
  


  (*In case mp_va is not the same that va*)
  intro H_no_mpva_va.
  
  (*We use the fact that s is a valid state*)
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.


  (*To use the H_valid_virt_s hypotesis, it's needed to*)
  (*consider cases of osi_ptaddr (pointer to PT), so the*)
  (*constraint to use H_valid_virt_s are met*)
  case (madd_eq osi_cpt_madd osi'_ptaddr).
  intro H_osi'_ptmadd.
  (*Case that osi_ptadd=pt_maddr; it means it changed*)
  cut(mp=(map_add vadd_eq osi_curr_map va ma)).
  (*relation between old and new current map*)
  intro H_mp_osi_curr_map.
  
  elim H_valid_virt_s with osi osi_cpt_madd 
    osi_curr_map osi_curr_page mp_va mp_ma;auto.
  intros pg H_pg H_acces.

  (*If the considered pointer a PT is cur_pt_madd,*)
  (*then the corresponding page is the current page*)

  cut (pg=osi_curr_page).
  intro H_pg_osi_curr.
  rewrite <- H_pg_osi_curr.
  rewrite H_osi'_osi;auto.
  rewrite <- H_mpma_cptmadd in H_pg.
  rewrite <- H_osi_curr_page in H_pg.
  inversion H_pg;auto.
  rewrite H_mp_osi_curr_map in H_mp_va_ma.
  rewrite lemma_map_add_apply_distinct in H_mp_va_ma;auto.
  rewrite H_update_osi_cpt in H_osi'_pt_map.
  rewrite H_osi'_ptmadd in H_osi'_pt_map.
  rewrite lemma_map_add_apply in H_osi'_pt_map.
  inversion H_osi'_pt_map.
  rewrite H0 in H_osi'_pt_cont.
  inversion H_osi'_pt_cont;auto.

  
  (*Case the pointer to PT (osi_ptaddr) is not osi_cpt_madd,*)
  (*that means its content doesn't change*)

  intro H_no_osi'_cptmadd.
  elim H_valid_virt_s with osi' osi'_ptaddr mp osi'_pt mp_va mp_ma;auto.
  intros pg H_pg H_acces.

  cut (pg=osi_curr_page). 
  intro H_pg_curr.
  rewrite <- H_pg_curr;assumption.

  rewrite <- H_mpma_cptmadd in H_pg.
  rewrite <- H_osi_curr_page in H_pg.
  inversion H_pg;auto.
  
  rewrite H_update_osi_cpt in H_osi'_pt_map.
  rewrite lemma_map_add_apply_distinct in H_osi'_pt_map;auto.


  (*Case that mp_ma is not the address of the current map*)

  intro H_no_mpma_cpt.


  (*We discriminate cases of mp_ma, whether it is the*)
  (*new one (ma)or not*)

  case (madd_eq ma mp_ma).
  (*Case mp_ma is the new one*)
  intro H_mpma_ma.

  (*The corresponding page is inferred from the pre-conditions.*)
  elim lemma_page_type_exists_page with (st_memory s) ma;auto.
  intro pg.
  intro H_pg.
  
  exists pg.


  rewrite H_update_osi_cpt.
  rewrite lemma_map_add_apply_distinct;auto.
  rewrite <- H_mpma_ma;auto.

  (*Now the accessibility constraint needs to be proved when mp_ma=ma*)
  
  (*Using pre-conditions*)

  split.
  rewrite H_osi'_osi.
  intro.
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_pg in H_ma_owner;auto.

  
  (*We consider cases of mp_va (pre-image in mp),  whether it is
  the same of va or not*)
  case (vadd_eq va mp_va).

  (*case mp_va=va*)
  intro H_mpva_va.
  rewrite <- H_mpva_va.
  intro.
  contradiction.

  (*case mp_va<>va*)
  intro H_no_mpva_va.


  (*We use the fact that s is a valid state*)
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.

  
  (*To use the H_valid_virt_s hypotesis, it's needed to*)
  (*consider cases of osi_ptaddr (pointer to PT), so the*)
  (*constrait to use H_valid_virt_s are met*)
  case (madd_eq osi_cpt_madd osi'_ptaddr).
  intro H_osi'_cptmadd.
  (*Case that osi_ptadd=pt_maddr; it means it changed*)
  cut(mp=(map_add vadd_eq osi_curr_map va ma)).
  (*relation between old and new current map*)
  intro H_mp_osi_curr_map.
  
  elim H_valid_virt_s with osi osi_cpt_madd 
    osi_curr_map osi_curr_page mp_va mp_ma;auto.
  intros pg' H_pg' H_acces.

  cut (pg=pg').
  intro H_pg_pg'.
  rewrite H_pg_pg'.
  apply H_acces.
  rewrite <- H_mpma_ma in H_pg'.
  rewrite <- H_pg' in H_pg.
  inversion H_pg;auto.
  rewrite H_mp_osi_curr_map in H_mp_va_ma.
  rewrite lemma_map_add_apply_distinct in H_mp_va_ma;auto.

  rewrite H_update_osi_cpt in H_osi'_pt_map.
  rewrite H_osi'_cptmadd in H_osi'_pt_map.
  rewrite lemma_map_add_apply in H_osi'_pt_map.
  inversion H_osi'_pt_map.
  rewrite H0 in H_osi'_pt_cont.
  inversion H_osi'_pt_cont;auto.

  (*case osi_ptaddr isn't pt_madd, it means osi_ptaddr doesn't change*)
  intro H_no_osi'_cptmadd.
  elim H_valid_virt_s with osi' osi'_ptaddr mp osi'_pt mp_va mp_ma;auto.
  intros pg' H_pg' H_acces.

  cut (pg=pg'). 
  intro H_pg_pg'.
  rewrite H_pg_pg'.
  apply H_acces.

  rewrite <- H_mpma_ma in H_pg'.
  rewrite <- H_pg' in H_pg.
  inversion H_pg;auto.
  
  rewrite H_update_osi_cpt in H_osi'_pt_map.
  rewrite lemma_map_add_apply_distinct in H_osi'_pt_map;auto.

  (*Proved accessibility when mp_ma=ma*)


  (*Case mp_ma isn't the new madd (ma)*)
  intro H_no_mpma_ma.


  (*the witness is the same page that mp_ma has assigned in s*)
  (*We use the fact that s is a velid state*)

  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.
  unfold valid_virtual_mapping in H_valid_virt_s.
  rewrite H_update_osi_cpt.
  rewrite lemma_map_add_apply_distinct;auto.

  
  (*To use the H_valid_virt_s hypotesis, it's needed to*)
  (*consider cases of osi_ptaddr (pointer to PT), so the*)
  (*constrait of H_valid_virt_s are met*)
  case (madd_eq osi'_ptaddr osi_cpt_madd).

  (*Case osi_ptaddr is not the one of the modified curr map*)
  2: intro H_no_osi_ptmadd.
  2: apply H_valid_virt_s with (osi:=osi') (pt_addr:=osi'_ptaddr) 
  (mp:=mp) (pt:=osi'_pt);auto.
  
  2: rewrite H_update_osi_cpt in H_osi'_pt_map.
  2: rewrite lemma_map_add_apply_distinct in H_osi'_pt_map;auto.

  (*Case osi_ptaddr is the one of the modified curr map*)
  intro H_osi'_cptmadd.

  (*mp is the current map updated*)

  cut(mp=(map_add vadd_eq osi_curr_map va ma)).
  intro H_mp_osi_curr_map.

  (*We consider cases of mp_va*)
  case (vadd_eq mp_va va).

  (*Case mp_va<>va*)
  2:intro H_no_mpva_va.
  2:apply H_valid_virt_s with (osi:=osi') 
    (pt_addr:=osi_cpt_madd) (mp:=osi_curr_map) (pt:=osi_curr_page);auto.
  2:rewrite H_osi'_osi;auto.
  2:rewrite H_mp_osi_curr_map in H_mp_va_ma.
  2:rewrite lemma_map_add_apply_distinct in H_mp_va_ma;auto.

  (*Case mp_va=va*)
  intro H_mpva_va.
  
  (*This case is absurd, because mp_va and va are the same*)
  (*and they can't have different images in the mp mapping*)
  cut (False).
  intro;contradiction.

  rewrite H_mp_osi_curr_map in H_mp_va_ma.
  rewrite <- H_mpva_va in H_mp_va_ma. 
  rewrite lemma_map_add_apply in H_mp_va_ma;auto.
  inversion H_mp_va_ma;auto.

  clear H_valid_virt_s.


  (*It is pending the proof mp and current map relation*)


  rewrite H_update_osi_cpt in H_osi'_pt_map.
  rewrite H_osi'_cptmadd in H_osi'_pt_map.
  rewrite lemma_map_add_apply in H_osi'_pt_map.
  inversion H_osi'_pt_map.
  rewrite H0 in H_osi'_pt_cont.
  inversion H_osi'_pt_cont;auto.


  (*Case case that the considered os in not osi*)
  intro H_osi'_ne_osi.

  intros H_osi'_cp_content H_osi'pt_owner.

  cut (osi'_ptaddr <> osi_cpt_madd).
  intro H_osi'pt_ne_osipt.
  rewrite H_update_osi_cpt in H_osi'_pt_map.
  rewrite lemma_map_add_apply_distinct in H_osi'_pt_map;auto.
  intros mp_va mp_ma H_mp_va_ma.


  (*Use the fact that s is a valid state*)
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.

  unfold valid_virtual_mapping in H_valid_virt_s.
  
  case (madd_eq mp_ma osi_cpt_madd).

  (*Case mp_ma <> current page maddr*)
  2: intros.
  2: rewrite H_update_osi_cpt.
  2: rewrite lemma_map_add_apply_distinct.
  2: apply H_valid_virt_s with (pt_addr:=osi'_ptaddr) 
    (mp:=mp) (pt:=osi'_pt); auto.
  2: auto.
 
  
  (*Case mp_ma=current page maddr is absurd*)
 
  intro H_mpma_osipt_abs.
 
  cut False.
  contradiction.

  elim H_valid_virt_s with osi' 
    osi'_ptaddr mp osi'_pt mp_va mp_ma; auto.
  intros pg H_pg_map H_owners.
  rewrite H_mpma_osipt_abs in H_pg_map.
  cut (pg = osi_curr_page).
  intro H_pg_osicp.
 
 
  rewrite H_pg_osicp in H_owners.
  (*We assume that the owner of osi's current page is osi*)
  cut (page_owned_by osi_curr_page = Os_PO osi).
  intro H_cp_owner_osi.
  
  unfold os_accessible in H_owners.
  destruct H_owners as (H_mp_va_acces,H_mp_va_no_acces).
  destruct (ctxt_vadd_accessible ctxt mp_va).
  cut (page_owned_by osi_curr_page = Os_PO osi').
  intro H_osi'_owns.
  rewrite H_cp_owner_osi in H_osi'_owns.
  inversion H_osi'_owns;auto.
  cut(true=true).
  intro H_true.
  apply (H_mp_va_acces H_true).
  trivial.
  cut (page_owned_by osi_curr_page = Hyp_PO).
  intro H_hyp_owns.
  rewrite H_cp_owner_osi in H_hyp_owns.
  inversion H_hyp_owns;auto.
  cut(false<>true).
  intro H_true.
  apply (H_mp_va_no_acces H_true).
  auto.
  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr.
  2:apply H_valid_s.
  unfold valid_current_page in H_valid_curr.
  unfold maddress_owner in H_valid_curr.
  unfold hypervisor_mapping_madd in H_ma_pa.
  destruct H_ma_pa as (p2m,H_p2m,H_ma_pa).
  elim H_valid_curr with osi os p2m.
  intro pt_madd'.
  intros H_pt_madd' H_pt'_owner.
  cut (pt_madd'=osi_cpt_madd).
  intro H_pt_pt'.
  rewrite H_pt_pt' in H_pt'_owner.
  destruct H_pt'_owner as (H_pt_ptype,H_pt_owner ).
  rewrite <- H_osi_curr_page in H_pt_owner.
  assumption.
  unfold os_current_vadd_mapping_madd in H_osi_cpt_madd.
  destruct H_osi_cpt_madd as (os', H_os'_osi,H_p2m').
  destruct H_p2m' as (p2m',H_p2m',H_os'_curr_page ).
  cut (os'=os).
  intro H_os'_os.
  cut (p2m'=p2m).
  intro H_p2m'_p2m.
  rewrite H_p2m'_p2m in H_os'_curr_page.
  rewrite H_os'_os in H_os'_curr_page.
  rewrite <- H_os'_curr_page in H_pt_madd'.
  inversion H_pt_madd';auto.
  unfold hypervisor_mapping in H_p2m'.
  rewrite <- H_p2m' in H_p2m.
  inversion H_p2m;auto.
  rewrite <- H_os'_osi in H_is_os.
  inversion H_is_os;auto.
  assumption.
  assumption.
  rewrite <- H_osi_curr_page in H_pg_map.
  inversion H_pg_map;auto.
  cut(maddress_owner (st_memory s')(osi_cpt_madd) 
    <> maddress_owner (st_memory s')(osi'_ptaddr)).
  intro H_owner.
  unfold not in H_owner.
  intro H_pt_osi'.
  rewrite H_pt_osi' in H_owner.
  apply H_owner.
  trivial.
  cut (maddress_owner (st_memory s') osi_cpt_madd = Os_PO osi).
  intro H_osi_cpt_madd_owner.
  rewrite H_osi_cpt_madd_owner.
  unfold maddress_owner.
  rewrite H_s'; simpl.
  rewrite <- H_osi'_pt_map.
  rewrite H_osi'pt_owner.
  intro H_Absurd; inversion H_Absurd; auto.
  
  unfold maddress_owner.
  rewrite H_s'; simpl.
  rewrite H_update_osi_cpt.
  rewrite lemma_map_add_apply.
  simpl.

  (*The owner of the osi_curr_page is osi*)
  (*We can use valid_currant_page to prove that*)
  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr.
  2:apply H_valid_s.
  
  unfold valid_current_page in H_valid_curr.

  unfold hypervisor_mapping_madd in H_ma_pa.
  destruct H_ma_pa as (p2m,H_p2m,H_ma_pa).
  
  elim H_valid_curr with osi os p2m.
  intro pt_madd'.
  intros H_ptmadd' H_ptmadd'_prop.
  cut (pt_madd'=osi_cpt_madd).
  intro H_pt_pt'.

  rewrite H_pt_pt' in H_ptmadd'. 
  rewrite H_pt_pt' in H_ptmadd'_prop.
  unfold maddress_owner in H_ptmadd'_prop.
  rewrite <- H_osi_curr_page in H_ptmadd'_prop.
  apply H_ptmadd'_prop.
  unfold os_current_vadd_mapping_madd in H_osi_cpt_madd.
  destruct H_osi_cpt_madd as (os',H_os'_osi,H_p2m').
  destruct H_p2m' as (p2m',H_p2m',H_os'_curr_pt).
  cut (os'=os).
  intro H_os'_os.
  cut (p2m'=p2m).
  intro H_p2m'_p2m.
  rewrite H_p2m'_p2m in H_os'_curr_pt.
  rewrite H_os'_os in H_os'_curr_pt.
  rewrite <- H_ptmadd' in H_os'_curr_pt.
  inversion H_os'_curr_pt;auto.
  unfold hypervisor_mapping in H_p2m'.
  rewrite <- H_p2m' in H_p2m.
  inversion H_p2m;auto.
  rewrite <- H_os'_osi in H_is_os.
  inversion H_is_os;auto.
  assumption.
  assumption.
Qed.


(*Lemma that proves valid_virtual_mapping constraint for new hyper action*)
      
Lemma valid_virtual_mapping_new_hyper: 

  forall va ma s s',
   valid_state ctxt s -> 
   pre_new_hyper va ma ctxt s ->
   post_new_hyper va ma s s' ->
   valid_virtual_mapping (st_memory s') ctxt.
Proof.

  intros va ma s s' H_valid_s H_pre_new H_post_new.


  unfold post_new_hyper in H_post_new.
  destruct H_post_new as (aos, H_is_aos,H_post_new).
  destruct H_post_new as (cur_pt_madd, H_cur_pt_madd, H_post_new).
  destruct H_post_new as (newmem, H_mem_pt_newmem, H_s').
  rewrite H_s'; simpl.

  (*We use the post-conditions, so the relation between*)
  (*s and s' component can be established*)

  (*More hypothesis, about ma pa and va extracted from preconditions*)
  unfold pre_new_hyper in H_pre_new.
  destruct H_pre_new as (H_hyper_running,H_pre_new).
  destruct H_pre_new as (H_ma_page_type,H_pre_new).
  destruct H_pre_new as (H_ma_owner,H_pre_new).
  destruct H_pre_new as (H_exists_p2m,H_no_acces_va).

  (*We extract the relation between memories*)
  unfold update_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem as (curr_page, H_pt_curr_page, H_upd_curr_pt).
  destruct H_upd_curr_pt as (curr_map, H_pt_curr_map,H_upd_curr_pt).

  (*Start Main Proof.*)

  unfold valid_virtual_mapping.
  intros osi osi_ptaddr mp osi_pt H_pt_map.

  (*Prove by cases, first on the os considered in *)
  (*valid_virtual_mapping, (osi). It can be the aos *)
  (*identifier, or not *)
  case (os_ident_eq osi (st_active_os s)).  


  (*Case osi=aos*)

  intro H_osi_aos.
  intros H_osi_pt_cont H_osi_pt_owner mp_va mp_ma H_mp_va_ma.

  (*The current page owner is aos. This is used several times*)
  assert (page_owned_by curr_page = Os_PO (st_active_os s)) 
    as H_owner_curr_pg.

  assert (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s))
    as H_valid_curr.
  apply H_valid_s.
  
  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  destruct H_cur_pt_madd as (aos',H_is_aos',H_cur_pt_madd ).
  assert (aos'=aos) as H_aos_aos'.
  rewrite <- H_is_aos' in H_is_aos.
  inversion H_is_aos;auto.
  rewrite H_aos_aos' in H_cur_pt_madd;clear aos' H_is_aos' H_aos_aos'.
  destruct H_cur_pt_madd as (p2m, H_p2m,H_cur_pt_madd).
  
  elim H_valid_curr with (st_active_os s) aos p2m;auto.
  intro cur_pt_madd'.
  intros H_cur_pt_madd'.
  cut (cur_pt_madd'=cur_pt_madd).
  intros H_pt_mad'_mad.
  rewrite H_pt_mad'_mad.
  clear H_pt_mad'_mad.
  intro H_cur_pt_mad_prop.
  destruct H_cur_pt_mad_prop as (H_cur_pt_mad_pt,H_cur_pt_mad_owner).
  unfold maddress_owner in H_cur_pt_mad_owner.
  rewrite <- H_pt_curr_page in H_cur_pt_mad_owner.
  assumption.
  rewrite <- H_cur_pt_madd in H_cur_pt_madd'.
  inversion H_cur_pt_madd';auto.


  (*Cases of mp_ma are considered, where mp_ma is the image 
  on mp, the mapping considered in valid_virtual_mapping.
  It can be the same of that being updated, current page 
  madd, or not*)
  case (madd_eq cur_pt_madd mp_ma).


  (*In case it is*)
  intro H_mpma_ptmadd.
 
  (*The corresponding page is the current page updated*) 
  exists (Page (PT_c (map_add vadd_eq curr_map va ma))
                     (page_owned_by curr_page)).

  (*Prove that is the corresponding in newmem*)
  rewrite H_upd_curr_pt.
  rewrite H_mpma_ptmadd.
  rewrite lemma_map_add_apply;auto.
  

  (*Now the accessibility constraint needs to be proved*)
  simpl.
  split.
  intro.
  rewrite H_osi_aos.
  assumption.


  (*To prove accessibility, cases of mp_va are considered*)
  case (vadd_eq va mp_va).

  (*In case mp_va=va*)
  intro H_mpva_va.
  
  assert (ma<>cur_pt_madd) as H_n_ma_currpt.
  unfold maddress_owner in H_ma_owner.
  elim lemma_page_type_exists_page with (st_memory s) ma;auto.
  intro ma_pg.
  intro H_ma_pg.
  rewrite <- H_ma_pg in H_ma_owner.
  unfold not.
  intro H_ma_curr.
  assert (ma_pg=curr_page) as H_mapg_currpg.
  rewrite H_ma_curr in H_ma_pg.
  rewrite <- H_pt_curr_page in H_ma_pg.
  inversion H_ma_pg;auto.
  rewrite H_mapg_currpg in H_ma_owner.
  rewrite  H_owner_curr_pg in H_ma_owner.
  inversion H_ma_owner.
  
  case (madd_eq osi_ptaddr cur_pt_madd).

  intro H_osipt_mpma.

  (*This case is absurd*)
  cut (ma=cur_pt_madd).
  intro H_ma_currpt.
  contradiction.

  rewrite H_mpma_ptmadd.

  rewrite H_upd_curr_pt in H_pt_map.
  rewrite <- H_osipt_mpma in H_pt_map.
  rewrite lemma_map_add_apply in H_pt_map;auto.
  inversion H_pt_map.
  rewrite H0 in H_osi_pt_cont.
  inversion H_osi_pt_cont.
  rewrite H1 in H_mp_va_ma.
  rewrite H_mpva_va in H_mp_va_ma.
  rewrite lemma_map_add_apply in H_mp_va_ma;auto.
  inversion H_mp_va_ma;auto.
  

  (*When ois_ptaddr <> curr_pt_madd*)
  intro H_osipt_no_mpma.
  
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.
  
  
  elim H_valid_virt_s with osi osi_ptaddr mp osi_pt mp_va mp_ma;auto.
  intros pg' H_pg' H_acces.

  cut (curr_page=pg'). 
  intro H_pg_pg'.
  rewrite H_pg_pg'.
  apply H_acces.

  rewrite <- H_mpma_ptmadd in H_pg'.
  rewrite <- H_pt_curr_page in H_pg'.
  inversion H_pg';auto.
  
  rewrite H_upd_curr_pt in H_pt_map.
  rewrite lemma_map_add_apply_distinct in H_pt_map;auto.


  (*In case mp_va is not the same that va*)
  intro H_no_mpva_va.
  
  (*We use the fact that s is a valid state*)
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.

  (*To use the H_valid_virt_s hypotesis, it's needed to*)
  (*consider cases of osi_ptaddr (pointer to PT), so the*)
  (*constraint to use H_valid_virt_s are met*)
  case (madd_eq cur_pt_madd osi_ptaddr).
  intro H_osi_ptmadd.

  (*Case that osi_ptadd=pt_maddr; it means it changed*)
  cut(mp=(map_add vadd_eq curr_map va ma)).
  (*relation between old and new current map*)
  intro H_mp_curr_map.
  
  elim H_valid_virt_s with (st_active_os s) cur_pt_madd curr_map curr_page mp_va mp_ma;auto.
  intros pg H_pg H_acces.
  
  (*If the considered pointer a PT is cur_pt_madd,*)
  (*then the corresponding page is the current page*)
  cut (pg=curr_page).
  intro H_pg_curr.
  rewrite <- H_pg_curr.
  apply H_acces.
  rewrite <- H_mpma_ptmadd in H_pg.
  rewrite <- H_pt_curr_page in H_pg.
  inversion H_pg;auto.
  rewrite H_mp_curr_map in H_mp_va_ma.
  rewrite lemma_map_add_apply_distinct in H_mp_va_ma;auto.
  rewrite H_upd_curr_pt in H_pt_map.
  rewrite H_osi_ptmadd in H_pt_map.
  rewrite lemma_map_add_apply in H_pt_map.
  inversion H_pt_map.
  rewrite H0 in H_osi_pt_cont.
  inversion H_osi_pt_cont;auto.


  (*Case the pointer to PT (osi_ptaddr) is not osi_cpt_madd,*)
  (*that means its content doesn't change*)
  intro H_no_osi_ptmadd.
  elim H_valid_virt_s with osi osi_ptaddr mp osi_pt mp_va mp_ma;auto.
  intros pg H_pg H_acces.

  cut (pg=curr_page). 
  intro H_pg_curr.
  rewrite <- H_pg_curr.
  apply H_acces.
 
  rewrite <- H_mpma_ptmadd in H_pg.
  rewrite <- H_pt_curr_page in H_pg.
  inversion H_pg;auto.
  
  rewrite H_upd_curr_pt in H_pt_map.
  rewrite lemma_map_add_apply_distinct in H_pt_map;auto.


  (*Case that mp_ma is not the address of the current map*)


  intro H_no_mpma_pt.

  (*We discriminate cases of mp_ma, whether it is the*)
  (*new one (ma)or not*)
  case (madd_eq ma mp_ma).

  (*Case mp_ma is the new one*)
  intro H_mpma_ma.

  (*The corresponding page is inferred from the pre-conditions*)
  elim lemma_page_type_exists_page with (st_memory s) ma;auto.
  intro pg.
  intro H_pg.
  
  exists pg.

  rewrite H_upd_curr_pt.
  rewrite lemma_map_add_apply_distinct;auto.
  rewrite <- H_mpma_ma;auto.

  (*Now the accessibility constraint needs to be proved when mp_ma=ma*)
  
  split.

  (*We consider cases of mp_va (pre-image in mp),  whether it is
  the same of va or not*)
  case (vadd_eq va mp_va).

  (*case mp_va=va*)
  intro H_mpva_va.
  rewrite <- H_mpva_va.
  intro.
  contradiction.

  (*case mp_va<>va*)
  intro H_no_mpva_va.


  (*We use the fact that s is a valid state*)
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.


  (*To use the H_valid_virt_s hypotesis, it's needed to*)
  (*consider cases of osi_ptaddr (pointer to PT), so the*)
  (*constrait to use H_valid_virt_s are met*)
  case (madd_eq cur_pt_madd osi_ptaddr).
  intro H_osi_ptmadd.

  (*Case that osi_ptadd=pt_maddr; it means it changed*)
  cut(mp=(map_add vadd_eq curr_map va ma)).
  (*relation between old and new current map*)
  intro H_mp_curr_map.
  
  elim H_valid_virt_s with (st_active_os s) cur_pt_madd curr_map curr_page mp_va mp_ma;auto.
  intros pg' H_pg' H_acces.

  cut (pg=pg'). 
  intro H_pg_pg'.
  rewrite H_pg_pg'.
  rewrite H_osi_aos.
  apply H_acces.
  rewrite <- H_mpma_ma in H_pg'.
  rewrite <- H_pg' in H_pg.
  inversion H_pg;auto.
  rewrite H_mp_curr_map in H_mp_va_ma.
  rewrite lemma_map_add_apply_distinct in H_mp_va_ma;auto.
  rewrite H_upd_curr_pt in H_pt_map.
  rewrite H_osi_ptmadd in H_pt_map.
  rewrite lemma_map_add_apply in H_pt_map.
  inversion H_pt_map.
  rewrite H0 in H_osi_pt_cont.
  inversion H_osi_pt_cont;auto.



  (*case osi_ptaddr isn't pt_madd, it means osi_ptaddr doesn't change*)
  intro H_no_osi_ptmadd.
  elim H_valid_virt_s with osi osi_ptaddr mp osi_pt mp_va mp_ma;auto.
  intros pg' H_pg' H_acces.

  cut (pg=pg'). 
  intro H_pg_pg'.
  rewrite H_pg_pg'.
  apply H_acces.

  rewrite <- H_mpma_ma in H_pg'.
  rewrite <- H_pg' in H_pg.
  inversion H_pg;auto.
  
  rewrite H_upd_curr_pt in H_pt_map.
  rewrite lemma_map_add_apply_distinct in H_pt_map;auto.


  (*We use pre-conditions to prove this*)
  intro.
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_pg in H_ma_owner;auto.

  (*Proved accessibility when mp_ma=ma*)


  (*Case mp_ma isn't the new madd (ma)*)
  intro H_no_mpma_ma.

  (*the witness is the same page that mp_ma has assigned in s*)
  (*We use the fact that s is a velid state*)
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.
  unfold valid_virtual_mapping in H_valid_virt_s.
  rewrite H_upd_curr_pt.
  rewrite lemma_map_add_apply_distinct;auto.


  (*To use the H_valid_virt_s hypotesis, it's needed to*)
  (*consider cases of osi_ptaddr (pointer to PT), so the*)
  (*constrait of H_valid_virt_s are met*)
  case (madd_eq cur_pt_madd osi_ptaddr).


  (*Case osi_ptaddr is not the one of the modified curr map*)
  2: intro H_no_osi_ptmadd.
  2: apply H_valid_virt_s with (osi:=osi) (pt_addr:=osi_ptaddr) 
  (mp:=mp) (pt:=osi_pt);auto.
  
  2: rewrite H_upd_curr_pt in H_pt_map.
  2: rewrite lemma_map_add_apply_distinct in H_pt_map;auto.

  
  (*Case osi_ptaddr is the one of the modified curr map*)
  intro H_osi_ptmadd.

  (*mp is the current map updated*)
  cut(mp=(map_add vadd_eq curr_map va ma)).
  intro H_mp_curr_map.

  (*We consider cases of mp_va*)
  case (vadd_eq mp_va va).

  (*Case mp_va<>va*)
  2:intro H_no_mpva_va.
  2:apply H_valid_virt_s with (osi:=osi) (pt_addr:=cur_pt_madd) (mp:=curr_map) (pt:=curr_page);auto.
  2:rewrite H_osi_aos;auto.
  2:rewrite H_mp_curr_map in H_mp_va_ma.
  2:rewrite lemma_map_add_apply_distinct in H_mp_va_ma;auto.

 
  (*Case mp_va=va*)
  intro H_mpva_va.

  (*This case is absurd, because mp_va and va are the same*)
  (*and they can't have different images in the mp mapping*)

  cut (False).
  intro;contradiction.

  rewrite H_mp_curr_map in H_mp_va_ma.
  rewrite <- H_mpva_va in H_mp_va_ma. 
  rewrite lemma_map_add_apply in H_mp_va_ma;auto.
  inversion H_mp_va_ma;auto.


  clear H_valid_virt_s.

  (*It is pending the proof mp and current map relation*)

  rewrite H_upd_curr_pt in H_pt_map.
  rewrite H_osi_ptmadd in H_pt_map.
  rewrite lemma_map_add_apply in H_pt_map.
  inversion H_pt_map.
  rewrite H0 in H_osi_pt_cont.
  inversion H_osi_pt_cont;auto.

 
  (*Case case that the considered os in not aos*)


  intro H_osi_ne_aos.

  intros H1 H2.

  cut (osi_ptaddr <> cur_pt_madd).
  intro.
  rewrite H_upd_curr_pt in H_pt_map.
  rewrite lemma_map_add_apply_distinct in H_pt_map.
  intros.

  (*Use the fact that s is a valid state*)
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.

  unfold valid_virtual_mapping in H_valid_virt_s.
  
  case (madd_eq ma0 cur_pt_madd).

  (*Case mp_ma <> current page maddr*)
  2: intros.
  2: rewrite H_upd_curr_pt.
  2: rewrite lemma_map_add_apply_distinct.
  2: apply H_valid_virt_s with (pt_addr:=osi_ptaddr) (mp:=mp) (pt:=osi_pt); auto.
  2: auto.
 
  (*Case mp_ma=current page maddr is Absurd*)
  
  intros. 
  cut False.
  contradiction.

  elim H_valid_virt_s with osi osi_ptaddr mp osi_pt va0 ma0; auto.
  intros pg H_pg_map H_owners.
  rewrite e in H_pg_map.
  cut (pg = curr_page).
  intro.
 
 
  rewrite H3 in H_owners.
  (*We assume that the owner of osi's current page is aos*)
  cut (page_owned_by curr_page = Os_PO (st_active_os s)).
  intro.
  
  unfold os_accessible in H_owners.
  destruct H_owners.
  destruct (ctxt_vadd_accessible ctxt va0).
  cut (page_owned_by curr_page = Os_PO osi).
  intro H_osi_owns.
  rewrite H4 in H_osi_owns.
  inversion H_osi_owns;auto.
  cut(true=true).
  intro H_true.
  apply (H5 H_true).
  trivial.
  cut (page_owned_by curr_page = Hyp_PO).
  intro H_hyp_owns.
  rewrite H4 in H_hyp_owns.
  inversion H_hyp_owns;auto.
  cut(false<>true).
  intro H_true.
  apply (H6 H_true).
  auto.
  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr.
  2:apply H_valid_s.
  unfold valid_current_page in H_valid_curr.
  unfold maddress_owner in H_valid_curr.

  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  destruct H_cur_pt_madd as (aos',H_is_aos',H_cur_pt_madd ).
  assert (aos'=aos) as H_aos_aos'.
  rewrite <- H_is_aos' in H_is_aos.
  inversion H_is_aos;auto.
  rewrite H_aos_aos' in H_cur_pt_madd;clear aos' H_is_aos' H_aos_aos'.
  destruct H_cur_pt_madd as (p2m, H_p2m,H_cur_pt_madd).
  
 

  elim H_valid_curr with (st_active_os s) aos p2m.
  intro pt_madd'.
  intros H_pt_madd' H_pt'_owner.
  cut (pt_madd'=cur_pt_madd).
  intro H_pt_pt'.
  rewrite H_pt_pt' in H_pt'_owner.
  destruct H_pt'_owner.
  rewrite <- H_pt_curr_page in H5.
  assumption.

  rewrite <- H_cur_pt_madd in H_pt_madd'.
  inversion H_pt_madd';auto.
  assumption.
  assumption.
  rewrite <- H_pt_curr_page in H_pg_map.
  inversion H_pg_map;auto.
  auto.
  cut(maddress_owner (st_memory s')(cur_pt_madd) <> maddress_owner (st_memory s')(osi_ptaddr)).
  intro H_owner.
  unfold not in H_owner.
  unfold not.
  intro H_pt_osi.
  rewrite H_pt_osi in H_owner.
  apply H_owner.
  trivial.
  cut (maddress_owner (st_memory s') cur_pt_madd = Os_PO (st_active_os s)).
  intro H_pt_madd_owner.
  rewrite H_pt_madd_owner.
  unfold maddress_owner.
  rewrite H_s'; simpl.
  rewrite <- H_pt_map.
  rewrite H2.
  unfold not.
  intro Absurd.
  inversion Absurd.
  auto.
  
  unfold maddress_owner.
  rewrite H_s'; simpl.
  rewrite H_upd_curr_pt.
  rewrite lemma_map_add_apply.
  simpl.

  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr.
  2:apply H_valid_s.
  
  unfold valid_current_page in H_valid_curr.

  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  destruct H_cur_pt_madd as (aos',H_is_aos',H_cur_pt_madd ).
  assert (aos'=aos) as H_aos_aos'.
  rewrite <- H_is_aos' in H_is_aos.
  inversion H_is_aos;auto.
  rewrite H_aos_aos' in H_cur_pt_madd;clear aos' H_is_aos' H_aos_aos'.
  destruct H_cur_pt_madd as (p2m, H_p2m,H_cur_pt_madd).
  
  elim H_valid_curr with (st_active_os s) aos p2m.
  intro pt_madd'.
  intros.
  cut (pt_madd'=cur_pt_madd).
  intro H_pt_pt'.

  rewrite H_pt_pt' in H. 
  rewrite H_pt_pt' in H0.
  unfold maddress_owner in H0.
  rewrite <- H_pt_curr_page in H0.
  apply H0.
  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  destruct H_cur_pt_madd.

  inversion H;auto.
  assumption.
  assumption.
Qed.


  (* Lemma that proves valid_current_page constraint
  for new trusted and untrusted actions*)

Lemma valid_current_page_new_trusted_untrusted :

  forall osi va pa s s',
  valid_state ctxt s ->
  post_all_new osi va pa None s s' ->

  valid_current_page (st_hypervisor s') (st_memory s') (st_oss s').

Proof.

  intros osi va pa  s s' H_valid_s H_post_new.

  (*Use the post-conditions, so the relation between*)
  (*s and s' component can be established*)

  destruct H_post_new as (ma,H_ma_pa,H_post_new).
  destruct H_post_new as (osi_cpt_madd, H_osi_cpt_madd,H_post_new).
  destruct H_post_new as (newmem, H_mem_pt_newmem, H_post_new).
  destruct H_post_new as (os, H_is_os,H_s').
  rewrite H_s'; simpl.

  unfold valid_current_page.
  intros osi' os' H_osi'_os' p2m' H_os'_p2m'.

  (*We divide the proof in cases, whether osi' is the same that osi or not*)
  case (os_ident_eq osi' osi).
  intro H_osi'_osi.

  assert (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s))
  as H_valid_curr_pt_s.
  apply H_valid_s.

  rewrite H_osi'_osi.

  unfold os_current_vadd_mapping_madd in H_osi_cpt_madd.
  destruct (H_osi_cpt_madd) as (os'',H_os'',H_os''_prop).

  assert (os''=os) as H_os''_os.
  rewrite <- H_is_os in H_os'';inversion H_os'';auto.
  rewrite H_os''_os in H_os''_prop;clear os'' H_os'' H_os''_os.
  destruct H_os''_prop as (p2m,H_p2m,H_osi_cpt).

  assert (p2m'=p2m) as H_p2m'_p2m.
  rewrite H_osi'_osi in H_os'_p2m'.
  unfold hypervisor_mapping in H_p2m.
  rewrite <- H_os'_p2m' in H_p2m;inversion H_p2m;auto.

  assert (os'=OS None (os_current_page os)) as H_os'.
  rewrite H_osi'_osi in H_osi'_os'.
  rewrite lemma_map_add_apply in H_osi'_os'.
  inversion H_osi'_os';auto.

  destruct  (H_valid_curr_pt_s) with (osi:=osi) (system:=os) (p2m:=p2m')
    as (osi_cpt',H_cpt',H_cpt_prop);auto.

  rewrite H_p2m'_p2m;auto.

  assert (osi_cpt'=osi_cpt_madd) as H_cpt'_cpt.
  rewrite H_p2m'_p2m in H_cpt'.
  rewrite <- H_osi_cpt in H_cpt'.
  inversion H_cpt';auto.
  rewrite H_cpt'_cpt in H_cpt_prop.
  exists osi_cpt_madd;auto.

  rewrite <- H_cpt'_cpt; rewrite H_os';auto.

  unfold update_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem as
   (osi_curr_page, H_osi_curr_page, H_mem_pt_newmem).
  destruct H_mem_pt_newmem as
   (osi_curr_map, H_osi_curr_map,H_update_osi_cpt).

  split.

  rewrite H_update_osi_cpt;unfold page_type.
  rewrite lemma_map_add_apply;auto.

  unfold maddress_owner.
  unfold maddress_owner in H_cpt_prop.
  rewrite H_update_osi_cpt.
  rewrite <- H_osi_curr_page in H_cpt_prop.
  rewrite lemma_map_add_apply;simpl.
  apply H_cpt_prop.

  (*In case that osi' <> osi, the current page is the*)
  (*same that osi' has in state s*)
  intro H_no_osi'_osi.

  assert (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s))
    as H_valid_curr_pt_s.
  apply H_valid_s.


  unfold valid_current_page in H_valid_curr_pt_s.

  elim H_valid_curr_pt_s with osi' os' p2m'.
  intros osi'_cpt H_osi'_cpt H_osi'_cpt_prop.
  exists osi'_cpt;auto.

  unfold update_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem 
    as (osi_curr_page, H_osi_curr_page, H_mem_pt_newmem).
  destruct H_mem_pt_newmem 
    as (osi_curr_map, H_osi_curr_map,H_update_osi_cpt).
  
  rewrite H_update_osi_cpt.
  unfold page_type.
  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct.
  apply H_osi'_cpt_prop.

  (*Pointer to current page is differen, in osi and osi'*)
  (*We prove it based on their different owners*)
  
  cut (maddress_owner (st_memory s) osi_cpt_madd = Os_PO osi).
  intro H_cpt_madd_owner.
  intro H_contrad.
  rewrite H_contrad in H_cpt_madd_owner.
  destruct H_osi'_cpt_prop as (H_osi'_cpt_ptype,H_osi'_cpt_owner).
  rewrite H_cpt_madd_owner in H_osi'_cpt_owner.
  inversion H_osi'_cpt_owner;auto.

  (*We can use valid_virtual_mapping on state s to get the owner*)
  (*of the current page pointer*)
  unfold os_current_vadd_mapping_madd in H_osi_cpt_madd. 
  destruct H_osi_cpt_madd  as (os'',H_is_os'',H_p2m).
  assert (os''=os) as H_os''_os.
  rewrite <- H_is_os  in H_is_os'';inversion H_is_os'';auto.
  rewrite H_os''_os in H_p2m;clear os'' H_is_os'' H_os''_os.
  destruct H_p2m as (osi_p2m, H_p2m,H_osi_cpt_madd).

  elim H_valid_curr_pt_s with osi os osi_p2m;auto.
  intro pt_madd'.
  intros H_ptmadd' H_ptmadd'_prop.
  cut (pt_madd'=osi_cpt_madd).
  intro H_same_pt_madd.
  rewrite <- H_same_pt_madd.
  apply H_ptmadd'_prop.
  
  rewrite <- H_osi_cpt_madd in H_ptmadd';inversion H_ptmadd';auto.
  
  (*The new OSS doen't change for osi'*)
  rewrite lemma_map_add_apply_distinct in H_osi'_os';auto.
  auto.
Qed.

(*Lemma that proves valid_current_page constraint for new hyper action*)

Lemma valid_current_page_new_hyper: 

  forall va ma s s',
   valid_state ctxt s -> 
   pre_new_hyper va ma ctxt s ->
   post_new_hyper va ma s s' ->
   valid_current_page (st_hypervisor s') (st_memory s') (st_oss s').
Proof.
  intros va ma s s' H_valid_s H_pre_new H_post_new.


  unfold post_new_hyper in H_post_new.
  destruct H_post_new as (aos, H_is_aos,H_post_new).
  destruct H_post_new as (cur_pt_madd, H_cur_pt_madd, H_post_new).
  destruct H_post_new as (newmem, H_mem_pt_newmem, H_s').


  (*Use the post-conditions, so the relation between*)
  (*s and s' component can be established*)
  rewrite H_s'; simpl.

  unfold valid_current_page.
  intros osi system H_osi_system p2m H_sys_p2m.

  (*We divide the proof in cases, whether osi' is the same that osi or not*)
  case (os_ident_eq osi (st_active_os s)).
  intro H_case_aos.
  exists cur_pt_madd.
  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  elim H_cur_pt_madd.
  intros system' H_system'_osi H_p2m'.
  destruct H_p2m' as (p2m',H_p2m',H_system'_cpt).
  cut (system=system').
  intro H_is_sys.
  cut (p2m'=p2m).
  intro H_is_p2m.
  rewrite H_is_p2m in H_system'_cpt.
  rewrite H_is_sys.
  assumption.
  
  unfold  hypervisor_mapping in H_p2m'.
  rewrite <- H_case_aos in H_p2m'.
  rewrite <- H_sys_p2m in H_p2m'.
  injection H_p2m'; intro; auto.
  rewrite H_case_aos in H_osi_system.
  rewrite <- H_osi_system in H_system'_osi.
  inversion H_system'_osi;auto.
  
  (*We extract the newmem definition*)
  generalize H_mem_pt_newmem.
  intro H_update_curr_pt.
  unfold update_pt_page in H_update_curr_pt.
  destruct H_update_curr_pt as (current_page, H_pt_current_page, H_update_curr_pt).
  destruct H_update_curr_pt as (current_map, H_pt_curr_map,H_update_curr_pt).
  auto.
  split.
  rewrite H_update_curr_pt.
  unfold page_type.
  rewrite lemma_map_add_apply;auto.

  unfold maddress_owner.
  rewrite H_update_curr_pt.
  rewrite lemma_map_add_apply.
  simpl.
  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr_pt_s.
  2:apply H_valid_s.


  unfold valid_current_page in H_valid_curr_pt_s.

  elim H_valid_curr_pt_s with osi aos p2m.
  intros pt_madd' H_osi_ptmadd' H_ptmadd'_prop.
  destruct H_ptmadd'_prop as (H_ptmadd'_ptype,H_ptmadd'_owner).
  cut (pt_madd'=cur_pt_madd).
  intro H_ptmadd'_cptmadd.
  unfold maddress_owner in H_ptmadd'_owner.
  rewrite H_ptmadd'_cptmadd in H_ptmadd'_owner.
  rewrite <- H_pt_current_page in H_ptmadd'_owner.
  assumption.
  clear H_valid_curr_pt_s.

  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  destruct H_cur_pt_madd as (aos',H_aos',H_p2m').
  destruct H_p2m' as (p2m',H_p2m',H_cur_pt_madd).
  cut (p2m'=p2m).
  intro H_p2m'_p2m.
  cut (os_current_page aos'=os_current_page aos).
  intro H_aos'_aos_cpt.
  rewrite H_p2m'_p2m in H_cur_pt_madd.
  rewrite H_aos'_aos_cpt in H_cur_pt_madd.
  rewrite <- H_osi_ptmadd' in H_cur_pt_madd.
  inversion H_cur_pt_madd;auto.
  rewrite <- H_is_aos in H_aos'.
  inversion H_aos';auto.

  unfold hypervisor_mapping in H_p2m'.
  rewrite H_case_aos in H_sys_p2m.
  rewrite <- H_sys_p2m in H_p2m'.
  inversion H_p2m';auto.
  rewrite H_case_aos.
  assumption.
  assumption.

  (*In case that osi <> aos, the current page is the*)
  (*same that osi has in state s*)
  intro H_case_no_aos.

  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr_pt_s.
  2:apply H_valid_s.


  unfold valid_current_page in H_valid_curr_pt_s.

  elim H_valid_curr_pt_s with osi system p2m.
  intros osi_pt_madd H_osi_cpt_madd H_osi_pt_prop.
  exists osi_pt_madd;auto.
  split.

  (*We extract the newmem definition*)
  generalize H_mem_pt_newmem.
  intro H_update_curr_pt.
  unfold update_pt_page in H_update_curr_pt.
  destruct H_update_curr_pt as (current_page, H_pt_current_page, H_update_curr_pt).
  destruct H_update_curr_pt as (current_map, H_pt_curr_map,H_update_curr_pt).
  
  rewrite H_update_curr_pt.
  unfold page_type.
  rewrite lemma_map_add_apply_distinct.
  apply H_osi_pt_prop.

  (*Tey are different because they have different owners*)
  

  cut (maddress_owner (st_memory s) cur_pt_madd = Os_PO (st_active_os s)).
  intro H_pt_madd_owner.
  intro H_contrad.
  rewrite H_contrad in H_pt_madd_owner.
  destruct H_osi_pt_prop as (H_osi_pt_ptype,H_osi_pt_owner).
  rewrite H_pt_madd_owner in H_osi_pt_owner.
  inversion H_osi_pt_owner;auto.

  (*For this we can use valid_virtual_mapping on state s to get the owner*)
  (*of the current page pointer*)
  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  destruct H_cur_pt_madd as (aos',H_is_aos',H_cur_pt_madd ).
  assert (aos'=aos) as H_aos_aos'.
  rewrite <- H_is_aos' in H_is_aos.
  inversion H_is_aos;auto.
  rewrite H_aos_aos' in H_cur_pt_madd;clear aos' H_is_aos' H_aos_aos'.
  destruct H_cur_pt_madd as (active_p2m, H_p2m,H_cur_pt_madd).


  elim H_valid_curr_pt_s with (st_active_os s) aos active_p2m.
  intros pt_madd' H_ptmadd' H_ptmadd'_prop.
  cut (pt_madd'=cur_pt_madd).
  intro H_same_pt_madd.
  rewrite <- H_same_pt_madd.
  apply H_ptmadd'_prop.
  
  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  destruct H_cur_pt_madd.
  inversion H_ptmadd';auto.

  assumption.
  assumption.
  3:auto.
  
  (*We extract the newmem definition*)
  generalize H_mem_pt_newmem.
  intro H_update_curr_pt.
  unfold update_pt_page in H_update_curr_pt.
  destruct H_update_curr_pt as (current_page, H_pt_current_page, H_update_curr_pt).
  destruct H_update_curr_pt as (current_map, H_pt_curr_map,H_update_curr_pt).
  
  rewrite H_update_curr_pt.
  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct.
  apply H_osi_pt_prop.

  (*They ae different because they have different owners*)
  
  cut (maddress_owner (st_memory s) cur_pt_madd = Os_PO (st_active_os s)).
  intro H_pt_madd_owner.
  intro H_contrad.
  rewrite H_contrad in H_pt_madd_owner.
  destruct H_osi_pt_prop as (H_osi_pt_ptype,H_osi_pt_owner).
  rewrite H_pt_madd_owner in H_osi_pt_owner.
  inversion H_osi_pt_owner;auto.

  (*Again, we use valid_virtual_mapping on state s to get the owner*)
  (*of the current page pointer*)

  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  destruct H_cur_pt_madd as (aos',H_is_aos',H_cur_pt_madd ).
  assert (aos'=aos) as H_aos_aos'.
  rewrite <- H_is_aos' in H_is_aos.
  inversion H_is_aos;auto.
  rewrite H_aos_aos' in H_cur_pt_madd;clear aos' H_is_aos' H_aos_aos'.
  destruct H_cur_pt_madd as (active_p2m, H_p2m,H_cur_pt_madd).

  elim H_valid_curr_pt_s with (st_active_os s) aos active_p2m.
  intros pt_madd' H_ptmadd' H_ptmadd'_prop.
  cut (pt_madd'=cur_pt_madd).
  intro H_same_pt_madd.
  rewrite <- H_same_pt_madd.
  apply H_ptmadd'_prop.
  
  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  destruct H_cur_pt_madd.
  inversion H_ptmadd';auto.
  assumption.
  assumption.
  
  (*The new OSS doen't change for osi*)
  assumption.
Qed.


  (* Lemma that proves va_has_valid_pa constraint
  for new trusted and untrusted actions*)

Lemma va_has_valid_pa_new_trusted_untrusted :

  forall osi va pa s s',
  valid_state ctxt s ->
  post_all_new osi va pa None s s' ->

  va_has_valid_pa (st_memory s') 
    (st_hypervisor s') ctxt.

Proof.

  intros osi va pa  s s' H_valid_s H_post_new.

  (*Use the post-conditions, so the relation between*)
  (*s and s' component can be established*)

  destruct H_post_new as (ma,H_ma_pa,H_post_new).
  destruct H_post_new 
    as (osi_cpt_madd, H_osi_cpt_madd, H_post_new).
  destruct H_post_new 
    as (newmem, H_mem_pt_newmem, H_post_new).
  destruct H_post_new 
    as (os, H_is_os,H_s').
  simpl in H_s'.
  rewrite H_s'; simpl.
  
  (*Start of the main proof*)
  unfold va_has_valid_pa.
  intro osi'.

  (*We extract the newmem definition, used sereval times*)
  unfold update_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem as
    (current_page, H_pt_current_page, H_mem_pt_newmem).
  destruct H_mem_pt_newmem  as
    (current_map, H_pt_curr_map,H_newmem_def).

  (*We use several taimes that the property holds in s*)
  assert (va_has_valid_pa (st_memory s) 
    (st_hypervisor s) ctxt) as H_va_has_pa_in_s.
  apply H_valid_s.

  (*We extract some os components from post-conditions*)
  destruct H_osi_cpt_madd as (system, H_system_osi,H_p2m).
  destruct H_p2m as (p2m,H_p2m,H_osi_cpt_madd).


  (*We must consider cases of osi'*)
  case (os_ident_eq osi' osi).

  (*In case osi' is osi*)
  intro H_osi'_osi.

  intros osi_pt_padd osi_pt_addr osi_pt_page 
    osi_mp osi_va osi_ma.
  intros H_osi_pt_page_addr H_pt_padd_madd 
    H_pt_page_cont H_osi_ma_va H_osi_va.

  (*We discriminate cases of the considered osi_ma*)
  case (madd_eq osi_ma ma).

  (*in case the considered osi_ma is ma, the new one*)
  intro H_osima_ma.

  (*the corresponding padd is pa*)
  exists pa.
  rewrite H_osima_ma.
  rewrite H_osi'_osi;auto.

  (*In case the considered osi_ma isn't ma*)
  intro H_osima_no_ma.
  
  (*We need to discriminate cases ef osi_ptadd*)
  case (madd_eq osi_pt_addr osi_cpt_madd).
  intro H_osipt_osicpt.

  (*We establish the relation betwen the pages*)
  assert (osi_pt_page=
              (Page (PT_c (map_add vadd_eq current_map va ma))
                    (page_owned_by current_page))
         ) as H_osipage_def.
  rewrite H_newmem_def in H_osi_pt_page_addr.
  rewrite H_osipt_osicpt in H_osi_pt_page_addr.
  rewrite lemma_map_add_apply in H_osi_pt_page_addr.
  inversion H_osi_pt_page_addr;auto.

  (*We establish the relation betwen the mappings*)
  assert (osi_mp = (map_add vadd_eq current_map va ma)) as H_osi_mp_def.
  rewrite H_osipage_def in H_pt_page_cont.
  simpl in H_pt_page_cont.
  inversion H_pt_page_cont;auto.

  (*osi_va can't be the same that va*)
  assert (osi_va <> va) as H_osiva_no_va.

  intro H_osiva_va_abs.
  
  rewrite H_osi_mp_def in H_osi_ma_va.
  rewrite H_osiva_va_abs in H_osi_ma_va.
  rewrite lemma_map_add_apply in H_osi_ma_va.
  inversion H_osi_ma_va;contradiction.
  
  elim H_va_has_pa_in_s with osi (os_current_page system)
    osi_pt_addr current_page (current_map) osi_va osi_ma;auto.

  intros osi_pa H_osipa.
  exists osi_pa;auto.
  
  auto.
  unfold hypervisor_mapping in H_p2m.
  subst osi';  exists p2m;auto.

  unfold hypervisor_mapping_madd in H_osipa.
  destruct H_osipa as [p2m' H_p2m'_map H_osi_ma_map].
  rewrite <- H_p2m in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map; auto.

  rewrite H_osi_mp_def in H_osi_ma_va.
  rewrite lemma_map_add_apply_distinct 
    in H_osi_ma_va;auto.
  rewrite <- H_osipt_osicpt in H_pt_current_page; auto.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  rewrite H_osipt_osicpt; auto.
  rewrite H_osi_mp_def in H_osi_ma_va.
  rewrite lemma_map_add_apply_distinct in H_osi_ma_va; auto.

  (*In case the considered pt madd is not osi cpt madd*)  
  intro H_osipt_no_osicpt.

  elim H_va_has_pa_in_s with osi osi_pt_padd
    osi_pt_addr osi_pt_page osi_mp osi_va osi_ma;auto.

  intros osi_pa H_osipa.
  exists osi_pa;auto.

  rewrite H_newmem_def in H_osi_pt_page_addr.
  rewrite lemma_map_add_apply_distinct
    in H_osi_pt_page_addr;auto.
  rewrite H_osi'_osi;auto.
  rewrite H_newmem_def in H_osi_pt_page_addr.
  rewrite lemma_map_add_apply_distinct in H_osi_pt_page_addr; auto.
  unfold hypervisor_mapping_madd in *.
  destruct H_pt_padd_madd as [p2m' H_p2m'_map H_osi_pt_addr_map].
  exists p2m'; auto.
  rewrite <- H_osi'_osi; auto.

  (*In case osi' is not osi*)
  intro H_osi'_no_osi.
  
  intros osi'_pt_padd osi'_pt_addr osi'_pt_page 
    osi'_mp osi'_va osi'_ma.
  intros H_osi'_pt_page_addr H_pt_padd_madd 
    H_pt_page_cont H_osi'_ma_va H_osi'_va.


  (*We must prove that the considered madd "osi'_pt_addr"*)
  (*can not be the same that osi current page madd.*)

  (*We prove this based in the fact that they have*)
  (*different owners. In one case osi, in the other osi'*)

  (*The owner of osi_pt_addr is osi'*)
  assert (maddress_owner (st_memory s) osi'_pt_addr 
    = Os_PO osi') as H_osi'pt_owner.

  destruct H_pt_padd_madd as (osi'_p2m,H_osi'p2m',
    H_osi'pt_osi'padd).

  assert (valid_hypervisor (st_hypervisor s') 
    (st_memory s) ) as H_valid_hyper.
  rewrite H_s'; simpl.
  apply H_valid_s.

  elim H_valid_hyper with osi' osi'_pt_padd osi'_p2m;auto.

  intro H_exist_ma.
  destruct H_exist_ma as (osi'_pt_madd, H_osi'madd,
    H_osi'madd_prop).
  assert (osi'_pt_addr=osi'_pt_madd) as H_same_pt_madd.
  rewrite <- H_osi'pt_osi'padd in H_osi'madd.
  inversion H_osi'madd;auto.

  rewrite <- H_same_pt_madd in H_osi'madd_prop;auto.

  intro H_absurd.
  rewrite <- H_osi'pt_osi'padd in H_absurd.
  inversion H_absurd;auto.
  

  (*The owner of osi_cpt_madd is osi*)
  assert (maddress_owner (st_memory s) osi_cpt_madd 
    = Os_PO osi) as H_osicpt_owner.

  assert (valid_current_page (st_hypervisor s) (st_memory s)
     (st_oss s)) as H_valid_curr_s. 
  apply H_valid_s.
  
  unfold valid_current_page in H_valid_curr_s.
  elim H_valid_curr_s with osi system p2m.
  intros osi_cpt_addr H_osicptaddr H_osicpt_prop.
  assert (osi_cpt_addr=osi_cpt_madd) as H_osi_cpt.
  rewrite <- H_osi_cpt_madd in H_osicptaddr.
  inversion H_osicptaddr;auto.
  rewrite H_osi_cpt in H_osicpt_prop.
  apply H_osicpt_prop.

  apply H_system_osi.

  apply H_p2m.
  rewrite H_s'; simpl; auto.

  (*The considered madd osi'_pt_addr is different from*)
  (*osi_cpt_madd because they have different owners*)

  assert (osi'_pt_addr<>osi_cpt_madd) 
    as H_osi'pt_no_osicpt.
  
  intro H_osi'pt_osicpt_abs.
  
  generalize H_valid_s; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall
      H_valid_exmd H_valid_osexmd H_valid_untosexmd
      H_valid_hyper H_valid_virtmap H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with osi system p2m as [pt_addr H_pt_addr_map
    [H_ptype H_pt_addr_owner]]; auto.
  rewrite <- H_pt_addr_map in H_osi_cpt_madd.
  injection H_osi_cpt_madd; intro; subst osi_cpt_madd.
  rewrite H in H_osi'pt_owner.
  rewrite H_osi'pt_owner in H_pt_addr_owner.
  injection H_pt_addr_owner; intro; apply H_osi'_no_osi; auto.

  (*Once we know osi'_pt isn't the osi cpt, we use*)
  (*the fact that s is a valid state*)
  elim H_va_has_pa_in_s with osi' osi'_pt_padd
    osi'_pt_addr osi'_pt_page osi'_mp 
      osi'_va osi'_ma;auto.

  intros osi'_pa H_osi'pa.
  exists osi'_pa;auto.

  rewrite H_newmem_def in H_osi'_pt_page_addr.
  rewrite lemma_map_add_apply_distinct 
    in H_osi'_pt_page_addr;auto.
Qed.


  (* Lemma that proves va_has_valid_pa constraint
  for new hyper action*)

Lemma va_has_valid_pa_new_hyper :

  forall va ma s s',
  valid_state ctxt s ->
  pre_new_hyper va ma ctxt s ->
  post_new_hyper va ma s s' ->

  va_has_valid_pa (st_memory s') 
    (st_hypervisor s') ctxt.

Proof.

  intros va ma s s' H_valid_s H_pre_new H_post_new.

  (*Use the post-conditions, so the relation between*)
  (*s and s' component can be established*)

  destruct H_post_new 
    as (aos, H_is_aos,H_post_new).
  destruct H_post_new 
    as (cur_pt_madd, H_cur_pt_madd, H_post_new).
  destruct H_post_new 
    as (newmem, H_mem_pt_newmem, H_s').
  simpl in H_s'; rewrite H_s'; simpl.
  
  (*Start of main proof*)
  unfold va_has_valid_pa.
  intro osi.

  (*We extract the newmem definition, used sereval times*)
  unfold update_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem as
    (current_page, H_pt_current_page, H_mem_pt_newmem).
  destruct H_mem_pt_newmem  as
    (current_map, H_pt_curr_map,H_newmem_def).

  (*We use several taimes that the property holds in s*)
  assert (va_has_valid_pa (st_memory s) 
    (st_hypervisor s) ctxt) as H_va_has_pa_in_s.
  apply H_valid_s.

  (*We extract some os components from post-conditions*)
  destruct H_cur_pt_madd as (aos', H_is_aos',H_p2m).
  assert (aos'=aos) as H_aos'_aos.
  rewrite <- H_is_aos in H_is_aos'.
  inversion H_is_aos';auto.
  rewrite H_aos'_aos in H_p2m.
  clear aos' H_is_aos' H_aos'_aos.
  destruct H_p2m as (p2m,H_p2m,H_cur_pt_madd).

  (*Ans some othes properties from preconditions*)
  destruct H_pre_new as (H_hyper_running,H_pre_new).
  destruct H_pre_new as (H_ma_ptype,H_pre_new).
  destruct H_pre_new as (H_ma_owner,H_pre_new).
  destruct H_pre_new as (H_exists_p2m,H_va_no_access).

  (*We must consider cases of osi*)
  case (os_ident_eq osi (st_active_os s)).

  (*In case osi is aos*)
  intro H_osi_aos.

  intros aos_pt_padd aos_pt_addr aos_pt_page 
    aos_mp aos_va aos_ma.
  intros H_aos_pt_page_addr H_pt_padd_madd 
    H_pt_page_cont H_aos_ma_va H_aos_va.

  (*we know that aos_va can't be va, because one*)
  (*is accessible and the other sin't*)
  assert (aos_va <> va) as H_aosva_no_va.
  intro H_aosva_va_abs.
  rewrite H_aosva_va_abs in H_aos_va;contradiction.  

  (*We need to discriminate cases ef osi_ptadd*)
  case (madd_eq aos_pt_addr cur_pt_madd).
  intro H_aospt_aoscpt.

  (*We establish the relation betwen the pages*)
  assert (aos_pt_page=
              (Page (PT_c (map_add vadd_eq current_map va ma))
                    (page_owned_by current_page))
         ) as H_aospage_def.
  rewrite H_newmem_def in H_aos_pt_page_addr.
  rewrite H_aospt_aoscpt in H_aos_pt_page_addr.
  rewrite lemma_map_add_apply in H_aos_pt_page_addr.
  inversion H_aos_pt_page_addr;auto.

  (*We establish the relation betwen the mappings*)
  assert (aos_mp = (map_add vadd_eq current_map va ma)) as H_aos_mp_def.
  rewrite H_aospage_def in H_pt_page_cont.
  simpl in H_pt_page_cont.
  inversion H_pt_page_cont;auto.

  elim H_va_has_pa_in_s with (st_active_os s) (os_current_page aos)
    aos_pt_addr current_page current_map aos_va aos_ma;auto.

  intros aos_pa H_aospa.
  rewrite H_osi_aos; 
  exists aos_pa;auto.
  
  rewrite H_aospt_aoscpt;auto.

  rewrite H_aospt_aoscpt.
  exists p2m;auto.
  
  rewrite H_aos_mp_def in H_aos_ma_va.
  rewrite lemma_map_add_apply_distinct 
    in H_aos_ma_va;auto.

  (*In case the considered pt madd is not osi cpt madd*)  
  intro H_aospt_no_aoscpt.

  elim H_va_has_pa_in_s with (st_active_os s) aos_pt_padd
    aos_pt_addr aos_pt_page aos_mp aos_va aos_ma;auto.

  intros aos_pa H_osipa.
  rewrite H_osi_aos;
  exists aos_pa;auto.

  rewrite H_newmem_def in H_aos_pt_page_addr.
  rewrite lemma_map_add_apply_distinct
    in H_aos_pt_page_addr;auto.

  rewrite <- H_osi_aos;auto.


  (*In case osi is not aos*)
  intro H_osi_no_aos.
  
  intros osi_pt_padd osi_pt_addr osi_pt_page 
    osi_mp osi_va osi_ma.
  intros H_osi_pt_page_addr H_pt_padd_madd 
    H_pt_page_cont H_osi_ma_va H_osi_va.


  (*We must prove that the considered madd "osi_pt_addr"*)
  (*can not be the same that aos current page madd.*)

  (*We prove this based in the fact that they have*)
  (*different owners. In one case aos, in the other osi*)

  (*The owner of osi_pt_addr is osi*)
  assert (maddress_owner (st_memory s) osi_pt_addr 
    = Os_PO osi) as H_osipt_owner.

  destruct H_pt_padd_madd as (osi_p2m,H_osip2m,
    H_osipt_osipadd).

  assert (valid_hypervisor (st_hypervisor s') 
    (st_memory s) ) as H_valid_hyper.
  rewrite H_s'; simpl; apply H_valid_s.

  elim H_valid_hyper with osi osi_pt_padd osi_p2m;auto.

  intro H_exist_ma.
  destruct H_exist_ma as (osi_pt_madd, H_osimadd,
    H_osimadd_prop).
  assert (osi_pt_addr=osi_pt_madd) as H_same_pt_madd.
  rewrite <- H_osipt_osipadd in H_osimadd.
  inversion H_osimadd;auto.

  rewrite <- H_same_pt_madd in H_osimadd_prop;auto.

  intro H_absurd.
  rewrite <- H_osipt_osipadd in H_absurd.
  inversion H_absurd;auto.
  
  rewrite H_s'; auto.

  (*The owner of osi_cpt_madd is osi*)
  assert (maddress_owner (st_memory s) cur_pt_madd 
    = Os_PO (st_active_os s)) as H_aoscpt_owner.

  assert (valid_current_page (st_hypervisor s) (st_memory s)
     (st_oss s)) as H_valid_curr_s. 
  apply H_valid_s.
  
  unfold valid_current_page in H_valid_curr_s.
  elim H_valid_curr_s with (st_active_os s) aos p2m.
  intros aos_cpt_addr H_aoscptaddr H_aoscpt_prop.
  assert (aos_cpt_addr=cur_pt_madd) as H_aos_cpt.
  rewrite <- H_cur_pt_madd in H_aoscptaddr.
  inversion H_aoscptaddr;auto.
  rewrite H_aos_cpt in H_aoscpt_prop.
  apply H_aoscpt_prop.

  apply H_is_aos.

  apply H_p2m.

  (*Once we know osi_pt isn't the aos cpt, we use*)
  (*the fact that s is a valid state*)
  elim H_va_has_pa_in_s with osi osi_pt_padd
    osi_pt_addr osi_pt_page osi_mp 
      osi_va osi_ma;auto.

  intros osi_pa H_osipa.
  exists osi_pa;auto.

  rewrite H_newmem_def in H_osi_pt_page_addr.
  rewrite lemma_map_add_apply_distinct 
    in H_osi_pt_page_addr;auto.
  intro H_osipt_aoscpt_abs.

  (*The considered madd osi_pt_addr is different from*)
  (*cur_pt_madd because they have different owners*)
  inversion H_aoscpt_owner.
  rewrite H_osipt_aoscpt_abs in H0.
  rewrite H_osipt_owner in H0.
  injection H0; auto.
Qed.


Theorem New_untrusted_secure : 
  forall (s s':state),
  forall osi va pa,
  valid_state ctxt s ->
  valid_exec ctxt s (New_untrusted osi va pa) s'->
  valid_state ctxt s'.
Proof.


  intros s s' osi va pa H_valid_s H_valid_exec.
  generalize H_valid_exec; intro H_valid_exec'.
  inversion_clear H_valid_exec'.
  simpl in H.
  simpl in H0.
  unfold post_new_untrusted in H0.

  assert (post_all_new osi va pa None s s') as H_post_new;auto.

  (*Use the post-conditions, so the relation between*)
  (*s and s' component can be established*)

  destruct H0 as (ma,H_ma_pa,H0).
  destruct H0 as (osi_cpt_madd, H_osi_cpt_madd, H0).
  destruct H0 as (newmem, H_mem_pt_newmem, H0).
  destruct H0 as (os, H_is_os,H_s').

  simpl in H_s'; rewrite H_s'; simpl.

  (*Start of the main proof*)
  unfold valid_state; simpl.
  
  do 10 (try split); try apply H_valid_s.
  

  (*First Constraint*)
  unfold trusted_os_not_hypercall.
  intros trusted_osi trusted_os H_is_trusted_osi H_trusted_os_map.
  
  (*We discriminate the cases of trusted_osi*)
  case (os_ident_eq trusted_osi osi).
  
  (*In caso trusted_osi is osi (the osi id)*)
  intro H_case_osi.
  (*The OS is updated in the OSS*)
  rewrite H_case_osi in H_trusted_os_map.
  rewrite lemma_map_add_apply in H_trusted_os_map.
  injection H_trusted_os_map; intro H_trusted_os_value;
    rewrite H_trusted_os_value; auto.

  (*Case trusted_osi <> osi*)
  intro H_case_not_osi.
  rewrite lemma_map_add_apply_distinct in H_trusted_os_map; auto.

  (*We use the fact that s is a valid state*)
  (*so the constraint we are proving holds in s*)

  cut (trusted_os_not_hypercall (ctxt_oss ctxt) (st_oss s)).
  intro H_trusted_os_s.
  2:apply H_valid_s.
  2:auto.
  apply (H_trusted_os_s trusted_osi trusted_os H_is_trusted_osi); auto.

  (*Second Constraint*)

  (*We discriminate the cases of the consideter OS (identified by osi)*)
  case (os_ident_eq osi (st_active_os s)).
  
  (*In case osi=aos, the OS is given en the post-conditions*)
  intro H_case_aos.
  exists (OS None (os_current_page os)).
  rewrite H_case_aos;rewrite lemma_map_add_apply;auto.
  trivial.

  (*In case osi<>aos, we use the fact that s is valid*)
  intro H_case_no_aos.
  unfold running_os_not_hypercall.
  rewrite lemma_map_add_apply_distinct;auto.
  apply H_valid_s.
  
  (*Sixth Constraint*)
  replace (st_hypervisor s) with (st_hypervisor s').
  replace newmem with (st_memory s').
  apply ( (valid_hyper_new)  
      va ma osi osi_cpt_madd s s');auto.
  rewrite H_s'; auto.
  rewrite H_s'; auto.
  rewrite H_s'; auto.
  rewrite H_s'; auto.

  (*Seventh Constraint*)
  replace newmem with (st_memory s').
  apply ((valid_virtual_mapping_new_trusted_untrusted) 
    osi va pa s s');auto .
  destruct H as (H_untrusted,H_pre).
  destruct H_pre as (ma',H_ma'_pa,H_pre).
  exists ma';auto.
  repeat split;apply H_pre.
  rewrite H_s'; auto.

  (*Eighth Constraint*)
  
  replace (st_hypervisor s) with (st_hypervisor s').
  replace newmem with (st_memory s').
  replace (map_add os_ident_eq (st_oss s) osi (OS None (os_current_page os)))
    with (st_oss s').
  apply ( (valid_current_page_new_trusted_untrusted)
    osi va pa s s');auto.
  rewrite H_s'; auto.
  rewrite H_s'; auto.
  rewrite H_s'; auto.

  (*Tenth Constraint*)  
  replace (st_hypervisor s) with (st_hypervisor s').
  replace newmem with (st_memory s').
  apply ( (va_has_valid_pa_new_trusted_untrusted)
    osi va pa s s');auto.
  rewrite H_s'; auto.
  rewrite H_s'; auto.

  rewrite <- H_s'; apply exec_valid_cache_tlb
    with ctxt s (New_untrusted osi va pa); auto.

Qed.

Theorem New_trusted_secure : 
  forall (s s':state),
  forall va pa,
  valid_state ctxt s ->
  valid_exec ctxt s (New_trusted va pa) s'->
  valid_state ctxt s'.
Proof.
  intros s s' va pa H_valid_s H_valid_exec.
  generalize H_valid_exec; intro H_valid_exec'.
  inversion_clear H_valid_exec'.
  simpl in H.
  simpl in H0.
  
  assert (post_all_new (st_active_os s) va pa None s s') 
    as H_pos_new;auto.

  (*Use the post-conditions, so the relation between*)
  (*s and s' component can be established*)

  destruct H0 as (ma,H_ma_pa,H0).
  destruct H0 as (cur_pt_madd, H_cur_pt_madd, H0).
  destruct H0 as (newmem, H_mem_pt_newmem, H0).
  destruct H0 as (aos, H_is_aos,H_s').

  (*Start the main Proof*)
  unfold valid_state.
  
  (*First Constraint*)
  split.
  unfold trusted_os_not_hypercall.
  intros trusted_osi trusted_os H_is_trusted_osi H_trusted_os_map.
  rewrite H_s' in H_trusted_os_map; simpl in H_trusted_os_map.
  
  (*We discriminate the cases of trusted_osi*)
  case (os_ident_eq trusted_osi (st_active_os s)).
  
  (*In caso trusted_osi is aos_osi (the active os id)*)
  intro H_case_aos.
  (*The OS is updated in the OSS*)
  rewrite H_case_aos in H_trusted_os_map.
  rewrite lemma_map_add_apply in H_trusted_os_map.
  injection H_trusted_os_map; intro H_trusted_os_value;
    rewrite H_trusted_os_value; auto.

  (*Case trusted_osi is not active*)
  intro H_case_not_aos.
  rewrite lemma_map_add_apply_distinct in H_trusted_os_map; auto.

  (*We use the fact that s is a valid state*)
  (*so the constraint we are proving holds in s*)
  cut (trusted_os_not_hypercall (ctxt_oss ctxt) (st_oss s)).
  intro H_trusted_os_s.
  2:apply H_valid_s.
  apply (H_trusted_os_s trusted_osi trusted_os H_is_trusted_osi); auto.

  (*Second Constraint*)
  split.
  unfold running_os_not_hypercall.
  exists (OS None (os_current_page aos)).
  rewrite H_s'; simpl.
  rewrite lemma_map_add_apply;auto.
  trivial.

  (*Third Constraint*)
  split.
  rewrite H_s'; simpl.
  apply H_valid_s.
  
  (*Fourth Constraint*)
  split.
  rewrite H_s'.
  apply H_valid_s.
  
  (*Fifth Constraint*)
  split.
  rewrite H_s'; apply H_valid_s.

  (*Sixth Constraint*)
  split.
  
  apply ( (valid_hyper_new)  
    va ma (st_active_os s) cur_pt_madd s s'); auto.
  rewrite H_s'; auto.
  rewrite H_s'; auto.
  
  (*Seventh Constraint*)
  split.

  apply ((valid_virtual_mapping_new_trusted_untrusted) 
    (st_active_os s) va pa s s');auto .
  destruct H as (ma',H_ma'_pa,H_pre).
  exists ma';auto.
  repeat split;apply H_pre.

  (*Eighth Constraint*)
  split.
  apply ( (valid_current_page_new_trusted_untrusted)
    (st_active_os s) va pa s s');auto.

  (*Ninth Constraint*)
  split.
  rewrite H_s'; apply H_valid_s.



  (*Tenth Constraint*)  
  split.
  apply ( (va_has_valid_pa_new_trusted_untrusted)
    (st_active_os s) va pa s s');auto.
  apply exec_valid_cache_tlb
    with ctxt s (New_trusted va pa); auto.
Qed.


Theorem New_hyper_secure : 
  forall (s s':state),
  forall va ma,
  valid_state ctxt s ->
  valid_exec ctxt s (New_hyper va ma) s'->
  valid_state ctxt s'.
Proof.
  intros s s' va ma H_valid_s H_valid_exec.
  generalize H_valid_exec; intro H_valid_exec'.
  inversion_clear H_valid_exec'.
  simpl in H.

  (*Use the post-conditions, so the relation between*)
  (*s and s' component can be established*)

  simpl in H0.
  assert (post_new_hyper va ma s s') as H_post_new;auto.
  unfold post_new_hyper in H0.
  destruct H0 as (aos, H_is_aos,H0).
  destruct H0 as (cur_pt_madd, H_cur_pt_madd, H0).
  destruct H0 as (newmem, H_mem_pt_newmem, H_s').

  (*Start Main Proof*)
  unfold valid_state.
  
  (*First Constraint*)
  split.
  rewrite H_s'; apply H_valid_s.

  (*Second Constraint*)
  split.
  rewrite H_s'; apply H_valid_s.

  (*Third Constraint*)
  split.
  rewrite H_s'; apply H_valid_s.
  
  (*Fourth Constraint*)
  split.
  rewrite H_s'.
  apply H_valid_s.
  
  (*Fifth Constraint*)
  split.
  rewrite H_s'.
  apply H_valid_s.

  (*Sixth Constraint*)
  split.
  apply ( (valid_hyper_new)  
    va ma (st_active_os s) cur_pt_madd s s');auto.
  rewrite H_s'; auto.
  rewrite H_s'; auto.
  
  (*Seventh Constraint*)
  split.

  apply ((valid_virtual_mapping_new_hyper) 
    va ma s s');auto .

  (*Eighth Constraint*)
  split.

  apply ((valid_current_page_new_hyper) va ma s s');auto.
 
  (*Ninth Constraint*)
  split.
  rewrite H_s';apply H_valid_s.

  (*Tenth Constraint*)  
  split.
  apply ((va_has_valid_pa_new_hyper) va ma s s');auto.

  apply exec_valid_cache_tlb with ctxt s (New_hyper va ma); auto.
Qed.


(*******************)
(***Del is Secure***)
(*******************)

(*Lemmas*)

(*Lemma that proves valid_virtual_mapping for del trusted*)
(*and untrusted operations*)

Lemma valid_virtual_mapping_del_trusted_untrusted: 

  forall osi system p2m osi_pt_madd va newmem s s',
  valid_state ctxt s ->
    
    Value _ system = map_apply os_ident_eq (st_oss s) osi ->
    hypervisor_mapping osi (st_hypervisor s) p2m ->
    Value _ osi_pt_madd = map_apply padd_eq p2m (os_current_page system) ->
    os_accessible ctxt va ->
    pre_all_del osi_pt_madd va s  -> 
    drop_va_pt_page osi_pt_madd va (st_memory s) newmem ->
    st_memory s' = newmem ->
  valid_virtual_mapping (st_memory s') ctxt.
Proof.
  intros osi system p2m osi_pt_madd va newmem s s' H_valid_s.
  intros H_is_osi H_p2m H_osi_pt_madd H_va_acces H_pre_del H_mem_pt_newmem H_newmem.

  (*Use the post-conditions, so the relation between*)
  (*s and s' component can be established*)

  (*Hypothesis extracted from the pre-conditions*)
  unfold pre_all_del in H_pre_del.
  destruct H_pre_del as (osi_cur_map, H_osi_cur_map, H_pre).
  destruct H_pre as (ma,H_ma_va).

  destruct H_osi_cur_map as (osi_cur_page, H_osi_cur_pg, H_osi_cur_map).

  unfold drop_va_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem as (osi_cur_page', H_osi_cur_pg', H_mem_pt_newmem).
  assert (osi_cur_page=osi_cur_page') as H_cur_page_page'.
  rewrite <- H_osi_cur_pg in H_osi_cur_pg'.
  inversion H_osi_cur_pg';auto.
  rewrite <- H_cur_page_page' in H_mem_pt_newmem.
  clear H_cur_page_page' osi_cur_page' H_osi_cur_pg'.
  destruct H_mem_pt_newmem as (old_cur_map, H_old_cur_map,H_mem_pt_newmem ).
  assert (old_cur_map=osi_cur_map) as H_old_curr_map.
  rewrite <- H_osi_cur_map in H_old_cur_map.
  inversion H_old_cur_map;auto.
  rewrite H_old_curr_map in H_mem_pt_newmem.
  clear H_old_curr_map H_old_cur_map old_cur_map.



  unfold valid_virtual_mapping.
  intros osi' osi'_ptaddr mp osi'_pt H_osi'_pt_map.

  (*Prove by cases, first on the os considered in *)
  (*valid_virtual_mapping, (osi'). It can be the  *)
  (*one involved in new action, or not (osi) *)
  case (os_ident_eq osi' osi).
  

  (*Case osi'=osi*)
  intro H_osi'_osi.
  intros H_osi'_pt_cont H_osi'_pt_owner mp_va mp_ma H_mp_va_ma.

  (*Current page owner is used several times on the proof*)
  assert (page_owned_by osi_cur_page = Os_PO osi) as H_owner_osi_cpt.

  assert (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)) 
    as H_valid_curr.
  apply H_valid_s.
  
  elim H_valid_curr with osi system p2m.
  intro osi_pt_madd'.
  intros H_osi_pt_madd' H_pt_madd'_prop.
  assert (osi_pt_madd'=osi_pt_madd) as H_cpt_pt'.
  rewrite <- H_osi_pt_madd in H_osi_pt_madd'.
  inversion H_osi_pt_madd';auto.

  rewrite H_cpt_pt' in H_osi_pt_madd'. 
  rewrite H_cpt_pt' in H_pt_madd'_prop.
  unfold maddress_owner in H_pt_madd'_prop.
  rewrite <- H_osi_cur_pg in H_pt_madd'_prop.
  apply H_pt_madd'_prop.
  assumption.
  assumption.
  (*End of curr_page owner demostration*)
  
  (*Cases of mp_ma are considered, where mp_ma is the image 
  on mp, the mapping considered in valid_virtual_mapping.
  It can be the same of that being updated, current page 
  madd, or not*)
  case (madd_eq osi_pt_madd mp_ma).
  
  (*In case it is*)
  intro H_mpma_ptmadd.
 
  (*The corresponding page is the current page updated*) 
  exists (Page (PT_c (map_drop vadd_eq osi_cur_map va))
                     (page_owned_by osi_cur_page)).

  (*Prove that is the corresponding in newmem*)
  rewrite H_newmem.
  rewrite H_mem_pt_newmem.
  rewrite H_mpma_ptmadd.
  rewrite lemma_map_add_apply;auto.


  (*Now the accessibility constraint needs to be proved*)
  simpl.
  
  
  (*To prove accessibility, cases of mp_va are considered*)
  (*mp_va (pre-image in mp) can be the same that va or not*)
  
  case (vadd_eq va mp_va).

  (*In case mp_va=va*)
  intro H_mpva_va.
  (*Pre-conditions are used, that ensures accessibility from aos*)
  rewrite H_mpva_va in H_va_acces.
  split.
  intro.
  rewrite H_osi'_osi.
    
  auto.

  unfold not.
  intro H_absurd.
  elim H_absurd;auto.
  


  (*In case mp_va is not the same that va*)
  intro H_no_va0_va.
  
  (*We use the fact that s is a valid state*)
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.

  (*To use the H_valid_virt_s hypotesis, it's needed to*)
  (*consider cases of osi_ptaddr (pointer to PT), so the*)
  (*constraint to use H_valid_virt_s are met*)
  case (madd_eq osi_pt_madd osi'_ptaddr).
  intro H_osi'_ptmadd.

  (*Case that osi_ptadd=pt_maddr; it means it changed*)
  cut(mp=(map_drop vadd_eq osi_cur_map va)).
  (*relation between old and new current map*)
  intro H_mp_osi_curr_map.
  
  elim H_valid_virt_s with osi osi_pt_madd osi_cur_map osi_cur_page mp_va mp_ma;auto.
  intros pg H_pg H_acces.

  (*If the considered pointer a PT is cur_pt_madd,*)
  (*then the corresponding page is the current page*)
  cut (pg=osi_cur_page).
  intro H_pg_osi_curr.
  rewrite <- H_pg_osi_curr.
  rewrite H_osi'_osi;auto.
  rewrite <- H_mpma_ptmadd in H_pg.
  rewrite <- H_osi_cur_pg in H_pg.
  inversion H_pg;auto.
  rewrite H_mp_osi_curr_map in H_mp_va_ma.
  rewrite map_drop_apply_distinct in H_mp_va_ma;auto.
  rewrite H_newmem in H_osi'_pt_map.
  rewrite H_mem_pt_newmem in H_osi'_pt_map.
  rewrite H_osi'_ptmadd in H_osi'_pt_map.
  rewrite lemma_map_add_apply in H_osi'_pt_map.
  
  inversion H_osi'_pt_map.
  rewrite H0 in H_osi'_pt_cont.
  inversion H_osi'_pt_cont;auto.

  (*Case the pointer to PT (osi_ptaddr) is not osi_cpt_madd,*)
  (*that means its content doesn't change*)
  intro H_no_osi'_cptmadd.
  elim H_valid_virt_s with osi' osi'_ptaddr mp osi'_pt mp_va mp_ma;auto.
  intros pg H_pg H_acces.

  cut (pg=osi_cur_page). 
  intro H_pg_curr.
  rewrite <- H_pg_curr;assumption.

  rewrite <- H_mpma_ptmadd in H_pg.
  rewrite <- H_osi_cur_pg in H_pg.
  inversion H_pg;auto.
  
  rewrite H_newmem in H_osi'_pt_map.
  rewrite H_mem_pt_newmem in H_osi'_pt_map.
  rewrite lemma_map_add_apply_distinct in H_osi'_pt_map;auto.


  (*Case that mp_ma is not the address of the current map*)

  intro H_no_mpma_cpt.

 
  (*The witness is the same it has assigned in s*)

  (*We use the fact that s is a valid state*)
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.
  unfold valid_virtual_mapping in H_valid_virt_s.
  rewrite H_newmem.
  rewrite H_mem_pt_newmem.
  rewrite lemma_map_add_apply_distinct;auto.

  (*To use the H_valid_virt_s hypotesis, it's needed to*)
  (*consider cases of osi_ptaddr (pointer to PT), so the*)
  (*constraint to use H_valid_virt_s are met*)
  case (madd_eq osi_pt_madd osi'_ptaddr).
  

  (*Case the pointer to PT (osi_ptaddr) is not osi_cpt_madd,*)
  (*that means its content doesn't change*)
  2: intro H_no_osi_ptmadd.
  2: apply H_valid_virt_s with (osi:=osi') (pt_addr:=osi'_ptaddr) 
  (mp:=mp) (pt:=osi'_pt);auto.
  
  2: rewrite H_newmem in H_osi'_pt_map.
  2: rewrite H_mem_pt_newmem in H_osi'_pt_map.
  2: rewrite lemma_map_add_apply_distinct in H_osi'_pt_map;auto.

  (*Case that osi_ptadd=pt_maddr; it means it changed*)
  intro H_osi'_cptmadd.

  (*mp is current map updated*)
  cut(mp=(map_drop vadd_eq osi_cur_map va)).
  intro H_mp_osi_curr_map.

  (*We consider cases of mp_va*)
  case (vadd_eq mp_va va).

  (*Case mp_va<>va*)
  2:intro H_no_mpva_va.
  2:apply H_valid_virt_s with (osi:=osi') (pt_addr:=osi_pt_madd) (mp:=osi_cur_map) (pt:=osi_cur_page);auto.
  2:rewrite H_osi'_osi;auto.
  2:rewrite H_mp_osi_curr_map in H_mp_va_ma.
  2:rewrite map_drop_apply_distinct in H_mp_va_ma;auto.

  (*Case mp_va=va*)
  intro H_mpva_va.
 
  (*This case is absurd, because mp_va and va are the same*)
  (*and they can't have different images in the mp mapping*)
  cut (False).
  intro;contradiction.

  rewrite H_mp_osi_curr_map in H_mp_va_ma.
  rewrite <- H_mpva_va in H_mp_va_ma. 
  rewrite map_drop_apply in H_mp_va_ma;auto.
  inversion H_mp_va_ma;auto.


  clear H_valid_virt_s.

  (*It is pending the proof mp and current map relation*)

  rewrite H_newmem in H_osi'_pt_map.
  rewrite H_mem_pt_newmem in H_osi'_pt_map.
  rewrite H_osi'_cptmadd in H_osi'_pt_map.
  rewrite lemma_map_add_apply in H_osi'_pt_map.
  inversion H_osi'_pt_map.
  rewrite H0 in H_osi'_pt_cont.
  inversion H_osi'_pt_cont;auto.

  (*Case case that the considered os in not osi*)

  intro H_osi_ne_aos.

  intros H1 H2.

  cut (osi'_ptaddr <> osi_pt_madd).
  intro.
  rewrite H_newmem in H_osi'_pt_map.
  rewrite H_mem_pt_newmem in H_osi'_pt_map.
  rewrite lemma_map_add_apply_distinct in H_osi'_pt_map.
  intros.

  (*Use the fact that s is a valid state*)
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.

  unfold valid_virtual_mapping in H_valid_virt_s.
  
  case (madd_eq ma0 osi_pt_madd).

  (*Case mp_ma <> current page maddr*)
  2: intros.
  2: rewrite H_newmem.
  2: rewrite H_mem_pt_newmem.
  2: rewrite lemma_map_add_apply_distinct.
  2: apply H_valid_virt_s with (pt_addr:=osi'_ptaddr) (mp:=mp) (pt:=osi'_pt); auto.
  2: auto.
 
  (*Case mp_ma=current page maddr is absurd*) 
 
  intros.
  cut False.
  contradiction.

  elim H_valid_virt_s with osi' osi'_ptaddr mp osi'_pt va0 ma0; auto.
  intros pg H_pg_map H_owners.
  rewrite e in H_pg_map.
  cut (pg = osi_cur_page).
  intro.
 
 
  rewrite H3 in H_owners.
  (*We assume that the owner of osi's current page is osi*)
  cut (page_owned_by osi_cur_page = Os_PO osi).
  intro.
  
  unfold os_accessible in H_owners.
  destruct H_owners.
  destruct (ctxt_vadd_accessible ctxt va0).
  cut (page_owned_by osi_cur_page = Os_PO osi').
  intro H_osi'_owns.
  rewrite H4 in H_osi'_owns.
  inversion H_osi'_owns;auto.
  cut(true=true).
  intro H_true.
  apply (H5 H_true).
  trivial.
  cut (page_owned_by osi_cur_page = Hyp_PO).
  intro H_hyp_owns.
  rewrite H4 in H_hyp_owns.
  inversion H_hyp_owns;auto.
  cut(false<>true).
  intro H_true.
  apply (H6 H_true).
  auto.

  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr.
  2:apply H_valid_s.
  unfold valid_current_page in H_valid_curr.
  unfold maddress_owner in H_valid_curr.

  elim H_valid_curr with osi system p2m.
  intro pt_madd'.
  intros H_pt_madd' H_pt'_owner.
  cut (pt_madd'=osi_pt_madd).
  intro H_pt_pt'.
  rewrite H_pt_pt' in H_pt'_owner.
  destruct H_pt'_owner.
  rewrite <- H_osi_cur_pg in H5.
  assumption.
  unfold os_current_vadd_mapping_madd in H_osi_pt_madd.
  rewrite <- H_osi_pt_madd in H_pt_madd'.
  inversion H_pt_madd';auto.
  assumption.
  assumption.
  rewrite <- H_osi_cur_pg in H_pg_map.
  inversion H_pg_map;auto.
  auto.
  cut(maddress_owner (st_memory s')(osi_pt_madd) <> maddress_owner (st_memory s')(osi'_ptaddr)).
  intro H_owner.
  unfold not in H_owner.
  unfold not.
  intro H_pt_osi'.
  rewrite H_pt_osi' in H_owner.
  apply H_owner.
  trivial.
  cut (maddress_owner (st_memory s') osi_pt_madd = Os_PO osi).
  intro H_osi_cpt_madd_owner.
  rewrite H_osi_cpt_madd_owner.
  unfold maddress_owner.
  rewrite <- H_osi'_pt_map.
  rewrite H2.
  unfold not.
  intro Absurd.
  inversion Absurd.
  auto.
  
  unfold maddress_owner.
  rewrite H_newmem.
  rewrite H_mem_pt_newmem .
  rewrite lemma_map_add_apply.
  simpl.

  (*The owner of the osi_curr_page is osi*)
  (*We can use valid_currant_page to prove that*)
  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr.
  2:apply H_valid_s.
  
  unfold valid_current_page in H_valid_curr.

  elim H_valid_curr with osi system p2m.
  intro pt_madd'.
  intros.
  cut (pt_madd'=osi_pt_madd).
  intro H_pt_pt'.

  rewrite H_pt_pt' in H. 
  rewrite H_pt_pt' in H0.
  unfold maddress_owner in H0.
  rewrite <- H_osi_cur_pg in H0.
  apply H0.
  unfold os_current_vadd_mapping_madd in H_osi_pt_madd.
  destruct H_osi_pt_madd.
  inversion H;auto.
  assumption.
  assumption.

Qed.


(*Lemmas*)

(*Lemma that proves va_has_valid_pa for del trusted*)
(*and untrusted operations*)

Lemma va_has_valid_pa_del_trusted_untrusted: 

  forall osi system p2m osi_pt_madd va newmem s s',
  valid_state ctxt s ->
    
    Value _ system = map_apply os_ident_eq (st_oss s) osi ->
    hypervisor_mapping osi (st_hypervisor s) p2m ->
    Value _ osi_pt_madd = map_apply padd_eq p2m (os_current_page system) ->
    os_accessible ctxt va ->
    pre_all_del osi_pt_madd va s  -> 
    drop_va_pt_page osi_pt_madd va (st_memory s) newmem ->
    st_memory s' = newmem ->
    st_hypervisor s' = st_hypervisor s ->
    
  va_has_valid_pa (st_memory s') (st_hypervisor s') ctxt.
Proof.

  intros osi system p2m osi_pt_madd va newmem s s' H_valid_s.
  intros H_is_osi H_p2m H_osi_pt_madd H_va_acces H_pre_del 
    H_mem_pt_newmem H_newmem H_same_hyper.

  (*Use the post-conditions, so the relation between*)
  (*s and s' component can be established*)

  (*Hypothesis extracted from the pre-conditions*)
  unfold pre_all_del in H_pre_del.
  destruct H_pre_del as (osi_cur_map, H_osi_cur_map, H_pre).
  destruct H_pre as (ma,H_ma_va).

  destruct H_osi_cur_map as (osi_cur_page, H_osi_cur_pg, H_osi_cur_map).

  unfold drop_va_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem as (osi_cur_page', H_osi_cur_pg', H_mem_pt_newmem).
  assert (osi_cur_page=osi_cur_page') as H_cur_page_page'.
  rewrite <- H_osi_cur_pg in H_osi_cur_pg'.
  inversion H_osi_cur_pg';auto.
  rewrite <- H_cur_page_page' in H_mem_pt_newmem.
  clear H_cur_page_page' osi_cur_page' H_osi_cur_pg'.
  destruct H_mem_pt_newmem as (old_cur_map, H_old_cur_map,H_mem_pt_newmem ).
  assert (old_cur_map=osi_cur_map) as H_old_curr_map.
  rewrite <- H_osi_cur_map in H_old_cur_map.
  inversion H_old_cur_map;auto.
  rewrite H_old_curr_map in H_mem_pt_newmem.
  clear H_old_curr_map H_old_cur_map old_cur_map.


  (*Start of the main proof*)
  unfold va_has_valid_pa.
  intro osi'.

  (*We use several taimes that the property holds in s*)
  assert (va_has_valid_pa (st_memory s) 
    (st_hypervisor s) ctxt) as H_va_has_pa_in_s.
  apply H_valid_s.

  (*We must consider cases of osi'*)
  case (os_ident_eq osi' osi).

  (*In case osi' is osi*)
  intro H_osi'_osi.

  intros osi_pt_padd osi_pt_addr osi_pt_page 
    osi_mp osi_va osi_ma.
  intros H_osi_pt_page_addr H_pt_padd_madd 
    H_pt_page_cont H_osi_ma_va H_osi_va.

  (*We need to discriminate cases ef osi_ptadd*)
  case (madd_eq osi_pt_addr osi_pt_madd).
  intro H_osipt_osicpt.

  (*We establish the relation betwen the pages*)
  assert (osi_pt_page=
              (Page (PT_c (map_drop vadd_eq osi_cur_map va))
                    (page_owned_by osi_cur_page))
         ) as H_osipage_def.
  rewrite H_newmem in H_osi_pt_page_addr.
  rewrite H_mem_pt_newmem in H_osi_pt_page_addr.
  rewrite H_osipt_osicpt in H_osi_pt_page_addr.
  rewrite lemma_map_add_apply in H_osi_pt_page_addr.
  inversion H_osi_pt_page_addr;auto.

  (*We establish the relation betwen the mappings*)
  assert (osi_mp = (map_drop vadd_eq osi_cur_map va))
    as H_osi_mp_def.
  rewrite H_osipage_def in H_pt_page_cont.
  simpl in H_pt_page_cont.
  inversion H_pt_page_cont;auto.

  (*osi_va can't be the same that va*)
  assert (osi_va <> va) as H_osiva_no_va.

  intro H_osiva_va_abs.
  
  rewrite H_osi_mp_def in H_osi_ma_va.
  rewrite H_osiva_va_abs in H_osi_ma_va.
  rewrite map_drop_apply in H_osi_ma_va.
  inversion H_osi_ma_va;contradiction.
  
  elim H_va_has_pa_in_s with osi (os_current_page system)
    osi_pt_addr osi_cur_page (osi_cur_map) osi_va osi_ma;auto.

  intros osi_pa H_osipa.
  rewrite H_osi'_osi; rewrite H_same_hyper.
  exists osi_pa;auto.
  
  rewrite H_osipt_osicpt;auto.

  rewrite H_osipt_osicpt.
  exists p2m;auto.
  
  rewrite H_osi_mp_def in H_osi_ma_va.
  rewrite map_drop_apply_distinct 
    in H_osi_ma_va;auto.

  (*In case the considered pt madd is not osi cpt madd*)  
  intro H_osipt_no_osicpt.

  elim H_va_has_pa_in_s with osi osi_pt_padd
    osi_pt_addr osi_pt_page osi_mp osi_va osi_ma;auto.

  intros osi_pa H_osipa.
  rewrite H_osi'_osi; rewrite H_same_hyper.
  exists osi_pa;auto.

  rewrite H_newmem in H_osi_pt_page_addr.
  rewrite H_mem_pt_newmem in H_osi_pt_page_addr.
  rewrite lemma_map_add_apply_distinct
    in H_osi_pt_page_addr;auto.

  rewrite <- H_same_hyper;rewrite <- H_osi'_osi;auto.

  (*In case osi' is not osi*)
  intro H_osi'_no_osi.
  
  intros osi'_pt_padd osi'_pt_addr osi'_pt_page 
    osi'_mp osi'_va osi'_ma.
  intros H_osi'_pt_page_addr H_pt_padd_madd 
    H_pt_page_cont H_osi'_ma_va H_osi'_va.


  (*We must prove that the considered madd "osi'_pt_addr"*)
  (*can not be the same that osi current page madd.*)

  (*We prove this based in the fact that they have*)
  (*different owners. In one case osi, in the other osi'*)

  (*The owner of osi_pt_addr is osi'*)
  assert (maddress_owner (st_memory s) osi'_pt_addr 
    = Os_PO osi') as H_osi'pt_owner.

  destruct H_pt_padd_madd as (osi'_p2m,H_osi'p2m',
    H_osi'pt_osi'padd).

  assert (valid_hypervisor (st_hypervisor s') 
    (st_memory s) ) as H_valid_hyper.
  rewrite H_same_hyper;apply H_valid_s.

  elim H_valid_hyper with osi' osi'_pt_padd osi'_p2m;auto.

  intro H_exist_ma.
  destruct H_exist_ma as (osi'_pt_madd, H_osi'madd,
    H_osi'madd_prop).
  assert (osi'_pt_addr=osi'_pt_madd) as H_same_pt_madd.
  rewrite <- H_osi'pt_osi'padd in H_osi'madd.
  inversion H_osi'madd;auto.

  rewrite <- H_same_pt_madd in H_osi'madd_prop;auto.

  intro H_absurd.
  rewrite <- H_osi'pt_osi'padd in H_absurd.
  inversion H_absurd;auto.
  

  (*The owner of osi_cpt_madd is osi*)
  assert (maddress_owner (st_memory s) osi_pt_madd 
    = Os_PO osi) as H_osicpt_owner.

  assert (valid_current_page (st_hypervisor s) (st_memory s)
     (st_oss s)) as H_valid_curr_s. 
  apply H_valid_s.
  
  unfold valid_current_page in H_valid_curr_s.
  elim H_valid_curr_s with osi system p2m.
  intros osi_cpt_addr H_osicptaddr H_osicpt_prop.
  assert (osi_cpt_addr=osi_pt_madd) as H_osi_cpt.
  rewrite <- H_osi_pt_madd in H_osicptaddr.
  inversion H_osicptaddr;auto.
  rewrite H_osi_cpt in H_osicpt_prop.
  apply H_osicpt_prop.

  assumption.

  assumption.

  (*The considered madd osi'_pt_addr is different from*)
  (*osi_cpt_madd because they have different owners*)

  assert (osi'_pt_addr<>osi_pt_madd) 
    as H_osi'pt_no_osicpt.
  
  intro H_osi'pt_osicpt_abs.

  rewrite H_osi'pt_osicpt_abs in H_osi'pt_owner.
  rewrite H_osi'pt_owner in H_osicpt_owner.
  inversion H_osicpt_owner;contradiction.


  (*Once we know osi'_pt isn't the osi cpt, we use*)
  (*the fact that s is a valid state*)
  elim H_va_has_pa_in_s with osi' osi'_pt_padd
    osi'_pt_addr osi'_pt_page osi'_mp 
      osi'_va osi'_ma;auto.

  intros osi'_pa H_osi'pa.
  rewrite H_same_hyper.
  exists osi'_pa;auto.

  rewrite H_newmem in H_osi'_pt_page_addr.
  rewrite H_mem_pt_newmem in H_osi'_pt_page_addr.
  rewrite lemma_map_add_apply_distinct 
    in H_osi'_pt_page_addr;auto.
  rewrite <- H_same_hyper;auto.

Qed.


(*******************)
(***Del is secure***)
(*******************)

Theorem Del_untrusted_secure:
  forall (s s':state),
  forall osi va,
  valid_state ctxt s ->
  valid_exec ctxt s (Del_untrusted osi va) s'->
  valid_state ctxt s'.
Proof.
  intros s s' osi va H_valid_s H_valid_exec.
  generalize H_valid_exec; intro H_valid_exec';
    inversion_clear H_valid_exec'.
  simpl in H.

  (*Use the post-conditions, so the relation between*)
  (*s and s' component can be established*)

  simpl in H0.
  unfold post_del_untrusted in H0.
  destruct H0 as (system, H_osi_system,H0).

  destruct H0 as (p2m,H_osi_p2m,H0).
  destruct H0 as (osi_pt_madd, H_osi_pt_madd, H0).
  destruct H0 as (newmem, H_mem_pt_newmem, H_s').

  (*Start Main Proof*)
  unfold valid_state.
  rewrite H_s'; simpl.

  (*First Constraint*)
  split.
  unfold trusted_os_not_hypercall.
  intros osi' os' H_osi'_trusted H_os'_map.
  case (os_ident_eq osi osi').
  intro H_osi_osi'.
  rewrite H_osi_osi' in H_os'_map.
  rewrite lemma_map_add_apply in H_os'_map.
  injection H_os'_map; intro H_os'_value; rewrite H_os'_value; auto.
  intro H_osi_no_osi'.
  cut (trusted_os_not_hypercall (ctxt_oss ctxt) (st_oss s)).
  intro H_trusted_no_hyp_s.
  2:apply H_valid_s.
  rewrite lemma_map_add_apply_distinct in H_os'_map; auto.
  apply (H_trusted_no_hyp_s osi' os' H_osi'_trusted); auto.
 

  (*Second Constraint*)
  split.
  unfold running_os_not_hypercall.
  case (os_ident_eq osi (st_active_os s)).
  intro H_osi_aos.
  exists (OS None (os_current_page system)).
  rewrite <- H_osi_aos.
  rewrite lemma_map_add_apply;auto.
  auto.
  intro H_osi_no_aos.
  cut (running_os_not_hypercall (st_active_os s) (st_aos_activity s) (st_oss s)).
  intro H_aos_no_hyp_s.
  2:apply H_valid_s.
  rewrite lemma_map_add_apply_distinct;auto.

  (*Third Constraint*)
  split.
  apply H_valid_s.
  
  (*Fourth Constraint*)
  split.
  apply H_valid_s.
   
  (*Fifth Constraint*)
  split.
  apply H_valid_s.


  (*Sixth Seventh and Eighth Constraints use the pre and post *)
  (*conditions extracted next*)
  unfold pre_del_untrusted in H.
  destruct H as (H_untrusted_osi, H_pre).
  destruct H_pre as (H_osi_some_hyper_del, H_pre).
  
  destruct H_pre as (system', H_system'_osi, H_pre).
  
  destruct H_pre as (p2m', H_p2m'_osi, H_pre).
  destruct H_pre as (sys'_pt_madd, H_sys'_pt_madd,H_pre).

  destruct H_pre as (H_pre_del,H_pre).
  destruct H_pre as (hyper_running,H_va_acces).
  
  assert (sys'_pt_madd=osi_pt_madd) as H_sys_sys'_pt.
  assert (system'=system) as H_sys'_sys.
  rewrite <- H_osi_system in H_system'_osi.
  inversion H_system'_osi;auto.
  assert (p2m'=p2m) as H_p2m'_p2m.
  unfold hypervisor_mapping in H_p2m'_osi.
  unfold hypervisor_mapping in H_osi_p2m.
  rewrite <- H_osi_p2m in H_p2m'_osi.
  inversion H_p2m'_osi;auto.
  rewrite H_sys'_sys in H_sys'_pt_madd.
  rewrite H_p2m'_p2m in H_sys'_pt_madd.
  rewrite <- H_osi_pt_madd in H_sys'_pt_madd.
  inversion H_sys'_pt_madd;auto.

  rewrite H_sys_sys'_pt in H_pre_del.  
  clear system' H_system'_osi p2m' H_p2m'_osi sys'_pt_madd 
                    H_sys'_pt_madd H_sys_sys'_pt.
  unfold pre_all_del in H_pre_del.
  destruct H_pre_del as (osi_cur_map, H_osi_cur_map, H_pre).
  destruct H_pre as (ma,H_ma_va).

  destruct H_osi_cur_map as (osi_cur_page, H_osi_cur_pg, H_osi_cur_map).

  unfold drop_va_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem as (osi_cur_page', H_osi_cur_pg', H_mem_pt_newmem).
  assert (osi_cur_page=osi_cur_page') as H_cur_page_page'.
  rewrite <- H_osi_cur_pg in H_osi_cur_pg'.
  inversion H_osi_cur_pg';auto.
  rewrite <- H_cur_page_page' in H_mem_pt_newmem.
  clear H_cur_page_page' osi_cur_page' H_osi_cur_pg'.
  destruct H_mem_pt_newmem as (old_cur_map, H_old_cur_map,H_mem_pt_newmem ).
  assert (old_cur_map=osi_cur_map) as H_old_curr_map.
  rewrite <- H_osi_cur_map in H_old_cur_map.
  inversion H_old_cur_map;auto.
  rewrite H_old_curr_map in H_mem_pt_newmem.
  clear H_old_curr_map H_old_cur_map old_cur_map.


  (*Sixth Constraint*)

  split.

  unfold valid_hypervisor.
  intros osi' any_pa osi'_p2m H_is_osi'_p2m.

  (*We use the fact that s is a valid state*)
  (*so the constraint we are proving holds in s*)
  cut (valid_hypervisor (st_hypervisor s) (st_memory s)).
  intro H_valid_hyper_s.
  2:apply H_valid_s.
  unfold valid_hypervisor in H_valid_hyper_s.
  elim H_valid_hyper_s with osi' any_pa osi'_p2m;auto.
  intro H_exist_in_s.
  cut (forall ma, maddress_owner (st_memory s) ma= maddress_owner (newmem) ma).
  intro H_same_owners.
  destruct H_exist_in_s as (s_ma,H_s_ma_pa,H_s_ma_owner).
  left.
  exists s_ma;auto.
  
  rewrite <-(H_same_owners s_ma);auto.
  intro s_ma.

  (*Proof for both cases of s_ma; wether it is osi current page or not*)
  case (madd_eq s_ma osi_pt_madd).
  intro H_case_osi_cpt.

  (*Rewrite with the data in newmem*)

  rewrite H_mem_pt_newmem.
  rewrite H_case_osi_cpt.
  unfold maddress_owner.
  rewrite lemma_map_add_apply.
  rewrite <- H_osi_cur_pg;trivial.
  intro H_case_no_pt.

  (*We use the fact that the data didn't change un memory*)
  unfold update_pt_page in H_mem_pt_newmem.
  rewrite H_mem_pt_newmem.

  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct;auto.


  (*Seventh Constraint*)
  split.
  replace newmem with (st_memory s').
  apply ((valid_virtual_mapping_del_trusted_untrusted) 
    osi system p2m osi_pt_madd va newmem s s'); auto.

  unfold   pre_all_del.
  exists osi_cur_map.
  exists osi_cur_page;auto.
  exists ma;auto.
  exists osi_cur_page;auto.
  exists osi_cur_map;auto.
  rewrite H_s'; auto.
  rewrite H_s'; auto.
  
  (*Eighth Constraint*)
  split.
 
  unfold valid_current_page.
  intros osi' system' H_osi'_system' p2m' H_sys'_p2m'.

  case (os_ident_eq osi' osi).
  intro H_osi'_osi.
  exists osi_pt_madd.
  
  assert (OS None (os_current_page system)=system') as H_sys_sys'.
  rewrite H_osi'_osi in H_osi'_system'.
  rewrite lemma_map_add_apply in H_osi'_system'.
  inversion  H_osi'_system';auto.
  rewrite <- H_sys_sys'.
  simpl.
  assert (p2m'=p2m) as H_p2m_p2m'.
  rewrite H_osi'_osi in H_sys'_p2m'.
  rewrite <- H_osi_p2m in H_sys'_p2m'.
  inversion H_sys'_p2m';auto.
  rewrite H_p2m_p2m'.
  assumption.

  split.
  rewrite H_mem_pt_newmem.
  unfold page_type.
  rewrite lemma_map_add_apply.
  simpl.
  trivial.

  unfold maddress_owner.
  rewrite H_mem_pt_newmem.
  rewrite lemma_map_add_apply.
  simpl.
  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr_pt_s.
  2:apply H_valid_s.


  unfold valid_current_page in H_valid_curr_pt_s.
  rewrite H_osi'_osi.
  
  destruct H_valid_curr_pt_s  with  osi system p2m
    as (pt_madd',H_ptmadd',H_ptmadd'_prop);auto.
  destruct H_ptmadd'_prop as (H_ptmdd'_ptype,H_ptmadd'_owner).
  cut (pt_madd'=osi_pt_madd).
  intro H_pt'_pt.
  unfold maddress_owner in H_ptmadd'_owner.
  rewrite H_pt'_pt in H_ptmadd'_owner.
  rewrite <- H_osi_cur_pg in H_ptmadd'_owner.
  assumption.
  clear H_valid_curr_pt_s.

  unfold os_current_vadd_mapping_madd in H_osi_pt_madd.
  destruct H_osi_pt_madd.
  inversion H_ptmadd';auto.



  intro H_no_osi'_osi.

  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr_pt_s.
  2:apply H_valid_s.

  unfold valid_current_page in H_valid_curr_pt_s.

  elim H_valid_curr_pt_s with osi' system' p2m'.
  intro osi'_pt_madd.
  intros H_ptmadd' H_ptmadd'_prop.
  exists osi'_pt_madd;auto.
  split.

  rewrite H_mem_pt_newmem.
  unfold page_type.
  rewrite lemma_map_add_apply_distinct.
  unfold page_type in H_ptmadd'_prop.
  apply H_ptmadd'_prop.

  (*they are different because they have diferent owners*)
  cut (maddress_owner (st_memory s) osi_pt_madd = Os_PO osi).
  intro H_cpt_madd_owner.
  intro H_contrad.
  rewrite H_contrad in H_cpt_madd_owner.
  destruct H_ptmadd'_prop as (H_ptmadd'_ptype,H_ptmadd'_owner).
  rewrite H_cpt_madd_owner in H_ptmadd'_owner.
  inversion H_ptmadd'_owner;auto.

  (*using the same constraint that we are proving*)
  elim H_valid_curr_pt_s with osi system p2m;auto.
  intro pt_madd'.
  intros H_pt_madd' H_pt_madd_prop.
  assert (pt_madd'=osi_pt_madd) as H_same_pt_madd.

  rewrite <- H_osi_pt_madd in H_pt_madd'.
  inversion H_pt_madd';auto.
  rewrite <- H_same_pt_madd.
  apply H_pt_madd_prop.
  
  (*newmem definition*)
  rewrite H_mem_pt_newmem.
  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct.
  unfold maddress_owner in H_ptmadd'_prop.
  apply H_ptmadd'_prop.

  (*they are different because they have different owners*)
  
  cut (maddress_owner (st_memory s) osi_pt_madd = Os_PO osi).
  intro H_cpt_madd_owner.
  intro H_contrad.
  rewrite H_contrad in H_cpt_madd_owner.
  destruct H_ptmadd'_prop as (H_ptmadd'_ptype,H_ptmadd'_owner).
  rewrite H_cpt_madd_owner in H_ptmadd'_owner.
  inversion H_ptmadd'_owner;auto.

  (*using the same constraint that we are profing, but for s instead*)
  elim H_valid_curr_pt_s with osi system p2m;auto.
  intros osi_pt_madd' H_osiptmadd' H_osiptmadd'_prop.
  cut (osi_pt_madd'=osi_pt_madd).
  intro H_same_pt_madd.
  rewrite <- H_same_pt_madd.
  apply H_osiptmadd'_prop.

  rewrite <- H_osi_pt_madd in H_osiptmadd'.
  inversion H_osiptmadd';auto.
  rewrite lemma_map_add_apply_distinct in H_osi'_system';auto.
  assumption.

  (*Ninth Constraint*)
  split.
  apply H_valid_s.
 
  (*Tenth Constraint*)
  generalize H_valid_s; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [H_cache_valid H_tlb_valie]].
  split.

  replace newmem with (st_memory s').
  replace (st_hypervisor s) with (st_hypervisor s').
  apply ((va_has_valid_pa_del_trusted_untrusted)
    osi system p2m osi_pt_madd va newmem s s');auto .

  unfold pre_all_del.
  exists osi_cur_map.
  exists osi_cur_page;auto.
  exists ma;auto.
  exists osi_cur_page;auto.
  exists osi_cur_map;auto.
  rewrite H_s'; auto.
  rewrite H_s'; auto.
  rewrite H_s'; auto.
  rewrite H_s'; auto.

  apply exec_valid_cache_tlb with ctxt s (Del_untrusted osi va); auto.
  rewrite H_s' in H_valid_exec; auto.
Qed.


Theorem Del_trusted_secure:
  forall (s s':state),
  forall va,
  valid_state ctxt s ->
  valid_exec ctxt s (Del_trusted va) s'->
  valid_state ctxt s'.
Proof.
  intros s s' va H_valid_s H_valid_exec.
  generalize H_valid_exec; intro H_valid_exec';
    inversion_clear H_valid_exec'.
  simpl in H.

  (*Use the post-conditions, so the relation between*)
  (*s and s' component can be established*)


  simpl in H0.
  unfold post_del_trusted in H0.
  destruct H0 as (pt_madd, H_is_ptmadd,H_post_del).
  
  unfold post_all_del in H_post_del.
  destruct H_post_del as (newmem,H_mem_pt_newmem,H_s').

  (*Start main proof*)
  unfold valid_state.

  unfold pre_del_trusted in H.
  destruct H as (pt_madd',H_is_ptmadd',H_pre).

  destruct H_pre as (H_pre_del,H_pre).
  destruct H_pre_del as (cur_map,H_cur_map,H_ma_va).
  destruct H_ma_va as (ma,H_ma_va).
  destruct H_pre as (H_os_running,H_pre).
  destruct H_pre as (H_trusted_aos,H_access_va).

  destruct H_is_ptmadd as (aos,H_is_aos,H_p2m).
  destruct H_p2m as (p2m,H_p2m,H_pt_madd).

  assert (pt_madd'=pt_madd) as H_ptmadd'_ptmadd.
  destruct H_is_ptmadd' as (aos',H_is_aos',H_p2m').
  assert (aos'=aos) as H_aos'_aos.
  rewrite <- H_is_aos in H_is_aos'.
  inversion H_is_aos';auto.
  rewrite H_aos'_aos in H_p2m'.
  destruct H_p2m' as (p2m',H_p2m',H_ptmadd').
  assert (p2m'=p2m) as H_p2m'_p2m.
  unfold hypervisor_mapping in H_p2m'.
  unfold hypervisor_mapping in H_p2m.
  rewrite <- H_p2m in H_p2m'.
  inversion H_p2m';auto.
  rewrite H_p2m'_p2m in H_ptmadd'.
  rewrite <- H_pt_madd in H_ptmadd'.
  inversion H_ptmadd';auto.

  rewrite H_ptmadd'_ptmadd in H_cur_map.
  destruct H_cur_map as (cur_page,H_cur_page,H_cur_map).
  
  clear pt_madd' H_is_ptmadd' H_ptmadd'_ptmadd.

  unfold drop_va_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem as (cur_page', H_cur_pg', H_mem_pt_newmem).
  assert (cur_page=cur_page') as H_cur_page_page'.
  rewrite <- H_cur_page in H_cur_pg'.
  inversion H_cur_pg';auto.
  rewrite <- H_cur_page_page' in H_mem_pt_newmem.
  clear H_cur_page_page' cur_page' H_cur_pg'.
  destruct H_mem_pt_newmem as (old_cur_map, H_old_cur_map,H_mem_pt_newmem ).
  assert (old_cur_map=cur_map) as H_old_curr_map.
  rewrite <- H_cur_map in H_old_cur_map.
  inversion H_old_cur_map;auto.
  rewrite H_old_curr_map in H_mem_pt_newmem.
  clear H_old_curr_map H_old_cur_map old_cur_map.

  do 10 (try split); try (rewrite H_s'; apply H_valid_s).

  (*Sixth Constraint*)
  rewrite H_s'; simpl.
  unfold valid_hypervisor.
  intros osi any_pa osi_p2m H_is_osi_p2m.


  (*We use the fact that s is a valid state*)
  (*so the constraint we are proving holds in s*)
  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) as H_valid_hyper_s.
  apply H_valid_s.
  unfold valid_hypervisor in H_valid_hyper_s.
  elim H_valid_hyper_s with osi any_pa osi_p2m;auto.
  intro H_exist_in_s.
  cut (forall ma, maddress_owner (st_memory s) ma= maddress_owner (newmem) ma).
  intro H_same_owners.
  destruct H_exist_in_s as (s_ma,H_s_ma_pa,H_s_ma_owner).
  left.
  exists s_ma;auto.
  
  rewrite <-(H_same_owners s_ma);auto.
  intro s_ma.

  (*Proof for both cases of s_ma; wether it is osi current page or not*)
  case (madd_eq s_ma pt_madd).
  intro H_case_cpt.

  (*Rewrite with the data in newmem*)

  rewrite H_mem_pt_newmem.
  rewrite H_case_cpt.
  unfold maddress_owner.
  rewrite lemma_map_add_apply.
  rewrite <- H_cur_page;trivial.
  intro H_case_no_pt.

  (*We use the fact that the data didn't change un memory*)
  unfold update_pt_page in H_mem_pt_newmem.
  rewrite H_mem_pt_newmem.

  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct;auto.

  (*Seventh Constraint*)
  apply ((valid_virtual_mapping_del_trusted_untrusted) 
    (st_active_os s) aos p2m pt_madd va newmem s s');auto .

  exists cur_map;auto.
  exists cur_page;auto.
  exists ma;auto.
  exists cur_page;auto.
  exists cur_map;auto.
  rewrite H_s'; auto.
  

  (*Eighth Constraint*)
  unfold valid_current_page.
  rewrite H_s'; simpl.
  intros osi system H_osi_system p2m' H_sys_p2m'.

  case (os_ident_eq osi (st_active_os s)).
  intro H_osi_aos.


  assert (p2m'=p2m) as H_p2m_p2m'.
  rewrite H_osi_aos in H_sys_p2m'.
  rewrite <- H_p2m in H_sys_p2m'.
  inversion H_sys_p2m';auto.
  rewrite H_p2m_p2m'.
  assert (aos=system) as H_aos_system.
  rewrite H_osi_aos in H_osi_system.
  rewrite <- H_is_aos in H_osi_system;inversion H_osi_system;auto.
  rewrite H_osi_aos.
  rewrite <- H_aos_system;auto.
  
  exists pt_madd;auto.
  
  split.
  rewrite H_mem_pt_newmem.
  unfold page_type.
  rewrite lemma_map_add_apply.
  simpl.
  trivial.

  unfold maddress_owner.
  rewrite H_mem_pt_newmem.
  rewrite lemma_map_add_apply.
  simpl.
  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr_pt_s.
  2:apply H_valid_s.


  unfold valid_current_page in H_valid_curr_pt_s.
  rewrite <- H_osi_aos.
  
  elim H_valid_curr_pt_s with (st_active_os s) aos p2m;auto.
  intros.
  destruct H0.
  cut (x=pt_madd).
  intro H_x_pt.
  unfold maddress_owner in H1.
  rewrite H_x_pt in H1.
  rewrite <- H_cur_page in H1.
  rewrite H_osi_aos.
  assumption.
  clear H_valid_curr_pt_s.

  unfold os_current_vadd_mapping_madd in H_pt_madd.
  destruct H_pt_madd.
  inversion H;auto.



  intro H_no_osi_aos.

  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr_pt_s.
  2:apply H_valid_s.


  unfold valid_current_page in H_valid_curr_pt_s.

  elim H_valid_curr_pt_s with osi system p2m'.
  intro osi_pt_madd.
  intros.
  exists osi_pt_madd;auto.

  split.

  
  rewrite H_mem_pt_newmem.
  unfold page_type.
  rewrite lemma_map_add_apply_distinct.
  unfold page_type in H0.
  apply H0.

  (*they are different because they have diferent owners*)
  

  cut (maddress_owner (st_memory s) pt_madd = Os_PO (st_active_os s)).
  intro H_cpt_madd_owner.
  intro H_contrad.
  rewrite H_contrad in H_cpt_madd_owner.
  destruct H0.
  rewrite H_cpt_madd_owner in H1.
  inversion H1.
  auto.

  (*using the same constraint that we are proving*)
  elim H_valid_curr_pt_s with (st_active_os s) aos p2m;auto.
  intro pt_madd'.
  intros H_pt_madd' H_pt_madd_prop.
  assert (pt_madd'=pt_madd) as H_same_pt_madd.

  rewrite <- H_pt_madd in H_pt_madd'.
  inversion H_pt_madd';auto.
  rewrite <- H_same_pt_madd.
  apply H_pt_madd_prop.
  
  rewrite H_mem_pt_newmem.
  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct.
  unfold maddress_owner in H0.
  apply H0.

  (*they are different because they have different owners*)
  
  cut (maddress_owner (st_memory s) pt_madd = Os_PO (st_active_os s)).
  intro H_cpt_madd_owner.
  intro H_contrad.
  rewrite H_contrad in H_cpt_madd_owner.
  destruct H0.
  rewrite H_cpt_madd_owner in H1.
  inversion H1.
  auto.

  (*using the same constraint that we are profing, but for s instead*)
  elim H_valid_curr_pt_s with (st_active_os s) aos p2m;auto.
  intro pt_madd'.
  intros.
  cut (pt_madd'=pt_madd).
  intro H_same_pt_madd.
  rewrite <- H_same_pt_madd.
  apply H2.

  rewrite <- H_pt_madd in H1.
  inversion H1;auto.

  assumption.
  assumption.

  (*Tenth Constraint*)
  apply ((va_has_valid_pa_del_trusted_untrusted)
    (st_active_os s) aos p2m pt_madd va newmem s s');auto.
  unfold pre_all_del; auto.
  exists cur_map;auto.
  exists cur_page;auto.
  exists ma;auto.
  exists cur_page;auto.
  exists cur_map;auto.
  rewrite H_s'; auto.
  rewrite H_s'; auto.
  apply exec_valid_cache_tlb with ctxt s (Del_trusted va); auto.
Qed.


Theorem Del_hyper_secure : 
  forall (s s':state),
  forall va,
  valid_state ctxt s ->
  valid_exec ctxt s (Del_hyper va) s'->
  valid_state ctxt s'.
Proof.
  intros s s' va H_valid_s H_valid_exec.
  generalize H_valid_exec; intro H_valid_exec'.
  inversion_clear H_valid_exec'.
  simpl in H.

  (*Use the post-conditions, so the relation between*)
  (*s and s' component can be established*)


  simpl in H0.
  unfold post_del_hyper in H0.
  destruct H0 as (cur_pt_madd, H_cur_pt_madd, H_post).
  unfold post_all_del in H_post.
  destruct H_post as (newmem, H_mem_pt_newmem, H_s').

  (*Start Main Proof*)
  unfold valid_state.

  do 10 (try split); try (rewrite H_s'; apply H_valid_s).
  

  (*Sixth Constraint*)
  rewrite H_s'.
  unfold valid_hypervisor.
  intros osi any_pa osi_p2m H_is_osi_p2m.

  (*We use the fact that s is a valid state*)
  (*so the constraint we are proving holds in s*)
  cut (valid_hypervisor (st_hypervisor s) (st_memory s)).
  intro H_valid_hyper_s.
  2:apply H_valid_s.
  unfold valid_hypervisor in H_valid_hyper_s.
  elim H_valid_hyper_s with osi any_pa osi_p2m;auto.
  intro H_exist_in_s.
  cut (forall ma, maddress_owner (st_memory s) ma= maddress_owner (newmem) ma).
  intro H_same_owners.
  destruct H_exist_in_s as (s_ma,H_s_ma_pa,H_s_ma_owner).
  left.
  exists s_ma;auto.
  
  simpl; rewrite <-(H_same_owners s_ma);auto.
  intro s_ma.

  (*Proof for both cases of s_ma; wether it is osi current page or not*)
  case (madd_eq s_ma cur_pt_madd).
  intro H_case_cur_pt.

  (*Rewrite with the data in newmem*)
  unfold update_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem as (current_page, H_pt_current_page, H_mem_pt_newmem).
  destruct H_mem_pt_newmem as (current_map, H_pt_curr_map,H_update_curr_pt).
  rewrite H_update_curr_pt.
  rewrite H_case_cur_pt.
  unfold maddress_owner.
  rewrite lemma_map_add_apply.
  rewrite <- H_pt_current_page;trivial.
  intro H_case_no_pt.

  (*We use the fact that the data didn't change un memory*)
  unfold update_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem as (current_page, H_pt_current_page, H_mem_pt_newmem).
  destruct H_mem_pt_newmem as (current_map, H_pt_curr_map,H_update_curr_pt).
  rewrite H_update_curr_pt.

  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct;auto.

  (*Seventh Constraint*)

  (*We complete the hypothesis extracted from pre and post conditions*)
  unfold pre_del_hyper in H.
  destruct H as (cur_pt_madd',H_curptmadd',H_pre).
  destruct H_pre as (H_pre_del,H_hyper_running_acces).
  destruct H_hyper_running_acces as (H_hyper_running,H_no_acces_va).
  destruct H_cur_pt_madd as (aos,H_is_aos,H_p2m).
  destruct H_p2m as (p2m,H_p2m,H_cur_page).
  unfold hypervisor_mapping in H_p2m.
  destruct H_curptmadd' as (aos',H_is_aos',H_p2m').
  destruct H_p2m' as (p2m',H_p2m',H_cur_page').
  unfold hypervisor_mapping in H_p2m'.
  assert (aos'=aos) as H_aos'_aos.
  rewrite <- H_is_aos in H_is_aos'.
  inversion H_is_aos';auto.
  assert (p2m'=p2m) as H_p2m'_p2m.
  rewrite <- H_p2m in H_p2m'.
  inversion H_p2m';auto.
  assert (cur_pt_madd=cur_pt_madd') as H_same_curptmadd.
  rewrite H_aos'_aos in H_cur_page'.
  rewrite H_p2m'_p2m in H_cur_page'.
  rewrite <- H_cur_page in H_cur_page'.
  inversion H_cur_page';auto.

  rewrite <- H_same_curptmadd in H_pre_del.

  clear cur_pt_madd' aos' H_is_aos' p2m' H_p2m' H_cur_page' H_aos'_aos 
    H_p2m'_p2m H_same_curptmadd.

  unfold pre_all_del in H_pre_del.
  destruct H_pre_del as (curr_map,H_pt_curr_map,H_ma).
  destruct H_pt_curr_map as (curr_page,H_pt_curr_page,H_pt_curr_map).
  destruct H_ma as (ma,H_ma_va).

  (*newmem assignment*)
  unfold drop_va_pt_page in H_mem_pt_newmem.
  destruct H_mem_pt_newmem as (curr_page', H_pt_curr_page', H_upd_curr_pt).
  destruct H_upd_curr_pt as (curr_map', H_pt_curr_map',H_upd_curr_pt).
  assert (curr_page'=curr_page) as H_same_curr_page.
  rewrite <- H_pt_curr_page in H_pt_curr_page'.
  inversion H_pt_curr_page';auto.
  assert (curr_map'=curr_map) as H_same_currmap.
  rewrite H_same_curr_page in H_pt_curr_map'.
  rewrite <- H_pt_curr_map in H_pt_curr_map'.
  inversion H_pt_curr_map';auto.
  rewrite H_same_currmap in H_upd_curr_pt.
  rewrite H_same_curr_page in H_upd_curr_pt.

  clear H_same_currmap H_same_curr_page curr_page' curr_map' H_pt_curr_page' H_pt_curr_map'.

  (*Start Main Proof.*)
  unfold valid_virtual_mapping.
  intros osi osi_ptaddr mp osi_pt H_osi_pt_map.


  (*Prove by cases, first on the os considered in *)
  (*valid_virtual_mapping, (osi). It can be the aos *)
  (*identifier, or not *)
  case (os_ident_eq osi (st_active_os s)).  


  (*Case osi=aos*)

  intro H_osi_aos.
  intros H_osi_pt_cont H_osi_pt_owner mp_va mp_ma H_mp_va_ma.

  (*The current page owner is aos. This is used several times*)
  assert (page_owned_by curr_page = Os_PO (st_active_os s)) 
    as H_owner_curr_pg.

  assert (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s))
    as H_valid_curr.
  apply H_valid_s.
  
  elim H_valid_curr with (st_active_os s) aos p2m;auto.
  intro cur_pt_madd'.
  intros H_cur_pt_madd'.
  cut (cur_pt_madd'=cur_pt_madd).
  intros H_pt_mad'_mad.
  rewrite H_pt_mad'_mad.
  clear H_pt_mad'_mad.
  intro H_cur_pt_mad_prop.
  destruct H_cur_pt_mad_prop as (H_cur_pt_mad_pt,H_cur_pt_mad_owner).
  unfold maddress_owner in H_cur_pt_mad_owner.
  rewrite <- H_pt_curr_page in H_cur_pt_mad_owner.
  assumption.
  rewrite <- H_cur_page in H_cur_pt_madd'.
  inversion H_cur_pt_madd';auto.


  (*Cases of mp_ma are considered, where mp_ma is the image 
  on mp, the mapping considered in valid_virtual_mapping.
  It can be the same of that being updated, current page 
  madd, or not*)
  case (madd_eq cur_pt_madd mp_ma).
  

  (*In case it is*)
  intro H_mpma_ptmadd.
 
  (*The corresponding page is the current page updated*)
  exists (Page (PT_c (map_drop vadd_eq curr_map va))
                     (page_owned_by curr_page)).

  (*Prove that is the corresponding in newmem*)
  rewrite H_s'; simpl.
  rewrite H_upd_curr_pt.
  rewrite H_mpma_ptmadd.
  rewrite lemma_map_add_apply;auto.

  

  (*Now the accessibility constraint needs to be proved*)
  simpl.
  split.
  intro.
  rewrite H_osi_aos.
  assumption.


  (*To prove accessibility, cases of mp_va are considered*)
  case (vadd_eq va mp_va).

  (*In case mp_va=va*)
  intro H_mpva_va.
  Focus 2.




  (*In case mp_va is not the same that va*)
  intro H_no_mpva_va.
  
  (*We use the fact that s is a valid state*)
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.

  (*To use the H_valid_virt_s hypotesis, it's needed to*)
  (*consider cases of osi_ptaddr (pointer to PT), so the*)
  (*constraint to use H_valid_virt_s are met*)
  case (madd_eq cur_pt_madd osi_ptaddr).

  intro H_osi_ptmadd.
  (*Case that osi_ptadd=pt_maddr; it means it changed*)
  cut(mp=(map_drop vadd_eq curr_map va)).
  (*relation between old and new current map*)
  intro H_mp_curr_map.
  
  elim H_valid_virt_s with (st_active_os s) cur_pt_madd curr_map curr_page mp_va mp_ma;auto.
  intros pg H_pg H_acces.
  
  (*If the considered pointer a PT is cur_pt_madd,*)
  (*then the corresponding page is the current page*)
  cut (pg=curr_page).
  intro H_pg_curr.
  rewrite <- H_pg_curr.
  apply H_acces.
  rewrite <- H_mpma_ptmadd in H_pg.
  rewrite <- H_pt_curr_page in H_pg.
  inversion H_pg;auto.
  rewrite H_mp_curr_map in H_mp_va_ma.
  rewrite map_drop_apply_distinct in H_mp_va_ma;auto.
  rewrite H_s' in H_osi_pt_map; simpl in H_osi_pt_map.
  rewrite H_upd_curr_pt in H_osi_pt_map.
  rewrite H_osi_ptmadd in H_osi_pt_map.
  rewrite lemma_map_add_apply in H_osi_pt_map.
  inversion H_osi_pt_map.
  rewrite H0 in H_osi_pt_cont.
  inversion H_osi_pt_cont;auto.


  (*Case the pointer to PT (osi_ptaddr) is not osi_cpt_madd,*)
  (*that means its content doesn't change*)
  intro H_no_osi_ptmadd.
  elim H_valid_virt_s with osi osi_ptaddr mp osi_pt mp_va mp_ma;auto.
  intros pg H_pg H_acces.

  cut (pg=curr_page). 
  intro H_pg_curr.
  rewrite <- H_pg_curr.
  apply H_acces.
 
  rewrite <- H_mpma_ptmadd in H_pg.
  rewrite <- H_pt_curr_page in H_pg.
  inversion H_pg;auto.
  
  rewrite H_s' in H_osi_pt_map; simpl in H_osi_pt_map.
  rewrite H_upd_curr_pt in H_osi_pt_map.
  rewrite lemma_map_add_apply_distinct in H_osi_pt_map;auto.


  case (madd_eq osi_ptaddr cur_pt_madd).
  intro H_osipt_curpt.

  assert (mp=(map_drop vadd_eq curr_map va)) as H_mp_curr_map.
  (*relaciono el current map viejo y el actualizado*)

  rewrite H_s' in H_osi_pt_map; simpl in H_osi_pt_map.
  rewrite H_upd_curr_pt in H_osi_pt_map.
  rewrite H_osipt_curpt in H_osi_pt_map.
  rewrite lemma_map_add_apply in H_osi_pt_map.
  inversion H_osi_pt_map.
  rewrite H0 in H_osi_pt_cont.
  inversion H_osi_pt_cont;auto.

  rewrite <- H_mpva_va in H_mp_va_ma.
  rewrite H_mp_curr_map in H_mp_va_ma.
  rewrite map_drop_apply in H_mp_va_ma.
  inversion H_mp_va_ma;auto.

  (*Case the pointer to PT (osi_ptaddr) is not osi_cpt_madd,*)
  (*that means its content doesn't change*)
  
  intro H_osipt_no_mpma.
  
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.
  
  elim H_valid_virt_s with osi osi_ptaddr mp osi_pt mp_va mp_ma;auto.
  intros pg' H_pg' H_acces.

  cut (curr_page=pg'). 
  intro H_pg_pg'.
  rewrite H_pg_pg'.
  apply H_acces.

  rewrite <- H_mpma_ptmadd in H_pg'.
  rewrite <- H_pt_curr_page in H_pg'.
  inversion H_pg';auto.
  
  rewrite H_s' in H_osi_pt_map; simpl in H_osi_pt_map.
  rewrite H_upd_curr_pt in H_osi_pt_map.
  rewrite lemma_map_add_apply_distinct in H_osi_pt_map;auto.

  
  (*Case that mp_ma is not the address of the current map*)

  intro H_no_mpma_cpt.

 
  (*the witness is the same page that mp_ma has assigned in s*)
  (*We use the fact that s is a velid state*)
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.
  unfold valid_virtual_mapping in H_valid_virt_s.
  rewrite H_s'; simpl.
  rewrite H_upd_curr_pt.
  rewrite lemma_map_add_apply_distinct;auto.


  (*To use the H_valid_virt_s hypotesis, it's needed to*)
  (*consider cases of osi_ptaddr (pointer to PT), so the*)
  (*constrait of H_valid_virt_s are met*)
  case (madd_eq cur_pt_madd osi_ptaddr).
  
  (*Case osi_ptaddr is the one of the modified curr map*)
  intro H_osi_cptmadd.

  (*Case osi_ptaddr is not the one of the modified curr map*)
  2: intro H_no_osi_ptmadd.
  2: apply H_valid_virt_s with (osi:=osi) (pt_addr:=osi_ptaddr) 
  (mp:=mp) (pt:=osi_pt);auto.
  
  2: rewrite H_s' in H_osi_pt_map; simpl in H_osi_pt_map.
  2: rewrite H_upd_curr_pt in H_osi_pt_map.
  2: rewrite lemma_map_add_apply_distinct in H_osi_pt_map;auto.

  (*mp is the current map updated*)

  cut(mp=(map_drop vadd_eq curr_map va)).
  intro H_mp_curr_map.

  (*We consider cases of mp_va*)
  case (vadd_eq mp_va va).


  (*Case mp_va<>va*)
  2:intro H_no_mpva_va.
  2:apply H_valid_virt_s with (osi:=osi) (pt_addr:=cur_pt_madd)
   (mp:=curr_map) (pt:=curr_page);auto.
  2:rewrite H_osi_aos;auto.
  2:rewrite H_mp_curr_map in H_mp_va_ma.
  2:rewrite map_drop_apply_distinct in H_mp_va_ma;auto.

  (*Case mp_va=va*)
  intro H_mpva_va.

  (*This case is absurd, because mp_va and va are the same*)
  (*and they can't have different images in the mp mapping*)

  cut (False).
  intro;contradiction.

  rewrite H_mp_curr_map in H_mp_va_ma.
  rewrite <- H_mpva_va in H_mp_va_ma. 
  rewrite map_drop_apply in H_mp_va_ma;auto.
  inversion H_mp_va_ma;auto.


  clear H_valid_virt_s.

  (*It is pending the proof mp and current map relation*)

  rewrite H_s' in H_osi_pt_map; simpl in H_osi_pt_map.
  rewrite H_upd_curr_pt in H_osi_pt_map.
  rewrite H_osi_cptmadd in H_osi_pt_map.
  rewrite lemma_map_add_apply in H_osi_pt_map.
  inversion H_osi_pt_map.
  rewrite H0 in H_osi_pt_cont.
  inversion H_osi_pt_cont;auto.

  (*Case case that the considered os in not aos*)


  intro H_osi_ne_aos.

  intros H1 H2.

  cut (osi_ptaddr <> cur_pt_madd).
  intro.
  rewrite H_s' in H_osi_pt_map; simpl in H_osi_pt_map.
  rewrite H_upd_curr_pt in H_osi_pt_map.
  rewrite lemma_map_add_apply_distinct in H_osi_pt_map.
  intros.

  (*Use the fact that s is a valid state*)
  cut (valid_virtual_mapping (st_memory s) ctxt).
  2:apply H_valid_s.
  intro H_valid_virt_s.

  unfold valid_virtual_mapping in H_valid_virt_s.
  
  case (madd_eq ma0 cur_pt_madd).

  (*Case mp_ma <> current page maddr*)
  2: intros.
  2: rewrite H_s'; simpl.
  2: rewrite H_upd_curr_pt.
  2: rewrite lemma_map_add_apply_distinct.
  2: apply H_valid_virt_s with (pt_addr:=osi_ptaddr) (mp:=mp) (pt:=osi_pt); auto.
  2: auto.
 
  (*Case mp_ma=current page maddr is Absurd*)
 
  intros.
  cut False.
  contradiction.

  elim H_valid_virt_s with osi osi_ptaddr mp osi_pt va0 ma0; auto.
  intros pg H_pg_map H_owners.
  rewrite e in H_pg_map.
  cut (pg = curr_page).
  intro.
 
 
  rewrite H3 in H_owners.
  (*We assume that the owner of osi's current page is aos*)
  cut (page_owned_by curr_page = Os_PO (st_active_os s)).
  intro.
  
  unfold os_accessible in H_owners.
  destruct H_owners.
  destruct (ctxt_vadd_accessible ctxt va0).
  cut (page_owned_by curr_page = Os_PO osi).
  intro H_osi_owns.
  rewrite H4 in H_osi_owns.
  inversion H_osi_owns;auto.
  cut(true=true).
  intro H_true.
  apply (H5 H_true).
  trivial.
  cut (page_owned_by curr_page = Hyp_PO).
  intro H_hyp_owns.
  rewrite H4 in H_hyp_owns.
  inversion H_hyp_owns;auto.
  cut(false<>true).
  intro H_true.
  apply (H6 H_true).
  auto.
  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr.
  2:apply H_valid_s.
  unfold valid_current_page in H_valid_curr.
  unfold maddress_owner in H_valid_curr.

 

  elim H_valid_curr with (st_active_os s) aos p2m.
  intro pt_madd'.
  intros H_pt_madd' H_pt'_owner.
  cut (pt_madd'=cur_pt_madd).
  intro H_pt_pt'.
  rewrite H_pt_pt' in H_pt'_owner.
  destruct H_pt'_owner.
  rewrite <- H_pt_curr_page in H5.
  assumption.

  rewrite <- H_cur_page in H_pt_madd'.
  inversion H_pt_madd';auto.
  assumption.
  assumption.
  rewrite <- H_pt_curr_page in H_pg_map.
  inversion H_pg_map;auto.
  auto.
  cut(maddress_owner (st_memory s')(cur_pt_madd) <> maddress_owner (st_memory s')(osi_ptaddr)).
  intro H_owner.
  unfold not in H_owner.
  unfold not.
  intro H_pt_osi.
  rewrite H_pt_osi in H_owner.
  apply H_owner.
  trivial.
  cut (maddress_owner (st_memory s') cur_pt_madd = Os_PO (st_active_os s)).
  intro H_pt_madd_owner.
  rewrite H_pt_madd_owner.
  unfold maddress_owner.
  rewrite <- H_osi_pt_map.
  rewrite H2.
  unfold not.
  intro Absurd.
  inversion Absurd.
  auto.
  
  unfold maddress_owner.
  rewrite H_s'; simpl.
  rewrite H_upd_curr_pt.
  rewrite lemma_map_add_apply.
  simpl.

  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr.
  2:apply H_valid_s.
  
  unfold valid_current_page in H_valid_curr.

  elim H_valid_curr with (st_active_os s) aos p2m.
  intro pt_madd'.
  intros.
  cut (pt_madd'=cur_pt_madd).
  intro H_pt_pt'.

  rewrite H_pt_pt' in H. 
  rewrite H_pt_pt' in H0.
  unfold maddress_owner in H0.
  rewrite <- H_pt_curr_page in H0.
  apply H0.
  unfold os_current_vadd_mapping_madd in H_cur_page.
  destruct H_cur_page.

  inversion H;auto.
  assumption.
  assumption.


  (*Eighth Constraint*)
  unfold valid_current_page.
  intros osi system H_osi_system p2m H_sys_p2m.

  (*We divide the proof in cases, whether osi' is the same that osi or not*)
  case (os_ident_eq osi (st_active_os s)).
  intro H_case_aos.
  exists cur_pt_madd.
  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  elim H_cur_pt_madd.
  intros.
  destruct H1.
  cut (system=x).
  intro H_is_sys.
  cut (x0=p2m).
  intro H_is_p2m.
  rewrite H_is_p2m in H2.
  rewrite H_is_sys.
  simpl.
  assumption.
  
  unfold  hypervisor_mapping in H1.
  rewrite <- H_case_aos in H1.
  rewrite H_s' in H_sys_p2m; simpl in H_sys_p2m.
  rewrite <- H_sys_p2m in H1.
  inversion H1.
  trivial.


  rewrite H_s' in H_osi_system; simpl in H_osi_system.
  rewrite H_case_aos in H_osi_system.
  rewrite <- H_osi_system in H0.
  inversion H0;auto.
  
  rewrite H_s'; simpl.
  
  (*We extract the newmem definition*)
  generalize H_mem_pt_newmem.
  intro H_update_curr_pt.
  unfold update_pt_page in H_update_curr_pt.
  destruct H_update_curr_pt as (current_page, H_pt_current_page, H_update_curr_pt).
  destruct H_update_curr_pt as (current_map, H_pt_curr_map,H_update_curr_pt).
  split.
  rewrite H_update_curr_pt.
  unfold page_type.
  rewrite lemma_map_add_apply.
  simpl.
  trivial.

  unfold maddress_owner.
  rewrite H_update_curr_pt.
  rewrite lemma_map_add_apply.
  simpl.
  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr_pt_s.
  2:apply H_valid_s.


  unfold valid_current_page in H_valid_curr_pt_s.

  elim H_valid_curr_pt_s with osi system p2m.
  intros.
  destruct H1.
  cut (x=cur_pt_madd).
  intro H_x_pt.
  unfold maddress_owner in H2.
  rewrite H_x_pt in H2.
  rewrite <- H_pt_current_page in H2.
  assumption.
  clear H_valid_curr_pt_s.

  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  destruct H_cur_pt_madd.
  destruct H4.
  cut (x1=p2m).
  intro H_x1_p2m.
  cut (os_current_page x0=os_current_page system).
  intro H_x0_No.
  rewrite H_x1_p2m in H5.
  rewrite H_x0_No in H5.
  rewrite <- H0 in H5.
  inversion H5.
  trivial.
  rewrite <- H_case_aos in H3.
  rewrite H_s' in H_osi_system; simpl in H_osi_system.
  rewrite <- H_osi_system in H3.
  inversion H3;auto.
  rewrite H_s' in H_sys_p2m; simpl in H_sys_p2m.

 
  unfold hypervisor_mapping in H4.
  rewrite H_case_aos in H_sys_p2m.
  rewrite <- H_sys_p2m in H4.
  inversion H4.
  trivial.
  rewrite H_s' in H_osi_system; simpl in H_osi_system; auto.
  
  rewrite H_s' in H_sys_p2m; auto.

  (*In case that osi <> aos, the current page is the*)
  (*same that osi has in state s*)
  intro H_case_no_aos.

  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr_pt_s.
  2:apply H_valid_s.


  unfold valid_current_page in H_valid_curr_pt_s.

  elim H_valid_curr_pt_s with osi system p2m.
  intro osi_pt_madd.
  intros .
  exists osi_pt_madd.
  assumption.
  rewrite H_s'; simpl.
  split.

  (*We extract the newmem definition*)
  generalize H_mem_pt_newmem.
  intro H_update_curr_pt.
  unfold update_pt_page in H_update_curr_pt.
  destruct H_update_curr_pt as (current_page, H_pt_current_page, H_update_curr_pt).
  destruct H_update_curr_pt as (current_map, H_pt_curr_map,H_update_curr_pt).
  
  rewrite H_update_curr_pt.
  unfold page_type.
  rewrite lemma_map_add_apply_distinct.
  unfold page_type in H1.
  apply H1.

  (*Tey are different because they have different owners*)
  

  cut (maddress_owner (st_memory s) cur_pt_madd = Os_PO (st_active_os s)).
  intro H_pt_madd_owner.
  intro H_contrad.
  rewrite H_contrad in H_pt_madd_owner.
  destruct H1.
  rewrite H_pt_madd_owner in H2.
  inversion H2.
  auto.

  (*For this we can use valid_virtual_mapping on state s to get the owner*)
  (*of the current page pointer*)
  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  destruct H_cur_pt_madd as (aos,H_is_aos,H_cur_pt_madd ).
  destruct H_cur_pt_madd as (active_p2m, H_p2m,H_cur_pt_madd).

  elim H_valid_curr_pt_s with (st_active_os s) aos active_p2m.
  intro pt_madd'.
  intros.
  cut (pt_madd'=cur_pt_madd).
  intro H_same_pt_madd.
  rewrite <- H_same_pt_madd.
  apply H3.
  
  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  destruct H_cur_pt_madd.
  destruct H3.
  inversion H2;auto.

  assumption.
  assumption.
  3:rewrite H_s' in H_sys_p2m; auto.
  
  (*We extract the newmem definition*)
  generalize H_mem_pt_newmem.
  intro H_update_curr_pt.
  unfold update_pt_page in H_update_curr_pt.
  destruct H_update_curr_pt as (current_page, H_pt_current_page, H_update_curr_pt).
  destruct H_update_curr_pt as (current_map, H_pt_curr_map,H_update_curr_pt).
  
  rewrite H_update_curr_pt.
  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct.
  unfold maddress_owner in H1.
  apply H1.

  (*They ae different because they have different owners*)
  
  cut (maddress_owner (st_memory s) cur_pt_madd = Os_PO (st_active_os s)).
  intro H_pt_madd_owner.
  intro H_contrad.
  rewrite H_contrad in H_pt_madd_owner.
  destruct H1.
  rewrite H_pt_madd_owner in H2.
  inversion H2.
  auto.

  (*Again, we use valid_virtual_mapping on state s to get the owner*)
  (*of the current page pointer*)

  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  destruct H_cur_pt_madd as (aos,H_is_aos,H_cur_pt_madd ).
  destruct H_cur_pt_madd as (active_p2m, H_p2m,H_cur_pt_madd).

  elim H_valid_curr_pt_s with (st_active_os s) aos active_p2m.
  intro pt_madd'.
  intros.
  cut (pt_madd'=cur_pt_madd).
  intro H_same_pt_madd.
  rewrite <- H_same_pt_madd.
  apply H3.
  
  unfold os_current_vadd_mapping_madd in H_cur_pt_madd.
  destruct H_cur_pt_madd.
  inversion H2;auto.
  assumption.
  assumption.
  
  (*The new OSS doen't change for aos*)
  rewrite H_s' in H_osi_system; auto.

  (*Tenth Constraint*)
  unfold va_has_valid_pa.
  intro osi.

  (*We extract the newmem definition, used sereval times*)
  destruct H_mem_pt_newmem as
    (current_page, H_pt_current_page, H_mem_pt_newmem).
  destruct H_mem_pt_newmem  as
    (current_map, H_pt_curr_map,H_newmem_def).
  simpl in H_newmem_def.

  (*We use several taimes that the property holds in s*)
  assert (va_has_valid_pa (st_memory s) 
    (st_hypervisor s) ctxt) as H_va_has_pa_in_s.
  apply H_valid_s.

  (*We extract some os components from post-conditions*)
  destruct H_cur_pt_madd as (aos, H_is_aos,H_p2m).
(*  destruct H_cur_pt_madd as (aos', H_is_aos',H_p2m).
  assert (aos'=aos) as H_aos'_aos.
  rewrite <- H_is_aos in H_is_aos'.
  inversion H_is_aos';auto.
  rewrite H_aos'_aos in H_p2m.
  clear aos' H_is_aos' H_aos'_aos. *)
  destruct H_p2m as (p2m,H_p2m,H_cur_pt_madd).

  (*Ans some othes properties from preconditions*)
  destruct H 
    as (aos_cpt_madd',H_aos_cpt_madd',H_pre_del).
  destruct H_pre_del as (H_pre_all_del,H_pre_del).
  destruct H_pre_del as (H_hyper_nunning, H_va_no_acces).

  (*We must consider cases of osi*)
  case (os_ident_eq osi (st_active_os s)).

  (*In case osi is aos*)
  intro H_osi_aos.

  intros aos_pt_padd aos_pt_addr aos_pt_page 
    aos_mp aos_va aos_ma.
  intros H_aos_pt_page_addr H_pt_padd_madd 
    H_pt_page_cont H_aos_ma_va H_aos_va.

  (*we know that aos_va can't be va, because one*)
  (*is accessible and the other sin't*)
  assert (aos_va <> va) as H_aosva_no_va.
  intro H_aosva_va_abs.
  rewrite H_aosva_va_abs in H_aos_va;contradiction.  

  (*We need to discriminate cases ef osi_ptadd*)
  case (madd_eq aos_pt_addr cur_pt_madd).
  intro H_aospt_aoscpt.

  (*We establish the relation betwen the pages*)
  assert (aos_pt_page=
              (Page (PT_c (map_drop vadd_eq current_map va))
                    (page_owned_by current_page))
         ) as H_aospage_def.
  rewrite H_s' in H_aos_pt_page_addr; simpl in H_aos_pt_page_addr.
  rewrite H_newmem_def in H_aos_pt_page_addr.
  rewrite H_aospt_aoscpt in H_aos_pt_page_addr.
  rewrite lemma_map_add_apply in H_aos_pt_page_addr.
  inversion H_aos_pt_page_addr;auto.

  (*We establish the relation betwen the mappings*)
  assert (aos_mp = (map_drop vadd_eq current_map va)) as H_aos_mp_def.
  rewrite H_aospage_def in H_pt_page_cont.
  simpl in H_pt_page_cont.
  inversion H_pt_page_cont;auto.

  elim H_va_has_pa_in_s with (st_active_os s) (os_current_page aos)
    aos_pt_addr current_page current_map aos_va aos_ma;auto.

  intros aos_pa H_aospa.
  rewrite H_osi_aos; rewrite H_s'; simpl.
  exists aos_pa;auto.
  
  rewrite H_aospt_aoscpt;auto.

  rewrite H_aospt_aoscpt.
  exists p2m;auto.
  
  rewrite H_aos_mp_def in H_aos_ma_va.
  rewrite map_drop_apply_distinct 
    in H_aos_ma_va;auto.

  (*In case the considered pt madd is not osi cpt madd*)  
  intro H_aospt_no_aoscpt.

  elim H_va_has_pa_in_s with (st_active_os s) aos_pt_padd
    aos_pt_addr aos_pt_page aos_mp aos_va aos_ma;auto.

  intros aos_pa H_osipa.
  rewrite H_osi_aos; rewrite H_s'; simpl.
  exists aos_pa;auto.

  rewrite H_s' in H_aos_pt_page_addr; simpl in H_aos_pt_page_addr.
  rewrite H_newmem_def in H_aos_pt_page_addr.
  rewrite lemma_map_add_apply_distinct
    in H_aos_pt_page_addr;auto.

  rewrite H_s' in H_pt_padd_madd; simpl in H_pt_padd_madd;
    rewrite <- H_osi_aos;auto.


  (*In case osi is not aos*)
  intro H_osi_no_aos.
  
  intros osi_pt_padd osi_pt_addr osi_pt_page 
    osi_mp osi_va osi_ma.
  intros H_osi_pt_page_addr H_pt_padd_madd 
    H_pt_page_cont H_osi_ma_va H_osi_va.


  (*We must prove that the considered madd "osi_pt_addr"*)
  (*can not be the same that aos current page madd.*)

  (*We prove this based in the fact that they have*)
  (*different owners. In one case aos, in the other osi*)

  (*The owner of osi_pt_addr is osi*)
  assert (maddress_owner (st_memory s) osi_pt_addr 
    = Os_PO osi) as H_osipt_owner.

  destruct H_pt_padd_madd as (osi_p2m,H_osip2m,
    H_osipt_osipadd).

  assert (valid_hypervisor (st_hypervisor s') 
    (st_memory s) ) as H_valid_hyper.
  rewrite H_s'; simpl ;apply H_valid_s.

  elim H_valid_hyper with osi osi_pt_padd osi_p2m;auto.

  intro H_exist_ma.
  destruct H_exist_ma as (osi_pt_madd, H_osimadd,
    H_osimadd_prop).
  assert (osi_pt_addr=osi_pt_madd) as H_same_pt_madd.
  rewrite <- H_osipt_osipadd in H_osimadd.
  inversion H_osimadd;auto.

  rewrite <- H_same_pt_madd in H_osimadd_prop;auto.

  intro H_absurd.
  rewrite <- H_osipt_osipadd in H_absurd.
  inversion H_absurd;auto.
  

  (*The owner of osi_cpt_madd is osi*)
  assert (maddress_owner (st_memory s) cur_pt_madd 
    = Os_PO (st_active_os s)) as H_aoscpt_owner.

  assert (valid_current_page (st_hypervisor s) (st_memory s)
     (st_oss s)) as H_valid_curr_s. 
  apply H_valid_s.
  
  unfold valid_current_page in H_valid_curr_s.
  elim H_valid_curr_s with (st_active_os s) aos p2m.
  intros aos_cpt_addr H_aoscptaddr H_aoscpt_prop.
  assert (aos_cpt_addr=cur_pt_madd) as H_aos_cpt.
  rewrite <- H_cur_pt_madd in H_aoscptaddr.
  inversion H_aoscptaddr;auto.
  rewrite H_aos_cpt in H_aoscpt_prop.
  apply H_aoscpt_prop.

  apply H_is_aos.

  apply H_p2m.

  (*The considered madd osi_pt_addr is different from*)
  (*cur_pt_madd because they have different owners*)

  assert (osi_pt_addr<>cur_pt_madd) 
    as H_osipt_no_aoscpt.
  
  intro H_osipt_aoscpt_abs.

  rewrite H_osipt_aoscpt_abs in H_osipt_owner.
  rewrite H_osipt_owner in H_aoscpt_owner.
  inversion H_aoscpt_owner;contradiction.


  (*Once we know osi_pt isn't the aos cpt, we use*)
  (*the fact that s is a valid state*)
  elim H_va_has_pa_in_s with osi osi_pt_padd
    osi_pt_addr osi_pt_page osi_mp 
      osi_va osi_ma;auto.

  intros osi_pa H_osipa.
  rewrite H_s'; simpl.
  exists osi_pa;auto.

  rewrite H_s' in H_osi_pt_page_addr; simpl in H_osi_pt_page_addr.
  rewrite H_newmem_def in H_osi_pt_page_addr.
  rewrite lemma_map_add_apply_distinct 
    in H_osi_pt_page_addr;auto.
  rewrite H_s' in H_pt_padd_madd; simpl in H_pt_padd_madd;auto.

  apply exec_valid_cache_tlb with ctxt s (Del_hyper va); auto.
Qed.
  

(**********************)
(***Switch is secure***)
(**********************)

Theorem Switch_secure : 
  forall (s s':state),
  forall osi,
  valid_state ctxt s ->
  valid_exec ctxt s (Switch osi) s'->
  valid_state ctxt s'.
Proof.
  intros s s' osi H_valid_s H_valid_exec.
  generalize H_valid_exec; intros H_valid_exec'.
  inversion_clear H_valid_exec'.
  simpl in H.
  destruct H as (H_hyper_running,H_osi_no_hcall).
  simpl in H0.
  unfold post_switch in H0.

  unfold valid_state; rewrite H0; simpl.
  do 10 (try split); try (apply H_valid_s).
  unfold running_os_not_hypercall.
  destruct H_osi_no_hcall as [system H_system_map H_no_hcall].
  exists system; auto.

  (*Fourth Constraint*)
  intro.
  (*It holds in s, we use the lemma "valid_hyper_exec_mode"*)
  cut (valid_hyper_exec_mode (st_aos_activity s) (st_exec_mode s)).
  intro H_to_prove.
  cut (st_aos_activity s = st_aos_activity s).
  intro H_trivial_act.
  cut (st_exec_mode s = st_exec_mode s).
  intro H_trivial_exec.
  elim (valid_hyper_exec_mode_inv s s
            (H_to_prove)
            (H_trivial_act)
            (H_trivial_exec)).
  assumption.
  assumption.
  trivial.
  trivial.
  apply H_valid_s.

  (*Fifth Constraint*)
  unfold valid_untrusted_os_exec_mode.
  intros H_not_trusted H_contradict.
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running in H_contradict.
  inversion H_contradict.

  rewrite <- H0;
    apply exec_valid_cache_tlb with ctxt s (Switch osi); auto.
Qed.


(***********************)
(***Lswitch is secure***)
(***********************)

Theorem Lswitch_trusted_secure : 
  forall (s s':state),
  forall pa,
  valid_state ctxt s ->
  valid_exec ctxt s (Lswitch_trusted pa) s'->
  valid_state ctxt s'.
Proof.
  intros s s' pa H_valid_s H_valid_exec.
  generalize H_valid_exec; intro H_valid_exec'.
  inversion_clear H_valid_exec'.
  destruct H0 as (aos, H_is_aos, H_s').
  
  simpl in H_s'.

  unfold valid_state.
  rewrite H_s'; simpl.

  (*First Constraint*)
  split.
  unfold trusted_os_not_hypercall.
  intros trusted_osi trusted_os H_is_trusted_osi H_trusted_os_map.
  case (os_ident_eq (st_active_os s) trusted_osi).
  intro H_osi_active.
  rewrite H_osi_active in H_trusted_os_map.
  rewrite lemma_map_add_apply in H_trusted_os_map.
  injection H_trusted_os_map; intro H_trusted_os_value; rewrite H_trusted_os_value; auto.
  intro H_osi_not_active.
  cut (trusted_os_not_hypercall (ctxt_oss ctxt) (st_oss s)).
  intro H_trusted_no_hyp_s.
  2:apply H_valid_s.
  rewrite lemma_map_add_apply_distinct in H_trusted_os_map; auto.
  apply (H_trusted_no_hyp_s trusted_osi trusted_os H_is_trusted_osi); auto.

  (*Second Constraint*)
  split.
  unfold running_os_not_hypercall.
  exists (OS None pa).
  rewrite lemma_map_add_apply.
  trivial.
  trivial.
  
  (*Third Constraint*)
  split.
  apply H_valid_s.
  
  (*Fourth Constraint*)
  split.
  apply H_valid_s.
  
  (*Fifth Constraint*)
  split.
  apply H_valid_s.

  (*Sixth Constraint*)
  split.
  apply H_valid_s.

  (*Seventh Constraint*)
  split.
  apply H_valid_s.

  (*Eighth Constraint*)
  split.
  unfold valid_current_page.
  intros osi' system' H_osi'_oss' p2m' H_osi'_p2m'.

  (*We discriminate the cases of trusted_osi*)
  case (os_ident_eq osi' (st_active_os s)).
  
  Focus 2.
  (*In case trusted_osi is not aos_osi*)
  intro H_case_not_aos.
  
  cut (  valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  2:apply H_valid_s.
  intro H_valid_page_s.
  rewrite lemma_map_add_apply_distinct in H_osi'_oss'.
  apply (H_valid_page_s osi' system' H_osi'_oss' p2m' H_osi'_p2m').
  auto.

  (*In case trusted_osi is aos_osi*)
  intro H_case_aos.

  rewrite H_case_aos in H_osi'_oss'.
  rewrite lemma_map_add_apply in H_osi'_oss'.
  
  (*We destrcut the pre-conditions, so we find the requested madd*)
  destruct H as (H_os_running,H).
  destruct H as (H_trusted_aos,H).
  destruct H as (ma,H_pa_ma,H_pt_owner).
  exists ma.

  (*2 out of the 3 properties are trivial*)
  Focus 2.
  split.
  apply H_pt_owner.
  rewrite H_case_aos.
  apply H_pt_owner.
  
  (*For the remaining we use the pa, ma, and memory mappings*)
  (*extracted from pre conditions*)
  
  elim H_pa_ma.
  intros p2m2 H_p2m2_aos H_pa_ma_p2m2.
  rewrite H_pa_ma_p2m2.
  cut (system'=OS None pa).
  intro H_sys_none_pa.
  2:inversion H_osi'_oss';trivial.
  rewrite H_sys_none_pa;simpl.
  
  (*p2m' is the same one extracted from pre-conditions*)
  cut (p2m'=p2m2).
  intro H_is_p2m';rewrite H_is_p2m';trivial.
  rewrite H_case_aos in H_osi'_p2m'.
  rewrite <- H_p2m2_aos in H_osi'_p2m'.
  inversion H_osi'_p2m';trivial.

  (*Ninth Constraint*)
  split.
  apply H_valid_s.

  (*Tenth Constraint*)
  split.
  apply H_valid_s.

  rewrite <- H_s'; apply exec_valid_cache_tlb with ctxt s (Lswitch_trusted pa); auto.
Qed.

Theorem Lswitch_untrusted_secure : 
  forall (s s':state),
  forall osi pa,
  valid_state ctxt s ->
  valid_exec ctxt s (Lswitch_untrusted osi pa) s'->
  valid_state ctxt s'.
Proof.

  intros s s' osi pa H_valid_s H_valid_exec.
  generalize H_valid_exec; intro H_valid_exec'.
  inversion_clear H_valid_exec'.
  destruct H0 as (aos, H_is_aos, H_s').
  simpl in H_s'.

  unfold valid_state.
  rewrite H_s'; simpl.
  do 10 (try split); try (apply H_valid_s).

  (*First Constraint*)
  unfold trusted_os_not_hypercall.
  intros trusted_osi trusted_os H_is_trusted_osi H_trusted_os_map.
  case (os_ident_eq osi trusted_osi).
  intro H_osi_trusted_osi.
  rewrite H_osi_trusted_osi in H_trusted_os_map.
  rewrite lemma_map_add_apply in H_trusted_os_map.
  injection H_trusted_os_map; intro H_trusted_os_value; rewrite H_trusted_os_value; auto.
  intro H_osi_not_trusted_osi.
  cut (trusted_os_not_hypercall (ctxt_oss ctxt) (st_oss s)).
  intro H_trusted_no_hyp_s.
  2:apply H_valid_s.
  rewrite lemma_map_add_apply_distinct in H_trusted_os_map; auto.
  apply (H_trusted_no_hyp_s trusted_osi trusted_os H_is_trusted_osi); auto.

  (*Second Constraint*)
  unfold running_os_not_hypercall.
  case (os_ident_eq (st_active_os s) (osi)).
  intro H_case_osi_aos.
  exists (OS None pa).
  rewrite <- H_case_osi_aos.
  rewrite lemma_map_add_apply.
  trivial.
  trivial.

  intro H_no_osi.
  rewrite lemma_map_add_apply_distinct.
  apply H_valid_s.
  auto.

  (*Eighth Constraint*)
  unfold valid_current_page.
  intros osi' system' H_osi'_oss' p2m' H_osi'_p2m'.

  (*We discriminate the cases of osi'*)
  case (os_ident_eq osi' osi).
  
  Focus 2.
  (*In case osi is not osi*)
  intro H_case_not_osi.
  
  cut (  valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  2:apply H_valid_s.
  intro H_valid_page_s.
  rewrite lemma_map_add_apply_distinct in H_osi'_oss'.
  apply (H_valid_page_s osi' system' H_osi'_oss' p2m' H_osi'_p2m').
  auto.

  (*In case osi is aos_osi*)
  intro H_case_osi.

  rewrite H_case_osi in H_osi'_oss'.
  rewrite lemma_map_add_apply in H_osi'_oss'.
  
  (*We destrcut the pre-conditions, so we find the requested madd*)

  simpl in H.
  unfold pre_lswitch_untrusted in H.
  destruct H as (H_untrusted_osi,H).
  destruct H as (OS_osi,H_os_is_osi,H).
  destruct H as (H_OS_hcall,H).
  destruct H as (H_hyper_running,H).
  unfold pre_all_lswitch in H.
  destruct H as (ma, H_ma_pa,H).
  destruct H as (H_some_PT,H_owner,H_exmd).
  exists ma.

  (*2 out of the 3 properties are trivial*)
  Focus 2.
  split.
  apply H_some_PT.
  rewrite H_case_osi.
  apply H_owner.
  
  (*For the remaining we use the pa, ma, and memory mappings*)
  (*extracted from pre conditions*)
  
  elim H_ma_pa.
  intros p2m2 H_p2m2_aos H_pa_ma_p2m2.
  rewrite H_pa_ma_p2m2.
  cut (system'=OS None pa).
  intro H_sys_none_pa.
  2:inversion H_osi'_oss';trivial.
  rewrite H_sys_none_pa;simpl.
  
  (*p2m' is the same one extracted from pre-conditions*)
  cut (p2m'=p2m2).
  intro H_is_p2m';rewrite H_is_p2m';trivial.
  rewrite H_case_osi in H_osi'_p2m'.
  rewrite <- H_p2m2_aos in H_osi'_p2m'.
  inversion H_osi'_p2m';trivial.

  rewrite <- H_s'; 
    apply exec_valid_cache_tlb with ctxt s (Lswitch_untrusted osi pa); auto.
Qed.

(*********************)
(***Hcall is secure***)
(*********************)

Theorem Hcall_secure : 
  forall (s s':state),
  forall hp,
  valid_state ctxt s ->
  valid_exec ctxt s (Hcall hp) s'->
  valid_state ctxt s'.
Proof.
  intros s s' hp H_valid_s H_valid_exec;
    generalize H_valid_exec; intro H_valid_exec'.
  inversion_clear H_valid_exec.
  simpl in H.
  unfold pre_hcall in H.
  destruct H as (H_os_running,H_untrusted_aos).
  simpl in H0.
  unfold post_hcall in H0.
  destruct H0 as (aos,H_is_aos,H_s').
  

  unfold valid_state.
  do 10 (try split); try (rewrite H_s'; apply H_valid_s).

  (*First Constraint*)
  unfold trusted_os_not_hypercall.
  intros trusted_osi trusted_os H_is_trusted_osi H_trusted_os_map.
  case (os_ident_eq (st_active_os s) trusted_osi).
  intro H_osi_active.
  unfold trusted_aos in H_untrusted_aos.
  contradiction H_untrusted_aos.
  rewrite H_osi_active; auto.
  intro H_osi_not_active.
  cut (trusted_os_not_hypercall (ctxt_oss ctxt) (st_oss s)).
  intro H_trusted_no_hyp_s.
  2:apply H_valid_s.
  rewrite H_s' in H_trusted_os_map; simpl in H_trusted_os_map.
  rewrite lemma_map_add_apply_distinct in H_trusted_os_map; auto.
  apply (H_trusted_no_hyp_s trusted_osi trusted_os H_is_trusted_osi); auto.

  (*Second Constraint*)
  unfold running_os_not_hypercall.
  exists (OS (Some hp) (os_current_page aos)).
  rewrite H_s'; simpl.
  rewrite lemma_map_add_apply;auto.
  rewrite H_s'; simpl.
  intro; discriminate.

  (*Third Constraint*)
  unfold valid_hyper_exec_mode.
  rewrite H_s'; auto.

  (*Fourth Constraint*)
  unfold valid_trusted_os_exec_mode.
  rewrite H_s'; simpl.
  intro H_trusted_aos.
  unfold trusted_aos in H_untrusted_aos.
  rewrite H_trusted_aos in H_untrusted_aos.
  cut False.
  intro H_False;elim H_False.
  elim H_untrusted_aos;auto.

  (*Fifth Constraint*)
  unfold valid_untrusted_os_exec_mode.
  intro H_redun;clear H_redun.
  rewrite H_s'; simpl.
  intro H_OSA_contr.
  inversion H_OSA_contr;auto.

  (*Eighth Constraint*)
  rewrite H_s'; simpl.
  unfold valid_current_page.
  intro osi.
  case (os_ident_eq osi (st_active_os s)).
  intro H_osi_aos.
  rewrite H_osi_aos.
  intros system H_is_system.
  assert (system=(OS (Some hp) (os_current_page aos))) as H_system_def.
  rewrite lemma_map_add_apply in H_is_system.
  inversion H_is_system;auto.
  rewrite H_system_def.
  simpl.
  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr_s.
  2:apply H_valid_s.
  apply H_valid_curr_s;auto.

  (*If they are different, it holds for validity of state s*)
  intro H_no_osi_aos.
  rewrite lemma_map_add_apply_distinct;auto.
  cut (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s)).
  intro H_valid_curr_s.
  2:apply H_valid_s.
  apply H_valid_curr_s;auto.

  apply exec_valid_cache_tlb with ctxt s (Hcall hp); auto.
Qed.

(************************)
(***Ret_ctrl is secure***)
(************************)

Theorem Ret_ctrl_secure : 
  forall (s s':state),
  valid_state ctxt s ->
  valid_exec ctxt s Ret_ctrl s'->
  valid_state ctxt s'.
Proof.
  intros s s' H_valid_s H_valid_exec;
    generalize H_valid_exec; intro H_valid_exec';
    inversion_clear H_valid_exec.
  simpl in H.
  unfold pre_ret_ctrl in H.
  simpl in H0.
  unfold post_ret_ctrl in H0.

  (*We establish the relation between s and s' components*)

  (*st_active_os*)
  elim H0 with Active_OS;auto.
  2:intro absurd;destruct absurd as (absurd1,absurd2).
  2:elim absurd1;split;discriminate.
  intro H_assign;destruct H_assign as (H_t,H_same_aos).
  clear H_t;simpl in H_same_aos.
  

  (*st_oss*)
  elim H0 with Guest_OSs;auto.
  2:intro absurd;destruct absurd as (absurd1,absurd2).
  2:elim absurd1;split;discriminate.
  intro H_assign;destruct H_assign as (H_t,H_same_oss).
  clear H_t;simpl in H_same_oss.
  

  (*exec_mode and aos_activity*)
  elim H0 with OS_Activity.
  intro absurd;destruct absurd as (absurd1,absurd2).
  destruct absurd1 as (H_true,absurd1).
  elim absurd1;auto.
  intros H_assign.
  destruct H_assign as (H_t, H_assign).
  clear H_t; destruct H_assign as (H_newexmd,H_newactiv).
  simpl in H_newexmd;simpl in H_newactiv.


  (*hypervisor*)
  elim H0 with Hypervisor_SC.
  2:intro absurd;destruct absurd as (absurd1,absurd2).
  2:elim absurd1;split;discriminate.
  intro H_assign;destruct H_assign as (H_t,H_same_hyper).
  clear H_t;simpl in H_same_hyper.
  

  (*memory*)
  elim H0 with System_Memory.
  2:intro absurd;destruct absurd as (absurd1,absurd2).
  2:elim absurd1;split;discriminate.
  intro H_assign;destruct H_assign as (H_t,H_same_mem).
  clear H_t;simpl in H_same_mem.
  
  clear H0.

  unfold valid_state.
  
  (*First Constraint*)
  split.
  rewrite <- H_same_oss.
  apply H_valid_s.

  (*Second Constraint*)
  split.
  rewrite <- H_same_aos; rewrite <- H_same_oss.
  unfold running_os_not_hypercall.
  assert (running_os_not_hypercall (st_active_os s) (st_aos_activity s) (st_oss s)) as H_run_no_H_s.
  apply H_valid_s.
  destruct H_run_no_H_s as (aos, H_aos, H_aos_prop).
  exists aos;auto.

  (*Third Constraint*)
  split.
  unfold valid_hyper_exec_mode.
  rewrite H_newactiv; rewrite H_newexmd.
  auto.

  (*Fourth Constraint*) 
  split.
  rewrite <- H_same_aos; rewrite H_newexmd.
  unfold valid_trusted_os_exec_mode.
  auto.
  
  (*Fifth Constraint*)
  split.
  rewrite <- H_same_aos; rewrite H_newexmd; rewrite H_newactiv.
  unfold valid_untrusted_os_exec_mode.
  intros H_untrusted_aos H_OSA_contr.
  inversion H_OSA_contr;auto.
  
  (*Sixth Constraint*)
  split.
  rewrite <- H_same_hyper; rewrite <- H_same_mem.
  apply H_valid_s.

  (*Seventh Constraint*)
  split.
  rewrite <- H_same_mem.
  apply H_valid_s.

  (*Eighth, Ninth and Tenth Constraint*)
  rewrite <- H_same_hyper; rewrite <- H_same_mem; rewrite <- H_same_oss.
  do 3 (try split); try (apply H_valid_s).
  apply exec_valid_cache_tlb with ctxt s Ret_ctrl; auto.
Qed.

(*********************)
(***Chmod is secure***)
(*********************)

Theorem Chmod_secure : 
  forall (s s':state),
  valid_state ctxt s ->
  valid_exec ctxt s Chmod s'->
  valid_state ctxt s'.
Proof.
  intros s s' H_valid_s H_valid_exec;
    generalize H_valid_exec; intro H_valid_exec';
    inversion_clear H_valid_exec.
  unfold pre_action in H.
  unfold pre_chmod in H.
  destruct H as (H_hyper_running, H_aos_no_hcall).
  simpl in H0.
  unfold post_chmod in H0.
  elim H0;clear H0.


  (*Trusted case*)
  
  intro H_post.
  destruct H_post as (H_trusted_aos,H_update_exmd_activ).


  (*We establish the relation between s and s' components*)

  (*st_active_os*)
  elim H_update_exmd_activ with Active_OS;auto.
  2:intro absurd;destruct absurd as (absurd1,absurd2).
  2:elim absurd1;split;discriminate.
  intro H_assign;destruct H_assign as (H_t,H_same_aos).
  clear H_t;simpl in H_same_aos.
  

  (*st_oss*)
  elim H_update_exmd_activ with Guest_OSs;auto.
  2:intro absurd;destruct absurd as (absurd1,absurd2).
  2:elim absurd1;split;discriminate.
  intro H_assign;destruct H_assign as (H_t,H_same_oss).
  clear H_t;simpl in H_same_oss.
  

  (*exec_mode and aos_activity*)
  elim H_update_exmd_activ with OS_Activity.
  intro absurd;destruct absurd as (absurd1,absurd2).
  destruct absurd1 as (H_true,absurd1).
  elim absurd1;auto.
  intros H_assign.
  destruct H_assign as (H_t, H_assign).
  clear H_t; destruct H_assign as (H_newexmd,H_newactiv).
  simpl in H_newexmd;simpl in H_newactiv.


  (*hypervisor*)
  elim H_update_exmd_activ with Hypervisor_SC.
  2:intro absurd;destruct absurd as (absurd1,absurd2).
  2:elim absurd1;split;discriminate.
  intro H_assign;destruct H_assign as (H_t,H_same_hyper).
  clear H_t;simpl in H_same_hyper.
  

  (*memory*)
  elim H_update_exmd_activ with System_Memory.
  2:intro absurd;destruct absurd as (absurd1,absurd2).
  2:elim absurd1;split;discriminate.
  intro H_assign;destruct H_assign as (H_t,H_same_mem).
  clear H_t;simpl in H_same_mem.
  
  clear H_update_exmd_activ.

  unfold valid_state.
  

  (*First Constraint*)
  split.
  rewrite <- H_same_oss.
  apply H_valid_s.


  (*Second Constraint*)
  split.
  rewrite <- H_same_oss; rewrite <- H_same_aos; rewrite H_newactiv.
  unfold running_os_not_hypercall.
  destruct H_aos_no_hcall as (aos, H_is_aos, H_aos_no_hcall).
  exists aos;auto.


  (*Third Constraint*)
  split.
  rewrite H_newexmd; rewrite H_newactiv.
  unfold valid_hyper_exec_mode.
  auto.


  (*Fourth Constraint*)
  split.
  rewrite <- H_same_aos;rewrite H_newexmd.
  unfold valid_trusted_os_exec_mode.
  auto.


  (*Fifth Constraint*)
  split.
  rewrite <- H_same_aos;rewrite H_newexmd;rewrite H_newactiv.
  unfold valid_untrusted_os_exec_mode.
  intro H_untrusted_aos.
  unfold trusted_aos in H_trusted_aos.
  rewrite H_trusted_aos in H_untrusted_aos.
  inversion H_untrusted_aos;auto.


  (*Sixth Constraint*)
  split.
  rewrite <- H_same_hyper; rewrite <- H_same_mem.
  apply H_valid_s.


  (*Seventh Constraint*)
  split.
  rewrite <- H_same_mem.
  apply H_valid_s.


  (*octavo subgoal*)
  rewrite <- H_same_hyper; rewrite <- H_same_mem; rewrite <- H_same_oss.
  do 3 (try split); try (apply H_valid_s).
  apply exec_valid_cache_tlb with ctxt s Chmod; auto.

  (*Caso Untrusted*)
  intro H_post.
  destruct H_post as (H_untrusted_aos,H_update_exmd_activ).


  (* We establish the relation between s and s' components *)

  (* st_active_os *)
  elim H_update_exmd_activ with Active_OS;auto.
  2:intro absurd;destruct absurd as (absurd1,absurd2).
  2:elim absurd1;split;discriminate.
  intro H_assign;destruct H_assign as (H_t,H_same_aos).
  clear H_t;simpl in H_same_aos.
  

  (*st_oss*)
  elim H_update_exmd_activ with Guest_OSs;auto.
  2:intro absurd;destruct absurd as (absurd1,absurd2).
  2:elim absurd1;split;discriminate.
  intro H_assign;destruct H_assign as (H_t,H_same_oss).
  clear H_t;simpl in H_same_oss.
  

  (*exec_mode and aos_activity*)
  elim H_update_exmd_activ with OS_Activity.
  intro absurd;destruct absurd as (absurd1,absurd2).
  destruct absurd1 as (H_true,absurd1).
  elim absurd1;auto.
  intros H_assign.
  destruct H_assign as (H_t, H_assign).
  clear H_t; destruct H_assign as (H_newexmd,H_newactiv).
  simpl in H_newexmd;simpl in H_newactiv.


  (*hypervisor*)
  elim H_update_exmd_activ with Hypervisor_SC.
  2:intro absurd;destruct absurd as (absurd1,absurd2).
  2:elim absurd1;split;discriminate.
  intro H_assign;destruct H_assign as (H_t,H_same_hyper).
  clear H_t;simpl in H_same_hyper.
  

  (*memory*)
  elim H_update_exmd_activ with System_Memory.
  2:intro absurd;destruct absurd as (absurd1,absurd2).
  2:elim absurd1;split;discriminate.
  intro H_assign;destruct H_assign as (H_t,H_same_mem).
  clear H_t;simpl in H_same_mem.
  
  clear H_update_exmd_activ.

  unfold valid_state.
  

  (*First Constraint*)
  split.
  rewrite <- H_same_oss.
  apply H_valid_s.


  (*Second Constraint*)
  split.
  rewrite <- H_same_oss; rewrite <- H_same_aos; rewrite H_newactiv.
  unfold running_os_not_hypercall.
  destruct H_aos_no_hcall as (aos, H_is_aos, H_aos_no_hcall).
  exists aos;auto.


  (*Third Constraint*)
  split.
  rewrite H_newexmd; rewrite H_newactiv.
  unfold valid_hyper_exec_mode.
  intro H_OSA_contr;inversion H_OSA_contr.


  (*Fourth Constraint*)
  split.
  rewrite <- H_same_aos;rewrite H_newexmd.
  unfold valid_trusted_os_exec_mode.
  intro H_trusted_aos.
  unfold trusted_aos in H_untrusted_aos.
  rewrite H_trusted_aos in H_untrusted_aos.
  unfold not in H_untrusted_aos; elim H_untrusted_aos;auto.


  (*Fifth Constraint*)
  split.
  rewrite <- H_same_aos;rewrite H_newexmd.
  unfold valid_untrusted_os_exec_mode.
  auto.


  (*Sixth Constraint*)
  split.
  rewrite <- H_same_hyper; rewrite <- H_same_mem.
  apply H_valid_s.


  (*Seventh Constraint*)
  split.
  rewrite <- H_same_mem.
  apply H_valid_s.


  (*Eighth, Ninth and Tenth Constraint*)
  rewrite <- H_same_hyper; rewrite <- H_same_mem;
    rewrite <- H_same_oss.
  do 3 (try split); try apply H_valid_s.
  apply exec_valid_cache_tlb with ctxt s Chmod; auto.
Qed.

(**************************)
(***Pin Unpin are secure***)
(**************************)

Theorem Page_pin_trusted_secure:
  forall (s s':state),
  forall pa pg_type,
  valid_state ctxt s ->
  valid_exec ctxt s (Page_pin_trusted pa pg_type) s'->
  valid_state ctxt s'.
Proof.
  intros s s' pa pg_type H_valid_s H_valid_exec;
    generalize H_valid_exec; intro H_valid_exec';
    inversion_clear H_valid_exec'.
  simpl in H.
  unfold pre_page_pin_trusted in H.
  destruct H as (H_os_runnig, H_pre).
  destruct H_pre as (H_trusted_aos, H_pre_pin).
  unfold pre_all_pin in H_pre_pin.
  destruct H_pre_pin as (H_free_pa, H_free_ma).
  clear H_free_ma. 
  destruct H_free_pa as (p2m, H_p2m,H_free_pa).
  
  simpl in H0.
  unfold post_page_pin_trusted in H0.
  unfold post_all_pin in H0.

  destruct H0 as (ma,H_post_pin).
  destruct H_post_pin as (ma_pg, H_ma_pg,H_post_pin).
  destruct H_post_pin as (H_pg_owner,H_post_pin).
  destruct H_post_pin as (H_pg_content,H_post_pin).
  destruct H_post_pin as (p2m',H_p2m',H_post_pin).
  assert (p2m=p2m') as H_p2m'_p2m.
  unfold hypervisor_mapping in H_p2m'.
  rewrite <- H_p2m in H_p2m';inversion H_p2m';auto.
  rewrite <- H_p2m'_p2m in H_post_pin.
  clear p2m' H_p2m' H_p2m'_p2m.
  destruct H_post_pin as (aos,H_is_aos,H_s'_def).
    
  (*We establish the relation between s and s' components*)

  assert (st_active_os s' = st_active_os s ) as H_same_aos.
  rewrite H_s'_def;auto.
  assert (st_oss s' = map_add os_ident_eq (st_oss s) (st_active_os s)
              (OS None (os_current_page aos))) as H_new_oss.
  rewrite H_s'_def;auto.
  assert (st_exec_mode s' = st_exec_mode s) as H_same_exmd.
  rewrite H_s'_def;auto.
  assert (st_aos_activity s' = st_aos_activity s) as H_same_activ.
  rewrite H_s'_def;auto.
  assert (st_hypervisor s' = (map_add os_ident_eq (st_hypervisor s) 
              (st_active_os s) (map_add padd_eq p2m pa ma))) as H_new_hyper.
  rewrite H_s'_def;auto.
  assert (st_memory s' = map_add madd_eq (st_memory s) ma
                (new_page (Some pg_type) (Os_PO (st_active_os s)))  ) as H_new_mem.
  rewrite H_s'_def;auto.

  clear H_s'_def.
  
  unfold valid_state.

  (*First Constraint*)
  split.
  unfold trusted_os_not_hypercall.
  intros trusted_osi trusted_os H_is_trusted_osi H_trusted_os_map.
  case (os_ident_eq (st_active_os s) trusted_osi).
  intro H_osi_active.
  rewrite H_new_oss in H_trusted_os_map.
  rewrite H_osi_active in H_trusted_os_map.
  rewrite lemma_map_add_apply in H_trusted_os_map.
  injection H_trusted_os_map; intro H_trusted_os_value; rewrite H_trusted_os_value; auto.
  intro H_osi_not_active.
  cut (trusted_os_not_hypercall (ctxt_oss ctxt) (st_oss s)).
  intro H_trusted_no_hyp_s.
  2:apply H_valid_s.
  rewrite H_new_oss in H_trusted_os_map.
  rewrite lemma_map_add_apply_distinct in H_trusted_os_map; auto.
  apply (H_trusted_no_hyp_s trusted_osi trusted_os H_is_trusted_osi); auto.

  (*Second Constraint*)
  split.
  rewrite H_same_aos; rewrite H_same_activ.
  unfold running_os_not_hypercall.
  exists (OS None (os_current_page aos));auto.
  rewrite H_new_oss.
  rewrite lemma_map_add_apply;auto.

  (*Third Constraint*)
  split.
  rewrite H_same_activ; rewrite H_same_exmd.
  apply H_valid_s.

  (*Fourth Constraint*)
  split.
  rewrite H_same_aos; rewrite H_same_exmd.
  apply H_valid_s.

  (*Fifth Constraint*)
  split.
  rewrite H_same_aos; rewrite H_same_activ; rewrite H_same_exmd.
  apply H_valid_s.

  (*Sixth Constraint*)
  split.
  
  unfold valid_hypervisor.
  intro osi.
  
  (*We discriminate the cases of osi*)
  case (os_ident_eq (st_active_os s) osi).


  (*In case osi is aos*)
  intro H_osi_aos.
  
  rewrite <- H_osi_aos.

  intro aos_pa.
  
  
  case (padd_eq aos_pa pa).
  
  
  (*In case osi is aos, and the considered padd is pa*)
  intro H_aospa_pa.
  
  rewrite H_aospa_pa.
  intros p2m' H_p2m'.
  left.
  
  exists ma.
  rewrite H_new_hyper in H_p2m'.
  rewrite lemma_map_add_apply in H_p2m'.
  inversion H_p2m';auto.
  rewrite lemma_map_add_apply;auto.

  rewrite H_new_mem.
  unfold maddress_owner.
  rewrite lemma_map_add_apply.
  unfold new_page.
  case pg_type;auto.


  (*In case osi is aos, and the considered padd is not pa*)
  
  intro H_no_aospa_pa.
  intros p2m'
  H_p2m'.

  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) as H_valid_hyper_s.
  apply H_valid_s.

  rewrite H_new_hyper in H_p2m'.
  rewrite lemma_map_add_apply in H_p2m'.
  inversion H_p2m'.

  elim H_valid_hyper_s with (st_active_os s) aos_pa p2m;auto.
  
  (*The hypothesis is an OR, we must prove both branches*)
  intro H_exist_ma'.

  (*Left branch of OR*)
  left.
  destruct H_exist_ma' as (ma', H_p2m_ma', H_ma'_owner).
  exists ma';auto.
  rewrite lemma_map_add_apply_distinct;auto.

  rewrite H_new_mem.
  
  case (madd_eq ma ma').

  intro H_ma'_ma.
  rewrite <- H_ma'_ma.
  unfold maddress_owner.
  rewrite lemma_map_add_apply.
  unfold new_page.
  case pg_type;auto.


  intro H_no_ma'_ma.

  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct;auto.


  (*Right branch of OR*)
  right.
  rewrite lemma_map_add_apply_distinct;auto.


  (*In case osi is not aos*)
  intro H_no_osi_aos.

  intros osi_pa osi_p2m H_osi_p2m.

  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) as H_valid_hyper_s.
  apply H_valid_s.

  elim H_valid_hyper_s with osi osi_pa osi_p2m;auto. 
  intro H_exists_ma'.


  (*The hypothesis is an OR, we must prove both branches*)  
  left.
  destruct H_exists_ma' as (ma', H_osi_p2m_ma', H_ma'_owner).
  exists ma';auto.
  
  rewrite H_new_mem.
    
  assert (ma'<>ma).
  (*diferente owner*)
  unfold not;intro H_ma'_ma_abs.
  unfold maddress_owner in H_ma'_owner.
  rewrite <- H_ma'_ma_abs in H_ma_pg.
  rewrite <- H_ma_pg in H_ma'_owner.
  rewrite H_pg_owner in H_ma'_owner.
  inversion H_ma'_owner.

  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct;auto.

  rewrite H_new_hyper in H_osi_p2m.
  rewrite lemma_map_add_apply_distinct in H_osi_p2m;auto.
  

  (*Seventh Constraint*)
  split.

  
  intro osi.

  (*We discriminate the cases of osi*)
  
  case (os_ident_eq osi (st_active_os s)).

  
  (*In case osi is aos*)
  intro H_osi_aos.
  
  rewrite H_osi_aos.
  intro aos_pt_madd.

  case (madd_eq aos_pt_madd ma).

  (*In case the considered madd is ma, the corresponding page
  is empty, if we suppose that a pin to a PT has been made.
  If we soppose that a pin to a RW page has been made, then
  there is no mapping to prove the property on*)

  intro H_aosptmadd_ma.
  rewrite H_aosptmadd_ma.


  intros ma_map' ma_page' H_ma_page'.
  rewrite H_new_mem in H_ma_page'.
  rewrite lemma_map_add_apply in H_ma_page'.
  inversion_clear H_ma_page'.
  case pg_type.
  simpl.
  
  intro H_ma_map'.
  inversion_clear H_ma_map'.
  intro H_temp;clear H_temp.
  simpl.
  intros ma_map_va ma_map_ma H_ma_va_ma_map.
  inversion H_ma_va_ma_map.

  (*RW type, doesn't met the hypothesis of the property*)
  simpl.
  intro H_ma_map'.
  inversion_clear H_ma_map'.

  (*In case the considered madd is not ma*)
  intro H_no_aosptmadd_ma.

  
  intros aos_map aos_page
    H_aos_pt_page H_aos_pg_map H_aos_pg_owner
      aos_map_va aos_map_ma H_aos_va_ma_map.
  
  assert (valid_virtual_mapping (st_memory s) ctxt) 
    as H_valid_virtual_s.
  apply H_valid_s;auto.

  
  assert (aos_map_ma <> ma) as H_no_aosmapma_ma.
  (*There can not be an aos_map that mets this, and has ma on it*)
  unfold not;intro H_aosmapma_ma_abs.
  rewrite H_aosmapma_ma_abs in H_aos_va_ma_map.

  rewrite H_new_mem in H_aos_pt_page.
  rewrite lemma_map_add_apply_distinct in H_aos_pt_page;auto.

  elim H_valid_virtual_s with (st_active_os s) aos_pt_madd aos_map 
    aos_page aos_map_va ma;auto.

  intros aos_map_ma_pg H_aos_map_ma_pg H_aos_map_ma_prop.

  assert (aos_map_ma_pg=ma_pg) as H_ma_aos_pg.
  rewrite <- H_aos_map_ma_pg in H_ma_pg.
  inversion H_ma_pg;auto.
  rewrite <- H_ma_aos_pg in H_pg_owner.

  rewrite H_pg_owner in H_aos_map_ma_prop.
  destruct H_aos_map_ma_prop as (H_va_acces,H_no_va_acces).

  unfold os_accessible in  H_va_acces.
  unfold os_accessible in  H_no_va_acces.

  destruct (ctxt_vadd_accessible ctxt aos_map_va).
  assert (No_Owner = Os_PO (st_active_os s)) as Absurd.
  apply  H_va_acces;auto.
  inversion Absurd.

  assert (No_Owner = Hyp_PO) as Absurd.
  apply  H_no_va_acces;auto.
  inversion Absurd.


  rewrite H_new_mem.
  rewrite lemma_map_add_apply_distinct;auto.

  rewrite H_new_mem in H_aos_pt_page.
  rewrite lemma_map_add_apply_distinct in H_aos_pt_page;auto.

  elim H_valid_virtual_s with (st_active_os s) aos_pt_madd aos_map 
    aos_page aos_map_va aos_map_ma;auto.

  intros aos_map_ma_pg H_aos_map_ma_pg H_aos_map_ma_prop.
  exists aos_map_ma_pg;auto.
  


  (*In case osi is not aos*)
  intro H_no_osi_aos.

  intros osi_pt_madd osi_map osi_page
    H_osi_pt_page H_osi_pg_map H_osi_pg_owner
      osi_map_va osi_map_ma H_osi_va_ma_map.
  
  assert (valid_virtual_mapping (st_memory s) ctxt) 
    as H_valid_virtual_s.
  apply H_valid_s;auto.

  assert (ma <> osi_pt_madd) as H_no_osiptmadd_ma.

  (*It can not be the same, because ma has owner oas in mem', not osi*)
  unfold not;intro H_osiptmadd_ma_abs.
  rewrite H_new_mem in H_osi_pt_page.
  rewrite H_osiptmadd_ma_abs in H_osi_pt_page.
  rewrite lemma_map_add_apply in H_osi_pt_page.
  assert (osi_page = (new_page (Some pg_type) (Os_PO (st_active_os s)))) as H_osi_page_def.
  inversion H_osi_pt_page;auto.
  assert (Os_PO (st_active_os s)=page_owned_by osi_page) as H_aos_owns_osi_pg.
  rewrite H_osi_page_def; case pg_type; auto.
  rewrite <- H_aos_owns_osi_pg in H_osi_pg_owner.
  inversion H_osi_pg_owner;auto.


  (*It can not be the same, because ma has owner oas in mem', not osi*)
  assert (ma <> osi_map_ma) as H_no_osimapma_ma.  

  unfold not;intro H_osimapma_ma_abs.
  
  rewrite H_new_mem in H_osi_pt_page.
  rewrite lemma_map_add_apply_distinct in H_osi_pt_page;auto.
  elim H_valid_virtual_s with osi osi_pt_madd osi_map 
    osi_page osi_map_va osi_map_ma;auto.

  intros osi_map_ma_pg H_osi_map_ma_pg H_osi_map_ma_prop.


  assert (osi_map_ma_pg=ma_pg) as H_ma_osi_pg.
  rewrite H_osimapma_ma_abs in H_ma_pg.
  rewrite <- H_osi_map_ma_pg in H_ma_pg.
  inversion H_ma_pg;auto.
  rewrite <- H_ma_osi_pg in H_pg_owner.

  rewrite H_pg_owner in H_osi_map_ma_prop.
  destruct H_osi_map_ma_prop as (H_va_acces,H_no_va_acces).

  unfold os_accessible in  H_va_acces.
  unfold os_accessible in  H_no_va_acces.

  destruct (ctxt_vadd_accessible ctxt osi_map_va).
  assert (No_Owner = Os_PO osi) as Absurd.
  apply  H_va_acces;auto.
  inversion Absurd.

  assert (No_Owner = Hyp_PO) as Absurd.
  apply  H_no_va_acces;auto.
  inversion Absurd.


  rewrite H_new_mem in H_osi_pt_page.
  rewrite lemma_map_add_apply_distinct in H_osi_pt_page;auto.
  
  rewrite H_new_mem.
  rewrite lemma_map_add_apply_distinct;auto.
 
  elim H_valid_virtual_s with osi osi_pt_madd osi_map 
    osi_page osi_map_va osi_map_ma;auto.

  intros osi_map_ma_pg H_osi_map_ma_pg H_osi_map_ma_prop.
  exists osi_map_ma_pg;auto.


  (*Eighth Constraint*)
  split.
  unfold valid_current_page.

  intro osi.

  (*We discriminate the cases of osi*)
  case (os_ident_eq osi (st_active_os s)).


  Focus 2.
  (*In case osi is not aos_osi*)
  intro H_no_osi_aos.
  
  intros system H_osi_system osi_p2m H_osi_p2m.

  rewrite H_new_oss in H_osi_system.
  rewrite lemma_map_add_apply_distinct in H_osi_system;auto.
  
  rewrite H_new_hyper in H_osi_p2m.
  rewrite lemma_map_add_apply_distinct in H_osi_p2m;auto.

  assert (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s) ) as H_valid_cur_s.
  apply H_valid_s.

  elim H_valid_cur_s with osi system osi_p2m;auto.
  intros osi_pt_madd H_osi_pt_madd H_osi_ptmadd_prop.
  exists osi_pt_madd;auto.

  rewrite H_new_mem.
  unfold maddress_owner.
  unfold page_type.
  rewrite lemma_map_add_apply_distinct;auto.

  (*They are different because they have different page owner o type*)
  unfold not;intro H_osiptmadd_ma_abs.
  destruct H_osi_ptmadd_prop as (H_osi_ptmadd_ptype,H_osi_ptmadd_owner).
  rewrite <- H_osiptmadd_ma_abs in H_osi_ptmadd_owner.
  unfold maddress_owner in H_osi_ptmadd_owner.
  rewrite <- H_ma_pg in  H_osi_ptmadd_owner.
  rewrite H_pg_owner in  H_osi_ptmadd_owner.
  inversion  H_osi_ptmadd_owner;auto.
  

  (*In case osi is aos_osi*)
  intro H_osi_aos.
  rewrite H_osi_aos.
  intros aos' H_is_aos' p2m' H_p2m'.

  assert (aos'=OS None (os_current_page aos)) as H_aos'_newaos.
  rewrite H_new_oss in  H_is_aos'.
  rewrite lemma_map_add_apply in H_is_aos'.
  inversion H_is_aos';auto.

  assert (p2m'=map_add padd_eq p2m pa ma) as H_p2m'_newp2m.
  rewrite H_new_hyper in H_p2m'.
  rewrite lemma_map_add_apply in H_p2m'.
  inversion H_p2m';auto.

  assert (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s) ) as H_valid_cur_s.
  apply H_valid_s.

  destruct H_valid_cur_s with  (st_active_os s) aos p2m 
    as (pt_madd,H_pt_madd,H_pt_prop);auto.
  
  rewrite H_p2m'_newp2m; rewrite H_aos'_newaos.
  simpl.

  (*In case pinned page isn't curr_page*)
  assert (pa<>(os_current_page aos)) as H_no_cur_page_pa.
  
  unfold not;intro H_pa_cur_page_abs.
  rewrite <- H_pa_cur_page_abs in H_pt_madd.
  rewrite H_free_pa in H_pt_madd.
  inversion H_pt_madd;auto.

  exists pt_madd;auto.
  rewrite lemma_map_add_apply_distinct;auto.
  
  rewrite H_new_mem.
  unfold maddress_owner.
  unfold page_type.
  rewrite lemma_map_add_apply_distinct;auto.

  (*They are different because they have different page owner o type*)
  unfold not;intro H_osiptmadd_ma_abs.
  destruct H_pt_prop as (H_osi_ptmadd_ptype,H_osi_ptmadd_owner).
  rewrite <- H_osiptmadd_ma_abs in H_osi_ptmadd_owner.
  unfold maddress_owner in H_osi_ptmadd_owner.
  rewrite <- H_ma_pg in  H_osi_ptmadd_owner.
  rewrite H_pg_owner in  H_osi_ptmadd_owner.
  inversion  H_osi_ptmadd_owner;auto.

  (*Ninth Constraint*)
  split.
  intro osi.
  
  (*We discriminate the cases of osi*)
  case (os_ident_eq osi (st_active_os s)).
  
  (*In case osi is aos_osi*)
  intro H_osi_aos.

  intros newp2m H_newp2m.
  
  (*We can assume injectivity in s*)
  assert (injective_hyper_mappings (st_hypervisor s)) as H_inject_s.
  apply H_valid_s.
  
  (*The considered p2m is that than pin operation is modifying to add <pa,ma>*)
  assert (newp2m=map_add padd_eq p2m pa ma) as H_newp2m_def.
  rewrite H_new_hyper in H_newp2m.
  rewrite H_osi_aos in H_newp2m.
  rewrite lemma_map_add_apply in H_newp2m.
  inversion H_newp2m;auto.


  unfold injective_mapping.
  assert  (map_at_index padd_eq newp2m pa ma) as H_pa_ma.
  unfold map_at_index.
  rewrite H_newp2m_def.
  rewrite lemma_map_add_apply;auto.

  cut (forall p', map_apply padd_eq p2m p' <> Value padd ma).
  (*There is no padd mapped to ma in p2m*)


  intro H_no_ma_p2m.

  intros pa1 pa2 H_pa1_pa2.

  case (padd_eq pa pa1).
  intro H_pa_pa1.

  case (padd_eq pa pa2).
  rewrite H_pa_pa1;auto.

  intro H_pa_na_pa2.
  elim H_no_ma_p2m with pa2;auto.
  rewrite <- H_pa_pa1 in H_pa1_pa2.
  rewrite H_newp2m_def in H_pa1_pa2.
  rewrite lemma_map_add_apply in H_pa1_pa2.
  rewrite lemma_map_add_apply_distinct in H_pa1_pa2;auto.

  intro H_pa_no_p1.
  
  case (padd_eq pa pa2).

  intro H_pa_pa2.
  elim H_no_ma_p2m with pa2;auto.
  rewrite <- H_pa_pa2 in H_pa1_pa2.
  rewrite H_newp2m_def in H_pa1_pa2.
  rewrite lemma_map_add_apply in H_pa1_pa2.
  rewrite lemma_map_add_apply_distinct in H_pa1_pa2;auto.
  elim H_no_ma_p2m with pa1;auto.

  intro H_pa_no_pa2.
  rewrite H_newp2m_def in H_pa1_pa2.
  rewrite lemma_map_add_apply_distinct in H_pa1_pa2;auto.
  rewrite lemma_map_add_apply_distinct in H_pa1_pa2;auto.
  apply ( (H_inject_s) (st_active_os s) (p2m) (H_p2m) (pa1) (pa2) );auto.

  (*We prove that no padd is maped to ma in p2m*)
  intro p'.
  intro H_p'_ma_abs.
  assert (valid_hypervisor (st_hypervisor s) (st_memory s))
    as H_valid_hyper_s.
  apply H_valid_s.
 
  unfold valid_hypervisor in H_valid_hyper_s.
  destruct H_valid_hyper_s with 
    (osi:=(st_active_os s)) (p2m:=p2m) (pa:=p');auto.
  elim H.
  intros ma' H_ma'.
  assert (ma'=ma) as H_ma'_ma.
  rewrite H_p'_ma_abs in H_ma'.
  inversion H_ma';auto.
  rewrite H_ma'_ma;clear H ma' H_ma' H_ma'_ma.
  intro H_ma_owner.
  
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_ma_pg in H_ma_owner.
  rewrite H_pg_owner in H_ma_owner.
  inversion H_ma_owner.
  rewrite H_p'_ma_abs in H;inversion H.

  (*In case osi is not aos*)
  intro H_osi_no_aos.
  rewrite H_new_hyper.
  rewrite lemma_map_add_apply_distinct;auto.
  apply H_valid_s.

  (*Tenth Constraint*)
  split.
  unfold va_has_valid_pa.
  intro osi.
  case (os_ident_eq (st_active_os s ) osi).
  
  (*Case aos*)
  intro H_osi_aos.
  intros aos_pt_padd aos_pt_addr aos_pt_page
    aos_mp aos_va aos_ma.
  intros H_aos_pt_addr H_aos_pt_padd_madd
    H_aos_mp_page_cont H_aos_va_ma H_aos_va.
  
  case (madd_eq aos_ma ma).
  
  (*Case aos_ma = ma*)
  intro H_aosma_ma.
  exists pa;auto.
  
  rewrite H_aosma_ma; rewrite H_new_hyper.
  unfold hypervisor_mapping_madd.
  rewrite H_osi_aos.
  rewrite lemma_map_add_apply.
  exists (map_add padd_eq p2m pa ma);auto.
  rewrite lemma_map_add_apply;auto.
  
  (*Case aos_ma <> ma*)
  intro H_aosma_no_ma.
  
  assert (va_has_valid_pa (st_memory s)
    (st_hypervisor s) ctxt) as H_va_has_pa_s.
  apply H_valid_s.


  case (madd_eq aos_pt_addr ma).

  (*Case aos_pt_addr=ma*)
  intro H_aosptaddr_ma.

  assert (aos_pt_page=new_page (Some pg_type)
    (Os_PO (st_active_os s))) as H_aos_pt_page_def.
  rewrite H_new_mem in H_aos_pt_addr.
  rewrite H_aosptaddr_ma in H_aos_pt_addr.
  rewrite lemma_map_add_apply in H_aos_pt_addr.
  inversion H_aos_pt_addr;auto.

  rewrite H_aos_pt_page_def in H_aos_mp_page_cont.
  simpl in H_aos_mp_page_cont.
  generalize H_aos_mp_page_cont.

  case pg_type.

  (*Case pg_type is PT*)
  intro H_mp_empty.
  assert (aos_mp = map_empty vadd madd)
    as H_aos_mp_emptyPT.
  inversion H_mp_empty;auto.
  clear H_mp_empty.
  rewrite H_aos_mp_emptyPT in H_aos_va_ma.
  inversion H_aos_va_ma.

  (*Case pg_type is RW*)
  intro H_abs;inversion H_abs.

  (*Case aos_pt_addr <> ma *)
  intro H_aosptmadd_no_ma.

  elim H_va_has_pa_s with osi aos_pt_padd aos_pt_addr
    aos_pt_page
    aos_mp aos_va aos_ma;auto.

  intros aos_pa H_aospa.
  exists aos_pa.


  rewrite H_new_hyper.
  rewrite H_osi_aos.
  unfold hypervisor_mapping_madd.
  rewrite lemma_map_add_apply.
  exists (map_add padd_eq p2m pa ma);auto.

  unfold hypervisor_mapping_madd in H_aospa.
  destruct H_aospa as (p2m',H_p2m',H_aospa).

  assert (p2m'=p2m) as H_p2m'_p2m.
  rewrite <- H_osi_aos in H_p2m'.
  rewrite <- H_p2m in H_p2m'.
  inversion H_p2m';auto.
  rewrite H_p2m'_p2m in H_aospa;auto.

  rewrite lemma_map_add_apply_distinct;auto.
  
  intro H_abs.
  rewrite <- H_abs in H_aospa.
  rewrite H_free_pa in H_aospa.
  inversion H_aospa.

  rewrite H_new_mem in H_aos_pt_addr.
  rewrite lemma_map_add_apply_distinct
    in H_aos_pt_addr;auto.

  exists p2m;auto.
  rewrite <- H_osi_aos;assumption.
  
  rewrite H_new_hyper in H_aos_pt_padd_madd.
  unfold hypervisor_mapping_madd in H_aos_pt_padd_madd.
  rewrite H_osi_aos in  H_aos_pt_padd_madd.
  rewrite lemma_map_add_apply in H_aos_pt_padd_madd.
  destruct  H_aos_pt_padd_madd as (p2m',H_p2m'_val,H_p2m').
  assert (p2m' = map_add padd_eq p2m pa ma) as H_p2m'_def.
  inversion H_p2m'_val;auto.
  rewrite H_p2m'_def in H_p2m'.
  rewrite lemma_map_add_apply_distinct in H_p2m';auto.

  intro H_abs.
  rewrite H_abs in H_p2m'.
  rewrite lemma_map_add_apply in H_p2m'.
  inversion H_p2m';contradiction.

  (*Case osi <> aos*)
  intro H_osi_no_aos.
  intros osi_pt_padd osi_pt_addr osi_pt_page
    osi_mp osi_va osi_ma.
  intros H_osi_pt_addr H_osi_pt_padd_madd
    H_osi_mp_page_cont H_osi_va_ma H_osi_va.
  
  assert (va_has_valid_pa (st_memory s)
    (st_hypervisor s) ctxt) as H_va_has_pa_s.
  apply H_valid_s.

  elim H_va_has_pa_s with osi osi_pt_padd osi_pt_addr
    osi_pt_page osi_mp osi_va osi_ma;auto.

  intros osi_pa H_osipa.
  exists osi_pa;auto.

  rewrite H_new_hyper; unfold hypervisor_mapping_madd.
  rewrite lemma_map_add_apply_distinct;auto.
  
  rewrite H_new_mem in H_osi_pt_addr.
  rewrite lemma_map_add_apply_distinct
    in H_osi_pt_addr;auto.

  (*ma is different form osi_pt_addr*)
  intro H_abs.
  rewrite H_abs in H_osi_pt_addr.
  rewrite lemma_map_add_apply in H_osi_pt_addr.
  assert (osi_pt_page = new_page (Some pg_type)
    (Os_PO (st_active_os s))) as H_osi_pt_page_def.
  inversion H_osi_pt_addr;auto.
  rewrite H_osi_pt_page_def in H_osi_mp_page_cont.
  generalize H_osi_mp_page_cont.

  case pg_type.

  (*Case pg_type is PT*)
  intro H_case_PT.

  assert (osi_mp = map_empty vadd madd) as H_osi_mp_def.
  inversion H_case_PT;auto.
  rewrite H_osi_mp_def in H_osi_va_ma.
  inversion H_osi_va_ma.

  intro H_case_RW.
  inversion H_case_RW;auto.

  
  rewrite H_new_hyper in H_osi_pt_padd_madd.
  unfold hypervisor_mapping_madd.
  unfold hypervisor_mapping_madd in H_osi_pt_padd_madd.
  rewrite lemma_map_add_apply_distinct
    in H_osi_pt_padd_madd;auto.

  apply exec_valid_cache_tlb with ctxt s (Page_pin_trusted pa pg_type); auto.
Qed.


Theorem Page_unpin_trusted_secure:
  forall (s s':state),
  forall pa,
  valid_state ctxt s ->
  valid_exec ctxt s (Page_unpin_trusted pa) s'->
  valid_state ctxt s'.
Proof.
  intros s s' pa H_valid_s H_valid_exec;
    generalize H_valid_exec; intro H_valid_exec';
    inversion_clear H_valid_exec'.
  simpl in H.
  unfold pre_page_unpin_trusted in H.
  destruct H as (H_os_running,H_pre).
  destruct H_pre as (H_trusted_aos,H_pre_unpin).

  unfold pre_all_unpin in H_pre_unpin.
  destruct H_pre_unpin as (aos, H_is_aos, H_pre_unpin).
  destruct H_pre_unpin as (H_no_cur_page_pa, H_pre_unpin).
  destruct H_pre_unpin as (p2m,H_p2m, H_pre_ma_pa).
  destruct H_pre_ma_pa as (ma, H_ma_pa, H_pre_ma_pa).
  destruct H_pre_ma_pa as (H_ma_ptype, H_ma_no_mapped).

  simpl in H0.
  unfold post_page_unpin_trusted in H0.
  unfold post_all_unpin in H0.

  destruct H0 as (p2m',H_p2m',H_post_unpin).
  destruct H_post_unpin as (ma',H_ma'_pa,H_post_unpin).
  assert (p2m'=p2m) as H_p2m'_p2m.
  unfold hypervisor_mapping in H_p2m'.
  unfold hypervisor_mapping in H_p2m.
  rewrite <-H_p2m in H_p2m';inversion H_p2m';auto.
  assert (ma'=ma) as H_ma'_ma.
  rewrite H_p2m'_p2m in H_ma'_pa.
  rewrite <- H_ma_pa in H_ma'_pa.
  inversion H_ma'_pa;auto.
  
  rewrite H_ma'_ma in H_post_unpin;rewrite H_p2m'_p2m in H_post_unpin.
  clear p2m' H_p2m' ma' H_ma'_pa H_ma'_ma H_p2m'_p2m.
  
  destruct H_post_unpin as (aos',H_is_aos',H_s'_def).
  assert (aos'=aos) as H_aos'_aos.
  rewrite <- H_is_aos in H_is_aos'.
  inversion H_is_aos';auto.
  rewrite H_aos'_aos in H_s'_def;clear aos' H_is_aos' H_aos'_aos.

  (*We establish the relation between s and s' components*)

  assert (st_active_os s' = st_active_os s ) as H_same_aos.
  rewrite H_s'_def;auto.
  assert (st_oss s' = map_add os_ident_eq (st_oss s) (st_active_os s)
              (OS None (os_current_page aos))) as H_new_oss.
  rewrite H_s'_def;auto.
  assert (st_exec_mode s' = st_exec_mode s) as H_same_exmd.
  rewrite H_s'_def;auto.
  assert (st_aos_activity s' = st_aos_activity s) as H_same_activ.
  rewrite H_s'_def;auto.
  assert (st_hypervisor s' = (map_add os_ident_eq (st_hypervisor s) 
              (st_active_os s) (map_drop padd_eq p2m pa))) as H_new_hyper.
  rewrite H_s'_def;auto.
  assert (st_memory s' = map_add madd_eq (st_memory s) ma
                (new_page None No_Owner)  ) as H_new_mem.
  rewrite H_s'_def;auto.

  clear H_s'_def.

  (*Injectivity of p2m holds, and it's used on the proof*)  
  assert (injective_mapping padd_eq p2m) as H_injective_p2m.
  assert (injective_hyper_mappings (st_hypervisor s)) as H_inject_s.
  apply H_valid_s.
  unfold injective_hyper_mappings in H_inject_s.
  apply ( (H_inject_s) (st_active_os s) (p2m));auto.

  unfold valid_state.

  (*First Constraint*)
  split.
  unfold trusted_os_not_hypercall.
  intros trusted_osi trusted_os H_is_trusted_osi H_trusted_os_map.
  case (os_ident_eq (st_active_os s) trusted_osi).
  intro H_osi_active.
  rewrite H_new_oss in H_trusted_os_map.
  rewrite H_osi_active in H_trusted_os_map.
  rewrite lemma_map_add_apply in H_trusted_os_map.
  injection H_trusted_os_map; intro H_trusted_os_value; rewrite H_trusted_os_value; auto.
  intro H_osi_not_active.
  cut (trusted_os_not_hypercall (ctxt_oss ctxt) (st_oss s)).
  intro H_trusted_no_hyp_s.
  2:apply H_valid_s.
  rewrite H_new_oss in H_trusted_os_map.
  rewrite lemma_map_add_apply_distinct in H_trusted_os_map; auto.
  apply (H_trusted_no_hyp_s trusted_osi trusted_os H_is_trusted_osi); auto.

  (*Second Constraint*)
  split.
  rewrite H_same_aos; rewrite H_same_activ.
  unfold running_os_not_hypercall.
  
  exists (OS None (os_current_page aos));auto.
  rewrite H_new_oss.
  rewrite lemma_map_add_apply;auto.

  (*Third Constraint*)
  split.
  rewrite H_same_activ; rewrite H_same_exmd.
  apply H_valid_s.

  (*Fourth Constraint*)
  split.
  rewrite H_same_aos; rewrite H_same_exmd.
  apply H_valid_s.

  (*Fifth Constraint*)
  split.
  rewrite H_same_aos; rewrite H_same_activ; rewrite H_same_exmd.
  apply H_valid_s.

  (*Sixth Constraint*)
  split.
  
  unfold valid_hypervisor.
  intro osi.
  
  (*We discriminate the cases of osi*)
  case (os_ident_eq osi  (st_active_os s)).


  (*In case osi is aos*)
  intro H_osi_aos.
  
  rewrite H_osi_aos.

  intro aos_pa.
  
  
  case (padd_eq aos_pa pa).
  
  (*In case osi is aos, and the considered padd is pa*)
  intro H_aospa_pa.
  
  rewrite H_aospa_pa.
  intros p2m' H_p2m'.
  right.
  
  rewrite H_new_hyper in H_p2m'.
  rewrite lemma_map_add_apply in H_p2m'.
  inversion H_p2m';auto.
  rewrite map_drop_apply;auto.

  (*In case osi is aos, and the considered padd is not pa*)

  intro H_no_aospa_pa.
  intros p2m' H_p2m'.

  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) 
    as H_valid_hyper_s.
  apply H_valid_s.

  rewrite H_new_hyper in H_p2m'.
  rewrite lemma_map_add_apply in H_p2m'.
  inversion H_p2m'.
  rewrite map_drop_apply_distinct;auto.

  elim H_valid_hyper_s with (st_active_os s) aos_pa p2m;auto.
  
  (*The hypothesis is an OR, we must prove both branches*)
  intro H_exist_ma'.

  (*Left branch of OR*)
  left.
  destruct H_exist_ma' as (ma', H_p2m_ma', H_ma'_owner).
  exists ma';auto.

  rewrite H_new_mem.
  
  assert (ma'<>ma) as H_no_ma'_ma.
  
  unfold not;intro H_ma'_ma_abs.
  unfold injective_mapping in H_injective_p2m.

  assert (aos_pa=pa) as H_aospa_pa_abs.
  rewrite H_ma'_ma_abs in H_p2m_ma'.
  rewrite H_ma_pa in H_p2m_ma'.
  elim ((H_injective_p2m) (pa) (aos_pa) (H_p2m_ma'));reflexivity.
  contradiction.

  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct;auto.


  (*In case osi is not aos*)

  intro H_no_osi_aos.

  intros osi_pa osi_p2m H_osi_p2m.

  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) 
    as H_valid_hyper_s.
  apply H_valid_s.

  elim H_valid_hyper_s with osi osi_pa osi_p2m;auto. 
  intro H_exists_ma'.


  (*The hypothesis is an OR, we must prove both branches*)  
  left.
  destruct H_exists_ma' as (ma', H_osi_p2m_ma', H_ma'_owner).
  exists ma';auto.
  
  rewrite H_new_mem.
  
  
  (*diferente owner*)
  assert (ma'<>ma) as H_ma'_no_ma.
  unfold not;intro H_ma'_ma_abs.
  
  (*For being ma and pa associated in the hypervisor mapping
  we can say that ma belongs to osi, using the property valid 
  hypervisor in s*)

  elim H_valid_hyper_s with (st_active_os s) pa p2m;auto. 
  intro H_exists_ma''.

  destruct H_exists_ma'' as (ma'', H_aos_p2m_ma'', H_ma''_owner).
  assert (ma''=ma) as H_ma''_ma.
  rewrite <- H_ma_pa in H_aos_p2m_ma'';inversion H_aos_p2m_ma'';auto.
  rewrite H_ma''_ma in H_ma''_owner.
  rewrite H_ma'_ma_abs in H_ma'_owner.
  rewrite H_ma'_owner in H_ma''_owner.
  inversion H_ma''_owner;auto.

  intro H_pa_not_pined_osi.
  rewrite <- H_ma_pa in H_pa_not_pined_osi.
  inversion H_pa_not_pined_osi;auto.


  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct;auto.

  rewrite H_new_hyper in H_osi_p2m.
  rewrite lemma_map_add_apply_distinct in H_osi_p2m;auto.
  

  (*Seventh Constraint*)
  split.

  intro osi.

  (*We discriminate the cases of osi*)
  
  case (os_ident_eq osi (st_active_os s)).

  
  (*In case osi is aos*)
  intro H_osi_aos.
  
  rewrite H_osi_aos.
  intro aos_pt_madd.
  
  (*We discriminate cases of the pt_madd considered*)  
  case (madd_eq aos_pt_madd ma).

  (*In case the considered ma is the same that ma, the
  corresponding page has type "Other", so it doesn't met
  the hypothesis of the property to be proved*)
  intro H_aosptmadd_ma.
  rewrite H_aosptmadd_ma.


  intros ma_map' ma_page' H_ma_page'.
  rewrite H_new_mem in H_ma_page'.
  rewrite lemma_map_add_apply in H_ma_page'.
  inversion_clear H_ma_page'.
  simpl.
  intro H_ma_map'_contr.
  inversion_clear H_ma_map'_contr.

  (*In case considered madd is not ma*)
  intro H_no_aosptmadd_ma.

  
  intros aos_map aos_page
    H_aos_pt_page H_aos_pg_map H_aos_pg_owner
      aos_map_va aos_map_ma H_aos_va_ma_map.
  
  assert (valid_virtual_mapping (st_memory s) ctxt) 
    as H_valid_virtual_s.
  apply H_valid_s;auto.

  
  assert (aos_map_ma <> ma) as H_no_aosmapma_ma.
  (*There can not be an aos_map that mets this, and has ma on it*)

  unfold not;intro H_aosmapma_ma_abs.

  rewrite H_aosmapma_ma_abs in H_aos_va_ma_map.
  unfold no_va_mapped_to_ma in H_ma_no_mapped.

  elim H_ma_no_mapped with aos_pt_madd aos_page aos_map_va aos_map;auto.
  rewrite H_new_mem in H_aos_pt_page.
  rewrite lemma_map_add_apply_distinct in H_aos_pt_page;auto.


  rewrite H_new_mem.
  rewrite lemma_map_add_apply_distinct;auto.

  rewrite H_new_mem in H_aos_pt_page.
  rewrite lemma_map_add_apply_distinct in H_aos_pt_page;auto.

  elim H_valid_virtual_s with (st_active_os s) aos_pt_madd aos_map 
    aos_page aos_map_va aos_map_ma;auto.

  intros aos_map_ma_pg H_aos_map_ma_pg H_aos_map_ma_prop.
  exists aos_map_ma_pg;auto.
  

  (*In case osi is not aos*)
  intro H_no_osi_aos.

  intros osi_pt_madd osi_map osi_page
    H_osi_pt_page H_osi_pg_map H_osi_pg_owner
      osi_map_va osi_map_ma H_osi_va_ma_map.
 
  assert (ma <> osi_pt_madd) as H_no_osiptmadd_ma.

  (*It can not be the same, because ma has owner oas in mem', not osi*)

  unfold not;intro H_osiptmadd_ma_abs.
  

  (*For being ma and pa associated in the hypervisor mapping
  we can say that ma belongs to osi, using the property valid 
  hypervisor in s*)
  
  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) 
    as H_valid_hyper_s.
  apply H_valid_s.

  elim H_valid_hyper_s with (st_active_os s) pa p2m;auto. 
  intro H_exists_ma'.

  destruct H_exists_ma' as (ma', H_aos_p2m_ma', H_ma'_owner).
  assert (ma'=ma) as H_ma'_ma.
  rewrite <- H_ma_pa in H_aos_p2m_ma';inversion H_aos_p2m_ma';auto.
  rewrite H_ma'_ma in H_ma'_owner.
  rewrite H_osiptmadd_ma_abs in H_ma'_owner.

  assert (maddress_owner (st_memory s) osi_pt_madd = No_Owner) 
    as H_osiptmadd_no_owner.

  rewrite H_new_mem in H_osi_pt_page.

  rewrite H_osiptmadd_ma_abs in H_osi_pt_page.
  rewrite lemma_map_add_apply in H_osi_pt_page.
  unfold maddress_owner.

  inversion H_osi_pt_page;auto.

  rewrite H0 in H_osi_pg_map.
  inversion H_osi_pg_map;auto.

  rewrite H_osiptmadd_no_owner in H_ma'_owner.
  inversion H_ma'_owner;auto.

  intro H_pa_not_pined_osi.
  rewrite <- H_ma_pa in H_pa_not_pined_osi.
  inversion H_pa_not_pined_osi;auto.


  (*It can not be the same, because ma has owner oas in mem', not osi*)
  assert (ma <> osi_map_ma) as H_no_osimapma_ma.

  unfold not;intro H_osimap_ma_abs.

  
  assert (valid_virtual_mapping (st_memory s) ctxt) 
    as H_valid_virtual_s.
  apply H_valid_s;auto.

  elim H_valid_virtual_s with osi osi_pt_madd 
    osi_map osi_page osi_map_va osi_map_ma;auto.

  2:rewrite H_new_mem in H_osi_pt_page.
  2:rewrite lemma_map_add_apply_distinct in H_osi_pt_page;auto.
  
  intros pg H_pg H_pg_prop.
  assert (maddress_owner (st_memory s) ma <> Os_PO (st_active_os s)) 
    as H_no_ma_owner_aos.
  
  rewrite <- H_osimap_ma_abs in H_pg.
  unfold maddress_owner. 
  rewrite <- H_pg.
  unfold not;intro H_pg_owner_abs.

  destruct H_pg_prop as (H_map_va_acces,H_map_va_no_acces).
  unfold os_accessible in H_map_va_no_acces.
  unfold os_accessible in H_map_va_acces.
  destruct (ctxt_vadd_accessible ctxt osi_map_va).
  rewrite H_pg_owner_abs in H_map_va_acces;auto.
  assert (Os_PO (st_active_os s) = Os_PO osi) as H_osi_aos_abs.
  apply H_map_va_acces;auto.
  inversion H_osi_aos_abs.
  rewrite  H0 in H_no_osi_aos;auto.
  
  rewrite H_pg_owner_abs in H_map_va_no_acces;auto.
  assert (Os_PO (st_active_os s) = Hyp_PO) as H_aos_hyper_abs.
  apply H_map_va_no_acces;auto.
  inversion H_aos_hyper_abs.


  (*For being ma and pa associated in the hypervisor mapping
  we can say that ma belongs to osi, using the property valid 
  hypervisor in s*)
  
  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) 
    as H_valid_hyper_s.
  apply H_valid_s.

  elim H_valid_hyper_s with (st_active_os s) pa p2m;auto. 
  intro H_exists_ma'.

  destruct H_exists_ma' as (ma', H_aos_p2m_ma', H_ma'_owner).
  assert (ma'=ma) as H_ma'_ma.
  rewrite <- H_ma_pa in H_aos_p2m_ma';inversion H_aos_p2m_ma';auto.
  rewrite H_ma'_ma in H_ma'_owner.
  contradiction.
  intro H_pa_not_pined_osi.
  rewrite <- H_ma_pa in H_pa_not_pined_osi.
  inversion H_pa_not_pined_osi;auto.
  

  rewrite H_new_mem in H_osi_pt_page.
  rewrite lemma_map_add_apply_distinct in H_osi_pt_page;auto.
  
  rewrite H_new_mem.
  rewrite lemma_map_add_apply_distinct;auto.
 
  assert (valid_virtual_mapping (st_memory s) ctxt) 
    as H_valid_virtual_s.
  apply H_valid_s;auto.

  elim H_valid_virtual_s with osi osi_pt_madd osi_map 
    osi_page osi_map_va osi_map_ma;auto.

  intros osi_map_ma_pg H_osi_map_ma_pg H_osi_map_ma_prop.
  exists osi_map_ma_pg;auto.


  (*Eighth Constraint*)
  split.
  unfold valid_current_page.

  intro osi.

  (*We discriminate the cases of osi*)
  case (os_ident_eq osi (st_active_os s)).


  Focus 2.
  (*In case osi is not aos_osi*)
  intro H_no_osi_aos.
  
  intros system H_osi_system osi_p2m H_osi_p2m.

  rewrite H_new_oss in H_osi_system.
  rewrite lemma_map_add_apply_distinct in H_osi_system;auto.
  
  rewrite H_new_hyper in H_osi_p2m.
  rewrite lemma_map_add_apply_distinct in H_osi_p2m;auto.

  assert (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s) ) 
    as H_valid_cur_s.
  apply H_valid_s.

  elim H_valid_cur_s with osi system osi_p2m;auto.
  intros osi_pt_madd H_osi_pt_madd H_osi_ptmadd_prop.
  exists osi_pt_madd;auto.

  rewrite H_new_mem.
  unfold maddress_owner.
  unfold page_type.
  rewrite lemma_map_add_apply_distinct;auto.

  (*They are different because they have different page owner o type*)
  unfold not;intro H_osiptmadd_ma_abs.
  destruct H_osi_ptmadd_prop as (H_osi_ptmadd_ptype,H_osi_ptmadd_owner).
  rewrite <- H_osiptmadd_ma_abs in H_osi_ptmadd_owner.

  (*For being ma and pa associated in the hypervisor mapping
  we can say that ma belongs to osi, using the property valid 
  hypervisor in s*)
  
  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) 
    as H_valid_hyper_s.
  apply H_valid_s.

  elim H_valid_hyper_s with (st_active_os s) pa p2m;auto. 
  intro H_exists_ma'.

  destruct H_exists_ma' as (ma', H_aos_p2m_ma', H_ma'_owner).
  assert (ma'=ma) as H_ma'_ma.
  rewrite <- H_ma_pa in H_aos_p2m_ma';inversion H_aos_p2m_ma';auto.
  rewrite H_ma'_ma in H_ma'_owner.
  rewrite H_ma'_owner in H_osi_ptmadd_owner.
  inversion H_osi_ptmadd_owner;auto.

  intro H_pa_not_pined_osi.
  rewrite <- H_ma_pa in H_pa_not_pined_osi.
  inversion H_pa_not_pined_osi;auto.


  (*In case osi is aos_osi*)
  intro H_osi_aos.
  rewrite H_osi_aos.
  intros aos' H_is_aos' p2m' H_p2m'.

  assert (aos'=OS None (os_current_page aos)) as H_aos'_newaos.
  rewrite H_new_oss in  H_is_aos'.
  rewrite lemma_map_add_apply in H_is_aos'.
  inversion H_is_aos';auto.

  assert (p2m'=map_drop padd_eq p2m pa) as H_p2m'_newp2m.
  rewrite H_new_hyper in H_p2m'.
  rewrite lemma_map_add_apply in H_p2m'.
  inversion H_p2m';auto.

  assert (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s) ) 
    as H_valid_cur_s.
  apply H_valid_s.

  destruct H_valid_cur_s with  (st_active_os s) aos p2m 
    as (pt_madd,H_pt_madd,H_pt_prop);auto.
  
  rewrite H_p2m'_newp2m; rewrite H_aos'_newaos.
  simpl.

  exists pt_madd;auto.
  rewrite map_drop_apply_distinct;auto.
  
  rewrite H_new_mem.
  unfold maddress_owner.
  unfold page_type.
  rewrite lemma_map_add_apply_distinct;auto.


  (*They are different becaus is pa can't be the padd 
  of the current page, then ma can't be de madd of the
  current page*)
  unfold not;intro H_aos'ptmadd_ma_abs.
  rewrite <- H_aos'ptmadd_ma_abs in H_pt_madd.
  rewrite H_ma_pa in H_pt_madd;auto.


  (*Ninth Constraint*)
  split.
  intro osi.
  
  (*We discriminate the cases of osi*)
  case (os_ident_eq osi (st_active_os s)).
  
  (*In case osi is aos_osi*)
  intro H_osi_aos.

  intros newp2m H_newp2m.
  
  (*We can assume injectivity in s*)
  assert (injective_hyper_mappings (st_hypervisor s)) as H_inject_s.
  apply H_valid_s.
  
  (*The considered p2m is that than unpin operation is modifying to drop <pa,ma>*)
  assert (newp2m=map_drop padd_eq p2m pa) as H_newp2m_def.
  rewrite H_new_hyper in H_newp2m.
  rewrite H_osi_aos in H_newp2m.
  rewrite lemma_map_add_apply in H_newp2m.
  inversion H_newp2m;auto.


  unfold injective_mapping.
  assert  (map_apply padd_eq newp2m pa = Error madd pa) as H_pa_ma.
  rewrite H_newp2m_def.
  rewrite map_drop_apply;auto.

  intros pa1 pa2 H_pa1_pa2.

  case (padd_eq pa pa1).
  intro H_pa_pa1.

  case (padd_eq pa pa2).
  rewrite H_pa_pa1;auto.

  intro H_pa_na_pa2.
  elim map_error_index_inv with padd padd_eq madd (p2m) (pa2) (pa);auto.
  rewrite <- H_pa_pa1 in H_pa1_pa2.
  rewrite H_newp2m_def in H_pa1_pa2.
  rewrite map_drop_apply in H_pa1_pa2.
  rewrite map_drop_apply_distinct in H_pa1_pa2;auto.

  intro H_pa_no_p1.
  
  case (padd_eq pa pa2).

  intro H_pa_pa2.
  elim map_error_index_inv with padd padd_eq madd (p2m) (pa1) (pa);auto.
  rewrite <- H_pa_pa2 in H_pa1_pa2.
  rewrite H_newp2m_def in H_pa1_pa2.
  rewrite map_drop_apply in H_pa1_pa2.
  rewrite map_drop_apply_distinct in H_pa1_pa2;auto.

  intro H_pa_no_pa2.
  rewrite H_newp2m_def in H_pa1_pa2.
  rewrite map_drop_apply_distinct in H_pa1_pa2;auto.
  rewrite map_drop_apply_distinct in H_pa1_pa2;auto.


  (*In case osi is not aos*)
  intro H_osi_no_aos.
  rewrite H_new_hyper.
  rewrite lemma_map_add_apply_distinct;auto.
  apply H_valid_s.

  (*Tenth Constraint*)
  split.
  unfold va_has_valid_pa.
  intro osi.
  case (os_ident_eq (st_active_os s ) osi).
  
  (*Case aos*)
  intro H_osi_aos.
  intros aos_pt_padd aos_pt_addr aos_pt_page
    aos_mp aos_va aos_ma.
  intros H_aos_pt_addr H_aos_pt_padd_madd
    H_aos_mp_page_cont H_aos_va_ma H_aos_va.
  
  assert (aos_pt_addr<>ma) as H_aosptadd_no_ma.
  intro H_abs_2.
  rewrite H_new_mem in H_aos_pt_addr.
  rewrite H_abs_2 in H_aos_pt_addr.
  rewrite lemma_map_add_apply in H_aos_pt_addr.
  assert (aos_pt_page=new_page None No_Owner)
    as H_aos_pt_page_def.
  inversion H_aos_pt_addr;auto.
  rewrite H_aos_pt_page_def in H_aos_mp_page_cont.
  inversion  H_aos_mp_page_cont.

  destruct H_aos_pt_padd_madd as (newp2m,H_newp2m,H_aos_pt_padd_madd).
  assert (newp2m=map_drop padd_eq p2m pa) as H_newp2m_def.
  rewrite H_new_hyper in H_newp2m.
  rewrite H_osi_aos in H_newp2m.
  rewrite lemma_map_add_apply in H_newp2m.
  inversion H_newp2m;auto.

  assert (pa <> aos_pt_padd) as H_aospa_no_pa.
  intro H_abs_3.
  rewrite H_newp2m_def in H_aos_pt_padd_madd.
  rewrite H_abs_3 in H_aos_pt_padd_madd.
  rewrite map_drop_apply in H_aos_pt_padd_madd.
  inversion H_aos_pt_padd_madd.

  assert (aos_ma<>ma) as H_aosma_no_ma.
  intro H_abs.
  elim H_ma_no_mapped with aos_pt_addr aos_pt_page aos_va aos_mp;auto.
  rewrite H_new_mem in H_aos_pt_addr.
  rewrite lemma_map_add_apply_distinct in H_aos_pt_addr;auto.
  assert (valid_hypervisor (st_hypervisor s)
    (st_memory s)) as H_valid_hyper_s.
  apply H_valid_s.
  elim H_valid_hyper_s with osi aos_pt_padd p2m.
  intro H_pt_madd.
  destruct H_pt_madd as (pt_madd,H_pt_madd,H_pt_madd_owner).
  assert (pt_madd=aos_pt_addr) as H_aospt_ma.
  rewrite H_newp2m_def in H_aos_pt_padd_madd.
  rewrite map_drop_apply_distinct in H_aos_pt_padd_madd;auto.
  rewrite <- H_aos_pt_padd_madd in H_pt_madd.
  inversion H_pt_madd;auto.
  unfold maddress_owner in H_pt_madd_owner.
  rewrite H_aospt_ma in H_pt_madd_owner.
  rewrite H_new_mem in H_aos_pt_addr.
  rewrite lemma_map_add_apply_distinct in H_aos_pt_addr;auto.
  rewrite <- H_aos_pt_addr in H_pt_madd_owner.
  rewrite H_osi_aos;assumption.
  intro H_madd_error.
  rewrite H_newp2m_def in H_aos_pt_padd_madd.
  rewrite map_drop_apply_distinct in
    H_aos_pt_padd_madd;auto.
  rewrite <- H_aos_pt_padd_madd in H_madd_error.
  inversion H_madd_error.
  rewrite <- H_osi_aos;assumption.
  rewrite H_abs in H_aos_va_ma;auto.
 
  (*We use the fact that the constraint holds in s*)
  assert (va_has_valid_pa (st_memory s)
    (st_hypervisor s) ctxt) as H_va_has_pa_s.
  apply H_valid_s.

  elim H_va_has_pa_s with osi aos_pt_padd aos_pt_addr
    aos_pt_page aos_mp aos_va aos_ma;auto.

  intros aos_pa H_aospa.
  destruct H_aospa as (p2m',H_p2m',H_aospa).
  assert (p2m'=p2m) as H_p2m'_p2m.
  rewrite <- H_osi_aos in H_p2m'.
  rewrite <- H_p2m in H_p2m'.
  inversion H_p2m';auto.

  exists aos_pa;auto.
  rewrite H_new_hyper.
  unfold hypervisor_mapping_madd.
  rewrite H_osi_aos.
  rewrite lemma_map_add_apply.
  exists (map_drop padd_eq p2m pa);auto.
  rewrite map_drop_apply_distinct;auto.
  rewrite H_p2m'_p2m in H_aospa;auto.

  intro H_aospa_abs.
  rewrite H_p2m'_p2m in H_aospa.
  rewrite <- H_aospa_abs in H_aospa.
  rewrite <- H_ma_pa in H_aospa.
  inversion H_aospa;contradiction.

  rewrite H_new_mem in H_aos_pt_addr.
  rewrite lemma_map_add_apply_distinct
    in H_aos_pt_addr;auto.
  
  exists p2m.
  rewrite <- H_osi_aos;assumption.
  rewrite H_newp2m_def in H_aos_pt_padd_madd.
  rewrite map_drop_apply_distinct
    in H_aos_pt_padd_madd;auto.

  (*Case osi is not aos*)
  intro H_osi_no_aos.
  
  intros osi_pt_padd osi_pt_addr osi_pt_page
    osi_mp osi_va osi_ma.
  intros H_osi_pt_addr H_osi_pt_padd_madd
    H_osi_mp_page_cont H_osi_va_ma H_osi_va.

  (*We prove osi_pt_addr <> ma*)
  assert (osi_pt_addr <> ma) as H_osipt_no_ma.
  intro H_osipt_abs.
  rewrite H_new_mem in H_osi_pt_addr.
  rewrite H_osipt_abs in H_osi_pt_addr.
  rewrite lemma_map_add_apply in H_osi_pt_addr.
  assert (osi_pt_page = Page Other_c No_Owner)
    as H_osi_pt_page_def.
  inversion H_osi_pt_addr;auto.
  rewrite H_osi_pt_page_def in
    H_osi_mp_page_cont.
  inversion H_osi_mp_page_cont.
 
  (*And now we just use the validity os s*)
  assert (va_has_valid_pa (st_memory s)
    (st_hypervisor s) ctxt) as H_va_has_pa_s.
  apply H_valid_s.

  elim H_va_has_pa_s with osi osi_pt_padd
    osi_pt_addr osi_pt_page osi_mp osi_va
  osi_ma;auto.
  
  intros osi_pa H_osi_pa.
  exists osi_pa.

  rewrite H_new_hyper;unfold hypervisor_mapping_madd.
  rewrite lemma_map_add_apply_distinct;auto.

  rewrite H_new_mem in H_osi_pt_addr.
  rewrite lemma_map_add_apply_distinct
    in H_osi_pt_addr;auto.

  rewrite H_new_hyper in H_osi_pt_padd_madd.
  unfold hypervisor_mapping_madd in H_osi_pt_padd_madd.
  rewrite lemma_map_add_apply_distinct in H_osi_pt_padd_madd;auto.

  apply exec_valid_cache_tlb with ctxt s (Page_unpin_trusted pa); auto.
Qed.

Theorem Page_pin_untrusted_secure:
  forall (s s':state),
  forall osi pa pg_type,
  valid_state ctxt s ->
  valid_exec ctxt s (Page_pin_untrusted osi pa pg_type) s'->
  valid_state ctxt s'.
Proof.
  intros s s' osi pa pg_type H_valid_s H_valid_exec;
    generalize H_valid_exec; intro H_valid_exec';
    inversion_clear H_valid_exec'.
  simpl in H.
  unfold pre_page_pin_untrusted in H.
  destruct H as (H_untrusted_osi,H_pre).
  destruct H_pre as (H_hypercall_pin,H_pre).
  destruct H_pre as (H_hyper_running,H_pre_pin).
  
  unfold pre_all_pin in H_pre_pin.
  destruct H_pre_pin as (H_free_pa, H_free_ma).
  clear H_free_ma. 
  destruct H_free_pa as (p2m, H_p2m,H_free_pa).
  
  simpl in H0.
  unfold post_page_pin_untrusted in H0.
  unfold post_all_pin in H0.

  destruct H0 as (ma,H_post_pin).
  destruct H_post_pin as (ma_pg, H_ma_pg,H_post_pin).
  destruct H_post_pin as (H_pg_owner,H_post_pin).
  destruct H_post_pin as (H_pg_content,H_post_pin).
  destruct H_post_pin as (p2m',H_p2m',H_post_pin).
  assert (p2m=p2m') as H_p2m'_p2m.
  unfold hypervisor_mapping in H_p2m'.
  rewrite <- H_p2m in H_p2m';inversion H_p2m';auto.
  rewrite <- H_p2m'_p2m in H_post_pin.
  clear p2m' H_p2m' H_p2m'_p2m.
  destruct H_post_pin as (system,H_is_system,H_s'_def).

  (*We establish the relation between s and s' components*)
 
  assert (st_active_os s' = st_active_os s ) as H_same_aos.
  rewrite H_s'_def;auto.
  assert (st_oss s' = map_add os_ident_eq (st_oss s) osi
              (OS None (os_current_page system))) as H_new_oss.
  rewrite H_s'_def;auto.
  assert (st_exec_mode s' = st_exec_mode s) as H_same_exmd.
  rewrite H_s'_def;auto.
  assert (st_aos_activity s' = st_aos_activity s) as H_same_activ.
  rewrite H_s'_def;auto.
  assert (st_hypervisor s' = (map_add os_ident_eq (st_hypervisor s) 
              osi (map_add padd_eq p2m pa ma))) as H_new_hyper.
  rewrite H_s'_def;auto.
  assert (st_memory s' = map_add madd_eq (st_memory s) ma
                (new_page (Some pg_type) (Os_PO osi))  ) as H_new_mem.
  rewrite H_s'_def;auto.

  clear H_s'_def.
  
  
  unfold valid_state.

  (*First Constraint*)
  split.
  unfold trusted_os_not_hypercall.
  intros trusted_osi trusted_os H_is_trusted_osi H_trusted_os_map.
  case (os_ident_eq osi trusted_osi).
  intro H_osi_trusted_osi.
  rewrite H_new_oss in H_trusted_os_map.
  rewrite H_osi_trusted_osi in H_trusted_os_map.
  rewrite lemma_map_add_apply in H_trusted_os_map.
  injection H_trusted_os_map; intro H_trusted_os_value; rewrite H_trusted_os_value; auto.
  intro H_osi_not_active.
  cut (trusted_os_not_hypercall (ctxt_oss ctxt) (st_oss s)).
  intro H_trusted_no_hyp_s.
  2:apply H_valid_s.
  rewrite H_new_oss in H_trusted_os_map.
  rewrite lemma_map_add_apply_distinct in H_trusted_os_map; auto.
  apply (H_trusted_no_hyp_s trusted_osi trusted_os H_is_trusted_osi); auto.

  (*Second Constraint*)
  split.
  rewrite H_same_aos; rewrite H_same_activ.
  unfold running_os_not_hypercall.
  
  (*We discriminate the cases of osi*)
  case (os_ident_eq (st_active_os s) osi).

  (*In case osi is aos*)
  intro H_aos_osi.

  exists (OS None (os_current_page system));auto.
  rewrite H_new_oss.
  rewrite H_aos_osi.
  rewrite lemma_map_add_apply;auto.

  (*In case osi is not aos*)
  intro H_no_aos_osi.

  rewrite H_new_oss.
  rewrite lemma_map_add_apply_distinct;auto.
  apply H_valid_s.

  (*Third Constraint*)
  split.
  rewrite H_same_activ; rewrite H_same_exmd.
  apply H_valid_s.

  (*Fourth Constraint*)
  split.
  rewrite H_same_aos; rewrite H_same_exmd.
  apply H_valid_s.

  (*Fifth Constraint*)
  split.
  rewrite H_same_aos; rewrite H_same_activ; rewrite H_same_exmd.
  apply H_valid_s.

  (*Sixth Constraint*)
  split.
  
  unfold valid_hypervisor.
  intro osi'.
  
  (*We discriminate the cases of osi'*)
  case (os_ident_eq osi osi').


  (*In case osi' is osi*)
  intro H_osi'_osi.
  
  rewrite <- H_osi'_osi.

  intro osi_pa.
  
  
  case (padd_eq osi_pa pa).
  
  
  (*In case osi' is osi, and the considered padd is pa*)
  intro H_osipa_pa.
  
  rewrite H_osipa_pa.
  intros p2m' H_p2m'.
  left.
  
  exists ma.
  rewrite H_new_hyper in H_p2m'.
  rewrite lemma_map_add_apply in H_p2m'.
  inversion H_p2m';auto.
  rewrite lemma_map_add_apply;auto.

  rewrite H_new_mem.
  unfold maddress_owner.
  rewrite lemma_map_add_apply.
  unfold new_page.
  case pg_type;auto.


  (*In case osi' is osi, and the considered padd is not pa*)
  intro H_no_osipa_pa.
  intros p2m'
  H_p2m'.

  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) 
    as H_valid_hyper_s.
  apply H_valid_s.

  rewrite H_new_hyper in H_p2m'.
  rewrite lemma_map_add_apply in H_p2m'.
  inversion H_p2m'.

  elim H_valid_hyper_s with osi osi_pa p2m;auto.
  
  (*The hypothesis is an OR, we must prove both branches*)
  intro H_exist_ma'.

  (*Left branch of OR*)
  left.
  destruct H_exist_ma' as (ma', H_p2m_ma', H_ma'_owner).
  exists ma';auto.
  rewrite lemma_map_add_apply_distinct;auto.

  rewrite H_new_mem.
  
  case (madd_eq ma ma').

  intro H_ma'_ma.
  rewrite <- H_ma'_ma.
  unfold maddress_owner.
  rewrite lemma_map_add_apply.
  unfold new_page.
  case pg_type;auto.


  intro H_no_ma'_ma.

  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct;auto.


  (*Right branch of OR*)
  right.
  rewrite lemma_map_add_apply_distinct;auto.


  (*In case osi' is not osi*)
  intro H_no_osi'_osi.

  intros osi'_pa osi'_p2m H_osi'_p2m.

  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) 
    as H_valid_hyper_s.
  apply H_valid_s.

  elim H_valid_hyper_s with osi' osi'_pa osi'_p2m;auto. 
  intro H_exists_ma'.


  (*The hypothesis is an OR, we must prove both branches*) 
  left.
  destruct H_exists_ma' as (ma', H_osi_p2m_ma', H_ma'_owner).
  exists ma';auto.
  
  rewrite H_new_mem.
    
  assert (ma'<>ma).
  (*diferente owner*)
  unfold not;intro H_ma'_ma_abs.
  unfold maddress_owner in H_ma'_owner.
  rewrite <- H_ma'_ma_abs in H_ma_pg.
  rewrite <- H_ma_pg in H_ma'_owner.
  rewrite H_pg_owner in H_ma'_owner.
  inversion H_ma'_owner.

  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct;auto.

  rewrite H_new_hyper in H_osi'_p2m.
  rewrite lemma_map_add_apply_distinct in H_osi'_p2m;auto.
  

  (*Seventh Constraint*)
  split.

  intro osi'.

  (*We discriminate the cases of osi'*)
  
  case (os_ident_eq osi' osi).

  
  (*In case osi' is osi*)
  intro H_osi'_osi.
  
  rewrite H_osi'_osi.
  intro osi_pt_madd.

  case (madd_eq osi_pt_madd ma).

  (*In case the considered madd is ma, the corresponding page
  is empty, if we suppose that a pin to a PT has been made.
  If we soppose that a pin to a RW page has been made, then
  there is no mapping to prove the property on*)

  intro H_osiptmadd_ma.
  rewrite H_osiptmadd_ma.


  intros ma_map' ma_page' H_ma_page'.
  rewrite H_new_mem in H_ma_page'.
  rewrite lemma_map_add_apply in H_ma_page'.
  inversion_clear H_ma_page'.
  case pg_type.
  simpl.
  
  intro H_ma_map'.
  inversion_clear H_ma_map'.
  intro H_temp;clear H_temp.
  simpl.
  intros ma_map_va ma_map_ma H_ma_va_ma_map.
  inversion H_ma_va_ma_map.

  (*RW type, doesn't met the hypothesis of the property*)
  simpl.
  intro H_ma_map'.
  inversion_clear H_ma_map'.

  (*In case the considered madd is not ma*)
  intro H_no_osiptmadd_ma.

  
  intros osi_map osi_page
    H_osi_pt_page H_osi_pg_map H_osi_pg_owner
      osi_map_va osi_map_ma H_osi_va_ma_map.
  
  assert (valid_virtual_mapping (st_memory s) ctxt) 
    as H_valid_virtual_s.
  apply H_valid_s;auto.

  
  assert (osi_map_ma <> ma) as H_no_osimapma_ma.
  (*There can not be an aos_map that mets this, and has ma on it*)

  unfold not;intro H_osimapma_ma_abs.
  rewrite H_osimapma_ma_abs in H_osi_va_ma_map.

  rewrite H_new_mem in H_osi_pt_page.
  rewrite lemma_map_add_apply_distinct in H_osi_pt_page;auto.

  elim H_valid_virtual_s with osi osi_pt_madd osi_map 
    osi_page osi_map_va ma;auto.

  intros osi_map_ma_pg H_osi_map_ma_pg H_osi_map_ma_prop.

  assert (osi_map_ma_pg=ma_pg) as H_ma_osi_pg.
  rewrite <- H_osi_map_ma_pg in H_ma_pg.
  inversion H_ma_pg;auto.
  rewrite <- H_ma_osi_pg in H_pg_owner.

  rewrite H_pg_owner in H_osi_map_ma_prop.
  destruct H_osi_map_ma_prop as (H_va_acces,H_no_va_acces).

  unfold os_accessible in  H_va_acces.
  unfold os_accessible in  H_no_va_acces.

  destruct (ctxt_vadd_accessible ctxt osi_map_va).
  assert (No_Owner = Os_PO osi) as Absurd.
  apply  H_va_acces;auto.
  inversion Absurd.

  assert (No_Owner = Hyp_PO) as Absurd.
  apply  H_no_va_acces;auto.
  inversion Absurd.


  rewrite H_new_mem.
  rewrite lemma_map_add_apply_distinct;auto.

  rewrite H_new_mem in H_osi_pt_page.
  rewrite lemma_map_add_apply_distinct in H_osi_pt_page;auto.

  elim H_valid_virtual_s with osi osi_pt_madd osi_map 
    osi_page osi_map_va osi_map_ma;auto.

  intros osi_map_ma_pg H_osi_map_ma_pg H_osi_map_ma_prop.
  exists osi_map_ma_pg;auto.
  


  (*In case osi' is osi*)
  intro H_no_osi'_osi.

  intros osi'_pt_madd osi'_map osi'_page
    H_osi'_pt_page H_osi'_pg_map H_osi'_pg_owner
      osi'_map_va osi'_map_ma H_osi'_va_ma_map.
  
  assert (valid_virtual_mapping (st_memory s) ctxt) 
    as H_valid_virtual_s.
  apply H_valid_s;auto.

  assert (ma <> osi'_pt_madd) as H_no_osi'ptmadd_ma.

  (*It can not be the same, because ma has owner osi in mem', not osi'*)
  unfold not;intro H_osi'ptmadd_ma_abs.
  rewrite H_new_mem in H_osi'_pt_page.
  rewrite H_osi'ptmadd_ma_abs in H_osi'_pt_page.
  rewrite lemma_map_add_apply in H_osi'_pt_page.
  assert (osi'_page = (new_page (Some pg_type) (Os_PO osi))) 
    as H_osi'_page_def.
  inversion H_osi'_pt_page;auto.
  assert (Os_PO osi=page_owned_by osi'_page) 
    as H_osi_owns_osi'_pg.
  rewrite H_osi'_page_def; case pg_type; auto.
  rewrite <- H_osi_owns_osi'_pg in H_osi'_pg_owner.
  inversion H_osi'_pg_owner;auto.


  (*It can not be the same, because ma has owner osi in mem', not osi'*)
  assert (ma <> osi'_map_ma) as H_no_osi'mapma_ma.  

  unfold not;intro H_osi'mapma_ma_abs.
  
  rewrite H_new_mem in H_osi'_pt_page.
  rewrite lemma_map_add_apply_distinct in H_osi'_pt_page;auto.
  elim H_valid_virtual_s with osi' osi'_pt_madd osi'_map 
    osi'_page osi'_map_va osi'_map_ma;auto.

  intros osi'_map_ma_pg H_osi'_map_ma_pg H_osi'_map_ma_prop.


  assert (osi'_map_ma_pg=ma_pg) as H_ma_osi'_pg.
  rewrite H_osi'mapma_ma_abs in H_ma_pg.
  rewrite <- H_osi'_map_ma_pg in H_ma_pg.
  inversion H_ma_pg;auto.
  rewrite <- H_ma_osi'_pg in H_pg_owner.

  rewrite H_pg_owner in H_osi'_map_ma_prop.
  destruct H_osi'_map_ma_prop as (H_va_acces,H_no_va_acces).

  unfold os_accessible in  H_va_acces.
  unfold os_accessible in  H_no_va_acces.

  destruct (ctxt_vadd_accessible ctxt osi'_map_va).
  assert (No_Owner = Os_PO osi') as Absurd.
  apply  H_va_acces;auto.
  inversion Absurd.

  assert (No_Owner = Hyp_PO) as Absurd.
  apply  H_no_va_acces;auto.
  inversion Absurd.


  rewrite H_new_mem in H_osi'_pt_page.
  rewrite lemma_map_add_apply_distinct in H_osi'_pt_page;auto.
  
  rewrite H_new_mem.
  rewrite lemma_map_add_apply_distinct;auto.
 
  elim H_valid_virtual_s with osi' osi'_pt_madd osi'_map 
    osi'_page osi'_map_va osi'_map_ma;auto.

  intros osi'_map_ma_pg H_osi'_map_ma_pg H_osi'_map_ma_prop.
  exists osi'_map_ma_pg;auto.


  (*Eighth Constraint*)
  split.
  unfold valid_current_page.

  intro osi'.

  (*We discriminate the cases of osi'*)
  case (os_ident_eq osi' osi).


  Focus 2.
  (*In case osi' is not osi*)
  intro H_no_osi'_osi.
  
  intros system' H_osi'_system osi'_p2m H_osi'_p2m.

  rewrite H_new_oss in H_osi'_system.
  rewrite lemma_map_add_apply_distinct in H_osi'_system;auto.
  
  rewrite H_new_hyper in H_osi'_p2m.
  rewrite lemma_map_add_apply_distinct in H_osi'_p2m;auto.

  assert (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s) ) as H_valid_cur_s.
  apply H_valid_s.

  elim H_valid_cur_s with osi' system' osi'_p2m;auto.
  intros osi'_pt_madd H_osi'_pt_madd H_osi'_ptmadd_prop.
  exists osi'_pt_madd;auto.

  rewrite H_new_mem.
  unfold maddress_owner.
  unfold page_type.
  rewrite lemma_map_add_apply_distinct;auto.

  (*They are different because they have different page owner o type*)
  unfold not;intro H_osi'ptmadd_ma_abs.
  destruct H_osi'_ptmadd_prop as (H_osi'_ptmadd_ptype,H_osi'_ptmadd_owner).
  rewrite <- H_osi'ptmadd_ma_abs in H_osi'_ptmadd_owner.
  unfold maddress_owner in H_osi'_ptmadd_owner.
  rewrite <- H_ma_pg in  H_osi'_ptmadd_owner.
  rewrite H_pg_owner in  H_osi'_ptmadd_owner.
  inversion  H_osi'_ptmadd_owner;auto.
  

  (*In case osi' is osi*)
  intro H_osi'_osi.
  rewrite H_osi'_osi.
  intros system' H_is_system' p2m' H_p2m'.

  assert (system'=OS None (os_current_page system)) as H_system'_newosi.
  rewrite H_new_oss in  H_is_system'.
  rewrite lemma_map_add_apply in H_is_system'.
  inversion H_is_system';auto.

  assert (p2m'=map_add padd_eq p2m pa ma) as H_p2m'_newp2m.
  rewrite H_new_hyper in H_p2m'.
  rewrite lemma_map_add_apply in H_p2m'.
  inversion H_p2m';auto.

  assert (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s) ) as H_valid_cur_s.
  apply H_valid_s.

  destruct H_valid_cur_s with  osi system p2m 
    as (pt_madd,H_pt_madd,H_pt_prop);auto.
  
  rewrite H_p2m'_newp2m; rewrite H_system'_newosi.
  simpl.


  (*In case pinned page isn't curr_page*)
  assert (pa<>(os_current_page system)) as H_no_cur_page_pa.
  
  unfold not;intro H_pa_cur_page_abs.
  rewrite <- H_pa_cur_page_abs in H_pt_madd.
  rewrite H_free_pa in H_pt_madd.
  inversion H_pt_madd;auto.

  exists pt_madd;auto.
  rewrite lemma_map_add_apply_distinct;auto.
  
  rewrite H_new_mem.
  unfold maddress_owner.
  unfold page_type.
  rewrite lemma_map_add_apply_distinct;auto.

  (*They are different because they have different page owner o type*)
  unfold not;intro H_osi'ptmadd_ma_abs.
  destruct H_pt_prop as (H_osi'_ptmadd_ptype,H_osi'_ptmadd_owner).
  rewrite <- H_osi'ptmadd_ma_abs in H_osi'_ptmadd_owner.
  unfold maddress_owner in H_osi'_ptmadd_owner.
  rewrite <- H_ma_pg in  H_osi'_ptmadd_owner.
  rewrite H_pg_owner in  H_osi'_ptmadd_owner.
  inversion  H_osi'_ptmadd_owner;auto.

  (*Ninth Constraint*)
  split.
  intro osi'.
  
  case (os_ident_eq osi' osi).
  
  (*We discriminate the cases of osi'*)
  intro H_osi'_osi.

  intros newp2m H_newp2m.
  
  (*We can assume injectivity in s*)
  assert (injective_hyper_mappings (st_hypervisor s)) as H_inject_s.
  apply H_valid_s.
  
  (*The considered p2m is that than pin operation is modifying to drop <pa,ma>*)
  assert (newp2m=map_add padd_eq p2m pa ma) as H_newp2m_def.
  rewrite H_new_hyper in H_newp2m.
  rewrite H_osi'_osi in H_newp2m.
  rewrite lemma_map_add_apply in H_newp2m.
  inversion H_newp2m;auto.


  unfold injective_mapping.
  assert  (map_at_index padd_eq newp2m pa ma) as H_pa_ma.
  unfold map_at_index.
  rewrite H_newp2m_def.
  rewrite lemma_map_add_apply;auto.

  cut (forall p', map_apply padd_eq p2m p' <> Value padd ma).
  (*There is no padd mapped to ma in p2m*)
 


  intro H_no_ma_p2m.

  intros pa1 pa2 H_pa1_pa2.

  case (padd_eq pa pa1).
  intro H_pa_pa1.

  case (padd_eq pa pa2).
  rewrite H_pa_pa1;auto.

  intro H_pa_na_pa2.
  elim H_no_ma_p2m with pa2;auto.
  rewrite <- H_pa_pa1 in H_pa1_pa2.
  rewrite H_newp2m_def in H_pa1_pa2.
  rewrite lemma_map_add_apply in H_pa1_pa2.
  rewrite lemma_map_add_apply_distinct in H_pa1_pa2;auto.

  intro H_pa_no_p1.
  
  case (padd_eq pa pa2).

  intro H_pa_pa2.
  elim H_no_ma_p2m with pa2;auto.
  rewrite <- H_pa_pa2 in H_pa1_pa2.
  rewrite H_newp2m_def in H_pa1_pa2.
  rewrite lemma_map_add_apply in H_pa1_pa2.
  rewrite lemma_map_add_apply_distinct in H_pa1_pa2;auto.
  elim H_no_ma_p2m with pa1;auto.

  intro H_pa_no_pa2.
  rewrite H_newp2m_def in H_pa1_pa2.
  rewrite lemma_map_add_apply_distinct in H_pa1_pa2;auto.
  rewrite lemma_map_add_apply_distinct in H_pa1_pa2;auto.
  apply ( (H_inject_s) osi (p2m) (H_p2m) (pa1) (pa2) );auto.

  (*We prove that no padd is maped to ma in p2m*)
  intro p'.
  intro H_p'_ma_abs.
  assert (valid_hypervisor (st_hypervisor s) (st_memory s))
    as H_valid_hyper_s.
  apply H_valid_s.
 
  unfold valid_hypervisor in H_valid_hyper_s.
  destruct H_valid_hyper_s with 
    (osi:=osi) (p2m:=p2m) (pa:=p');auto.
  elim H.
  intros ma' H_ma'.
  assert (ma'=ma) as H_ma'_ma.
  rewrite H_p'_ma_abs in H_ma'.
  inversion H_ma';auto.
  rewrite H_ma'_ma;clear H ma' H_ma' H_ma'_ma.
  intro H_ma_owner.
  
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_ma_pg in H_ma_owner.
  rewrite H_pg_owner in H_ma_owner.
  inversion H_ma_owner.
  rewrite H_p'_ma_abs in H;inversion H.

  (*In case osi is not aos*)
  intro H_osi_no_aos.
  rewrite H_new_hyper.
  rewrite lemma_map_add_apply_distinct;auto.
  apply H_valid_s.

  (*Tenth Constraint*)
  split.
  unfold va_has_valid_pa.
  intro osi'.
  case (os_ident_eq osi osi').
  
  (*Case osi*)
  intro H_osi'_osi.
  intros osi_pt_padd osi_pt_addr osi_pt_page
    osi_mp osi_va osi_ma.
  intros H_osi_pt_addr H_osi_pt_padd_madd
    H_osi_mp_page_cont H_osi_va_ma H_osi_va.
  
  case (madd_eq osi_ma ma).
  
  (*Case osi_ma = ma*)
  intro H_osima_ma.
  exists pa;auto.
  
  rewrite H_osima_ma; rewrite H_new_hyper.
  unfold hypervisor_mapping_madd.
  rewrite H_osi'_osi.
  rewrite lemma_map_add_apply.
  exists (map_add padd_eq p2m pa ma);auto.
  rewrite lemma_map_add_apply;auto.
  
  (*Case osi_ma <> ma*)
  intro H_osima_no_ma.
  
  assert (va_has_valid_pa (st_memory s)
    (st_hypervisor s) ctxt) as H_va_has_pa_s.
  apply H_valid_s.


  case (madd_eq osi_pt_addr ma).

  (*Case osi_pt_addr=ma*)
  intro H_osiptaddr_ma.

  assert (osi_pt_page=new_page (Some pg_type)
    (Os_PO osi)) as H_osi_pt_page_def.
  rewrite H_new_mem in H_osi_pt_addr.
  rewrite H_osiptaddr_ma in H_osi_pt_addr.
  rewrite lemma_map_add_apply in H_osi_pt_addr.
  inversion H_osi_pt_addr;auto.

  rewrite H_osi_pt_page_def in H_osi_mp_page_cont.
  simpl in H_osi_mp_page_cont.
  generalize H_osi_mp_page_cont.

  case pg_type.

  (*Case pg_type is PT*)
  intro H_mp_empty.
  assert (osi_mp = map_empty vadd madd)
    as H_osi_mp_emptyPT.
  inversion H_mp_empty;auto.
  clear H_mp_empty.
  rewrite H_osi_mp_emptyPT in H_osi_va_ma.
  inversion H_osi_va_ma.

  (*Case pg_type is RW*)
  intro H_abs;inversion H_abs.

  (*Case osi_pt_addr <> ma *)
  intro H_osiptmadd_no_ma.

  elim H_va_has_pa_s with osi' osi_pt_padd osi_pt_addr
    osi_pt_page osi_mp osi_va osi_ma;auto.

  intros osi_pa H_osipa.
  exists osi_pa.

  rewrite H_new_hyper.
  rewrite H_osi'_osi.
  unfold hypervisor_mapping_madd.
  rewrite lemma_map_add_apply.
  exists (map_add padd_eq p2m pa ma);auto.

  unfold hypervisor_mapping_madd in H_osipa.
  destruct H_osipa as (p2m',H_p2m',H_osipa).

  assert (p2m'=p2m) as H_p2m'_p2m.
  rewrite <- H_osi'_osi in H_p2m'.
  rewrite <- H_p2m in H_p2m'.
  inversion H_p2m';auto.
  rewrite H_p2m'_p2m in H_osipa.

  rewrite lemma_map_add_apply_distinct;auto.
  
  intro H_abs.
  rewrite <- H_abs in H_osipa.
  rewrite H_free_pa in H_osipa.
  inversion H_osipa.

  rewrite H_new_mem in H_osi_pt_addr.
  rewrite lemma_map_add_apply_distinct
    in H_osi_pt_addr;auto.

  exists p2m;auto.
  rewrite <- H_osi'_osi;assumption.
  
  rewrite H_new_hyper in H_osi_pt_padd_madd.
  unfold hypervisor_mapping_madd in H_osi_pt_padd_madd.
  rewrite H_osi'_osi in  H_osi_pt_padd_madd.
  rewrite lemma_map_add_apply in H_osi_pt_padd_madd.
  destruct  H_osi_pt_padd_madd as (p2m',H_p2m'_val,H_p2m').
  assert (p2m' = map_add padd_eq p2m pa ma) as H_p2m'_def.
  inversion H_p2m'_val;auto.
  rewrite H_p2m'_def in H_p2m'.
  rewrite lemma_map_add_apply_distinct in H_p2m';auto.

  intro H_abs.
  rewrite H_abs in H_p2m'.
  rewrite lemma_map_add_apply in H_p2m'.
  inversion H_p2m';contradiction.

  (*Case osi' <> osi*)
  intro H_osi'_no_osi.
  intros osi'_pt_padd osi'_pt_addr osi'_pt_page
    osi'_mp osi'_va osi'_ma.
  intros H_osi'_pt_addr H_osi'_pt_padd_madd
    H_osi'_mp_page_cont H_osi'_va_ma H_osi'_va.
  
  assert (va_has_valid_pa (st_memory s)
    (st_hypervisor s) ctxt) as H_va_has_pa_s.
  apply H_valid_s.

  elim H_va_has_pa_s with osi' osi'_pt_padd osi'_pt_addr
    osi'_pt_page osi'_mp osi'_va osi'_ma;auto.

  intros osi'_pa H_osi'pa.
  exists osi'_pa;auto.

  rewrite H_new_hyper; unfold hypervisor_mapping_madd.
  rewrite lemma_map_add_apply_distinct;auto.
  
  rewrite H_new_mem in H_osi'_pt_addr.
  rewrite lemma_map_add_apply_distinct
    in H_osi'_pt_addr;auto.

  (*ma is different form osi'_pt_addr*)
  intro H_abs.
  rewrite H_abs in H_osi'_pt_addr.
  rewrite lemma_map_add_apply in H_osi'_pt_addr.
  assert (osi'_pt_page = new_page (Some pg_type)
    (Os_PO osi)) as H_osi'_pt_page_def.
  inversion H_osi'_pt_addr;auto.
  rewrite H_osi'_pt_page_def in H_osi'_mp_page_cont.
  generalize H_osi'_mp_page_cont.

  case pg_type.

  (*Case pg_type is PT*)
  intro H_case_PT.

  assert (osi'_mp = map_empty vadd madd) as H_osi'_mp_def.
  inversion H_case_PT;auto.
  rewrite H_osi'_mp_def in H_osi'_va_ma.
  inversion H_osi'_va_ma.

  intro H_case_RW.
  inversion H_case_RW;auto.
  
  rewrite H_new_hyper in H_osi'_pt_padd_madd.
  unfold hypervisor_mapping_madd.
  unfold hypervisor_mapping_madd in H_osi'_pt_padd_madd.
  rewrite lemma_map_add_apply_distinct
  in H_osi'_pt_padd_madd;auto.

  apply exec_valid_cache_tlb with ctxt s (Page_pin_untrusted osi pa pg_type); auto.
Qed.


Theorem Page_unpin_untrusted_secure:
  forall (s s':state),
  forall osi pa,
  valid_state ctxt s ->
  valid_exec ctxt s (Page_unpin_untrusted osi pa) s'->
  valid_state ctxt s'.
Proof.
  intros s s' osi pa H_valid_s H_valid_exec;
    generalize H_valid_exec; intro H_valid_exec';
    inversion_clear H_valid_exec'.
  simpl in H.
  unfold pre_page_unpin_untrusted in H.
  destruct H as (H_untrusted_osi,H_pre).
  destruct H_pre as (H_hypercall_unpin,H_pre).
  destruct H_pre as (H_hyper_running,H_pre_unpin).
  
  unfold pre_all_unpin in H_pre_unpin.
  destruct H_pre_unpin as (system, H_is_system, H_pre_unpin).
  destruct H_pre_unpin as (H_no_cur_page_pa, H_pre_unpin).
  destruct H_pre_unpin as (p2m, H_p2m, H_pre_ma_pa).
  destruct H_pre_ma_pa as (ma, H_ma_pa, H_pre_ma_pa).
  destruct H_pre_ma_pa as (H_ma_ptype, H_ma_no_mapped).

  simpl in H0.
  unfold post_page_unpin_untrusted in H0.
  unfold post_all_unpin in H0.

  destruct H0 as (p2m',H_p2m',H_post_unpin).
  destruct H_post_unpin as (ma',H_ma'_pa,H_post_unpin).
  assert (p2m'=p2m) as H_p2m'_p2m.
  unfold hypervisor_mapping in H_p2m'.
  unfold hypervisor_mapping in H_p2m.
  rewrite <-H_p2m in H_p2m';inversion H_p2m';auto.
  assert (ma'=ma) as H_ma'_ma.
  rewrite H_p2m'_p2m in H_ma'_pa.
  rewrite <- H_ma_pa in H_ma'_pa.
  inversion H_ma'_pa;auto.
  
  rewrite H_ma'_ma in H_post_unpin;rewrite H_p2m'_p2m in H_post_unpin.
  clear p2m' H_p2m' ma' H_ma'_pa H_ma'_ma H_p2m'_p2m.
  
  destruct H_post_unpin as (system',H_is_system',H_s'_def).
  assert (system'=system) as H_sys'_sys.
  rewrite <- H_is_system in H_is_system'.
  inversion H_is_system';auto.
  rewrite H_sys'_sys in H_s'_def;clear system' H_is_system' H_sys'_sys.

  (*We establish the relation between s and s' components*)

  assert (st_active_os s' = st_active_os s ) as H_same_aos.
  rewrite H_s'_def;auto.
  assert (st_oss s' = map_add os_ident_eq (st_oss s) osi
              (OS None (os_current_page system))) as H_new_oss.
  rewrite H_s'_def;auto.
  assert (st_exec_mode s' = st_exec_mode s) as H_same_exmd.
  rewrite H_s'_def;auto.
  assert (st_aos_activity s' = st_aos_activity s) as H_same_activ.
  rewrite H_s'_def;auto.
  assert (st_hypervisor s' = (map_add os_ident_eq (st_hypervisor s) 
              osi (map_drop padd_eq p2m pa))) as H_new_hyper.
  rewrite H_s'_def;auto.
  assert (st_memory s' = map_add madd_eq (st_memory s) ma
                (new_page None No_Owner)  ) as H_new_mem.
  rewrite H_s'_def;auto.

  clear H_s'_def.
  
  (*Injectivity of p2m holds, and it's used on the proof*)  
  assert (injective_mapping padd_eq p2m) as H_injective_p2m.
  assert (injective_hyper_mappings (st_hypervisor s)) as H_inject_s.
  apply H_valid_s.
  unfold injective_hyper_mappings in H_inject_s.
  apply ( (H_inject_s) osi (p2m));auto.


  unfold valid_state.

  (*First Constraint*)
  split.
  unfold trusted_os_not_hypercall.
  intros trusted_osi trusted_os H_is_trusted_osi H_trusted_os_map.
  case (os_ident_eq osi trusted_osi).
  intro H_osi_trusted_osi.
  rewrite H_new_oss in H_trusted_os_map.
  rewrite H_osi_trusted_osi in H_trusted_os_map.
  rewrite lemma_map_add_apply in H_trusted_os_map.
  injection H_trusted_os_map; intro H_trusted_os_value; rewrite H_trusted_os_value; auto.
  intro H_osi_not_active.
  cut (trusted_os_not_hypercall (ctxt_oss ctxt) (st_oss s)).
  intro H_trusted_no_hyp_s.
  2:apply H_valid_s.
  rewrite H_new_oss in H_trusted_os_map.
  rewrite lemma_map_add_apply_distinct in H_trusted_os_map; auto.
  apply (H_trusted_no_hyp_s trusted_osi trusted_os H_is_trusted_osi); auto.

  (*Second Constraint*)
  split.
  rewrite H_same_aos; rewrite H_same_activ.
  unfold running_os_not_hypercall.
  
  (*We discriminate the cases of osi*)
  case (os_ident_eq (st_active_os s) osi).

  (*In case osi is aos*)
  intro H_aos_osi.

  exists (OS None (os_current_page system));auto.
  rewrite H_new_oss.
  rewrite H_aos_osi.
  rewrite lemma_map_add_apply;auto.

  (*In case osi is not aos*)
  intro H_no_aos_osi.

  rewrite H_new_oss.
  rewrite lemma_map_add_apply_distinct;auto.
  apply H_valid_s.

  (*Third Constraint*)
  split.
  rewrite H_same_activ; rewrite H_same_exmd.
  apply H_valid_s.

  (*Fourth Constraint*)
  split.
  rewrite H_same_aos; rewrite H_same_exmd.
  apply H_valid_s.

  (*Fifth Constraint*)
  split.
  rewrite H_same_aos; rewrite H_same_activ; rewrite H_same_exmd.
  apply H_valid_s.

  (*Sixth Constraint*)
  split.
  
  unfold valid_hypervisor.
  intro osi'.
  
  (*We discriminate the cases of osi*)
  case (os_ident_eq osi osi').


  (*In case osi' is osi*)
  intro H_osi'_osi.
  
  rewrite <- H_osi'_osi.

  intro osi_pa.
  
  
  case (padd_eq osi_pa pa).
  
  
  (*In case osi is aos, and the considered padd is pa*)

  intro H_osipa_pa.
  
  rewrite H_osipa_pa.
  intros p2m' H_p2m'.
  right.
  
  rewrite H_new_hyper in H_p2m'.
  rewrite lemma_map_add_apply in H_p2m'.
  inversion H_p2m';auto.
  rewrite map_drop_apply;auto.

  (*In case osi is aos, and the considered padd is not pa*)

  intro H_no_osipa_pa.
  intros p2m' H_p2m'.

  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) 
    as H_valid_hyper_s.
  apply H_valid_s.

  rewrite H_new_hyper in H_p2m'.
  rewrite lemma_map_add_apply in H_p2m'.
  inversion H_p2m'.
  rewrite map_drop_apply_distinct;auto.

  elim H_valid_hyper_s with osi osi_pa p2m;auto.
  
  (*The hypothesis is an OR, we must prove both branches*)
  intro H_exist_ma'.

  left.
  destruct H_exist_ma' as (ma', H_p2m_ma', H_ma'_owner).
  exists ma';auto.

  rewrite H_new_mem.
  
  assert (ma'<>ma) as H_no_ma'_ma.

  unfold not;intro H_ma'_ma_abs.
  unfold injective_mapping in H_injective_p2m.

  assert (osi_pa=pa) as H_osipa_pa_abs.
  rewrite H_ma'_ma_abs in H_p2m_ma'.
  rewrite H_ma_pa in H_p2m_ma'.
  elim ((H_injective_p2m) (pa) (osi_pa) (H_p2m_ma'));reflexivity.
  contradiction.

  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct;auto.


  (*In case osi is not aos*)

  intro H_no_osi'_osi.

  intros osi'_pa osi'_p2m H_osi'_p2m.

  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) 
    as H_valid_hyper_s.
  apply H_valid_s.

  elim H_valid_hyper_s with osi' osi'_pa osi'_p2m;auto. 
  intro H_exists_ma'.


  (*The hypothesis is an OR, we must prove both branches*)    
  left.
  destruct H_exists_ma' as (ma', H_osi_p2m_ma', H_ma'_owner).
  exists ma';auto.
  
  rewrite H_new_mem.
  
  
  (*Se prueba en base a que ma' y ma son distintas, porque
  tienen diferente owner, uno es osi, y otro es osi'*)
  assert (ma'<>ma) as H_ma'_no_ma.
  unfold not;intro H_ma'_ma_abs.
  
  (*For being ma and pa associated in the hypervisor mapping
  we can say that ma belongs to osi, using the property valid 
  hypervisor in s*)

  elim H_valid_hyper_s with osi pa p2m;auto. 
  intro H_exists_ma''.

  destruct H_exists_ma'' as (ma'', H_osi_p2m_ma'', H_ma''_owner).
  assert (ma''=ma) as H_ma''_ma.
  rewrite <- H_ma_pa in H_osi_p2m_ma'';inversion H_osi_p2m_ma'';auto.
  rewrite H_ma''_ma in H_ma''_owner.
  rewrite H_ma'_ma_abs in H_ma'_owner.
  rewrite H_ma'_owner in H_ma''_owner.
  inversion H_ma''_owner;auto.

  intro H_pa_not_pined_osi.
  rewrite <- H_ma_pa in H_pa_not_pined_osi.
  inversion H_pa_not_pined_osi;auto.


  unfold maddress_owner.
  rewrite lemma_map_add_apply_distinct;auto.

  rewrite H_new_hyper in H_osi'_p2m.
  rewrite lemma_map_add_apply_distinct in H_osi'_p2m;auto.
  

  (*Seventh Constraint*)
  split.

  intro osi'.

  (*We discriminate the cases of osi*)
  
  case (os_ident_eq osi' osi).

  
  (*En caso que si sea el sistema operativo actual*)
  intro H_osi'_osi.
  
  rewrite H_osi'_osi.
  intro osi_pt_madd.

  (*We discriminate cases of the pt_madd considered*) 
  case (madd_eq osi_pt_madd ma).

  (*In case the considered ma is the same that ma, the
  corresponding page has type "Other", so it doesn't met
  the hypothesis of the property to be proved*)

  intro H_osiptmadd_ma.
  rewrite H_osiptmadd_ma.


  intros ma_map' ma_page' H_ma_page'.
  rewrite H_new_mem in H_ma_page'.
  rewrite lemma_map_add_apply in H_ma_page'.
  inversion_clear H_ma_page'.
  simpl.
  intro H_ma_map'_contr.
  inversion_clear H_ma_map'_contr.

  (*In case considered madd is not ma*)
  intro H_no_osiptmadd_ma.

  
  intros osi_map osi_page
    H_osi_pt_page H_osi_pg_map H_osi_pg_owner
      osi_map_va osi_map_ma H_osi_va_ma_map.
  
  assert (valid_virtual_mapping (st_memory s) ctxt) 
    as H_valid_virtual_s.
  apply H_valid_s;auto.

  
  assert (osi_map_ma <> ma) as H_no_osimapma_ma.
  (*There can not be an aos_map that mets this, and has ma on it*)

  unfold not;intro H_osimapma_ma_abs.

  rewrite H_osimapma_ma_abs in H_osi_va_ma_map.
  unfold no_va_mapped_to_ma in H_ma_no_mapped.

  elim H_ma_no_mapped with osi_pt_madd osi_page osi_map_va osi_map;auto.
  rewrite H_new_mem in H_osi_pt_page.
  rewrite lemma_map_add_apply_distinct in H_osi_pt_page;auto.


  rewrite H_new_mem.
  rewrite lemma_map_add_apply_distinct;auto.

  rewrite H_new_mem in H_osi_pt_page.
  rewrite lemma_map_add_apply_distinct in H_osi_pt_page;auto.

  elim H_valid_virtual_s with osi osi_pt_madd osi_map 
    osi_page osi_map_va osi_map_ma;auto.

  intros osi_map_ma_pg H_osi_map_ma_pg H_osi_map_ma_prop.
  exists osi_map_ma_pg;auto.
  


  (*In case osi' is not osi*)
  intro H_no_osi'_osi.

  intros osi'_pt_madd osi'_map osi'_page
    H_osi'_pt_page H_osi'_pg_map H_osi'_pg_owner
      osi'_map_va osi'_map_ma H_osi'_va_ma_map.
 
  assert (ma <> osi'_pt_madd) as H_no_osi'ptmadd_ma.

  (*It can not be the same, because ma has owner osi in mem', not osi'*)
  unfold not;intro H_osi'ptmadd_ma_abs.
  

  (*For being ma and pa associated in the hypervisor mapping
  we can say that ma belongs to osi, using the property valid 
  hypervisor in s*)
  
  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) 
    as H_valid_hyper_s.
  apply H_valid_s.

  elim H_valid_hyper_s with osi pa p2m;auto. 
  intro H_exists_ma'.

  destruct H_exists_ma' as (ma', H_osi_p2m_ma', H_ma'_owner).
  assert (ma'=ma) as H_ma'_ma.
  rewrite <- H_ma_pa in H_osi_p2m_ma';inversion H_osi_p2m_ma';auto.
  rewrite H_ma'_ma in H_ma'_owner.
  rewrite H_osi'ptmadd_ma_abs in H_ma'_owner.

  assert (maddress_owner (st_memory s) osi'_pt_madd = No_Owner) 
    as H_osi'ptmadd_no_owner.

  rewrite H_new_mem in H_osi'_pt_page.

  rewrite H_osi'ptmadd_ma_abs in H_osi'_pt_page.
  rewrite lemma_map_add_apply in H_osi'_pt_page.
  unfold maddress_owner.

  inversion H_osi'_pt_page;auto.

  rewrite H0 in H_osi'_pg_map.
  inversion H_osi'_pg_map;auto.

  rewrite H_osi'ptmadd_no_owner in H_ma'_owner.
  inversion H_ma'_owner;auto.

  intro H_pa_not_pined_osi.
  rewrite <- H_ma_pa in H_pa_not_pined_osi.
  inversion H_pa_not_pined_osi;auto.


  (*No puede ser porque pertenece a osi en mem', no a osi'*)
  assert (ma <> osi'_map_ma) as H_no_osi'mapma_ma.

  unfold not;intro H_osi'map_ma_abs.

  
  assert (valid_virtual_mapping (st_memory s) ctxt) 
    as H_valid_virtual_s.
  apply H_valid_s;auto.

  elim H_valid_virtual_s with osi' osi'_pt_madd 
    osi'_map osi'_page osi'_map_va osi'_map_ma;auto.

  2:rewrite H_new_mem in H_osi'_pt_page.
  2:rewrite lemma_map_add_apply_distinct in H_osi'_pt_page;auto.
  
  intros pg H_pg H_pg_prop.
  assert (maddress_owner (st_memory s) ma <> Os_PO osi) 
    as H_no_ma_owner_osi.
  
  rewrite <- H_osi'map_ma_abs in H_pg.
  unfold maddress_owner. 
  rewrite <- H_pg.
  unfold not;intro H_pg_owner_abs.

  destruct H_pg_prop as (H_map_va_acces,H_map_va_no_acces).
  unfold os_accessible in H_map_va_no_acces.
  unfold os_accessible in H_map_va_acces.
  destruct (ctxt_vadd_accessible ctxt osi'_map_va).
  rewrite H_pg_owner_abs in H_map_va_acces;auto.
  assert (Os_PO osi = Os_PO osi') as H_osi'_osi_abs.
  apply H_map_va_acces;auto.
  inversion H_osi'_osi_abs.
  rewrite  H0 in H_no_osi'_osi;auto.
  
  rewrite H_pg_owner_abs in H_map_va_no_acces;auto.
  assert (Os_PO osi = Hyp_PO) as H_osi_hyper_abs.
  apply H_map_va_no_acces;auto.
  inversion H_osi_hyper_abs.


  (*For being ma and pa associated in the hypervisor mapping
  we can say that ma belongs to osi, using the property valid 
  hypervisor in s*)
  
  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) 
    as H_valid_hyper_s.
  apply H_valid_s.

  elim H_valid_hyper_s with osi pa p2m;auto. 
  intro H_exists_ma'.

  destruct H_exists_ma' as (ma', H_osi_p2m_ma', H_ma'_owner).
  assert (ma'=ma) as H_ma'_ma.
  rewrite <- H_ma_pa in H_osi_p2m_ma';inversion H_osi_p2m_ma';auto.
  rewrite H_ma'_ma in H_ma'_owner.
  contradiction.
  intro H_pa_not_pined_osi.
  rewrite <- H_ma_pa in H_pa_not_pined_osi.
  inversion H_pa_not_pined_osi;auto.
  

  rewrite H_new_mem in H_osi'_pt_page.
  rewrite lemma_map_add_apply_distinct in H_osi'_pt_page;auto.
  
  rewrite H_new_mem.
  rewrite lemma_map_add_apply_distinct;auto.
 
  assert (valid_virtual_mapping (st_memory s) ctxt) 
    as H_valid_virtual_s.
  apply H_valid_s;auto.

  elim H_valid_virtual_s with osi' osi'_pt_madd osi'_map 
    osi'_page osi'_map_va osi'_map_ma;auto.

  intros osi'_map_ma_pg H_osi'_map_ma_pg H_osi'_map_ma_prop.
  exists osi'_map_ma_pg;auto.


  (*Eighth Constraint*)
  split.
  unfold valid_current_page.

  intro osi'.

  (*We discriminate the cases of osi*)
  case (os_ident_eq osi' osi).


  Focus 2.
  (*In case osi' is not osi*)
  intro H_no_osi'_osi.
  
  intros system' H_osi'_system osi'_p2m H_osi'_p2m.

  rewrite H_new_oss in H_osi'_system.
  rewrite lemma_map_add_apply_distinct in H_osi'_system;auto.
  
  rewrite H_new_hyper in H_osi'_p2m.
  rewrite lemma_map_add_apply_distinct in H_osi'_p2m;auto.

  assert (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s) ) as H_valid_cur_s.
  apply H_valid_s.

  elim H_valid_cur_s with osi' system' osi'_p2m;auto.
  intros osi'_pt_madd H_osi'_pt_madd H_osi'_ptmadd_prop.
  exists osi'_pt_madd;auto.

  rewrite H_new_mem.
  unfold maddress_owner.
  unfold page_type.
  rewrite lemma_map_add_apply_distinct;auto.

  (*They are different because they have different page owner o type*)
  unfold not;intro H_osi'ptmadd_ma_abs.
  destruct H_osi'_ptmadd_prop as (H_osi'_ptmadd_ptype,H_osi'_ptmadd_owner).
  rewrite <- H_osi'ptmadd_ma_abs in H_osi'_ptmadd_owner.

  (*For being ma and pa associated in the hypervisor mapping
  we can say that ma belongs to osi, using the property valid 
  hypervisor in s*)
  
  assert (valid_hypervisor (st_hypervisor s) (st_memory s)) 
    as H_valid_hyper_s.
  apply H_valid_s.

  elim H_valid_hyper_s with osi pa p2m;auto. 
  intro H_exists_ma'.

  destruct H_exists_ma' as (ma', H_osi_p2m_ma', H_ma'_owner).
  assert (ma'=ma) as H_ma'_ma.
  rewrite <- H_ma_pa in H_osi_p2m_ma';inversion H_osi_p2m_ma';auto.
  rewrite H_ma'_ma in H_ma'_owner.
  rewrite H_ma'_owner in H_osi'_ptmadd_owner.
  inversion H_osi'_ptmadd_owner;auto.

  intro H_pa_not_pined_osi.
  rewrite <- H_ma_pa in H_pa_not_pined_osi.
  inversion H_pa_not_pined_osi;auto.


  (*In case osi' is osi*)
  intro H_osi'_osi.
  rewrite H_osi'_osi.
  intros system' H_is_system' p2m' H_p2m'.

  assert (system'=OS None (os_current_page system)) as H_system'_newosi.
  rewrite H_new_oss in  H_is_system'.
  rewrite lemma_map_add_apply in H_is_system'.
  inversion H_is_system';auto.

  assert (p2m'=map_drop padd_eq p2m pa) as H_p2m'_newp2m.
  rewrite H_new_hyper in H_p2m'.
  rewrite lemma_map_add_apply in H_p2m'.
  inversion H_p2m';auto.

  assert (valid_current_page (st_hypervisor s) (st_memory s) (st_oss s) ) 
    as H_valid_cur_s.
  apply H_valid_s.

  destruct H_valid_cur_s with  osi system p2m 
    as (pt_madd,H_pt_madd,H_pt_prop);auto.
  
  rewrite H_p2m'_newp2m; rewrite H_system'_newosi.
  simpl.

  exists pt_madd;auto.
  rewrite map_drop_apply_distinct;auto.
  
  rewrite H_new_mem.
  unfold maddress_owner.
  unfold page_type.
  rewrite lemma_map_add_apply_distinct;auto.


  (*They are different becaus is pa can't be the padd 
  of the current page, then ma can't be de madd of the
  current page*)
  unfold not;intro H_osi'ptmadd_ma_abs.
  rewrite <- H_osi'ptmadd_ma_abs in H_pt_madd.
  rewrite H_ma_pa in H_pt_madd.
  inversion  H_pt_madd;auto.


  (*Ninth Constraint*)
  split.
  intro osi'.
  
  case (os_ident_eq osi' osi).
  
  (*We discriminate the cases of osi*)
  intro H_osi'_osi.

  intros newp2m H_newp2m.
  
  (*We can assume injectivity in s*)
  assert (injective_hyper_mappings (st_hypervisor s)) as H_inject_s.
  apply H_valid_s.
  
  (*The considered p2m is that than unpin operation is modifying to drop <pa,ma>*)
  assert (newp2m=map_drop padd_eq p2m pa) as H_newp2m_def.
  rewrite H_new_hyper in H_newp2m.
  rewrite H_osi'_osi in H_newp2m.
  rewrite lemma_map_add_apply in H_newp2m.
  inversion H_newp2m;auto.


  unfold injective_mapping.
  assert  (map_apply padd_eq newp2m pa = Error madd pa) as H_pa_ma.
  rewrite H_newp2m_def.
  rewrite map_drop_apply;auto.

  intros pa1 pa2 H_pa1_pa2.

  case (padd_eq pa pa1).
  intro H_pa_pa1.

  case (padd_eq pa pa2).
  rewrite H_pa_pa1;auto.

  intro H_pa_na_pa2.
  elim map_error_index_inv with padd padd_eq madd (p2m) (pa2) (pa);auto.
  rewrite <- H_pa_pa1 in H_pa1_pa2.
  rewrite H_newp2m_def in H_pa1_pa2.
  rewrite map_drop_apply in H_pa1_pa2.
  rewrite map_drop_apply_distinct in H_pa1_pa2;auto.

  intro H_pa_no_p1.
  
  case (padd_eq pa pa2).

  intro H_pa_pa2.
  elim map_error_index_inv with padd padd_eq madd (p2m) (pa1) (pa);auto.
  rewrite <- H_pa_pa2 in H_pa1_pa2.
  rewrite H_newp2m_def in H_pa1_pa2.
  rewrite map_drop_apply in H_pa1_pa2.
  rewrite map_drop_apply_distinct in H_pa1_pa2;auto.

  intro H_pa_no_pa2.
  rewrite H_newp2m_def in H_pa1_pa2.
  rewrite map_drop_apply_distinct in H_pa1_pa2;auto.
  rewrite map_drop_apply_distinct in H_pa1_pa2;auto.


  (*In case osi' is not osi*)
  intro H_osi'_no_osi.
  rewrite H_new_hyper.
  rewrite lemma_map_add_apply_distinct;auto.
  apply H_valid_s.

  (*Tenth Constraint*)
  split.
  unfold va_has_valid_pa.
  intro osi'.
  case (os_ident_eq osi osi').
  
  (*Case osi*)
  intro H_osi'_osi.
  intros osi_pt_padd osi_pt_addr osi_pt_page
    osi_mp osi_va osi_ma.
  intros H_osi_pt_addr H_osi_pt_padd_madd
    H_osi_mp_page_cont H_osi_va_ma H_osi_va.
  
  assert (osi_pt_addr<>ma) as H_osiptadd_no_ma.
  intro H_abs_2.
  rewrite H_new_mem in H_osi_pt_addr.
  rewrite H_abs_2 in H_osi_pt_addr.
  rewrite lemma_map_add_apply in H_osi_pt_addr.
  assert (osi_pt_page=new_page None No_Owner)
    as H_osi_pt_page_def.
  inversion H_osi_pt_addr;auto.
  rewrite H_osi_pt_page_def in H_osi_mp_page_cont.
  inversion  H_osi_mp_page_cont.

  destruct H_osi_pt_padd_madd as (newp2m,H_newp2m,H_osi_pt_padd_madd).
  assert (newp2m=map_drop padd_eq p2m pa) as H_newp2m_def.
  rewrite H_new_hyper in H_newp2m.
  rewrite H_osi'_osi in H_newp2m.
  rewrite lemma_map_add_apply in H_newp2m.
  inversion H_newp2m;auto.

  assert (pa <> osi_pt_padd) as H_osipa_no_pa.
  intro H_abs_3.
  rewrite H_newp2m_def in H_osi_pt_padd_madd.
  rewrite H_abs_3 in H_osi_pt_padd_madd.
  rewrite map_drop_apply in H_osi_pt_padd_madd.
  inversion H_osi_pt_padd_madd.

  assert (osi_ma<>ma) as H_osima_no_ma.
  intro H_abs.
  elim H_ma_no_mapped with osi_pt_addr osi_pt_page osi_va osi_mp;auto.
  rewrite H_new_mem in H_osi_pt_addr.
  rewrite lemma_map_add_apply_distinct in H_osi_pt_addr;auto.
  assert (valid_hypervisor (st_hypervisor s)
    (st_memory s)) as H_valid_hyper_s.
  apply H_valid_s.
  elim H_valid_hyper_s with osi' osi_pt_padd p2m.
  intro H_pt_madd.
  destruct H_pt_madd as (pt_madd,H_pt_madd,H_pt_madd_owner).
  assert (pt_madd=osi_pt_addr) as H_osipt_ma.
  rewrite H_newp2m_def in H_osi_pt_padd_madd.
  rewrite map_drop_apply_distinct in H_osi_pt_padd_madd;auto.
  rewrite <- H_osi_pt_padd_madd in H_pt_madd.
  inversion H_pt_madd;auto.
  unfold maddress_owner in H_pt_madd_owner.
  rewrite H_osipt_ma in H_pt_madd_owner.
  rewrite H_new_mem in H_osi_pt_addr.
  rewrite lemma_map_add_apply_distinct in H_osi_pt_addr;auto.
  rewrite <- H_osi_pt_addr in H_pt_madd_owner.
  rewrite H_osi'_osi;assumption.
  intro H_madd_error.
  rewrite H_newp2m_def in H_osi_pt_padd_madd.
  rewrite map_drop_apply_distinct in
    H_osi_pt_padd_madd;auto.
  rewrite <- H_osi_pt_padd_madd in H_madd_error.
  inversion H_madd_error.
  rewrite <- H_osi'_osi;assumption.
  rewrite H_abs in H_osi_va_ma;auto.
 
  (*We use the fact that the constraint holds in s*)
  assert (va_has_valid_pa (st_memory s)
    (st_hypervisor s) ctxt) as H_va_has_pa_s.
  apply H_valid_s.

  elim H_va_has_pa_s with osi' osi_pt_padd osi_pt_addr
    osi_pt_page osi_mp osi_va osi_ma;auto.

  intros osi_pa H_osipa.
  destruct H_osipa as (p2m',H_p2m',H_osipa).
  assert (p2m'=p2m) as H_p2m'_p2m.
  rewrite <- H_osi'_osi in H_p2m'.
  rewrite <- H_p2m in H_p2m'.
  inversion H_p2m';auto.

  exists osi_pa;auto.
  rewrite H_new_hyper.
  unfold hypervisor_mapping_madd.
  rewrite H_osi'_osi.
  rewrite lemma_map_add_apply.
  exists (map_drop padd_eq p2m pa);auto.
  rewrite map_drop_apply_distinct;auto.
  rewrite H_p2m'_p2m in H_osipa;auto.

  intro H_osipa_abs.
  rewrite H_p2m'_p2m in H_osipa.
  rewrite <- H_osipa_abs in H_osipa.
  rewrite <- H_ma_pa in H_osipa.
  inversion H_osipa;contradiction.

  rewrite H_new_mem in H_osi_pt_addr.
  rewrite lemma_map_add_apply_distinct
    in H_osi_pt_addr;auto.
  
  exists p2m.
  rewrite <- H_osi'_osi;assumption.
  rewrite H_newp2m_def in H_osi_pt_padd_madd.
  rewrite map_drop_apply_distinct
    in H_osi_pt_padd_madd;auto.

  (*Case osi' is not osi*)
  intro H_osi'_no_osi.
  
  intros osi'_pt_padd osi'_pt_addr osi'_pt_page
    osi'_mp osi'_va osi'_ma.
  intros H_osi'_pt_addr H_osi'_pt_padd_madd
    H_osi'_mp_page_cont H_osi'_va_ma H_osi'_va.

  (*We prove osi'_pt_addr <> ma*)
  assert (osi'_pt_addr <> ma) as H_osi'pt_no_ma.
  intro H_osi'pt_abs.
  rewrite H_new_mem in H_osi'_pt_addr.
  rewrite H_osi'pt_abs in H_osi'_pt_addr.
  rewrite lemma_map_add_apply in H_osi'_pt_addr.
  assert (osi'_pt_page = Page Other_c No_Owner)
    as H_osi'_pt_page_def.
  inversion H_osi'_pt_addr;auto.
  rewrite H_osi'_pt_page_def in
    H_osi'_mp_page_cont.
  inversion H_osi'_mp_page_cont.
 
  (*And now we just use the validity os s*)
  assert (va_has_valid_pa (st_memory s)
    (st_hypervisor s) ctxt) as H_va_has_pa_s.
  apply H_valid_s.

  elim H_va_has_pa_s with osi' osi'_pt_padd
    osi'_pt_addr osi'_pt_page osi'_mp osi'_va
  osi'_ma;auto.
  
  intros osi'_pa H_osi'_pa.
  exists osi'_pa.

  rewrite H_new_hyper;unfold hypervisor_mapping_madd.
  rewrite lemma_map_add_apply_distinct;auto.

  rewrite H_new_mem in H_osi'_pt_addr.
  rewrite lemma_map_add_apply_distinct
    in H_osi'_pt_addr;auto.

  rewrite H_new_hyper in H_osi'_pt_padd_madd.
  unfold hypervisor_mapping_madd in H_osi'_pt_padd_madd.
  rewrite lemma_map_add_apply_distinct in H_osi'_pt_padd_madd;auto.

  apply exec_valid_cache_tlb with ctxt s (Page_unpin_untrusted osi pa); auto.
Qed.


Theorem valid_state_invariant : forall (s s':state) (a:action),
  valid_state ctxt s ->
  valid_exec ctxt s a s' ->
  valid_state ctxt s'.
Proof.
  induction a.
  apply silent_secure.
  apply Read_secure.
  apply Read_hyper_secure.
  apply Write_secure.
  apply Write_Hyper_secure.
  apply New_untrusted_secure.
  apply New_trusted_secure.
  apply New_hyper_secure.
  apply Del_untrusted_secure.
  apply Del_trusted_secure.
  apply Del_hyper_secure.
  apply Switch_secure.
  apply Lswitch_trusted_secure.
  apply Lswitch_untrusted_secure.
  apply Hcall_secure.
  apply Ret_ctrl_secure.
  apply Chmod_secure.
  apply Page_pin_trusted_secure.
  apply Page_unpin_trusted_secure.
  apply Page_pin_untrusted_secure.
  apply Page_unpin_untrusted_secure.
Qed.    


End Secure_State.
